unit Equation;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Buttons, Equmenu, ExtCtrls, StdCtrls, Parser, Grids;

type
  TEquationForm = class(TForm)
    ShapeBut: TSpeedButton;
    TestEdit: TEdit;
    ResultEdit: TEdit;
    TestBut: TButton;
    ConstrBut: TButton;
    MathParser: TMathParser;
    CalcBut: TButton;
    ValueGrid: TStringGrid;
    CheckBut: TButton;
    DummyEdit: TEdit;
    SaveBut: TButton;
    OpenBut: TButton;
    CurveImg: TImage;
    CurveBut: TButton;
    NameLabel: TLabel;
    NameEdit: TEdit;
    CurrentLabel: TLabel;
    MultiLabel: TLabel;
    AdderLabel: TLabel;
    WarnEdit: TEdit;
    EquivEdit: TEdit;
    PointerEdit: TEdit;
    procedure ShapeButClick(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure TestButClick(Sender: TObject);
    procedure ConstrButClick(Sender: TObject);
    procedure CalcButClick(Sender: TObject);
    procedure CheckButClick(Sender: TObject);
    procedure TestEditEnter(Sender: TObject);
    procedure MathParserParseError(Sender: TObject; ParseError: Integer);
    procedure MathParserGetVar(Sender: TObject; VarName: String;
              var Value: Extended; var Found: Boolean);
    procedure OpenButClick(Sender: TObject);
    procedure SaveButClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CurveButClick(Sender: TObject);
    procedure NameEditChange(Sender: TObject);
    procedure CurrentLabelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure MultiLabelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure AdderLabelMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure EquivEditChange(Sender: TObject);
    procedure PointerEditChange(Sender: TObject);
  private
    { Private declarations }
    CurrIndx : integer;   {index for I/Iset variable}
    MultIndx : integer;   {index for time multiplier}
    AddeIndx : integer;   {index for time adder}
  public
    { Public declarations }
    procedure EquationRectangle;
    procedure ExplanationRectangle;
    procedure FillEquivList;
    procedure FindEquivName(lbl: TLabel; var found: boolean; var indx: integer);
  end;

type
  TpForm = ^TForm;

var
  EquationForm : TEquationForm;


implementation

{$R *.DFM}

uses
  {values}routines, Imagopas, PQEquation;

const
  cmax = 100.0;    {max. current for curve}
  cmin = 1.0;      {min. current for curve} {Start with a larger, avoid /0 !}
  ln10 = 2.302585; {ln(10)}
  npoint = {21}41; {number of points for curve}    {ALUSSA enemmn??}
  tmax = 1000.0;   {max. time for curve}
  tmin = 0.01;     {min. time for curve}


var
  cmult     : single;                     {multiplier for calculation of current points}
  PartStr   : TStringList;                {temporary, for part strings}
  TestL2    : TStringList;
  VarList   : TStringList;                {names and values as text, for saving}
  EquivList : TStringList;                {list of equivalent names, see below}
  values    : array[0..9] of single;      {array for parameter values. Dim >= ValueGrid.rowcount !!}
  currs     : array[0..npoint] of single; {values of current}
  temps     : array[0..npoint] of single; {corresponding values of time}
{NOTE: There are three parameters for the inverse time char. (or similar) of a PGQ-relay
 in the ValueGrid of PQForm (of PQRelay.pas): Set current, Time multiplier (B) and
 Time adder (C). The list EquivList tells, which variables in the (human readable) equation
 correspond to these variables.
 But NOTE, the Set current is not used as such in the equations. It is used to calculate
 a normalized current I/Iset. The symbol 'M' is often used in the equations for this
 The lines of the EquivList _must_ be like  "name in ValueGrid of PQform;index of the name
 in the ValueGrid of this form". The three (so far?) lines are thus something like
 Current/set current;0           (meaning that I/Iset = M)
 Time Multiplier (B);2
 Time Adder (C);3                (the indices 0, 2, 3 are as examples only)
 The semicolon is used as separator because of compatibility with some procedures}


procedure TEquationForm.MathParserParseError(Sender: TObject; ParseError: Integer);
var
  Msg : string;
begin
  case ParseError of
    1 : Msg := 'Parser stack overflow.';
    2 : Msg := 'Bad cell range.';
    3 : Msg := 'Unknown variable or missing operator or..?';
    4 : Msg := 'Expected operator.';
    5 : Msg := 'Expected opening parenthesis.';
    6 : Msg := 'Expected operator or closing parenthesis.';
    7 : Msg := 'Invalid numeric expression.';
  end; { case }
  Msg := Msg + ' Position in string: ' + IntToStr(MathParser.Position);
  MessageDlg(Msg, mtError, [mbOk], 0);
  TestEdit.SelStart := Pred(MathParser.Position);
  TestEdit.SelLength := 0;
end;

procedure TEquationForm.MathParserGetVar(Sender: TObject; VarName: String;
  var Value: Extended; var Found: Boolean);
{try to use values given in the grid. No difference between upper and lower case }
var
  ii      : integer;
  GrNamU  : string;       {variable name in grid in upper case}
  VarNamU : string;       {variable names in upper case}
begin
  value := 0;
  found := false;         {initialize}
  VarNamU := uppercase(VarName);
  for ii := 0 to ValueGrid.rowcount-1 do begin
    GrNamU := uppercase(ValueGrid.cells[0,ii]);
    if (GrNamU = VarNamU) then begin     {variable found in the grid}
      Value := values[ii];               {pick value, must have been written there}
      found := true;
      break;
    end;
  end;
end;

procedure TEquationForm.ShapeButClick(Sender: TObject);
begin
  EquationMenuForm.visible := ShapeBut.down;
end;

procedure TEquationForm.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  EquationMenuForm.StartDrawing(x,y);
  {dummy edit is needed to get focus to the form!}
  Dummyedit.visible := true;            {must be visible...}
  ActiveControl := DummyEdit;           {..when getting focus, but...}
end;

procedure TEquationForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  {draw the shape by floowing the mouse}
  EquationMenuForm.FinishDrawingWithMouse(x,y);
  DummyEdit.visible := false;           {...focus to the form, when dummyedit becomes invisible!}
end;

procedure TEquationForm.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if (ssLeft in Shift) then
    EquationMenuForm.DrawWithMouse(x,y);
end;

procedure TEquationForm.FormCreate(Sender: TObject);
{transfer pointers, initialize the grid for variable values}
var
  ii : integer;
begin
//EquationMenuForm.PutpMainForm(longint(@EquationForm),longint(@ShapeBut),longint(@VarList), {..moved to...}
//                              longint(@EquivList),longint(@EquivEdit));  {...PointerEditChange..}
  {initialize the grid}
  Valuegrid.cells[0,0] := 'M';
  Valuegrid.cells[1,0] := '1.01';
  for ii := 1 to ValueGrid.rowcount-1 do begin
    Valuegrid.cells[0,ii] := char(ii+96);       {a,b,c,...}
    ValueGrid.cells[1,ii] := '1';               {initialize to nonzerozero}
  end;
  EquivList := TStringList.create;
  VarList := TStringList.create;
  PartStr := TStringList.create;
  TestL2 := TStringList.create;
  {rectangle around the picture }
  with (CurveImg.canvas) do begin
    pen.color := clBlack;
    pen.width := 1;
    brush.style := bsClear;
    rectangle(0,0,CurveImg.width,CurveImg.height);
  end;
  {multiplier for calculation of current points}
  cmult := exp((1.0/(npoint-1))*ln(cmax/cmin));
  {rectangle around the equation area}
  EquationRectangle;
  ExplanationRectangle;
  {create the objects for working with the equations}
  PQEquForm.CreateEquationObjects;
  {- - - - - - - - - - - - - - -}
  {add flag that form created}
  Imagoo.ReadyEdit.text := Imagoo.ReadyEdit.text+'E';
end;

procedure TEquationForm.FormPaint(Sender: TObject);
{...}
begin
  EquationMenuForm.DrawAll;
  EquationRectangle;
  ExplanationRectangle;
end;

procedure TEquationForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
{move the graphical elements}
var
  dx,dy : integer;
begin
  dx := 0;
  dy := 0;
  if (key = 38) then dy := -16
  else if (key = 39) then dx := 16
  else if (key = 40) then dy := 16
  else if (key = 37) then dx := -16;
  if (dx <> 0) or (dy <> 0) then EquationMenuForm.MoveAll(dx,dy);
end;

procedure TEquationForm.TestButClick(Sender: TObject);
{for tests}
var
  ii : integer;
begin
//  {check removal of spaces}
//  ResultEdit.text := '!'+WithoutLTspaces(TestEdit.text)+'!';
//  TestL1.clear;
//  TestL2.clear;
//  TestL1.add('eka');
//  TestL1.add('toka');
//  TestL2.add('kolmas');
//  TestL2.add('neljs');
//  TestL1.addstrings(TestL2);
//  for ii := 0 to TestL1.count-1 do showmessage(TestL1[ii]);
  EquationRectangle;
end;

procedure TEquationForm.ConstrButClick(Sender: TObject);
{construct an equation of one line from the "graphical" equation}
var
  equ : string;
begin
  TestEdit.text := '';
  EquationMenuForm.ConstructEquation(equ);
  TestEdit.text := equ;
end;

procedure TEquationForm.CalcButClick(Sender: TObject);
{parse and calculate the value of the expression in testedit}
begin
  {transform the equation from the "human readable form" to the "computer readable form"}
  ConstrButClick(sender);  
  {first, check and copy variables to local array}
  CheckButClick(sender);
  {parse and calculate}
  with MathParser do begin
    ParseString := TestEdit.text;
    Parse;
    if ParseError then
      ResultEdit.Text := '#Error'
    else
      ResultEdit.Text := FloatToStrF(ParseValue, ffGeneral, 15, 2);
  end;
end;

procedure TEquationForm.CheckButClick(Sender: TObject);
{check the values in the grid, and collect to an array}
var
  apu,apv : single;    {temporary}
  apw,apx : single;    {temporary}
  errc    : integer;   {error flag}
  ii      : integer;   {index}
  ipu     : integer;   {temporary}
  line    : string;    {temporary}
  st1,st2 : string;    {temporary}
  st3,st4 : string;    {temporary}
  val     : single;    {value of the variable}
begin
  ValueGrid.color := clWhite;
  for ii := 0 to ValueGrid.rowcount-1 do begin
    line := ValueGrid.cells[1,ii];
    if (line <> '') then begin      {empty cells not considered}
      ValuesFromLine(line,1,false,ipu, val,apv,apw,apx,st1,st2,st3,st4,errc); {one parameter a time}
      if (errc <> 0) then begin       {error in some value}
        ValueGrid.color := clYellow;
        values[ii] := 1;              {set some value}
      end else begin
        values[ii] := val;
      end;
    end else begin                    {empty cell, probably not used}
      values[ii] := 1;                {set some value}
    end;
  end;
end;

procedure TEquationForm.TestEditEnter(Sender: TObject);
begin
  EquationForm.keypreview := false;
end;

procedure TEquationForm.OpenButClick(Sender: TObject);
{read the equation in graphical form, and the variables from a file}
var
  ii  : integer;
  jj  : integer;
  nam : string;  {name of variable}
  val : string;  {value of variable}
begin
  {clear first  (SAVE somewhere for "cancel", or undo !!??}
  NameEdit.text := '';
  TestEdit.text := '';
  ResultEdit.text := '';
  EquationMenuForm.DeleteAll;                   {clear the equation}
  CurveImg.picture := nil;                      {clear the picture}
  with (CurveImg.canvas) do begin               {rectangle around the picture}
    pen.color := clBlack;
    pen.width := 1;
    brush.style := bsClear;
    rectangle(0,0,CurveImg.width,CurveImg.height);
  end;
  for ii := 0 to ValueGrid.rowcount-1 do begin  {clear the parameters}
    ValueGrid.cells[0,ii] := '';
    ValueGrid.cells[1,ii] := '';
  end;
  EquationMenuForm.OpenButClick(sender);
  {move variables to the grid}
  if (VarList.count > 1) then begin
    for ii := 0 to (VarList.count div 2)-1 do begin
      if (ii < ValueGrid.rowcount) then begin
        jj := 2*ii;
        nam := VarList[jj];            {name}
        val := VarList[jj+1];          {value}
        ValueGrid.cells[0,ii] := nam;  {name to grid}
        ValueGrid.cells[1,ii] := val;
      end else begin
        break;
      end;
    end;
  end;
  {get name of the equation}
  EquationMenuForm.GetEquationName(nam);
  NameEdit.text := nam;
end;

procedure TEquationForm.SaveButClick(Sender: TObject);
{save the graphical equation and the variables (parameters) to file}
var
  ii   : integer;  {index}
  line : string;  {one line of text}
begin
  {find equivalent variables, to get the latest (this unit and the unit for PQrelays)}
  FillEquivList;
  {cannot continue, if variable for current/set current not found !}
  if (not WarnEdit.visible) then begin
    {construct the equation, to get the latest}
    ConstrButClick(sender);
    EquationMenuForm.PutEquation(TestEdit.text);     {move the equation to data to be saved}



    {collect name + value separated by ; on a list}
    VarList.clear;
    for ii := 0 to ValueGrid.rowcount-1 do begin
      line := ValueGrid.cells[0,ii]+';'+ValueGrid.cells[1,ii];
      VarList.add(line);
    end;
    {save the graphical equation and the variable values to file }
    EquationMenuForm.SaveButClick(sender);
  end;
end;

procedure TEquationForm.FormDestroy(Sender: TObject);
begin
  EquivList.free;
  VarList.free;
  PartStr.free;
  TestL2.free;
end;

procedure TEquationForm.CurveButClick(Sender: TObject);
{calculate time as a function of current, draw curve}
const
 initcur = 1.01;                   {initial value for current}
var
  apu,apv     : single;   {temporary}
  ipoint      : integer;  {index for points}
  ix,iy       : integer;  {coordinates}
  lcmin,lcmax : single;   {log(10) of min, max current}
  lgcurr      : single;   {log of current}
  lgtime      : single;   {log of time}
  ltmin,ltmax : single;   {log(10) of min, max time}
begin
  ConstrButClick(sender);          {construct one-line equation from the graphical eq.}
  CheckButClick(sender);           {check parameters, fill array}
  FillEquivList;                   {find variable names for curve and PQrelay}
  {can continue only, if variable for I/Iset found}
  if (CurrIndx >= 0) then begin
    Values[CurrIndx] := initcur;            {1.01 for avoiding /0}

    {calculate first value}
    with MathParser do begin
      ParseString := TestEdit.text;
      Parse;
      if ParseError then begin
        ResultEdit.Text := '#Error';
        currs[0] := 1;
        temps[0] := 1;    {just some value}
      end else begin
        currs[0] := initcur;
        temps[0] := ParseValue;
        ResultEdit.Text := FloatToStrF(ParseValue, ffGeneral, 15, 2);
      end;
    end;
    {calculate other values}
    values[CurrIndx] := 1;                         {initialize current}
    for ipoint := 1 to npoint-1 do begin
      values[CurrIndx] := cmult*values[CurrIndx];  {increase current}
      {calculation, (and parsing every time, not actually necessary?)}
      with MathParser do begin
        Parse;
        if ParseError then begin
          ResultEdit.Text := '#Error';
          currs[ipoint] := 1;
          temps[ipoint] := 1;    {just some value}
        end else begin
          currs[ipoint] := values[CurrIndx];
          temps[ipoint] := ParseValue;
        end;
      end;
    end;
    {draw the curve as log-log plot}
    lcmin := ln(cmin)/ln10;               {log of min. current}
    lcmax := ln(cmax)/ln10;
    ltmin := ln(tmin)/ln10;               {log of min. time}
    ltmax := ln(tmax)/ln10;
    {set image properties}
    CurveImg.picture := nil;              {clear first}
    CurveImg.canvas.rectangle(0,0,CurveImg.width,CurveImg.height);
    CurveImg.canvas.pen.color := clBlack;
    CurveImg.canvas.pen.color := clBlue;
    CurveImg.canvas.pen.width := 1;
    CurveImg.canvas.brush.style := bsClear;
    {draw point to point}
    for ipoint := 0 to npoint-1 do begin
      if (currs[ipoint] > 0) then lgcurr := ln(currs[ipoint])/ln10 else lgcurr := 0;
      if (temps[ipoint] > 0) then lgtime := ln(temps[ipoint])/ln10 else lgtime := 0;
      {calculate corresponding point in the image}
      apu := (lgcurr-lcmin)*CurveImg.width/(lcmax-lcmin);     {current}
      apv := (lgtime-ltmin)*CurveImg.height/(ltmax-ltmin);    {time}
      ix := round(apu+0.5);
      iy := CurveImg.height-round(apv+0.5);
      {draw straight line between points}
      if (ipoint = 0) then CurveImg.canvas.moveto(ix,iy)
      else CurveImg.canvas.lineto(ix,iy);
    end;
  end else begin
    {variable for I/Iset not found...}
  end;
end;

procedure TEquationForm.EquationRectangle;
{draw a rectangle around the area where equation written}
const
  rx = 4;
  ry = 40;
  rw = 394;
  rh = 200;
begin
  with canvas do begin
    pen.width := 1;
    pen.color := clBlack;
    pen.style := psSolid;
    brush.style := bsClear;
    rectangle(rx,ry,rw,rh);
    font.name := 'arial';
    font.color := clBlue;
    font.size := 10;
    textout(rx+2,ry+2,'equation');
  end;
end;

procedure TEquationForm.ExplanationRectangle;
{draw a rectangle around the area where the explanations are stored}
const
  rx = 266;
  ry = 208;
  rw = 394;
  rh = 316;
begin
  with canvas do begin
    pen.width := 1;
    pen.color := clGray;
    pen.style := psSolid;
    brush.style := bsClear;
    rectangle(rx,ry,rw,rh);
    font.name := 'arial';
    font.color := clBlue;
    font.size := 9;
    textout(rx+2,ry+2,'explanations');
    textout(rx+2,ry+18,'drag with mouse!');
  end;
end;

procedure TEquationForm.NameEditChange(Sender: TObject);
begin
  EquationMenuForm.PutEquationName(NameEdit.text);
end;

procedure TEquationForm.CurrentLabelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
{move the label}
const
  x0  : integer = 0;    {old coordinates}
  y0  : integer = 0;
begin
  if (ssLeft in Shift) then begin        {left button pressed}
    CurrentLabel.left := CurrentLabel.left+X-x0;
    CurrentLabel.top  := CurrentLabel.top +Y-y0;
  end else begin
    x0 := X;
    y0 := Y;
  end;
  {set color, if used as explanation}
  if (CurrentLabel.left < ValueGrid.left+ValueGrid.width+16) then
    CurrentLabel.font.color := clGreen
  else
    CurrentLabel.font.color := clSilver;
end;

procedure TEquationForm.MultiLabelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
{move the label}
const
  x0  : integer = 0;    {old coordinates}
  y0  : integer = 0;
begin
  if (ssLeft in Shift) then begin        {left button pressed}
    MultiLabel.left := MultiLabel.left+X-x0;
    MultiLabel.top  := MultiLabel.top +Y-y0;
  end else begin
    x0 := X;
    y0 := Y;
  end;
  {set color, if used as explanation}
  if (MultiLabel.left < ValueGrid.left+ValueGrid.width+16) then
    MultiLabel.font.color := clGreen
  else
    MultiLabel.font.color := clSilver;
end;

procedure TEquationForm.AdderLabelMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
{move the label}
const
  x0  : integer = 0;    {old coordinates}
  y0  : integer = 0;
begin
  if (ssLeft in Shift) then begin        {left button pressed}
    AdderLabel.left := AdderLabel.left+X-x0;
    AdderLabel.top  := AdderLabel.top +Y-y0;
  end else begin
    x0 := X;
    y0 := Y;
  end;
  {set color, if used as explanation}
  if (AdderLabel.left < ValueGrid.left+ValueGrid.width+16) then
    AdderLabel.font.color := clGreen
  else
    AdderLabel.font.color := clSilver;
end;

procedure TEquationForm.FindEquivName(lbl: TLabel; var found: boolean; var indx: integer);
{find the variable name that is on the same height as the label
 lbl    the label (component)
 found  true, when the equivalent is found
 indx   the index of the variable in the value grid}
var
  grh    : integer;  {height of cell+grid line}
  grx    : integer;  {right coord. of value grid}
  line   : string;
  lx,ly  : integer;  {coordinates of label}
  midlbl : integer;  {height (elevation) of midpoint of label}
begin
  found := false;
  lx := lbl.left;                           {copy to locals}
  ly := lbl.top;
  midlbl := ly+(lbl.height div 2);          {y-coord. of midpoint}
  grx := ValueGrid.left+ValueGrid.width;    {right edge of value grid}
  {height of one row}
  grh := ValueGrid.defaultRowHeight+ValueGrid.GridLineWidth;
  {near the grid for variables?}
  if (lx >= grx-4) and (lx <= grx+20) then begin
    indx := (midlbl-ValueGrid.top) div grh;                 {row at the height of the label}
//  EquivList.add(lbl.caption+';'+ValueGrid.cells[0,indx]); {save the equivalence}
    EquivList.add(lbl.caption+';'+Inttostr(indx));          {save the equivalence}
    found := true;
  end else begin
    found := false;
    indx := -1;
  end;
end;

procedure TEquationForm.FillEquivList;
{fill the list defining the equivalencies of the variable names in the value grid
of this program and in the value grid of the PQrelay-form PQForm.
The equivalencies are determined by the positions of the explanations
(labels) that can be moved to the grid}
var
  FoundCurr : boolean;  {variable for I/Iset found}
  FoundMult : boolean;  {variable for time multiplier found}
  FoundAdde : boolean;  {variable for time adder found}
begin
  EquivList.clear;
  {variable for normalized current, I/Iset}
  FindEquivName(CurrentLabel,FoundCurr, CurrIndx);
  {variable for time multiplier}
  FindEquivName(MultiLabel,FoundMult, MultIndx);
  {variable for time adder}
  FindEquivName(AdderLabel,FoundAdde, AddeIndx);
  {give warning, if variable for I/Iset not found (the others are not so important!}
  if (not FoundCurr) then begin
    WarnEdit.text := 'Variable for Current/Set current not found! Drag text!';
    WarnEdit.visible := true;
  end else begin
    WarnEdit.visible := false;
  end;
end;

procedure TEquationForm.EquivEditChange(Sender: TObject);
{a change here indicates that a new curve has been read (opened), with
 new indices for the equivalent variables (variables here and parameters
 in PQrelays). Put the values of the indices, and the texts to correct places
 Rext changed in RecallEquivalencies of EquMenu.pas, when starting and reading equations}
var
  cnt   : integer;  {number of equivalencies}
  expla : string;   {explanation}
  grh   : integer;  {height of row in grid}
  ii    : integer;  {index}
  indx  : integer;  {index of explanation}
  line  : string;   {one line of text}
  pcnt  : integer;  {number of part strings}
  txt   : string;   {temporary}
begin
{Old values of indices ????}
  cnt := EquivList.count;
  line := EquivList[cnt-1];
  GetPartStrings(line, longint(@PartStr)); {separate the explanation and index}
  {there should be two strings: explanation and index}
  pcnt := PartStr.count;                   {number of part strings}
  if (pcnt >= 2) then begin
    expla := PartStr[0];                   {explanation}
    txt := PartStr[1];                     {index}
    {index should be integer}
    if (IsInteger(txt,indx)) then begin
      grh := ValueGrid.defaultRowHeight+ValueGrid.GridLineWidth; {height of row}
      {is integer, but which explanation}
      if (expla = CurrentLabel.caption) then begin
        CurrIndx := indx;
        {put the explanation to the right place}
        CurrentLabel.left := ValueGrid.left+ValueGrid.width+1;
        CurrentLabel.top := ValueGrid.top+indx*grh;
        CurrentLabel.font.color := clGreen;
      end else if (expla = MultiLabel.caption) then begin
        MultIndx := indx;
        {put the explanation to the right place}
        MultiLabel.left := ValueGrid.left+ValueGrid.width+1;
        MultiLabel.top := ValueGrid.top+indx*grh;
        MultiLabel.font.color := clGreen;
      end else if (expla = AdderLabel.caption) then begin
        AddeIndx := indx;
        {put the explanation to the right place}
        AdderLabel.left := ValueGrid.left+ValueGrid.width+1;
        AdderLabel.top := ValueGrid.top+indx*grh;
        AdderLabel.font.color := clGreen;
      end;
    end;
  end;
end;

procedure TEquationForm.PointerEditChange(Sender: TObject);
{send pointers to EquMenu, after the form has been created.
 Text is changed in FormCreate of Equmenu                            }
begin
 if (PointerEdit.text = 'created') then begin
   EquationMenuForm.PutpMainForm(longint(@EquationForm),longint(@ShapeBut),longint(@VarList),
                              longint(@EquivList),longint(@EquivEdit));
   EquationMenuForm.InitializeGraphic;                           
 end;
end;

end.
