unit TapChanger;

interface

uses
  SysUtils,Dialogs,Classes, Messages, Skemadef, Nodes, Compdef, elplex, Cmatmul,
  matmul, Utildef, Imagodef, LoadFlow, Emtp, GenVolt;

procedure LoadTapCalculation(ipSet, calc : longint; var success: boolean);
procedure UpdateTapDisplay(pSet : TpCompSet);
procedure ShowTap(pComp : TpComp);

implementation

uses
  Imagopas;

type
{define an object for the calculations with the transformers}
  TpTrafoCalc = ^TTrafoCalc;            {pointer to the object for calculations}
  TTrafoCalc  = class(TObject)          {the set of nodes}
  procedure Initialize(ipSet: longint); {initializes the variables}
  procedure FinalizeAlmost;             {free memory, etc, except for the derivatives}
  procedure GetCalcMemory;
  procedure FreeCalcMemory;
  procedure ReadTapLimits;
  procedure ReadSecondaryVoltages;
  procedure CheckSmallDiagonalElements;
  procedure SolveNet;
  procedure dVdTap;                     {calculate the deriv. ofvoltage wrt tap position}
  procedure ConstantTap(matr : integer);
  procedure ConstructRhs(var averr : single);
  procedure Backsub;
  procedure Decompose;
  procedure DecompGra;
  procedure BacksubGra;
  procedure RemoveConstraintsIf(var rem : boolean);
  procedure NewTapPositions(var avtap : single; var NewLim : boolean);
  procedure FindParallelTrafos;
  procedure ModifyMatrForParallelTrafos(matr: integer);
  procedure ModifyRhsForParallelTrafos(itr : integer);
  function  SecondaryVoltage(ii : integer): single;
  private
   nnod_t      : integer;    {number of nodes}
   ntrafo_t    : integer;    {number of transformers}
   paa_t       : pSingle;    {pointer  to coeff.matrix}
   pbb_t       : pSingle;    {pointer to right hand side}
   pgra_t      : pSingle;    {pointer to coeff.matrix without limits (~gradient)}
   pgrb_t      : pSingle;    {pointer to rhs without constraints}
   pww_t       : pSingle;    {pointer to working space}
   pIndx_t     : pInteger;   {pointer to book-keeping of row exchanges im paa_t}
   pGrIndx_t   : pInteger;   {poointer to book-keeping of row exchanges in pGra_t}
   pSmall_t    : pBoolean;   {pointer to array indicating (too) small derivatives}
   pdTap_t     : pComplex;   {pointer to limits in tap position, x= max, y = min}
   pAtLimit_t  : pInteger;   {pointer to flags for tap being at limit}
   pCoeff_t    : pComplex;   {copy of coefficient matrix (admittance matrix)}
   pdVdT_t     : pComplex;   {pointer to derivatives d(voltage)/d(Tap); .x = prim, .y = sec.}
   pParal_t    : pInteger;   {pointer to parallel trafos}
   pRhs_t      : pComplex;   {pointer to copy of fight hand side}
   pSet        : TpCompSet;  {pointer to the common}
   pVolt0_t    : pComplex;   {pointer to trafo voltages, taps = 0; .x = prim, .y = sec.}
   pV2_t       : pSingle;    {pointer to secondary (system) voltages}
   pVrV_t      : pSingle;    {pointer to voltage ratios, for parallel trafos}
   TrafoList_t : TList;      {list for pointers to the trafo}

end;

procedure TTrafoCalc.Initialize(ipSet : longint);
{initialize the object for the calculations (get.memory etc.}
var
  errc      : integer;   {error flag}
  jcomp     : integer;   {counts components}
  nT2       : integer;   {no of trafos squared}
  pComp     : TpComp;    {pointer to a component}
begin
  pSet := TpCompSet(ipSet);          {pointer to the common}
  TrafoList_t := TList.Create;
  {find all trafos, put pointers to the list}
  nTrafo_t := 0;
  errc := 0;                         {initialize error flag}
  pComp := nil;
  jcomp := 0;
  {scan all components to find the transformers}
  while (errc = 0) do begin                  {to last component, or first error}
    pSet^.NextComp(pComp, errc,jcomp,false); {pointer to next component}
    if (errc = 0) then begin                 {not error, but maybe last comp.}
      {is a trafo, but not a 3-winding?}
      if (isTrafo(pComp)) and not (pComp^.hint = gTrafo3W) then begin
        inc(nTrafo_t);                {count trafos}
        TrafoList_t.add(pComp);       {add trafo-pointer to list}
        pComp^.PutZpar(0,0,0,1);      {initialize tap position}
      end;
    end;
  end;
  nT2 := nTrafo_t*nTrafo_t;
  Imagoo.NodeButClick(nil);          {deduce nodes}
  nnod_t := pSet^.appl_nn;           {number of nodes}
  if (nnod_t > 0) and (nTrafo_t > 0) then begin         {there are nodes and transformers}
    {reserve mem for curr.trafos and relays (if there is any)}
    pSet^.GetMemForCurrTrAndRelays;
    {reserve and initialize memory for matrices}
    ReserveMat(nnod_t,pAdmat,pWs,pRhs,pIndex);
    {get memory for saving the coefficient matrix}
    GetMem(pCoeff_t,nnod_t*nnod_t*sizeOf(Complex));
    GetMem(pRhs_t,nnod_t*sizeOf(Complex));
    GetMem(pVolt0_t,nTrafo_t*sizeOf(complex));
    GetMem(pdVdT_t,nT2*sizeOf(complex));
  end;

end;

procedure TTrafoCalc.FinalizeAlmost;
{free memory, etc. to release the object}
begin
{ TrafoList_t.free; later}
  if (nnod_t > 0) then begin
    FreeMat(nnod_t);
    FreeMem(pCoeff_t,nnod_t*nnod_t*sizeOf(Complex));
    FreeMem(pRhs_t,nnod_t*sizeOf(Complex));
{   FreeMem(pVolt0_t,nTrafo_t*sizeOf(single));}
{   FreeMem(pdVdT_t,nTrafo_t*nTrafo_t*sizeOf(single));   later}
    pSet^.FreeMemForCurrTrAndRelays;
  end;
end;

procedure TTrafoCalc.SolveNet;
{construct the admittance matrix and solve the net}
var
  apu       : single;    {temporary}
  Dodd      : single;    {keeps track on permutations}
  errcom    : boolean;   {true, if there are unknown components}
  freq      : single;    {frequency}
  ii        : integer;   {index}
  pComp     : TpComp;    {pointer to a component}
  pVolt     : pComplex;   {pointer to voltages}
  Sngl      : Integer;   {flag for singular matrix}
  Tout      : Complex;   {test output from matrix decomposition}
  Volt      : complex;   {trafo voltage}
begin
  freq := pSet^.appl_hz;             {frequency}
  {construct the extended admittance matrix}
  ConstrMatr(pSet, pAdmat, pRhs, nnod_t, freq, Steady, TapDeriv, false, errcom);
  {copy admittance matrix and right hand side to be reused}
  CopyCmat(pAdmat,pCoeff_t,nnod_t);
  CopyCvec(pRhs,pRhs_t,nnod_t);
  {solve voltages, but only if no unknown components}
  if (not(errcom)) then begin
    CLUdecomp(pAdmat,pWs,nnod_t,pIndex,Dodd,Sngl,Tout); {decompose first}
    {back substitution, note result OVERWRITES rhs !!}
    CLUbacksub(pAdmat,nnod_t,pIndex,pRhs);              {backsubstitute}
{..}SaveAndDisplayResults(pSet,nnod_t,freq);            {as a test...test..}
    {save voltages, when tap = 0}
    pVolt := pVolt0_t;                      {initialize pointer}
    for ii := 0 to nTrafo_t-1 do begin      {all trafos}
      pComp :=TrafoList_t.items[ii];
      GetVolt(pComp,pRhs,2,nnod_t,Volt);    {secondary voltage}
      apu := Cabs(Volt);
      pVolt^.y:= apu;                       {save abs.value of voltage}
      GetVolt(pComp,pRhs,1,nnod_t,Volt);    {primary voltage}
      apu := Cabs(Volt);
      pVolt^.x:= apu;                       {save abs.value}
      inc(pVolt);
    end;
{showRmat1(pVolt0_t,nTrafo_t,1);}
  end else begin                            {there was an unknown component}
    { do nothing ? }
  end;
end;

procedure TTrafoCalc.dVdTap;
{calculate the derivatives of secondary voltage wrt tap position}
const
  dTap = 0.01;       {a small change in tap position}
var
  apu,apv : single;    {temporary}
  Dodd    : single;    {keeps track on permutations}
  dVdT    : complex;   {derivative of voltage wrt tap position}
  ElVar   : TElVar;    {for transferring parameters}
  ii      : integer;   {index}
  itr     : integer;   {index}
  MatPro  : TCompProc; {procedure for constructing adm.matr.}
  pdVdT   : pComplex;  {pointer to derivative matrix}
  pComp   : TpComp;    {pointer to a trafo}
  pPrev   : TpComp;    {pointer to previous trafo}
  pTrafo  : TpComp;    {pointer to a trafo}
  pVolt   : pComplex;  {pointer to voltage, when tap = 0}
  Sngl    : Integer;   {flag for singular matrix}
  Tout    : Complex;   {test output from matrix decomposition}
  Volt    : Complex;   {secondary voltage when tap = dTap}
  Volt0   : Complex;   {magnitude of voltage, when tap = 0; x = prim, y = sec}
begin
  {prepare progress bar }
  Imagoo.TapBar.max := nTrafo_t;
  Imagoo.TapBar.width := 400;
  Imagoo.TapBar.top := 50;
  Imagoo.TapBar.left := 500;
  Imagoo.TapBar.position := 0;
  Imagoo.TapBar.visible := true;
  {put problem parameters to one combined variable. Pointer copied later}
  ElVar.paa := pAdmat;
  ElVar.pbb := pRhs;
  ElVar.nn := nnod_t;
  ElVar.freq := pSet^.appl_hz;;
  ElVar.probl := TapDeriv;
  ElVar.fphase := Steady;
  ElVar.initst := false;
  {start calculation}
  pPrev := nil;                          {no previous trafo, yet}
  for itr := 0 to nTrafo_t -1 do begin   {scal all trafos}
    CopyCmat(pCoeff_t,pAdmat,nnod_t);    {original coefficient matrix}
    CopyCvec(pRhs_t,pRhs,nnod_t);        {original right hand side}
    pComp := TrafoList_t.items[itr];     {pointer to a trafo}
    pComp^.PutZpar(dTap,0,0,1);          {a small change tap position}
    if (pPrev <> nil) then pPrev^.PutZpar(0,0,0,1); {reset tap position of prev. trafo}
    {change the admittance matrix for the new tap position}
    ElVar.pComp := pComp;                {copy also the pointer to the component}
    pComp^.GetMatPro(MatPro);            {get procedure for constructing matrix}
    MatPro(longint(@ElVar));             {update admittance matrix for this trafo}
    CLUdecomp(pAdmat,pWs,nnod_t,pIndex,Dodd,Sngl,Tout); {decompose first}
    {back substitution, note result OVERWRITES rhs !!}
    CLUbacksub(pAdmat,nnod_t,pIndex,pRhs);              {backsubstitute}
    {get voltages, when tap = dTap}
    pVolt := pVolt0_t;                          {initialize pointer}
    for ii := 0 to nTrafo_t-1 do begin          {all trafos}
      Volt0.y := pVolt^.y;                      {magnitude of sec.voltage at tap = 0}
      Volt0.x := pVolt^.x;
      pTrafo :=TrafoList_t.items[ii];
      GetVolt(pTrafo,pRhs,2,nnod_t,Volt);       {secondary voltage}
      apu := Cabs(Volt);
      dVdT.y := (apu-Volt0.y)/dTap;             {derivative of sec.voltage}
      GetVolt(pTrafo,pRhs,1,nnod_t,Volt);       {primary voltage}
      apu := Cabs(Volt);
      dVdT.x := (apu-Volt0.x)/dTap;             {derivative of prim.voltage}
      PutCMat(pdVdT_t,ii+1,itr+1,nTrafo_t,nTrafo_t,dVdT); {save derivative}
      inc(pVolt);
    end;
    pPrev := pComp;                      {save pointer to previous trafo}
    Imagoo.TapBar.StepIt;
  end;
{showRmat1(pdVdT_t,nTrafo_t,nTrafo_t);}
Imagoo.TapBar.visible := false;
end;

procedure TTrafoCalc.FindParallelTrafos;
{find the trafos that are parallel with some other trafo.
 One parallel trafo, at most. Trafos are "chained"}
const
  eps = 0.01;         {criterium for small difference}
var
  apu,apv : single;   {temporary}
  cpu,cpv : complex;  {temporary}
  diff    : single;   {difference of coefficients}
  dVj     : single;   {dV/dTap of trafo}
  dVmax   : single;   {max. dV/dTap of parallel trafos}
  ii,jj   : integer;
  imax    : integer;  {index of the trafo with the max dV/dTap, of parallel trafos}
  IsPar   : boolean;  {there are parallel trafo(s)}
  jpar    : integer;  {indesx of a parallel trafo}
  jtr     : integer;  {index of trafo}
  kk      : integer;
  nn      : integer;  {no of trafos}
  pComp   : TpComp;   {pointer to a trafo}
  pInd    : pInteger; {temp.pointer to parallel trafos}
  pParal  : pInteger; {pointer to the parallel trafos}
  pww     : pSingle;  {working space (flag for checked dV/dTap)}
  sum     : single;   {sum of coefficients}
begin
  nn := nTrafo_t;          {no of trafos to local}
  pParal := pParal_t;
  for ii := 1 to nn do begin
    pParal^ := 0;         {initialize}
    inc(pParal);
  end;
  if (nn > 1) then begin             {must have more than one}
    for ii := 1 to nn do begin       {scan all trafos}
      for jj := ii+1 to nn do begin  {all trafos below the trafo under study}
        diff := 0;
        sum := 0;
        for kk := 1 to nn do begin   {all elements in one row of the dV/dTap matrix}
          GetCMat(pdVdT_t,ii,kk,nn,nn,cpu);  {trafo under study}
          GetCMat(pdVdT_t,jj,kk,nn,nn,cpv);  {trafo of comparison}
          diff := diff+abs(cpu.y-cpv.y);     {*.y = deriv. of sec. voltage}
          sum := sum+abs(cpu.y)
        end;
        {if relative difference small enough => parallel}
        if (sum > 0) then begin                {should be...}
          apu := diff/sum;                     {relative difference}
          if (apu < eps) then begin            {small difference => parallel}
            pParal := pParal_t;
            inc(pParal,jj-1);                  {save index of the trafo..}
            pParal^:= ii;                      {...this trafo is parallel to}
            break;                             {one parallel trafo is enough}
          end;
        end;
      end;
    end;
    {keep the equation for the trafo with the largest diagonal dV/dTap,
     the others are parallel with this one}
    pww := pww_t;
    pInd := pIndx_t;
    pParal := pParal_t;
    for ii := 1 to nn do begin       {initialize flags}
      pww^ := 0;
      pInd^ := pParal^;              {copy to temporary array}
      inc(pww);
      inc(pInd);
      inc(pParal);
    end;
    {searh for the largest dV/dTap for each group of parallel trafos. Start from the end}
    for ii := 1 to nn do begin
      jtr := nn-ii+1;                          {start from the end}
      pww := pww_t; inc(pww,jtr-1);            {flag: checked or not}
      if (pww^ = 0) then begin                 {not checked yet}
        GetCMat(pdVdT_t,jtr,jtr,nn,nn,cpu);    {initialize max dV/dTap}
        dVmax := cpu.y;                        {deriv. of secondary voltage}
        imax := jtr;                           {initialize index for max dV/dTap}
        pInd := pIndx_t; inc(pInd,jtr-1);      {get index for...}
        jpar := pInd^;                         {...a parallel trafo}
        IsPar := false;                        {not yet parallel trafos}
        while (jpar > 0) do begin              {check the whole chain of parallel trafos}
          IsPar := true;                       {one parallel trafo, at least}
          pww := pww_t; inc(pww,jpar-1);pww^:=1; {set flag, that this handled}
          GetCMat(pdVdT_t,jpar,jpar,nn,nn,cpu); {dV/dTap of the parallel trafo}
          dVj := cpu.y;
          if (dVj > dVmax) then begin          {largest, so far}
            dVmax := dVj;                      {new largest dV/dTap}
            imax := jpar;                      {new index of the trafo with the largest dV/dTap}
          end;
          pInd := pIndx_t; inc(pInd,jpar-1);  {get index for...}
          jpar := pInd^;                       {...the next parallel trafo}
        end;
        if (IsPar) then begin                  {if there were parallel trafos,...}
          {...set the index of the largest dV/dTap to the parallel trafos}
          jpar := jtr;                         {this trafo is the first parallel trafo}
          pInd := pIndx_t; inc(pInd,jpar-1);
          while (jpar > 0) do begin            {all parallel trafors in this group}
            pParal := pParal_t; inc(pParal,jpar-1);
            if (jpar <> imax) then
              pParal^ := imax                  {index of the trafo this is parallel to}
            else pParal^:= 0;                  {the master-parallel trafo}
            pInd := pIndx_t; inc(pInd,jpar-1); {get index for ...}
            jpar := pInd^;                     {...the next parallel trafo}
          end;
        end;
      end;
    end;
  end;
end;

procedure TTrafoCalc.ModifyMatrForParallelTrafos(matr: integer);
{if a trafo is parallel with another trafo, modify the equation for the trafo
 to force the tap positions to be equal
 matr   1 = actual coefficient matrix paa_t
        2 = coef.matrix for checking constraints pgra_t}
var
  apu,apv : single;    {temporary}
  cpu     : complex;   {temporary}
  cVi,cVp : single;    {calculated prim. voltage of this and parallel trafo}
  dVdTi   : complex;   {dVolt/dTap; x = prim, y = sec}
  dVdTp   : complex;   {dVolt/dTap of the parallel trafo}
  ii,jj   : integer;   {indices}
  ipar    : integer;   {index of (eventual) parallel trafo}
  nn      : integer;   {no of trafos}
  pcf     : pSingle;   {pointer to coeff.matrix}
  pComp   : TpCOmp;    {pointer to trafo under study}
  pTrafo  : TpComp;    {pointer to the main parallel trafo}
  tVi,tVp : single;    {theoretical secondary voltage}
  V1i,V2i : single;    {prim. secondary voltage (parameter) of this trafo}
  V1p,V2p : single;    {prim. secondary voltage (parameter) of parallel trafo}
begin
  nn := nTrafo_t;               {copy to local}
  if(matr = 1) then pcf := paa_t else pcf := pgra_t;
  if (nn > 1) then begin        {otherwise no parallel trafos}
    for ii := 1 to nn do begin
      GetImat(pParal_t,ii,1,nn,1,ipar);
      GetCmat(pdVdT_t,ii,ii,nn,nn,dVdTi);        {derivative}
      if (ipar > 0) and (ipar <= nn) then begin  {there is a parallel trafo}
        GetCmat(pdVdT_t,ipar,ipar,nn,nn,dVdTp);
        pComp := TrafoList_t.items[ii-1];        {pointer to this trafo}
        pTrafo := TrafoList_t.items[ipar-1];     {pointer to the parallel trafo}
        {modify the equation}
        for jj := 1 to nn do begin  {first, clear all elements on the row of this trafo}
          PutMat(pcf,ii,jj,nn,nn,0)
        end;
        {set the tap positions so that currents are equal (if impedances are)}
        PutMat(pcf,ii,ii,nn,nn,1);     {one to the diagonal}
        {must calculate ...}
        pComp^.GetPar(gTrafo_V1,V1i);  {prim.voltage of this trafo (parameter)}
        pComp^.GetPar(gTrafo_V2,V2i);  {sec. voltage of this trafo}
        pTrafo^.GetPar(gTrafo_V1,V1p); {prim.voltage of the parallel trafo}
        pTrafo^.GetPar(gTrafo_V2,V2p); {sec.voltage of the parallel trafo}
        if (V1i > 0) and (V2i > 0) and (V1p > 0) and (V2p > 0) then begin {reasonable param.}
          pComp^.GetVar(gTrafo_V1r,cpu.x);     {real part of prim. voltage}
          pComp^.GetVar(gTrafo_V1i,cpu.y);     {imag part of prom. voltage}
          cVi := Cabs(cpu);                    {calculated prim. voltage of this trafo}
          pTrafo^.GetVar(gTrafo_V1r,cpu.x);    {real part of prim. voltage}
          pTrafo^.GetVar(gTrafo_V1i,cpu.y);    {imag part of prom. voltage}
          cVp := Cabs(cpu);                    {calculated prim. voltage of the parall. trafo}
          tVi := (cVi+dVdTi.x)*V2i/V1i;        {"theoretical" secondary voltage of this trafo}
          tVp := (cVp+dVdTp.x)*V2p/V1p;        {"theoretical" secondary voltage of the parallel trafo}
          if (tVi > 0) and (tVp > 0) then begin
            apu := tVp/tVi;
            apv := (cVp*(V2p/V1p)-cVi*(V2i/V1i))/((V2i/V1i)*(dVdTi.x+cVi));
            PutMat(pcf,ii,ipar,nn,nn,-apu);
            PutMat(pVrV_t,ii,1,nn,1,apv);  {save voltage ratio for the rhs}
          end else begin
            PutMat(pcf,ii,ipar,nn,nn,-1);  {tap positions are the same}
          end;
        end else begin                     {error case}
          PutMat(pcf,ii,ipar,nn,nn,-1);    {tap positions are the same}
        end;
      end;
    end;
  end;
end;

procedure TTrafoCalc.ModifyRhsForParallelTrafos(itr : integer);
{at first iteration, put ratio of voltages - 1 to the rhs, if parallel trafos
 else put zero to the rhs, if parallel trafos, because
 equation is: set taps so that currents are (approximately) equal
 itr   iteration }
var
  apu    : single;
  ii     : integer;
  ipar   : integer;   {index of the parallel trafo}
  nn     : integer;   {no of trafos}
begin
  nn := nTrafo_t;               {copy to local}
  if (nn > 1) then begin        {otherwise no parallel trafos}
    for ii := 1 to nn do begin
      GetImat(pParal_t,ii,1,nn,1,ipar);
      if (ipar > 0) then begin
        if (itr = 1) then begin
          getmat(pVrV_t,ii,1,nn,1,apu); {get voltage ratio}
          {apu := apu-1;                {because the change in tap calculated...}
        end else begin
          apu := 0;                     {..constant added only at first iteration}
        end;
        putmat(pbb_t,ii,1,nn,1,apu);    {parallel trafo, ratio of voltages -1 to rhs}
      end;
    end;
  end;
end;

procedure TTrafoCalc.GetCalcMemory;
{get memory for the calculation of the tap positions}
var
  ii,jj    : integer;
  nn       : integer;   {number of transformers}
  paa      : psingle;
  pbb      : psingle;
  pww      : psingle;
  pIndx    : pinteger;
  pSmall   : pboolean;
  pdtap    : pcomplex;
  pAtLimit : pinteger;
  pV2      : psingle;
  pgra     : psingle;
  pgrb     : psingle;
  pParal   : pInteger;
  pVrV     : pSingle;
begin
  nn := nTrafo_t;            {dim of matrices}
  GetMem(paa_t,nn*nn*sizeOf(single));
  GetMem(pbb_t,nn*sizeOf(single));
  GetMem(pww_t,nn*sizeOf(single));
  GetMem(pIndx_t,nn*sizeOf(integer));
  getMem(pSmall_t,nn*sizeOf(boolean));
  getMem(pdtap_t,nn*sizeOf(complex));
  GetMem(pAtLimit_t,nn*sizeOf(integer));
  GetMem(pV2_t,nn*sizeOf(single));
  GetMem(pGra_t,nn*nn*sizeOf(single));
  GetMem(pGrIndx_t,nn*sizeOf(integer));
  GetMem(pGrb_t,nn*sizeOf(single));
  GetMem(pParal_t,nn*sizeOf(integer));
  GetMem(pVrV_t,nn*sizeOf(single));
  {clear initially}
  paa := paa_t;
  pbb := pbb_t;
  pww := pww_t;
  pIndx := pIndx_t;
  pSmall := pSmall_t;
  pdtap := pdtap_t;
  pAtLimit := pAtLimit_t;
  pV2 := pV2_t;
  pgra := pgra_t;
  pgrb := pgrb_t;
  pParal := pParal_t;
  pVrV := pVrV_t;
  for ii := 1 to nn do begin
    for jj := 1 to nn do begin
      paa^ := 0;
      pgra^:= 0;
      inc(paa);
      inc(pgra);
    end;
    pbb^ := 0;
    pww^ := 0;
    pIndx^ := 0;
    pSmall^ := false;
    pdtap^.x := 0;
    pdtap^.y := 0;
    pAtLimit^ := 0;
    pV2^ := 0;
    pgrb^:= 0;
    pParal^:= 0;
    pVrV^:= 1;
    inc(pbb);
    inc(pww);
    inc(pIndx);
    inc(pSmall);
    inc(pdtap);
    inc(pAtLimit);
    inc(pV2);
    inc(pgrb);
    inc(pParal);
    inc(pVrV);
  end;
end;

procedure TTrafoCalc.FreeCalcMemory;
{free memory for the calculation of the tap positions}
var
  nn : integer;   {number of transformers}
begin
  nn := nTrafo_t;            {dim of matrices}
  FreeMem(paa_t,nn*nn*sizeOf(single));
  FreeMem(pbb_t,nn*sizeOf(single));
  FreeMem(pww_t,nn*sizeOf(single));
  FreeMem(pIndx_t,nn*sizeOf(integer));
  FreeMem(pSmall_t,nn*sizeOf(boolean));
  FreeMem(pdtap_t,nn*sizeOf(complex));
  FreeMem(pAtLimit_t,nn*sizeOf(integer));
  FreeMem(pV2_t,nn*sizeOf(single));
  FreeMem(pgra_t,nn*nn*sizeOf(single));
  FreeMem(pGrIndx_t,nn*sizeOf(integer));
  FreeMem(pGrb_t,nn*sizeOf(single));
  FreeMem(pParal_t,nn*sizeOf(integer));
  FreeMem(pVrV_t,nn*sizeOf(single));
end;

procedure TTrafoCalc.ReadTapLimits;
{read the tap limit parameter for the trafos, put to array
 Note: limit given in %, abs. value of the limit given, the magnitude of the
       limit is the same in the up and down direction}
var
  apu   : single;    {temporary}
  ii    : integer;   {index}
  pComp : TpComp;    {pointer to the trafo}
  pdtap : pComplex;  {local pointer to the limit array}
begin
  pdTap := pdTap_t;                            {pointer to array of tap-limits}
  for ii := 0 to nTrafo_t-1 do begin
    pComp := TrafoList_t.items[ii];      {pointer to a trafo}
    {get the parameter for the tap limit, given in relative transformation ratio}
    if      (pComp^.hint = 'Transformator'    ) then pComp^.GetPar(gYY_tmin,apu)
    else if (pComp^.hint = 'Transformator Y D') then pComp^.GetPar(gYD_tmin,apu)
    else if (pComp^.hint = 'Transformator D Y') then pComp^.GetPar(gDY_tmin,apu)
    else if (pComp^.hint = 'Transformator D D') then pComp^.GetPar(gDD_tmin,apu)
    else apu := 0.85;
    {change to +-fraction of sec. voltage, check also that limit reasonable}
    if (apu > 1.50) then pdTap^.x := 1/1.5-1      {min.parameter becomes max limit}
    else if (apu > 0.5) then pdTap^.x:= 1/apu-1
    else pdTap^.x:= 1/0.5-1;
    {then the max parameter becomes the min limit}
    if      (pComp^.hint = 'Transformator'    ) then pComp^.GetPar(gYY_tmax,apu)
    else if (pComp^.hint = 'Transformator Y D') then pComp^.GetPar(gYD_tmax,apu)
    else if (pComp^.hint = 'Transformator D Y') then pComp^.GetPar(gDY_tmax,apu)
    else if (pComp^.hint = 'Transformator D D') then pComp^.GetPar(gDD_tmax,apu)
    else apu := 0.85;
    {change to +-fraction of sec. voltage, check also that limit reasonable}
    if (apu > 1.50) then pdTap^.y := 1/1.5-1      {max. parameter becomes min limit}
    else if (apu > 0.5) then pdTap^.y:= 1/apu-1
    else pdTap^.y:= 1/0.5-1;
    {next trafo}
    inc(pdTap);
  end;
end;

procedure TTrafoCalc.ReadSecondaryVoltages;
{get the secondary system voltages of the trafos
 if the system voltage is not given, use the secondary voltage}
var
  ii    : integer; {index}
  pComp : TpComp;  {pointer to the trafo}
  pV2   : pSingle; {pointer to the voltages}
  V2    : single;  {the secondary voltage}
  Vs2   : single;  {secondary system voltage}
begin
  pV2 := pV2_t;
  for ii := 0 to nTrafo_t-1 do begin
    pComp := TrafoList_t.items[ii];      {pointer to a trafo}
    pComp^.GetPar(gTrafo_V2,V2);         {secondary voltage}
    if (V2 < 1) then V2 := 1 else V2 := V2/sqrt(3);
    {get the secondary system voltage}
    if      (pComp^.hint = 'Transformator'    ) then pComp^.GetPar(gYY_Vs2,Vs2)
    else if (pComp^.hint = 'Transformator Y D') then pComp^.GetPar(gYD_Vs2,Vs2)
    else if (pComp^.hint = 'Transformator D Y') then pComp^.GetPar(gDY_Vs2,Vs2)
    else if (pComp^.hint = 'Transformator D D') then pComp^.GetPar(gDD_Vs2,Vs2)
    else Vs2 := 1;    {error case}
    if (Vs2 <= 1) then Vs2 := V2 else Vs2 := Vs2/sqrt(3);  {use secondary volt. if system volt. not given}
    pV2^:= Vs2;
    inc(pV2);
  end;
end;

procedure TTrafoCalc.CheckSmallDiagonalElements;
{a small derivative dV(i)/dTap(i) => tap has little effect => nearly singular
 => set flag}
var
  apu    : single;      {temporary}
  cpu    : complex;     {temporary}
  dVdT   : single;      {derivative of voltage wrt tap position}
  itra   : integer;     {index for trafo}
  nn     : integer;     {number of trafos}
  pSmall : pBoolean;    {local pointer to flag for small derivative}
  pVolt0 : pComplex;    {pointer to secondary voltage, when tap = 0}

begin
  nn := nTrafo_t;                         {nuber of trafos}
  pVolt0 := pVolt0_t;                     {pointer to sec. voltages when tap = 0}
  pSmall := pSmall_t;                     {copy to local}
  for itra := 1 to nn do begin
    GetCMat(pdVdT_t,itra,itra,nn,nn,cpu);   {diagonal elements}
    dVdT := cpu.y;
    if (pVolt0^.y > 1) then begin           {reasonable secondary voltage}
      apu := dVdT/pVolt0^.y;                {relative derivative}
      if (apu < 0.05) then pSmall^:= true {small diagonal element}
      else pSmall^:= false;
    end else begin                        {small voltage}
      pSmall^:= true;
    end;
    inc(pVolt0);
    inc(pSmall);
  end;
end;

procedure TTrafoCalc.NewTapPositions(var avtap : single; var NewLim : boolean);
{add the changes in the tap positions, calculate average change (for convergence),
 check tap position against limits
 avtap    average change in the tap positions
 NewLim   set flag, if the tap of a new trafo is at limit}
var
  apu      : single;     {temporary}
  apv,apw  : single;     {temporary}
  ii       : integer;    {index}
  pAtLimit : pInteger;   {pointer to flags for tap at upper (1) or lower (-1) boundary}
  pComp    : TpComp;     {pointer to a trafo}
  pdTap    : pComplex;   {pointer to the tap limit}
  pvar     : pSingle;    {pointer to a variable (on the right hand side)}
begin
  {new tap positions}
  pAtLimit := pAtLImit_t;                      {pointer to flag for tap at limit}
  pdTap := pdTap_t;                            {pointer to tap limit}
  pvar := pbb_t;                               {initialize for readingthe rhs}
  avtap := 0;                                  {initialize average change in tap pos.}
  NewLim := false;
  for ii := 0 to nTrafo_t-1 do begin
    pComp := TrafoList_t.items[ii];            {pointer to a trafo}
    pComp^.GetZpar(apu,apv,apw);               {apu = old tap position, others not needed}
    apu := apu+pvar^;                          {new tap position}
    {check against limits}
    if (apu > pdTap^.x) then begin
      pComp^.PutZpar(pdTap^.x,0,0,1);          {too large}
      if (pAtLimit^ <> 1) then begin           {was not at limit}
        pAtLimit^:= 1;                         {set flag for upper limit}
        NewLim := true;                        {set flag for new tap at limit}
      end;
    end else if (apu > pdTap^.y) then begin
      pComp^.PutZpar(apu,0,0,1);               {set new tap position}
    end else begin                             {too small}
      pComp^.PutZpar(pdTap^.y,0,0,1);
      if (pAtLimit^<> -1) then begin           {was not at limit}
        pAtLimit^ := -1;                       {set flag for lower limit}
        NewLim := true;                        {set flag: new trafo at limit}
      end;
    end;
    avtap := avtap+abs(pVar^)/nTrafo_t;        {average change in tap pos.}
    inc(pvar);
    inc(pdTap);
    inc(pAtLimit);
  end;
end;

function TTrafoCalc.SecondaryVoltage(ii : integer) : single;
{get the secondary voltage of transformer ii}
var
  pV2  : pSingle;   {pointer to the voltage}
begin
  if (ii > 0) and (ii <= nTrafo_t) then begin
    pV2 := pV2_t;
    inc(pV2,ii-1);
    SecondaryVoltage := pV2^;
  end else begin
    SecondaryVoltage := 1;       {just some value}
  end;
end;

procedure TTrafoCalc.ConstantTap(matr : integer);
{if there were small diagonal elements in the dV/dTap-matrix, the matrix
 is almost singular, but not if there is a parallel trafo.
 If tap is at limit, the tap must not change.
 In both cases, modify the matrix so that the tap position is not changed
 the matrix to be modified: 1 = paa_t, 2 = pgra_t}
var
  ii       : integer;    {index}
  itra     : integer;    {index for the trafo}
  nn       : integer;    {number of trafos}
  pAtLimit : pInteger;   {pointer to tap at limit}
  pParal   : pInteger;   {pointer to parallel trafo}
  pSmall   : pBoolean;   {pointer to flag for small diag.element}
  pmat     : psingle;    {pointer to the matrix to be modified}
begin
  nn := nTrafo_t;      {copy no of trafos to local}
  pSmall := pSmall_t;
  pParal := pParal_t;
  pAtLimit := pAtLimit_t;
  if (matr = 1) then pmat := paa_t else pmat := pgra_t;
  for itra := 1 to nn do begin
    {tap limits neglected for pgra_t}
    if ((pSmall^) and (pParal^ = 0)) or ((pAtLimit^ <> 0) and (matr = 1)) then begin
      for ii := 1 to nn do begin
        PutMat(pmat,itra,ii,nn,nn,0);     {zeroes to the row...}
        putMat(pmat,ii,itra,nn,nn,0);     {..and column of the small element...}
      end;
      PutMat(pmat,itra,itra,nn,nn,1);     {..but one to the diagonal}
    end;
    inc(pSmall);
    inc(pAtLimit);
    inc(pParal);
  end;
end;

procedure TTrafoCalc.Decompose;
{decompose the coefficient matrix paa_t}
var
  Dodd    : single;        {permutations flag}
  sngl    : integer;       {singularity flag}
  Tout    : single;        {test output}
begin
  LUdecomp(paa_t ,pww_t, nTrafo_t, pIndx_t, Dodd, Sngl, Tout);
end;

procedure TTrafoCalc.DecompGra;
{decompose the coefficient matrix pGra_t}
var
  Dodd    : single;        {permutations flag}
  sngl    : integer;       {singularity flag}
  Tout    : single;        {test output}
begin
  LUdecomp(pgra_t ,pww_t, nTrafo_t, pGrIndx_t, Dodd, Sngl, Tout);
end;

procedure TTrafoCalc.ConstructRhs(var averr : single);
{construct the right hand side, and calculate the average relative error
 pbb_t contains the solution for tha change in the tap position
 averr   the average, relative error}
var
  cpu      : complex;   {temporary}
  ii       : integer;   {index}
  pAtLimit : pInteger;  {flag for tap at limit}
  pComp    : TpComp;    {pointer to a trafo}
  pSmall   : pBoolean;  {pointer to flags for small (diagonal) elements}
  pvar     : pSingle;   {pointer to a variable}
  Volt     : single;    {magnitude of calculated secondary voltage}
  Vs2      : single;    {nominal secondary voltage}
begin
  averr := 0;                           {initialize average error}
  {check secondary voltages, adjust tap, and calculate once more}
  pvar := pbb_t;                        {initialize for constructing the rhs}
  pSmall := pSmall_t;                   {flag for small elements}
  pAtLimit := pAtLimit_t;               {flag for tap at limit}
  for ii := 0 to nTrafo_t-1 do begin
    pComp := TrafoList_t.items[ii];     {pointer to a trafo}
    pComp^.GetVar(gTrafo_V2r,cpu.x);    {real part of secondary voltage}
    pComp^.GetVar(gTrafo_V2i,cpu.y);    {imag part of secondary voltage}
    Volt := Cabs(cpu);                  {abs.value of voltage must be used}
    {get the nominal secondary voltage}
    Vs2 := SecondaryVoltage(ii+1);
    if (pSmall^) or (pATLimit^ <> 0) then
      pVar^:= 0                  {no changes in this tap, if small dV/dtap, or at limit}
    else
      pvar^ := Vs2-Volt;         {construct the rhs}
    averr := averr+abs(Vs2-Volt)/(Vs2*nTrafo_t);     {average error}
    inc(pvar);
    inc(pSmall);
    inc(pAtLimit);
  end;
end;

procedure TTrafoCalc.Backsub;
begin
  LUbacksub(paa_t, nTrafo_t, pIndx_t, pbb_t);
end;

procedure TTrafoCalc.BacksubGra;
{find the solution without the limits = find "gradient"}
begin
  CopyMat(pbb_t,pgrb_t,nTrafo_t,1);   {copy rhs}
  LUbacksub(pgra_t, nTrafo_t, pGrIndx_t, pgrb_t);
end;

procedure TTrafoCalc.RemoveConstraintsIf(var rem : boolean);
{if a tap is at limit, and is moving towards feasible region, the constraint
 (tap on limit) can be removed
 rem   flag that a constraing has been removed}
var
  ii,jj    : integer;   {index}
  pAtLimit : pInteger;  {pointer to  flags for tap at limit}
  pAtLim   : pinteger;
  pgrb     : pSingle;   {pointer to rhs without limits}
  pParal   : pInteger;  {pointer to parallel trafos}
  pPar     : pInteger;
begin
  pgrb := pgrb_t;
  pAtLimit := pAtLimit_t;
  pParal := pParal_t;
  rem := false;
  for ii := 1 to nTrafo_t do begin
    if (((pAtLimit^ > 0) and (pgrb^ < -0.001)) or    {at upper limit, moving down}
        ((pAtLimit^ < 0) and (pgrb^ >  0.001))) and  {at lower limit, moving up}
        (pParal^= 0) then begin                      {and not a "sub-parallel" trafo}
       pAtLimit^:= 0;       {remove "at limit"-flag}
       rem := true;
       {remove constraint also for eventual "sub-parallel" trafos}
       pAtLim := pAtLimit_t;
       pPar := pParal_t;
       for jj := 1 to nTrafo_t do begin
         if (pPar^= ii) then pAtLim^:= 0;  {remove constr. for a parallel trafo}
         inc(pAtLim);
         inc(pPar);
       end;
    end;
    inc(pgrb);
    inc(pAtLimit);
    inc(pParal);
  end;
end;

procedure ShowTap(pComp : TpComp);
{show the tap position in the image attached to the trafo
 (called also in FormDeactivate of parameter.pas)
 pComp  pointer to the component}
var
  apu,apv : single;
  ipu     : integer; {temporary}
  tap     : single;  {tap position, in fractions (around zero)}
  txt     : string;  {temporary text}
  x1,x2   : integer; {x-coord. of nodes}
  y1,y2   : integer; {y-coord. of nodes}
begin
  if (pComp^.pImg_s <> nil) and (isTrafo(pComp)) then begin
    pComp^.GetZpar(tap,apu,apv);                  {get tap position}
    {tap displayed either as % change in sec. voltage, or relative value in prim.volt}
    if (Imagoo.TapItem.checked) then begin
      txt := TwoDecimalsAtMost(100*tap)+'%';        {tap in percentage}
    end else begin
      txt := ThreeDecimals(1/(1+tap));              {tap as a relative prim.voltage}
    end;
//    pComp^.pImg_s^.picture := nil;                {image cleared in AmpageLimits, of elplex.pas}
    if (Imagoo.TapItem.checked) then begin
      pComp^.pImg_s^.canvas.font.color := clFuchsia;
    end else begin
      pComp^.pImg_s^.canvas.font.color := clBlack;
    end;
    {if horizontal trafo, and no ampage limit displayed, write tap near to the trafo}
    pComp^.GetNxy(1,x1,y1);
    pComp^.GetNxy(2,x2,y2);
    if (y1 = y2) then begin      {horizontal}
      {.tag contains the flag for the ampage limit, set in AmpageLimits of elplex.pas}
      if (pComp^.pImg_s^.tag > 0) and (Imagoo.AmpageItem.checked) then
        ipu := 0 else ipu := pComp^.pImg_s^.height div 2;
    end else begin               {vertical}
      ipu := 0;
    end;
    {clear the tap-part}
    with pComp^.pImg_s^.canvas do begin
      pen.color := clWhite;
      pen.width := 1;
      brush.color := clWhite;
      brush.style := bsSolid;
      rectangle(0,ipu,pComp^.pImg_s^.width,ipu+pComp^.pImg_s^.height div 2);
    end;
    {write the tap position}
    pComp^.pImg_s^.canvas.font.name := 'symbol';
    pComp^.pImg_s^.canvas.font.size := 8;
    pComp^.pImg_s^.canvas.textout(0,{0}ipu,txt);
    pComp^.pImg_s^.visible := true;
  end;
end;

procedure UpdateTapDisplay(pSet : TpCompSet);
{update the display of the tap position after a change in the options}
var
  errc  : integer;   {error flag}
  jcomp : integer;  {counts components}
  pComp : TpComp;   {pointer to a trafo}
begin
  errc := 0;                         {initialize error flag}
  pComp := nil;
  jcomp := 0;
  {scan all components to find the transformers}
  while (errc = 0) do begin                 {to last component, or first error}
    pSet^.NextComp(pComp, errc,jcomp,true); {pointer to next component}
    if (errc = 0) then begin                {not error, but maybe last comp.}
     {is a trafo, but not a 3-winding?}
      if (isTrafo(pComp)) and not (pComp^.hint = gTrafo3W) then begin
        ShowTap(pComp);                     {display the tap position }
      end;
    end;
  end;
end;

procedure LoadTapCalculation(ipSet, calc: longint;  var success: boolean);
{solve the load flow problem with tap changers
 ipSet    pointer to the "common" as integer
 calc     1 = load flow, 2 = adjust voltages
 success  flag for a successful calculation}
const
  epsilon = 0.005; {convergence criteria for average relative error }
  miter = 8;       {max. number of iterations}
  tapeps = 0.0005; {convergence criteria for average change in tap position}
var
  averr    : single;        {average error}
  avtap    : single;        {average change in tap position}
  converg  : boolean;       {true when converged}
  Dodd     : single;        {permutations flag}
  ii,iter  : integer;       {index}
  NewLimit : boolean;       {tap of a new trafo at limit}
  nn       : integer;       {dim of matrices = no of trafos}
  pComp    : TpComp;        {pointer to a trafo}
  removed  : boolean;       {constraing removed}
  sngl     : integer;       {singularity flag}
  Tout     : single;        {test output}
  Trafo    : TTrafoCalc;    {object for the calculations}
begin

  Trafo := TTrafoCalc.create;        {create object for the calculations}
  Trafo.initialize(ipSet);           {reserve memory, etc}
  if (Trafo.nTrafo_t > 0) then begin {there are transformers}
    nn := Trafo.nTrafo_t;            {dim of matrices}
    Trafo.SolveNet;                  {solve net with taps at 0}
    Unconnected(Trafo.pSet);         {determine unconnected loads}
    Trafo.dVdTap;                    {calculate derivatives}
    Trafo.GetCalcMemory;             {get memory for the calculation}
    Trafo.CheckSmallDiagonalElements;   {small diagonal elements dV/dTap inicate singularity}
    Trafo.FindParallelTrafos;
    Trafo.FinalizeAlmost;            {free memory, except for the derivatives and list}
    Trafo.ReadTapLimits;             {read the tap limit parameters}
    Trafo.ReadSecondaryVoltages;     {read the secondary system voltages}
    {"unlimited" coeff. matrix is needed to check the direction of the tap change
     This is needed to see, if a tap at a limit is moving back to allowed direction}
    imgMat(Trafo.pdVdT_t,Trafo.pgra_t,nn);
    {if there were small elements, or tap at limit, the tap must not be changed}
    Trafo.ConstantTap(2);
    Trafo.ModifyMatrForParallelTrafos(2);  {same tap for parallel trafos}
    Trafo.DecompGra;
    {set all taps initially to zero}
    for ii := 0 to Trafo.nTrafo_t-1 do begin
      pComp := Trafo.TrafoList_t.items[ii];      {pointer to a trafo}
      pComp^.PutZpar(0,0,0,1);             {zero tap position}
    end;
    if (calc = 1) then LoadFlowCalculation(ipSet,TapLoad, success)  {load flow}
    else SolveNetWithGivenOutputVoltages(ipSet,TapVolt);            {adjust voltage}
    converg := false;                       {not yet converged}
    NewLimit := false;
    for iter := 1 to miter do begin
      Trafo.ConstructRhs(averr);             {rhs for the change in tap pos.}
      Trafo.ModifyRhsForParallelTrafos(iter);
{showRmat1(Trafo.pbb_t,nn,1);
showmessage(' ');}
      Trafo.BacksubGra;                      {solve without limits}
      Trafo.RemoveConstraintsIf(removed);    {moving away from limits?}
      {coefficient matrix at first iteration, or if new tap at limit}
      if (iter = 1) or (NewLimit) or (removed) then begin
        {decompose the derivative matrix, copy first}
        imgMat(Trafo.pdVdT_t,Trafo.paa_t,nn);
        {if there were small elements, or tap at limit, the tap must not be changed}
        Trafo.ConstantTap(1);
        Trafo.ModifyMatrForParallelTrafos(1);  {(nearly) same tap for parallel trafos}
{if (iter = 1) then showRmat1(Trafo.paa_t,nn,nn);}
        Trafo.decompose;
      end;
      {the actual solution}
      Trafo.Backsub;                         {solve the change in tap pos.}
      Trafo.NewTapPositions(avtap,NewLimit); {add the change to the old tap position, check}
      if (calc = 1) then LoadFlowCalculation(ipSet,TapLoad, success)  {try once more}
      else SolveNetWithGivenOutputVoltages(ipSet,TapVolt);
      if (averr < epsilon) or (avtap < tapeps) then begin     {converged}
        converg := true;
        break;              {converged}
      end;
    end;
    {show tap positions near to the trafos}
    for ii := 0 to Trafo.nTrafo_t-1 do begin
      pComp := Trafo.TrafoList_t.items[ii];      {pointer to a trafo}
      ShowTap(pComp);                            {display the tap value}
    end;
    {tell, if not converged}
    if (not converg) then begin
      show_message('not converged');
    end;
    Trafo.FreeCalcMemory;
  end else begin             {no trafos, normal load flow or adjust volt. calculation}
    if (calc = 1) then LoadFlowCalculation(ipSet,SolvingLoadFlow, success)  {try once more}
    else SolveNetWithGivenOutputVoltages(ipSet,AdjustVolt);
    {SolvingLoadFlor, AdjustVolt are of type "Solve", defined in elplex.pas}
  end;
  {free the remaining reserved memory}
  FreeMem(Trafo.pdVdT_t,Trafo.nTrafo_t*Trafo.nTrafo_t*sizeOf(single));
  FreeMem(Trafo.pVolt0_t,Trafo.nTrafo_t*sizeOf(complex));
  Trafo.TrafoList_t.free;
  Trafo.free;
end;

end.
