unit TimeDiff;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Skemadef, routines, PQrelay, imagodef, Recloser, emtp, Userdef,
  utildef, shortcir, Buttons, skemproc;

type
  TTimeDiffForm = class(TForm)
    CurveALabel: TLabel;
    CurveAEdit: TEdit;
    TimeAEdit: TEdit;
    SALabel: TLabel;
    CurveBLabel: TLabel;
    CurveBEdit: TEdit;
    TimeBEdit: TEdit;
    SBLabel: TLabel;
    CurrentLabel: TLabel;
    CurrentEdit: TEdit;
    DeltaLabel: TLabel;
    TimeLabel: TLabel;
    TimeDiffEdit: TEdit;
    SDtLabel: TLabel;
    AmpLabel: TLabel;
    GraphCurrEdit: TEdit;
    KeyBut: TRadioButton;
    GraphBut: TRadioButton;
    HelpMemo: TMemo;
    ClearBut: TButton;
    HelpBut: TSpeedButton;
    CloseBut: TButton;
    HintAEdit: TEdit;
    HintBEdit: TEdit;
    FromLabel: TLabel;
    procedure FormHide(Sender: TObject);
    procedure CurveAEditChange(Sender: TObject);
    procedure RelayFromIndex(line: string; var pComp: TpComp; var rr,curve: integer);
    procedure CurveBEditChange(Sender: TObject);
    procedure TimeDifference;
    procedure CurrentEditChange(Sender: TObject);
    procedure GraphCurrEditChange(Sender: TObject);
    procedure KeyButClick(Sender: TObject);
    procedure GraphButClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ClearCompPointers;
    procedure RemoveRelay(pRelay: TpComp);
    procedure UpdateRelayIndex;
    procedure CollectModulesIf(pRelay: TpComp; AorB : char);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure RemoveModule(pRelay: TpComp; AorB: char);
    procedure HelpButClick(Sender: TObject);
    procedure ClearButClick(Sender: TObject);
    procedure CloseButClick(Sender: TObject);
    procedure GetModuleIndices(var rrA, rrB: integer);
    procedure GetRelayPointers(var pCompA, pCompB: TpComp);
  private
    { Private declarations }
    AList_t  : TList;          {list of simple relays connected to the same current trafo}
    BList_t  : TList;
    pCompA_t : TpComp;         {pointer to the first relay}
    pCompB_t : TpComp;         {pointer to the second relay}
    rrA_t    : integer;        {index of the module of the first relay}
    rrB_t    : integer;        {index of the module of the second relay}
    procedure WithITorDT(name: string; pComp: TpComp; rr: integer; var withIT, withDT: boolean);
  public
    { Public declarations }
  end;

var
  TimeDiffForm: TTimeDiffForm;

implementation

uses Relcurr, Units, Imagopas;

{$R *.DFM}

const
  TDWidth = 280;  {nominal width of the form, without the help memo}

procedure TTimeDiffForm.FormHide(Sender: TObject);
begin
  RelayCurr.TimeDiffItem.checked := false;
end;

procedure TTimeDiffForm.ClearCompPointers;
{needed in imagopas.pas}
begin
  pCompA_t := nil;
  pCompB_t := nil;
  rrA_t := 0;
  rrB_t := 0;
end;

procedure TTimeDiffForm.RemoveModule(pRelay: TpComp; AorB: char);
{check and remove a relay module from the calculations (when deleted)
 pRelay  pointer to the relay
 AorB    flag to the relay}
var
  ii    : integer;
  nr    : integer;   {number of relays}
  pComp : TpComp;    {pointer to relay A or B}
  pList : TpList;    {pointer to the list of modules of the relay}
begin
  if (AorB = 'A') then begin
    pList := @AList_t;
    pComp := pCompA_t;
  end else begin
    pList := @BList_t;
    pComp := pCompB_t;
  end;
  nr := pList^.count;             {number of modules}
  if (nr > 0) then begin          {there are modules}
    for ii := 0 to nr-1 do begin
      if (pRelay = pList^.items[ii]) then begin   {is a module in the list}
        pList^.delete(ii);                       {remove from the list}
        pList^.pack;
        {must still check the relay itself}
        if (pRelay = pComp) then begin
          {remove it, but replace with another from the list, if there is any}
{TM ei toimi !! ???}
          if (nr > 1) then pComp := Alist_t.items[0]   {just any from the list}
          else pComp := nil;
        end;
        break;
      end;
    end;
  end;
end;

procedure TTimeDiffForm.RemoveRelay(pRelay: TpComp);
{remove a relay from the calculations (when deleted)
 pRelay  pointer to the relay}
begin
  {check the "simple" relay modules first}
  if (isRelayModule(pRelay)) then begin   {in shortcir.pas}
    RemoveModule(pRelay, 'A');
    RemoveModule(pRelay, 'B');
  end else begin                    {some other component}
    if (pRelay = pCompA_t) then begin
      pCompA_t := nil;
      rrA_t := 0;
      CurveAEdit.text := '';
    end;
    if (pRelay = pCompB_t) then begin
      pCompB_t := nil;
      rrB_t := 0;
      CurveBEdit.text := '';
    end;
  end;
end;

procedure TTimeDiffForm.RelayFromIndex(line: string; var pComp: TpComp; var rr, curve: integer);
{get the relay pointer and module index from the index of the curve, given as a string
 line   the index of the curve, as a string
 pComp  pointer to the relay
 rr     index of the relay module
 curve  index of the curve, from line}
var
  ii    : integer;   {index}
  isNr  : boolean;   {character is a number}
  len   : integer;   {length of line }
  ll    : longint;   {temporary}
  nRel  : integer;   {dimension of arrays = max. no of curves-1}
  mm    : integer;   {index of the module}
  txt   : string;    {temporary}
begin
  nRel := RelayCharact.GetNComp;   {dimension of arrays}
  len := length(line);
  {continue only, if there is some text}
  if (len > 0) then begin          {there is some text}
    if (line[1] >= '0') and (line[1] <= '9') then begin
      isNr := true;                  {initialize}
      {at most two numbers accepted, checked}
      if (len >= 2) then begin
        if (line[2] >= '0') and (line[2] <= '9') then begin
          txt := line[1]+line[2];
        end else begin               {some other characted, discarded}
          txt := line[1];
        end;
      end else begin                 {only one char = number}
        txt := line[1];
      end;
    end else begin                   {the first character is not a number}
      isNr := false;
      txt := '';
    end;
    {continue only, if an integer found}
    if (isNr) then begin
      curve := strtoint(txt);        {to integer = index of the curve}
      {continue only, if reasonable index}
      if (curve > 0) and (curve <= nRel+1) then begin  {array index goes 0..nRel}
        ll := RelayCharact.CurveDat[curve-1].ipComp;   {pointer to the relay, as integer}
        mm := RelayCharact.CurveDat[curve-1].rr;       {index of the module}
        {continue only, if reasonable pointer and index}
        if (ll <> 0) and (mm > 0) then begin
          pComp := TpComp(ll);
          rr := mm;              {check that reasonable, <= 3 or <= 6 ??}
        end else begin
          pComp := nil;
          rr := 0;
        end;
      end else begin
        pComp := nil;
        rr := 0;
        curve := 0;  
      end;
    end else begin
      pComp := nil;
      rr := 0;
      curve := 0;
    end;
  end else begin    {no identifier given}
    pComp := nil;
    rr := 0;
  end;
end;

procedure TTimeDiffForm.UpdateRelayIndex;
{when a new relay is added to the graph (coordination graphics), the indices
 of the relays change. Update the index in the edit
 Used in TImageExt.UpdateRelayLists of relcurr.pas}
const
  cnt : integer = 0;
var
  ii : integer;
  ll : longint;
  nr : integer;
begin
  nr := high(RelayCharact.CurveDat);
  for ii := 0 to nr do begin                     {scan the whole list}
    if (pCompA_t <> nil) then begin
      ll := longint(pCompA_t);                   {pointer as integer}
      if (ll = RelayCharact.CurveDat[ii].ipComp) and
         (rrA_t = RelayCharact.CurveDat[ii].rr)  then begin {relay found in the list}
        CurveAEdit.text := inttostr(ii+1);
      end;
    end;
    {the same for the second relay}
    if (pCompB_t <> nil) then begin
      ll := longint(pCompB_t);                   {pointer as integer}
      if (ll = RelayCharact.CurveDat[ii].ipComp) and
         (rrB_t = RelayCharact.CurveDat[ii].rr)  then begin {relay found in the list}
        CurveBEdit.text := inttostr(ii+1);
      end;
    end;
  end;
end;

function ColorOfCurve(pComp: TpComp; rr: integer) : TColor;
{find the color of the curve in the graph}
var
  apu  : single;
  ic   : integer;
  ll   : longint;
  name : string;
  pCT  : TpComp;   {pointer to the current trafo connected to the relay}
begin
  {set color of text = same as in graph, differently for simple relays and combined rel.}
  if (isRelayModule(pComp)) then begin  {simple relay (module)}
    pComp^.GetipComp(ll);
    if (ll <> 0) then begin
      pCT := TpComp(ll);         {pointer to the current trafo connected to the comp}
      ic := pCT^.int5_s;         {index for color, set when CT created, in skemproc.pas}
      if (ic >= 0) and (ic <= high(RelayCharact.ColorArr)) then
        ColorOfCurve := RelayCharact.ColorArr[ic]
      else
        ColorOfCurve := clBlack;
    end else begin               {no current trafo, something wrong}
      ColorOfCurve := clBlue;
    end;
  end else begin
    pComp^.GetName(name);
    if (name = gPQrel) or (name = gEMrel) then pComp^.GetPar(gPQ_color[rr],apu)
    else if (name = gRCrel) or (name = gUDrel) then pComp^.GetPar(gRC_color[rr],apu)
    else apu := 1;
    ic := round(apu);
    if (ic >= Low(gColor)) and (ic <= High(gColor)) then
      ColorOfCurve := gColor[ic]
    else ColorOfCurve := clRed;
  end;
end;

procedure TTimeDiffForm.CollectModulesIf(pRelay: TpComp; AorB : char);
{if the relay is a simple relay, consisting of discrete modules, must
 collect the modules that belong together = connected to the same node, or same trafo
 pRelay   pointer to the relay
 AorB     flag of the relay}
var
  errc  : integer;   {error flag}
  errn  : integer;   {error flag}
  jcomp : integer;   {counts components}
  kk    : integer;   {temporary (node)}
  ll    : longint;   {temporary}
  pComp : TpComp;    {pointer to a component}
  pList : TpList;    {pointer to the list of modules}
  pSet  : TpCompSet; {pointer to the common}
  rnode : integer;   {node of relay}
begin
  if (pRelay <> nil) then begin
    pRelay^.GetipSet(ll);                          {pointer to common as integer}
    pSet := TpCompSet(ll);                         {pointer to the common}
    if (isRelayModule(pRelay)) then begin          {a simple relay? in shortcir.pas}
      {find relays connected to this relay, add to the list or either relay A or B}
      if (AorB = 'A') then pList := @AList_t else pList := @BList_t;
      pList^.clear;  {clear initially}
      {scan all components, check type, connection}
      pRelay^.getNode(1,rnode,errn);               {get index of node}
      pComp := nil;
      errc := 0;
      jcomp := 0;
      while (errc = 0) do begin                  {scan all components}
        pSet^.NextComp(pComp,errc,jcomp,true);   {get pointer to next component}
        if (errc = 0) then begin                 {component found}
          if (isRelayModule(pComp)) then begin   {a simple relay module? in shortcir.pas}
            pComp^.getNode(1,kk,errn);           {get node}
            if (kk=rnode) then begin             {same node ?}
              pList^.add(pComp);                 {connected to the relay}
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure TTimeDiffForm.CurveAEditChange(Sender: TObject);
{get the pointer to the relay of the curve, and the index of the relay module}
var
//ix,iy: integer;
//lbl: string;
  curve : integer;   {index of the curve, from text}
  pComp : TpComp;    {pointer to the relay}
  rr    : integer;   {index of relay module}
begin
  RelayFromIndex(CurveAEdit.text, pComp, rr,curve);
  {continue, if reasonable relay and index}
  if (pComp <> nil) and (rr > 0) then begin
//pComp^.GetLabel(lbl,ix,iy);
//showmessage(lbl);
    pCompA_t := pComp;
    rrA_t := rr;
    {for simple relays, must collect modules}
    CollectModulesIf(pComp,'A');
    CurveALabel.font.color := ColorOfCurve(pCompA_t,rrA_t);
    {calculate a new time difference}
    TimeDifference;
  end;
end;

procedure TTimeDiffForm.CurveBEditChange(Sender: TObject);
var
//ix,iy: integer;
//lbl: string;
  curve : integer;   {index of the curve, from text}
  pComp : TpComp;    {pointer to the relay}
  rr    : integer;   {index of relay module}
begin
  RelayFromIndex(CurveBEdit.text, pComp, rr,curve);
  {continue, if reasonable relay and index}
  if (pComp <> nil) and (rr > 0) then begin
    pCompB_t := pComp;
    rrB_t := rr;
    {for simple relays, must collect modules}
    CollectModulesIf(pComp,'B');
    {set color of text = same as in graph}
    CurveBLabel.font.color := ColorOfCurve(pCompB_t,rrB_t);
    {calculate a new time difference}
    TimeDifference;
  end;
end;

procedure VoltageLevel(pComp: TpComp; var Volt : single);
{get the voltage level of the relay by reading the voltage level of
 the connected current trafo
 pComp   pointer to the relay
 Volt    the voltage level, = 0 if not defined}
var
  ll    : longint;   {temporary}
  pCurr : TpComp;    {pointer to the relay}
begin
  {voltage level of the relay, for transforming the current}
  pComp^.GetipComp(ll);       {pointer to current trafo, as integer}
  if (ll <> 0) then begin
    pCurr := TpComp(ll);
    pCurr^.GetVoltLev(Volt);  {voltage level of the current trafo}
  end else begin
    Volt := 0;                {not defined, use 0 as a flag}
  end;
end;

procedure CoeffForPlane(pComp : TpComp; rr: integer; var coef : single);
{the value of the current must be multiplied by a coefficient so that the
 calculated time difference is the same as that one read rom the graph.
 The coefficient depends on the "plane": phase plane or neg.seq.plane
 In the phase plane:
   the current parameters of the P and G graphs are used as such,
   but the current parameters of the Q-graph are divided by sqrt(3) (moved to the left)
 In the neg.seq.plane
   the current parameters of the P and G graphs are multiplied by sqrt(3) (moved to the right)
   but the current parameters of the Q graph are used as such
 pComp  pointer to the relay
 rr     index for the module type  }

var
  name : string;       {name = type of the relay}
  pgq  : integer;      {index of the module type 1 = P, 2 = G, 3 = Q}
begin
  {the module index determines the module type, but differently for different relays}
  pComp^.GetName(name);
  if (name = gPQrel) or (name = gEMrel) then begin          {PGQ-relay, or electromech.}
    pgq := rr;                                              {the index gives directly the type}
  end else if (name = gRCrel) or (name = gUDrel) then begin {recloser or user defined}
    pgq := (rr+1) div 2;                                    {1,2 = P, 3,4 = G, 5,6 = Q}
  end else begin                                            {simple relay}
    pgq := 1;                                               {only P}
  end;
  {determine the coefficient}
  if (RelayCurr.PhaseItem_rc.checked) then begin            {phase plane}
    if (pgq = 1) then coef := 1
    else if (pgq = 2) then coef := 1
    else coef := sqrt(3);                                         {to compensate}
  end else if (RelayCurr.NegSeqItem_rc.checked) then begin  {neg.seq. plane}
    if (pgq = 1) then coef := 1/sqrt(3)
    else if (pgq = 2) then coef := 1/sqrt(3)
    else coef := 1;
  end else begin
    coef := 1;
  end;
end;

procedure TimeOfSimpleRelay(pRelay: TpComp; pList: TpList; Voltlev: single; curr: single;
          var actime: single);
{get the activation time for the given current for a "simple" relay consisting of
 one or more modules
 pRelay   pointer to the relay (one of the modules)
 pList    pointer to list of modules
 VoltLev  voltage level of the coord.graph
 curr     the current for which time calculated
 actime   the activation time}
const
  mdef = 20; {max.number of def.time modules in a simple relay. Large enough?}
var
  alf,bet : single;     {parameters of the inv.time module, for different types}
  apu     : single;     {temporary}
  cFirst  : single;     {first current in the table of the UD-relay (not used)}
  curve   : integer;    {inv.time curve within a group (not used)}
  fact    : single;     {pickup factor for inv.time relay (not used)}
  group   : integer;    {index of the group of curves in use (not used)}
  ii      : integer;    {index}
  Iss     : single;     {current param.of the inverse time module}
  kk      : integer;
  ktm     : single;     {time multiplier of inv.time relay}
  mname   : string;     {name = type of a relay module}
  na      : integer;    {number of modules in the simple relay}
  ns      : integer;    {number of modules in calculations}
  pComp   : TpComp;     {pointer to a module}
  planecf : single;     {coefficient for the "plane"}
  relcur  : single;     {relative current, not used}
  Volt    : single;     {voltage level of relay }
  Idefs   : array[1..mdef] of single;   {currents of def.time modules of simple relays}
  tdefs   : array[1..mdef] of single;   {times of def.time modules of simple relays}
begin
  VoltageLevel(pRelay, Volt);
  {clear parameters first}
  for ii := 1 to mdef do begin
    Idefs[ii] := 0;
    tdefs[ii] := 0;
  end;
  Iss := 0;
  ktm := 0;
  fact := 0;                        {these and the following are not needed}
  group := 0;
  curve := 0;
  alf := 0;
  bet := 0;
  cFirst := 1;
  na := pList^.count;                    {number of modules}
  if (na < mdef) then ns := na-1 else ns := mdef-1;   {at most mdef modules included}
  {get parameters of all modules}
  kk := 1;                               {index of def.time modules}
  for ii := 0 to ns do begin
    pComp := pList^.items[ii];          {pointer to a module}
    pComp^.GetName(mName);               {name = type of the module}
    if (mName = gRelay) then begin       {definite time module}
      pComp^.GetPar(gRelay_I,Idefs[kk]); {current-parameter}
      pComp^.GetPar(gRelay_t,tdefs[kk]); {time-parameter}
      inc(kk);
    end else begin                       {inverse time module (there should be only one)}
      pComp^.GetPar(gInv_I,Iss);         {get the relay starting current}
      pComp^.GetPar(gInv_k,ktm);         {get the time multiplier}
      InvTimePar(mName, alf,bet);        {get the parameters for this type of relay}
    end;
  end;
  {calculate the time at the given current}
  {take the voltage level into account}
  if (Volt > 0) and (VoltLev > 0) then apu := curr*VoltLev/Volt else apu := curr;
  {take the "plane" into account: phase plane, or neg.seq.plane}
  CoeffForPlane(pRelay,1,planecf);       {coefficient to compensate for the "plane"}
  TimeFromCurrent(pRelay,planecf*apu, Idefs,Tdefs, mdef,Iss, fact, ktm, alf,bet,0, cFirst,
                  group,curve, 1, true, true, actime, relcur);   {in PQRelay.pas}
end;

procedure TTimeDiffForm.WithITorDT(name: string; pComp: TpComp; rr: integer; var withIT, withDT: boolean);
{determines, if the inverse time and/or the definite time part of the char.curve is used
 name    name = type of the relay
 pComp   pointer to the component
 rr      index of module considered (1..3 or 1..6, depending on relay type
 withIT  inverse time part is used
 withDT  definite time part is used}
var
  apu : single;
begin
  if (name = gPQrel) or (name = gEMrel) then begin  {PGQ- or electromech.relay}
    if (PQForm.visible) then begin
      {parameters of this relay shown?}
      if (PQForm.tag = longint(pComp)) then begin
        withIT := PQRelay.GridCalc.RelayWithIT(rr);
        withDT := PQRelay.GridCalc.RelayWithDT(rr);
      end else begin                      {not this relay, must check parameters}
        pComp^.GetPar(gPQ_show[rr],apu);
        withIT := (apu = 1) or (apu = 2);
        withDT := (apu = 1) or (apu = 3);
      end;
    end else begin
      pComp^.GetPar(gPQ_show[rr],apu);
      withIT := (apu = 1) or (apu = 2);
      withDT := (apu = 1) or (apu = 3);
    end;
    {recloser or user def.relay}
  end else if (name = gRCrel) or (name = gUDrel) then begin
    if (RecloserForm.visible) then begin
      {parameters of this relay shown?}
      if (RecloserForm.tag = longint(pComp)) then begin
        withIT := Recloser.RCGridCalc.RelayWithIT(rr);
        withDT := Recloser.RCGridCalc.RelayWithDT(rr);
      end else begin                      {not this relay, must check parameters}
        pComp^.GetPar(gRC_show[rr],apu);
        withIT := (apu = 1) or (apu = 2);
        withDT := (apu = 1) or (apu = 3);
      end;
    end else begin
      pComp^.GetPar(gRC_show[rr],apu);
      withIT := (apu = 1) or (apu = 2);
      withDT := (apu = 1) or (apu = 3);
    end;
  end else begin
    withIT := true;
    withDT := true;
  end;
end;

procedure IndicateCurveParts(pEdit: TpEdit; withIT, withDT: boolean);
{show "warning"-texts for displayed curve parts
 pEdit   edit, where text
 withIT  inverse time part shown
 withDT  definite time part shown}
begin
  if (withIT and withDT) then pEdit^.text := 'complete curve'
  else if (withIT) then pEdit^.text := 'only inv. time'
  else if (withDT) then pEdit^.text := 'only def.time'
  else pEdit^.text := 'no curve';       {should not be possible}
end;

procedure TTimeDiffForm.TimeDifference;
{calculate the time difference for the given relays and the given current}
const
  ndef = 4;  {number of definite time modules}
var
  actime    : single;     {activation time of the first relay}
  apu       : single;
  bb,cc     : single;     {time multiplier, adder; or alpha, beta for an inv.time module}
  bctime    : single;     {activation time of the second relay}
  cFirst    : single;     {first current in the table of the UD-relay}
  combrel   : boolean;    {true for combined relays (PGQ, recloser, etc)}
  conA,conB : single;     {connection of the CT: wye or delta}
  curr      : single;     {current}
  currA     : single;     {current of relay A, when transformation ratio included}
  currB     : single;     {current of relay B, when transformation ratio included}
  currtxt   : string;     {the current as text (string)}
  curve     : integer;    {inv.time curve within a group}
  err       : integer;    {error flag}
  fact      : single;     {pickup factor for inv.time relay}
  group     : integer;    {index of the group of curves in use}
  IprA,IprB : single;     {primary current of current trafo}
  ipu,ipv   : integer;    {temporary}
  IscA,IscB : single;     {sec.current of the current trafo}
  Iss       : single;     {current param.of the inverse time module}
  ktm       : single;     {time multiplier of inv.time relay}
  modA,modB : single;     {the mode  0 = use the prim current = current as such
                                     1 = use the secondary current of the CT
                                     2 = use the relative secondary current = current/nominal current}
  nameA     : string;     {name = type of the relay}
  nameB     : string;
  planecf   : single;     {coefficient for the "plane"}
  plMult    : single;     {multiplier for the "plane"}
  ratA,ratB : single;     {transformation ratio of the CT}
  relcur    : single;     {relative current, not needed}
  tmin      : single;     {minimum time}
  txt       : string;     {temporary}
  VoltA     : single;     {voltage level of relay A}
  VoltB     : single;     {voltage level of relay B}
  Voltlev   : single;     {the voltage level of the graph = latest clicked relay}
  withIT    : boolean;    {inverse time part of curve shown}
  withDT    : boolean;    {definite time part of curve shown}
  Idef      : array[1..ndef] of single;   {currents of def.time modules}
  tdef      : array[1..ndef] of single;   {times of def.time modules}
begin
  RelayCharact.GetVoltLev(Voltlev);     {the voltage level of the graph}
  {relays and modules given ?}
  if (pCompA_t <> nil) and (pCompB_t <> nil) and
     (rrA_t > 0) and (rrB_t > 0) then begin
    {select the source of the current: keyboard or graph (mouse)}
    if (KeyBut.checked) then currtxt := CurrentEdit.text
    else currtxt := GraphCurrEdit.text;
    {check that current reasonable}
    chkfloat(currtxt,err,txt);
    {continue, if no errors}
    if (err = 0) then begin
      {take units into account}
      if (UnitsForm.KiloAmperBut.checked) then apu := 1000 else apu := 1;
      curr := apu*strtofloat(txt);    {current from string}
      {need object for calculations, different for different relays}
      {The first relay first. . . . . . . . . }
      {take the transformation ratio of the CT into account}
      GetModeAndCTmainPar(pCompA_t, conA, IprA, IscA, modA, ratA);    {in skemproc.pas}
      if (modA = 0) then currA := curr                 {primary current used}
      else if (modA = 1) then currA := curr/ratA       {secondary current used}
      else if (modA = 2) then currA := curr/(IscA*ratA)  {relative sec.current used}
      else currA := curr;                              {error case}
      {voltage level of the relay, for transforming the current}
      VoltageLevel(pCompA_t, VoltA);
      {some routines depend on the type of the relay}
      pCompA_t^.getName(nameA);
      {PGQ relay or electromech.relay, or recloser relay or user defined relay}
      combrel := (nameA = gPQrel) or (nameA = gEMrel) or
                 (nameA = gRCrel) or (nameA = gUDrel);
      {for UD-relays needed fist current}
      if (nameA = gUDrel) then begin                 {user defined recloser relay}
        ipu := (rrA_t +1) div 2;                     {relay module (phase, ground, seq.)}
        GetUDFirst(pCompA_t, ipu, ipv, cFirst, apu); {first entry in table}
      end else begin
        cFirst := 1;                                 {not needed, but give some value}
      end;
      {differently for combined and simple relays}
      if (combrel) then begin
        {get parameters of the first relay}
        ParametersFromComp(pCompA_t, rrA_t,Idef[1],tdef[1],Idef[2],tdef[2],
                           Idef[3],tdef[3],Idef[4],tdef[4],
                           Iss, fact, ktm, bb, cc,tmin, cFirst, group,curve);
        {get the activation time of the first relay}
        {take the voltage level into account}
        if (VoltA > 0) and (VoltLev > 0) then {apu}plMult := currA*VoltLev/VoltA else {apu}plMult := currA;
        {take the "plane" into account: phase plane, or neg.seq.plane}
        CoeffForPlane(pCompA_t,rrA_t,planecf);   {coefficient to compensate for the "plane"}
        {check, perhaps only inverse time part or def.time part of the curve visible}
        {if form visible, take parameters from form, else from component}
        WithITorDT(nameA, pCompA_t, rrA_t, withIT, withDT);
        TimeFromCurrent(pCompA_t,planecf*plMult{apu}, Idef,Tdef, ndef,Iss, fact, ktm, bb, cc, tmin,cFirst,
                        group,curve, rrA_t, withIT, withDT, actime, relcur);   {in PQRelay.pas}
        {indicate which part(s) of curve shown}
        IndicateCurveParts(@HintAEdit, withIT, withDT);
      end else begin                             {simple relay}
        TimeOfSimpleRelay(pCompA_t, @AList_t, Voltlev, currA, actime);
        HintBEdit.text := 'complete curve';
      end;
      {next, the second relay . . . . . . . . . }
      {take the transformation ratio of the CT into account}
      GetModeAndCTmainPar(pCompB_t, conB, IprB, IscB, modB, ratB);    {in skemproc.pas}
      if (modB = 0) then currB := curr                 {primary current used}
      else if (modB = 1) then currB := curr/ratB       {secondary current used}
      else if (modB = 2) then currB := curr/(IscB*ratB)  {relative sec.current used}
      else currB := curr;                              {error case}
      {voltage level of the relay, for transforming the current}
      VoltageLevel(pCompB_t, VoltB);
      pCompB_t^.getName(nameB);
      {PGQ relay or electromech.relay, or recloser relay or user defined relay}
      combrel := (nameB = gPQrel) or (nameB = gEMrel) or
                 (nameB = gRCrel) or (nameB = gUDrel);
      {for UD-relays needed fist current}
      if (nameB = gUDrel) then begin                  {user defined recloser relay}
        ipu := (rrB_t +1) div 2;                     {relay module (phase, ground, seq.)}
        GetUDFirst(pCompB_t, ipu, ipv, cFirst, apu); {first entry in table}
      end else begin
        cFirst := 1;
      end;
      {differently for combined and simple relays}
      if (combrel) then begin
        {get parameters of the second relay}
        ParametersFromComp(pCompB_t, rrB_t,Idef[1],tdef[1],Idef[2],tdef[2],
                           Idef[3],tdef[3],Idef[4],tdef[4],
                           Iss, fact, ktm, bb, cc,tmin, cFirst, group,curve);
        {get the activation time of the second relay}
        CoeffForPlane(pCompB_t,rrB_t,planecf);   {coefficient to compensate for the "plane"}
        if (VoltB > 0) and (VoltLev > 0) then {apu}plMult := currB*VoltLev/VoltB else plMult := currB;
        {check, perhaps only inverse time part or def.time part of the curve visible}
        {if form visible, take parameters from form, else from component}
        WithITorDT(nameB, pCompB_t, rrB_t, withIT, withDT);
        TimeFromCurrent(pCompB_t,planecf*plMult{apu}, Idef,Tdef, ndef,Iss, fact, ktm, bb, cc,tmin, cFirst,
                        group,curve, rrB_t, withIT, withDT, bctime, relcur);    {in PQrelay.pas}
        {indicate which part(s) of curve shown}
        IndicateCurveParts(@HintBEdit, withIT, withDT);
      end else begin     {a simple relay}
        TimeOfSimpleRelay(pCompB_t, @BList_t, Voltlev, currB, bctime);
        HintBEdit.text := 'complete curve';
      end;
      {display the results . . . . . . . . . . , if reasonable}
      if (actime > 0) and (actime <> gMaxTime) then
        TimeAEdit.text := ThreeDecimals(actime)      {display the time}
      else TimeAEdit.text := 'out of range';         {not defined, out of range}
      if (bctime > 0) and (bctime <> gMaxTime) then
        TimeBEdit.text := ThreeDecimals(bctime)      {display the time}
      else TimeBEdit.text := 'out of range';
      if (actime > 0) and (actime <> gMaxTime) and
         (bctime > 0) and (bctime <> gMaxTime) then begin
        apu := actime-bctime;
        TimeDiffEdit.text := ThreeDecimals(apu);     {display time difference}
      end else begin
        TimeDiffEdit.text := '***';
      end;
    end;
  end else begin         {not reasonable relays, or module indices}
    TimeAEdit.text := '';
    TimeBEdit.text := '';
    TimeDiffEdit.text := '';
  end;
end;

procedure TTimeDiffForm.CurrentEditChange(Sender: TObject);
begin
  {calculate the time difference}
  TimeDifference;
end;

procedure TTimeDiffForm.GraphCurrEditChange(Sender: TObject);
begin
  {calculate the time difference}
  TimeDifference;
end;

procedure TTimeDiffForm.KeyButClick(Sender: TObject);
{indicate which current in use}
begin
  CurrentEdit.enabled := true;
  CurrentEdit.color := clWhite;
  GraphCurrEdit.color := clSilver;
  GraphCurrEdit.text := '';
end;

procedure TTimeDiffForm.GraphButClick(Sender: TObject);
{indicate which current in use}
begin
  CurrentEdit.enabled := false;
  CurrentEdit.color := clSilver;
  CurrentEdit.text := '';
  GraphCurrEdit.color := clWhite;
end;

procedure TTimeDiffForm.FormActivate(Sender: TObject);
{recalculate the time difference with the present parameters}
begin
  CurveAEditChange(Sender);
  CurveBEditChange(Sender);
end;

procedure TTimeDiffForm.FormCreate(Sender: TObject);
begin
  TimeDiffForm.width := HelpMemo.left{TDwidth};
  HelpMemo.visible := false;
  AList_t := TList.Create;
  BList_t := TList.Create;
  {add flag that form created}
  Imagoo.ReadyEdit.text := Imagoo.ReadyEdit.text+'d';
end;

procedure TTimeDiffForm.FormDestroy(Sender: TObject);
begin
  AList_t.clear;
  AList_t.free;
  BList_t.clear;
  BList_t.free;
end;

procedure TTimeDiffForm.HelpButClick(Sender: TObject);
{display/hide the help memo}
begin
  if (HelpBut.down) then begin
    TimeDiffForm.width := HelpMemo.left+HelpMemo.width+8;
    HelpMemo.visible := true;
  end else begin
    TimeDiffForm.width := HelpMemo.left{TDwidth};
    HelpMemo.visible := false;
  end;
end;

procedure TTimeDiffForm.ClearButClick(Sender: TObject);
begin
  CurveAEdit.text := '';
  CurveBEdit.text := '';
  CurveALabel.font.color := clBlack;
  CurveBLabel.font.color := clBlack;
  TimeAEdit.text := '';
  TimeBEdit.text := '';
  CurrentEdit.text := '';
  GraphCurrEdit.text := '';
  TimeDiffEdit.text := '';
  ClearCompPointers;
end;

procedure TTimeDiffForm.CloseButClick(Sender: TObject);
begin
  TimeDiffForm.visible := false;
end;

procedure TTimeDiffForm.GetModuleIndices(var rrA, rrB: integer);
begin
  rrA := rrA_t;
  rrb := rrB_t;
end;

procedure TTimeDiffForm.GetRelayPointers(var pCompA, pCompB: TpComp);
begin
  pCompA := pCompA_t;
  pCompB := pCompB_t;
end;

end.
