unit atp;
{the unit saves the diagram as *.atp file for the ATP program}
interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls,skemadef,imagodef,shortcir,Earthflt,
  Compdef,Cmatmul, Emtp, GenProc, Routines, utildef, Matmul, Twoshort;


procedure ImpedanceATP(ipPar : longint);
procedure CapacitorATP(ipPar : longint);
procedure SourceATP(ipPar : longint);
procedure GeneratorATP(ipPar : longint);
procedure FaultATP(ipPar : longint);
procedure LoadATP(ipPar : longint);
procedure UnsymmLoadATP(ipPar : longint);
procedure LineATP(ipPar : longint);
procedure AsynchMotATP(ipPar : longint);
procedure TrafoYYATP(ipPar : longint);
procedure TrafoYDATP(ipPar : longint);
procedure TrafoDYATP(ipPar : longint);
procedure TrafoDDATP(ipPar : longint);
procedure Trafo3WATP(ipPar : longint);
procedure OpenLineATP(ipPar : longint);
procedure MutualLineATP(ipPar : longint);
procedure AtpFile(ipSet : longint; FileName : string);

implementation

uses
  imagopas, elplex;

procedure ImpedanceATP(ipPar : longint);
{add the lines for an impedance in the ATP *.atp format}
var
  cc    : single;        {capacitance}
  ix,iy : integer;       {temporary}
  lbl   : string;        {label (name) of the component}
  linea : string;        {ATP data for impedance in phase a}
  lineb : string;
  linec : string;
  namea : string;        {upstream node in phase a}
  nameb : string;        {upstream node in phase b}
  namec : string;        {upstream node in phase c}
  namer : string;        {downstream node in phase a}
  names : string;        {downstream node in phase b}
  namet : string;        {downstream node in phase c}
  n1,n2 : integer;       {nodes}
  pComp : TpComp;        {pointer to the impedance}
  pPar  : TpEdat;        {pointer to the data}
  rr    : single;        {resistance}
  x1    : single;        {reactance}
  xx    : single;        {reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);       {pointer to data}
    pComp := pPar^.pComp;        {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);  {name}
    pComp^.GetPar(gImped_R,rr);  {get value of resistor}
    pComp^.GetPar(gImped_X,x1);  {get value of reactance}
    {reactance may also be capacitive}
    if (x1 >= 0) then begin
      xx := x1;
      cc := 0;                   {no capacitances}
    end else begin
      xx := 0;
      cc := -1.e6/x1;            {admittance, in uS}
    end;
    {add explanatory text}
    pPar^.pIlist^.add('C  impedance  '+lbl+'  - - - - - - - - - - - - - ');
    pPar^.pIlist^.add('C <nod1><nod2><ref1><ref2>< R  >< X  >< uS >');
    {node names}
    PhaseNames(pComp,1, pPar, n1,namea,nameb,namec);
    PhaseNames(pComp,2, pPar, n2,namer,names,namet);
    ImpedanceLines(namea, nameb, namec, namer, names, namet, false,rr, xx, cc,
                   linea, lineb, linec);    {in emtp.pas}
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
  end;
end;

procedure CapacitorATP(ipPar : longint);
{add the lines for an impedance in the ATP *.atp format}
var
  cc    : complex;        {admittance of the capacitor}
  first : boolean;       {flag for first impedance}
  freq  : single;        {frequency}
  ix,iy : integer;       {temporary}
  lbl   : string;        {label (name) of the component}
  linea : string;        {ATP data for cap. in phase a}
  lineb : string;
  linec : string;
  namea : string;        {upstream node in phase a}
  nameb : string;        {upstream node in phase b}
  namec : string;        {upstream node in phase c}
  namer : string;        {downstream node in phase a}
  names : string;        {downstream node in phase b}
  namet : string;        {downstream node in phase c}
  n1,n2 : integer;       {nodes}
  pComp : TpComp;        {pointer to the component}
  pPar  : TpEdat;        {pointer to the data}
  pSet  : TpCompSet;     {pointer to the common}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);       {pointer to data}
    pSet := TpCompSet(pPar^.ipSet);     {pointer to the common}
    freq := pSet^.appl_Hz;       {frequency}
    first := (pPar^.pIlist.count <= 1); {empty, when first impedance}
    pComp := pPar^.pComp;        {pointer to the cap.}
    pComp^.GetLabel(lbl,ix,iy);  {name}
    AdmCap(pComp, freq, cc);     {admittance of the cap.}
    {add explanatory text}
    pPar^.pIlist^.add('C  capacitor  '+lbl+'  - - - - - - - - - - - - - ');
    {node names}
    PhaseNames(pComp,1, pPar, n1,namea,nameb,namec);
    PhaseNames(pComp,2, pPar, n2,namer,names,namet);
    ImpedanceLines(namea, nameb, namec, namer, names, namet, first,0,0, cc.y*1e6,
                   linea, lineb, linec);       {admittance in uS}{in emtp.pas}
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
  end;
end;

procedure SourceATP(ipPar : longint);
{add the lines for a voltage source, eventually with some internal impedance.
 Note: The earthing impedance is not included (not possible in EMTP)
 Is possible in ATP, ADD !!!}
var
  busxa   : string;      {name for the Bus-X node in the source}
  busxb   : string;      {name for the Bus-X node in the source}
  busxc   : string;      {name for the Bus-X node in the source}
  cc      : single;      {capacitance}
  cpu     : complex;     {temporary}
  err     : integer;     {error flag}
  fii     : single;      {phase angle}
  freq    : single;      {frequency}
  ipset   : integer;     {pointer to the common, as integer}
  ix,iy   : integer;     {temporary}
  lbl     : string;      {label (name) of the component}
  linea   : string;      {ATP data for impedance in phase a}
  lineb   : string;
  linec   : string;
  namea   : string;      {name of node in phase a}
  nameb   : string;      {name of node in phase b}
  namec   : string;      {name of node in phase c}
  namer   : string;      {name of source node in phase a}
  names   : string;      {name of source node in phase b}
  namet   : string;      {name of source node in phase c}
  namey   : string;      {name of neutral node}
  n1,n2   : integer;     {node}
  nn      : integer;     {neutral node}
  pComp   : TpComp;      {pointer to the component}
  pPar    : TpEdat;      {pointer to the data}
  pSet    : TpCompSet;   {pointer to the common}
  R0      : single;      {zero sequence resistance}
  Rg      : single;      {earthing resistance}
  Rn      : single;      {total neutral point resistance}
  rr      : single;      {internal resistance}
  txa,txb : string;      {temporary}
  txc     : string;      {temporary}
  Vpn     : single;      {phase to neutral voltage}
  X0      : single;      {zero sequence reactance}
  Xg      : single;      {earthing reactance}
  Xn      : single;      {total neutral point reactance}
  xx      : single;      {internal reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);           {pointer to data}
    ipSet := pPar^.ipSet;            {pointer to the common, as integer}
    pSet := TpCompSet(ipSet);        {pointer to the common}
    pComp := pPar^.pComp;            {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);      {name}
    pComp^.getNode(1,n1,err);        {node where source connected}
    pComp^.getNode(3,nn,err);        {neutral node}
    NodeNames(n1,namea,nameb,namec); {node of the source}
//  NodeNamez(nn,busxa,busxb,busxc); {the hidden nodes Bus-X}
busxa := '      ';
busxb := '      ';
busxc := '      ';
    {get parameters}
    pComp^.GetVar(gFeed_V1r,cpu.x);
    pComp^.GetVar(gFeed_V1i,cpu.y);
    Vpn := Cabs(cpu);
    fii := 180*Cangle(cpu)/pi;
    pComp^.GetPar(gACg_R,rr);        {get internal resistance}
    pComp^.GetPar(gACg_X,xx);        {get internal reactance}
    {earthing impedance}
    pComp^.GetPar(gFeed_Rg,Rg);      {grounding resistance}
    pComp^.GetPar(gFeed_Xg,Xg);      {grounding reactance}
    {zero sequence impedance}
    pComp^.GetPar(gFeed_R0,R0);      {resistance}
    pComp^.GetPar(gFeed_X0,X0);      {reactance}
    {include part of zero seq. impedance in the earthing impedance}
    if (R0 > rr) then Rn := (R0-rr)/3+Rg  {ATP does not like negative values !}
    else              Rn := Rg;
    if (X0 > xx) then Xn := (X0-xx)/3+Xg
    else              Xn := Xg;
{test}
//    Rn := (R0-rr)/3+Rg;
//    Xn := (X0-xx)/3+Xg;
{endtest}
    NodeNames(nn,txa,txb,txc);       {prel.name for neutral node (txb,txc not needed)}
//  if (Rg > 0) or (Xg > 0) then namey := txa
    if (Rn <> 0) or (Xn <> 0) then namey := txa
    else                                 namey := '      ';
    freq := pSet^.Appl_Hz;           {frequency}
    cc := 0;
    {if the internal impedance is nonzero, must add additional nodes}
    if (xx > 0) or (rr > 0) then begin
      pComp^.getNode(2,n2,err);        {"node" for current}
      NodeNames(n2,namer,names,namet); {extra node, if nonzero internal impedance }
     {add the impecance to the network}
      ImpedanceLines(namer,names,namet,namea,nameb,namec, false, rr,xx,cc,
                     linea, lineb,linec); {in emtp.pas}
      {add the lines to the output list for impedances}
      pPar^.pIlist^.add('C internal impedance of feeder '+lbl);
      pPar^.pIlist^.add(linea);
      pPar^.pIlist^.add(lineb);
      pPar^.pIlist^.add(linec);
    end else begin
      namer := namea;
      names := nameb;
      namet := namec;
    end;
    {add large impedance from source to ground, else ATP dislikes}
    ImpedanceLines(namer,names,namet,'      ','      ','      ', false, 1.e6,0.0,0.0,
                   linea, lineb,linec);
    {add the lines to the output list for impedances}
//    pPar^.pIlist^.add('C large impedances from feeder to ground (ATP needs this) '+lbl);
//    pPar^.pIlist^.add(linea);
//    pPar^.pIlist^.add(lineb);
//    pPar^.pIlist^.add(linec);
    {generate lines for source} {Tstart must be 0 (or larger), ATP does not line if < 0 !!??}
    linea := '14'+namer+'  '+TenChar(Vpn)+TenChar(freq)+TenChar(fii)+
             '          '+'          '+'  0.0     '+' 32.0     ';   {long enough ?}
    lineb := '14'+names+'  '+TenChar(Vpn)+TenChar(freq)+TenChar(fii-120.0)+
             '          '+'          '+'  0.0     '+' 32.0     ';
    linec := '14'+namet+'  '+TenChar(Vpn)+TenChar(freq)+TenChar(fii+120.0)+
             '          '+'          '+'  0.0     '+' 32.0     ';
    pPar^.pVlist^.add('C  feeder  '+lbl+'  - - - - - - - - - - - ');
    pPar^.pVList^.add(linea);
    pPar^.pVList^.add('18'+namey+'         1.0'+'      '+'      '+busxa);
    pPar^.pVList^.add(lineb);
    pPar^.pVList^.add('18'+namey+'         1.0'+'      '+'      '+busxb);
    pPar^.pVlist^.add(linec);
    pPar^.pVList^.add('18'+namey+'         1.0'+'      '+'      '+busxc);
    {earthing impedance, if there is any}
    if (Rn <> 0) or (Xn <> 0) then begin
      txa := '  '+namey+'      '+'      '+'      '+SixSignedChar(Rn)+SixSignedChar(Xn);
      pPar^.pIlist^.add(txa);
    end;
  end;
end;

procedure GeneratorATP(ipPar : longint);
{add the lines for a generator in the ATP *.atp format}
{MAADOITUSimpedanssi puuttuu viel !}
const
  BigRes = 1.0e6;        {big resistance, or impedance}
var
  agline: single;        {field current for 1000 volt}
  apu   : single;        {temporary}
  err   : integer;       {error flag}
  fii   : single;        {phase angle of output voltage, in degrees}
  freq  : single;        {frequency (50 or 60 Hz)}
  gPar  : GMset;         {generator parameters}
  ii    : integer;       {index}
  ipSet : longint;       {pointer to the "common" as integer}
  ix,iy : integer;       {temporary}
  lbl   : string;        {label (name) of the component}
  linea : string;        {one line of data}
  lineb : string;        {one line of data}
  linec : string;        {one line of data}
  name  : string;        {name of component}
  namea : string;        {name of node in phase a}
  nameb : string;        {name of node in phase b}
  namec : string;        {name of node in phase c}
  namey : string;        {name of neutral node}
  namex : string;        {name of tacs-node}
  nd,nn : integer;       {downstream node, neutral nodes}
  ngp   : single;        {number of parallel generators}
  pComp : TpComp;        {pointer to the component}
  pImped: TpComp;        {pointer to the (eventual) earthing impedance}
  pPar  : TpEdat;        {pointer to the data}
  pSet  : TpCompSet;     {pointer to the "common"}
  PP    : PhysPar;       {the set of physical parameters (inductances, resist.)}
  Rg    : single;        {the earthing resistance, in series with Xg}
  texta : string;        {temporary}
  textb : string;
  textc : string;
  Vcx   : complex;       {output voltage}
  Vpp   : single;        {abs.value of output voltage}
  Xg    : single;        {the earthing reactance, in series with Rg}
  XL    : single;        {leakage reactance}
  X0    : single;        {zero sequence reactance}
  xs,ys : integer;       {coordinates of the neutral point}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);         {pointer to data}
    ipSet := pPar^.ipSet;          {pointer to the common, as integer}
    pSet := TpCompSet(ipSet);      {pointer to the common}
    pComp := pPar^.pComp;          {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);    {name}
(*  pComp^.getNode(1,nd,err);      {upstream node} *)
    pComp^.getNode(3,nn,err);      {downstream node}
    freq := pSet^.appl_Hz;         {net frequency}
    pComp^.GetVar(gGen_V2r,Vcx.x); {real part of of output voltage}
    pComp^.GetVar(gGen_V2i,Vcx.y); {imag part of voltage}
    Vpp := Cabs(Vcx);              {abs.value of output voltage}
    fii := Cangle(Vcx)*180/pi;            {phase}
    ngp := nGenP(pComp);           {number of paraller generators}
    pComp^.GetName(name);
    if (name = gGen) then begin    {generator}
      GetGenPar(pComp,false,gPar);
    end else begin                 {synch. motor}
      GetSMotPar(pComp,false,gPar);
    end;
    SolveGenPar(gPar,freq,PP);        {solve physical parameters}
    apu := PP.Ld-1.5*PP.Mf*PP.Mf/PP.Lff; {leakage reactance, is this CORRECT ??}
    if (apu <= 0.0001) then XL := 0.0001
    else XL := apu;
    pComp^.GetPar(gGen_X0,X0);       {get zero seq. reactance}
    X0 := X0/ngp;
    if (X0 <= 0) then X0 := 0.001;   {ATP requires a value > 0}
    {the earthing impedance (if there is one) is a separate component}
    pComp^.getNxy(3,xs,ys);          {coordinates of neutral point}
    FindZg(ipSet, {nn,}xs,ys, pImped); {find the eventual earthing impedance}
    if (pImped <> nil) then begin    {impedance found}
      pImped^.GetPar(gGrImped_R,Rg); {grounding resistance}
      pImped^.GetPar(gGrImped_X,Xg); {grounding reactance}
    end else begin
      Rg := BigRes;                  {must have some resistance}
      Xg := BigRes;
    end;
    pPar^.pGlist^.add('C  synchronous machine '+lbl+' - - - - - - - - - - - ');
    {node names as text}
    PhaseNames(pComp,1, pPar, nd,namea,nameb,namec);
(*  NodeNames(nd,namea,nameb,namec);    {gen. output node} *)
    if (Rg > 0) or (Xg > 0) then begin  {there is an earthing impedance}
      NodeNames(nn,namey,textb,textc);  {name of star point (textb, textc not needed)}
    end else begin
      namey := '      ';                {directly to ground = spaces}
    end;
    {name for the tacs node for generator voltage}
    namex := 'XX'+inttostr(nn);
    if (length(namex) < 6) then begin
      for ii := length(namex)+1 to 6 do namex := namex+' ';
    end;
    {comment line, to mark generator data}
    if (name = gGen) then begin         {generator}
      pPar^.pGlist^.add('C  .   .   .   . Generator, constant speed assumed');
    end else begin
      pPar^.pGlist^.add('C  .   . Synchronous motor, constant speed assumed');
    end;
    {construct first line}
    linea := '59'+namea+'  '+TenChar(Vpp)+TenChar(freq)+TenChar(fii);
    {second line, base impedance is set to 1 ohm}
    lineb := '  '+nameb;
    {third line}
    linec := '  '+namec;
    {add these lines to output}
    pPar^.pGlist^.add(linea);
    pPar^.pGlist^.add(lineb);
    pPar^.pGlist^.add(linec);
    pPar^.pGlist^.add('PARAMETER FITTING             3.');  {use short circuit time constants}
    {flags, units, no of poles, ref.values}
    agline := 1000.0/(2.0*pi*freq*PP.Mf);       {field current for 1000 V}
    texta := ' 1 1  1  2        1.        1.        1.        1.'+TenChar(agline);
    pPar^.pGlist^.add(texta);
    pPar^.pGlist^.add('BLANK');
    texta := 'C   Ra   ><   XL   ><   Xd   ><   Xq   ><   Xd  ><   Xq  ><   Xd"  ><   Xq"  >';
    textb := TenChar(gPar.Rs) +TenChar(XL)      +TenChar(gPar.Xd  )+TenChar(gPar.Xd)+  {Xd = Xq}
             TenChar(gPar.Xdd)+TenChar(gPar.Xqd)+TenChar(gPar.Xd2d)+TenChar(gPar.Xq2d);
    pPar^.pGlist^.add(texta);                {explanations}
    pPar^.pGlist^.add(textb);                {impedances X..}
    texta :='C   Td  ><   Tq  ><   Td"  ><   Tq"  ><   X0   ><   Rg   ><   Xg   ><   Xcan >';
    textb := TenChar(gPar.Tdd)+TenChar(gPar.Tqd)+TenChar(gPar.Td2d)+TenChar(gPar.Tq2d)+
             TenChar(X0)+TenChar(Rg)+TenChar(Xg)+TenChar(XL);
    pPar^.pGlist^.add(texta);                {explanations}
    pPar^.pGlist^.add(textb);                {time constants, etc}
    {mechanical data, large inertia}
    texta := ' 1               1.0      1.E3                                     50.';
    pPar^.pGlist^.add(texta);
    pPar^.pGlist^.add('BLANK');
    pPar^.pGlist^.add('  11');
    pPar^.pGlist^.add('  21');
    pPar^.pGlist^.add('  31');
    pPar^.pGlist^.add('  41');
    pPar^.pGlist^.add('  51');
    pPar^.pGlist^.add('BLANK');
    pPar^.pGlist^.add('71'+namex);           {for tacs}
    pPar^.pGlist^.add('  FINISH');
    {for tacs = for generator voltage}
    pPar^.pTlist^.add('11'+namex+
    '          1.                                               -1.      100.');
  end;
end;

procedure FaultATP(ipPar : longint);
{add the lines for a fault = switch in the ATP *.atp format}
var
  err   : integer;       {error flag}
  linea : string;        {ATP data for impedance in phase a}
  lineb : string;
  linec : string;
  namea : string;        {name of node in phase a}
  nameb : string;        {name of node in phase b}
  namec : string;        {name of node in phase c}
  n1    : integer;       {node}
  pComp : TpComp;        {pointer to the fault}
  pPar  : TpEdat;        {pointer to the data}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);           {pointer to data}
    pComp := pPar^.pComp;            {pointer to the impedance}
    pComp^.getNode(1,n1,err);        {node where fault connected}
    NodeNames(n1,namea,nameb,namec); {node of the fault}
    {generate lines for the fault, assumed from node 1 to earth}
    {closing time set to 0.1 sec, opening time = 1000 s}
    linea := '  '+namea+'      '+' 1.5      '+' 1000.0   '+' 1.e6     ';
    lineb := '  '+nameb+'      '+' 1.5      '+' 1000.0   '+' 1.e6     ';
    linec := '  '+namec+'      '+' 1.5      '+' 1000.0   '+' 1.e6     ';
    {write to list of lines}
    pPar^.pSlist^.add('C  three phase fault as switches- - - - - - - - - - - - ');
    pPar^.pSlist^.add(linea);
    pPar^.pSlist^.add(lineb);
    pPar^.pSlist^.add(linec);
    {other faults "commented" away}
    linea := 'C'+'  '+namea+'      '+' 1.5      '+' 1000.0   '+' 1.e6     ';
    pPar^.pSlist^.add('C  line to earth fault as a switch- - - - - - - - - - - - ');
    pPar^.pSlist^.add(linea);
    linea := 'C'+'  '+namea+nameb+' 1.5      '+' 1000.0   '+' 1.e6     ';
    pPar^.pSlist^.add('C  line to line fault as a switch- - - - - - - - - - - - ');
    pPar^.pSlist^.add(linea);
  end;
end;

procedure OpenLineATP(ipPar : longint);
{add the lines for an open-line-fault = switch in the ATP *.atp format}
var
  err   : integer;       {error flag}
  linea : string;        {ATP data for impedance in phase a}
  lineb : string;
  linec : string;
  namea : string;        {name of first node in phase a}
  nameb : string;        {name of first node in phase b}
  namec : string;        {name of first node in phase c}
  namex : string;        {name of second node in phase a}
  namey : string;        {name of second node in phase b}
  namez : string;        {name of second node in phase c}
  n1,n2 : integer;       {node}
  pComp : TpComp;        {pointer to the fault}
  pPar  : TpEdat;        {pointer to the data}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);           {pointer to data}
    pComp := pPar^.pComp;            {pointer to the impedance}
    pComp^.getNode(1,n1,err);        {node where fault connected}
    NodeNames(n1,namea,nameb,namec); {node of the fault}
    pComp^.getNode(2,n2,err);        {node where fault connected}
    NodeNames(n2,namex,namey,namez); {node of the fault}
    {generate lines for the fault as switches, opens only in phase c}
    {closing time set to -1 sec, opening time = 1.0 or 1000 s}
    linea := '  '+namea+namex+'-1.0      '+' 1000.0   '+' 1.e6     ';
    lineb := '  '+nameb+namey+'-1.0      '+' 1000.0   '+' 1.e6     ';
    linec := '  '+namec+namez+'-1.0      '+' 1.0      '+' 1.e6     ';
    {write to list of lines}
    pPar^.pSlist^.add('C  open-line-fault as switches- - - - - - - - - - - - ');
    pPar^.pSlist^.add(linea);
    pPar^.pSlist^.add(lineb);
    pPar^.pSlist^.add(linec);
  end;
end;

procedure LoadATP(ipPar : longint);
{add the lines for a load in the ATP *.atp format
 The load is described as an impedance. The size of that impedance
 is calculated using the voltage and current of the load}
var
  cc    : single;        {capacitance}
  cpu   : complex;       {temporary}
  err   : integer;       {error flag}
  first : boolean;       {flag for first impedance}
  Icx   : complex;       {load current}
  ix,iy : integer;       {temporary}
  lbl   : string;        {label (name) of the component}
  linea : string;        {ATP data for  phase a}
  lineb : string;
  linec : string;
  namea : string;        {upstream node in phase a}
  nameb : string;        {upstream node in phase b}
  namec : string;        {upstream node in phase c}
  namey : string;        {downstream node = neutral node}
  n1,n2 : integer;       {nodes}
  pComp : TpComp;        {pointer to the impedance}
  pPar  : TpEdat;        {pointer to the data}
  Vcx   : complex;       {load voltage}
  rr    : single;        {resistance}
  textb : string;        {temporary}
  textc : string;
  x1    : single;        {temp.reactance}
  xx    : single;        {reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);       {pointer to data}
    first := (pPar^.pIlist.count <= 1); {empty, when first impedance}
    pComp := pPar^.pComp;        {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);  {name}
    pComp^.getNode(1,n1,err);    {upstream node}
    pComp^.getNode(3,n2,err);    {neutral node}
    {get the voltage and current of the load}
    pComp^.GetVar(gLoad_Vr,Vcx.x); {real part of voltage to comp.}
    pComp^.GetVar(gLoad_Vi,Vcx.y); {imag.part of voltage}
    pComp^.GetVar(gLoad_Ir,Icx.x); {real part of current}
    pComp^.GetVar(gLoad_Ii,Icx.y); {imag.part of current}
    {calculate impedance}
    if ((Vcx.x <> 0) or (Vcx.y <> 0)) and ((Icx.x <> 0) or (Icx.y <> 0)) then begin
      cpu := Cdiv(Vcx,Icx);
      if (cpu.x >= 0) then rr := cpu.x    {resistance}
      else                 rr := 0.001;   {some value}
      x1 := cpu.y;                        {reactance, checked later}
    end else begin         {voltage or current zero, must use nominal values}
      NominalLoadPar(pComp, rr,x1);
    end;
    {reactance might also be capacitive}
    if (x1 >= 0.001) then begin
      xx := x1;
      cc := 0;                   {no capacitances}
    end else if (x1 < -0.001) then begin
      xx := 0;
      cc := -1.e6/x1;            {admittance, in uS}
    end else begin
      xx := 0;
      cc := 0;
    end;
    {add explanatory text}
    pPar^.pIlist^.add('C  Load  '+lbl+'  - - - - - - - - - - - - - ');
    {node names}
    NodeNames(n1,namea,nameb,namec);  {upstream node}
    NodeNames(n2,namey,textb,textc);  {downstream node}
    ImpedanceLines(namea, nameb, namec, namey, namey, namey, first,rr, xx, cc,
                   linea, lineb, linec);
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
  end;
end;

procedure UnsymmLoadATP(ipPar : longint);
{add the lines for an unsymmetrcial load = three different impedances
 in the EMTP *.dat format}
var
  c1,c2,c3 : single;        {capacitive admittance, in uS}
  cpu      : complex;       {temporary}
  err      : integer;       {error flag}
  first    : boolean;       {flag for first impedance}
  ix,iy    : integer;       {temporary}
  lbl      : string;        {label (name) of the component}
  linea    : string;        {EMTP data for  phase a}
  lineb    : string;
  linec    : string;
  namea    : string;        {upstream node in phase a}
  nameb    : string;        {upstream node in phase b}
  namec    : string;        {upstream node in phase c}
  namey    : string;        {downstream node = neutral node}
  n1       : integer;       {nodes}
  pComp    : TpComp;        {pointer to the component}
  pPar     : TpEdat;        {pointer to the data}
  rr       : single;        {temporary resistance}
  r1,r2,r3 : single;        {resistance}
  textb    : string;        {temporary}
  textc    : string;
  xx       : single;        {temporary reactance}
  x1,x2,x3 : single;        {inductive reactance, ohms}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);       {pointer to data}
    first := (pPar^.pIlist.count <= 1); {empty, when first impedance}
    pComp := pPar^.pComp;        {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);  {name}
    pComp^.getNode(1,n1,err);    {upstream node}
    {get the first impedance}
    pComp^.GetPar(1,rr);
    pComp^.GetPar(2,xx);
    CheckedImedance(rr,xx, r1,x1,c1);  {get resist., induct.react., and cap. sucept.}
    {get the second impedance}
    pComp^.GetPar(3,rr);
    pComp^.GetPar(4,xx);
    CheckedImedance(rr,xx, r2,x2,c2);  {get resist., induct.react., and cap. sucept.}
    {get the second impedance}
    pComp^.GetPar(5,rr);
    pComp^.GetPar(6,xx);
    CheckedImedance(rr,xx, r3,x3,c3);  {get resist., induct.react., and cap. sucept.}
    {add explanatory text}
    pPar^.pIlist^.add('C  Unsymmetrical Load  '+lbl+'  - - - - - - - - - - - - - ');
    {node names}
    NodeNames(n1,namea,nameb,namec);  {upstream node}
    NodeNamez(n1,namey,textb,textc);  {downstream node = neutral node}
    ThreeImpedLines(namea, nameb, namec, namey, namey, namey, first,r1,x1,c1,
                    r2,x2,c2,r3,x3,c3,linea, lineb, linec);
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
    {add large impedance from neutral point to ground, else ATP dislikes}
    ImpedanceLines(namey,namey,namey,'      ','      ','      ', false, 1.e6,0.0,0.0,
                   linea, lineb,linec);   {only linea needed}
    pPar^.pIlist^.add(linea);
  end;
end;

procedure WriteATPline(pComp: TpComp; r1,x1,r0,x0,cc: single; nc1,nc2 : integer; pPar: TpEdat);
{calculate and write the transmission line parameters
 pComp   pointer to the tr.line component
 r1,x1   pos.seq. resistance, reactance
 r0,x0   zero seq. resistance ,reactance
 cc      checked susceptance in uS
 nc1,nc2 nodes within the component (1,2 etc)
 pPar    pointer to the data}
var
  ix,iy   : integer;   {temporary}
  lbl     : string;    {label (name) of the component}
  linea   : string;    {ATP data for tr.line in phase a}
  lineb   : string;
  linec   : string;
  namea   : string;    {upstream node in phase a}
  nameb   : string;    {upstream node in phase b}
  namec   : string;    {upstream node in phase c}
  namer   : string;    {downstream node in phase a}
  names   : string;    {downstream node in phase b}
  namet   : string;    {downstream node in phase c}
  n1,n2   : integer;   {node indices in the network}
  rii,rij : single;    {real part of diag. and off-diag elements}
  xii,xij : single;    {imag.part of diag. and off diag. elements}
begin
  {convert to matrix elements, user pos.seq. and zero seq.}
  rii := (r0+2*r1)/3.0;
  xii := (x0+2*x1)/3.0;
  rij := (r0-r1)/3.0;
  xij := (x0-x1)/3.0;
  {add explanatory text}
  pComp^.GetLabel(lbl,ix,iy);  {name}
  pPar^.pIlist^.add('C  transmission line   '+lbl+' - - - - - - - - - - - - ');
  {node names}
  PhaseNames(pComp,nc1, pPar, n1,namea,nameb,namec);
  PhaseNames(pComp,nc2, pPar, n2,namer,names,namet);
  {generate the lines of text for the ATP *.dat file}
  inc(pPar^.branch);                 {increment branch counter}
  if (pPar^.branch < 48) then begin  {at most 50 transmission lines (!)}
    linea := '1 '+namea+namer+'      '+'      '+
             SixChar(rii)+SixChar(xii)+SixChar(cc);
    inc(pPar^.branch);
    lineb := '2 '+nameb+names+'      '+'      '+
             SixSignedChar(rij)+SixSignedChar(xij)+'      '+
             SixChar(rii)+SixChar(xii)+SixChar(cc);
    inc(pPar^.branch);
    linec := '3 '+namec+namet+'      '+'      '+
             SixSignedChar(rij)+SixSignedChar(xij)+'      '+
             SixSignedChar(rij)+SixSignedChar(xij)+'      '+
             SixChar(rii)+SixChar(xii)+SixChar(cc);
  end;
  {add the lines to the output list for impedances}
  pPar^.pIlist^.add(linea);
  pPar^.pIlist^.add(lineb);
  pPar^.pIlist^.add(linec);
end;

procedure LineATP(ipPar : longint);
{add the lines for a transmission line in the ATP *.atp format
 The line is modeled with a pi-circuit, with no coupling between lines}
var
  apu     : single;    {temporary}
  cap     : single;    {susceptance/km of the capacitances}
  cc      : single;    {checked susceptance in uS}
  cpu     : complex;   {temporary}
  len     : single;    {length of tr.line (km)}
  nnod    : integer;   {number of nodes}
  pComp   : TpComp;    {pointer to the impedance}
  par     : single;    {number of parallel lines}
  pPar    : TpEdat;    {pointer to the data}
  rpm     : single;    {relative position of the middle node (if any)}
  r1,r0   : single;    {pos., zero seq resistance}
  x1,x0   : single;    {pos., zero seq. reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);       {pointer to data}
    pComp := pPar^.pComp;        {pointer to the tr.line}
    pComp^.GetNnod(nnod);
    pComp^.GetPar(gTrliS_len,len);   {length of the line (m or km)}
    LineImped(pComp, len, 1.0, cpu); {impedance of the line}
    r1 := cpu.x;
    x1 := cpu.y;
    {capacitance between lines and neutral (ground)}
    pComp^.GetPar(gTrliS_S,par);       {number of parallel lines}
    pComp^.GetPar(gTrliS_S,cap);       {get value of susceptance /km}
    if (par > 0) then cap := cap*par;
    if (cap < 0.0) then cap := 0;
    cc  := Cap*len*1e6;                {total suxc. of capacitances in uS}
    LineZeroImped(pComp, len,cpu);     {zero seq.impedance}
    r0 := cpu.x;
    x0 := cpu.y;
    {calculate and write the atp-parameters}
    if (nnod = 2) then begin                               {"simple line", two nodes}
      WriteATPline(pComp, r1, x1, r0, x0, cc, 1, 2, pPar);
    end else if (nnod = 3) then begin                      {additional node in the middle}
      pComp^.GetPar(gTrliS_s1,apu);                        {relative position of the middle point}
      {check the middle point}
      if (apu < 0.001) then rpm := 0.001
      else if (apu < 0.999) then rpm := apu
      else rpm := 0.999;
      {line as two lines}
      WriteATPline(pComp, rpm*r1, rpm*x1, rpm*r0, rpm*x0, rpm*cc, 1, 3, pPar);
      WriteATPline(pComp, (1-rpm)*r1, (1-rpm)*x1, (1-rpm)*r0, (1-rpm)*x0, (1-rpm)*cc, 3, 2, pPar);
    end;
  end;
end;

procedure MutualLineATP(ipPar : longint);
{add the lines for the mutual impedance (transmission line)-component
 in the ATP *.atp format
 The lines are modeled with a pi-circuit, BUT with NO coupling between lines
 That is: as with several independent transmission line components}
var
  apu      : single;    {temporary}
  bb       : single;    {susceptance/km of the capacitances}
  cc       : single;    {checked susceptance in uS}
  cpu      : complex;   {temporary}
  iline    : integer;   {index}
  ip       : integer;   {index of first parameter in the set}
  ju,jd,jm : integer;   {index of node, within the component}
  len      : single;    {length of tr.line (km)}
  nline    : integer;   {number of transmission lines}
  nnod     : integer;   {number of nodes}
  pComp    : TpComp;    {pointer to the impedance}
  par      : single;    {number of parallel lines}
  pos      : single;    {not used}
  pPar     : TpEdat;    {pointer to the data}
  rpm      : single;    {relative position of the middle node (if any)}
  rr,xx    : single;    {resistance/length, reactance/length}
  rr0,xx0  : single;    {zero seq. resistance/length, react./length}
  r1,r0    : single;    {pos., zero seq resistance}
  x1,x0    : single;    {pos., zero seq. reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);             {pointer to data}
    pComp := pPar^.pComp;              {pointer to the tr.line}
    pComp^.GetPar(1,apu);              {number of transmission lines}
    nline := round(apu);
    pComp^.GetPar(gMutuStrt-1,apu);  {relative position of middle node}
    {check the middle point}
    if (apu < 0.001) then rpm := 0.001
    else if (apu < 0.999) then rpm := apu
    else rpm := 0.999;
    {all lines, if reasonable number}
    if (nline > 0) and (nline <= gMutualMax) then begin  {reasonable number, should be ! }
      ip := gMutuStrt;                         {index of first parameter}
      for iline := 1 to nline do begin
        {get nodes within the component}
        ju := (iline-1)*5+1;                   {get upstream node of line ii}
        jd := (iline-1)*5+2;                   {get downstream node of line ii}
        jm := (iline-1)*5+3;                   {get middle node of line ii}
        {get parameters}
        GetMutualLinePar(pComp, ip, len,rr,xx,rr0,xx0,bb,pos);   {ip updated in the procedure}
        r1 := rr*len;
        x1 := xx*len;
        {capacitance between lines and neutral (ground)}
        cc  := bb*len*1e6;                     {total susc. of capacitances in uS}
        {zero seq. impedance}
        r0 := rr0*len;
        x0 := xx0*len;
        {calculate and write the atp-parameters}
        {line as two lines}
        WriteATPline(pComp, rpm*r1, rpm*x1, rpm*r0, rpm*x0, rpm*cc, ju, jm, pPar);
        WriteATPline(pComp, (1-rpm)*r1, (1-rpm)*x1, (1-rpm)*r0, (1-rpm)*x0, (1-rpm)*cc, jm, jd, pPar);
      end;
    end;
  end;
end;

procedure InitialCurrent(Rr,Rs,X2,X3,slip : single; volt, curr : complex;
                       var statcur, rotcur : Ttrois);
{calculates the initial rotor (and stator) currents in the form
 needed in ATPdraw, when the voltage and stator current in phase a are known
 Rr,Rs   stator, rotor resistance
 X2,X3   impedances (rotor, magnetization)
 slip    slip
 volt    stator phase to neutral voltage of phase a
 curr    stator current of phase a
 statcur stator current of phases a,b,c
 rotcur  Park transformed rotor currents d,q,0}
const
  n3 = 3;            {dimension of Park's matrix, etc}
  sqrt2 = 1.414214;
  sqrt3 = 1.732051;
var
  ACs,ACr  : single;      {abs.value of stator, rotor current}
  alpha    : single;      {phase of stator current}
  apu,apv  : single;      {temporary}
  beta     : single;      {phase of rotor current}
  ca,cb,cc : complex;     {stator currents}
  cosf     : single;      {cos(120)}
  cpu      : complex;     {temporary}
  Cr       : array[1..3] of complex;  {rotor current}
  Cpr      : array[1..3] of complex;  {Park transformed rotor current}
  gam      : single;      {phase of stator voltage}
  pP       : pComplex;    {pointer to Park's matrix}
  sinf     : single;      {sin(120)}
  sqrt23   : single;      {sqrt(2/3)}
  x,y      : single;      {temporary}
  wLr      : single;      {stator reactance}
  wLsr     : single;      {stator-rotor mutual reactance}
begin
  getmem(pP,n3*n3*SizeOf(complex));   {get memory for Park's matrix}
  {generate the Park's matrix}
  sqrt23 := sqrt2/sqrt3;
  putCMat(pP,1,1,n3,n3, cplx( sqrt23    ,0));
  putCMat(pP,1,2,n3,n3, cplx(-0.5*sqrt23,0));
  putCMat(pP,1,3,n3,n3, cplx(-0.5*sqrt23,0));
  putCMat(pP,2,1,n3,n3, cplx( 0.0       ,0));
  putCMat(pP,2,2,n3,n3, cplx(-0.5*sqrt2 ,0));
  putCMat(pP,2,3,n3,n3, cplx( 0.5*sqrt2 ,0));
  putCMat(pP,3,1,n3,n3, cplx( 1/sqrt3   ,0));
  putCMat(pP,3,2,n3,n3, cplx( 1/sqrt3   ,0));
  putCMat(pP,3,3,n3,n3, cplx( 1/sqrt3   ,0));
  { sin and cos of 120 degrees}
  sinf := 0.5*sqrt3;
  cosf := -0.5;
  { stator currents }
  ca := curr;            {known current of phase a}
  { stator currents of  phases b and c}
  cb := cplx(ca.x*cosf+ca.y*sinf, -ca.x*sinf+ca.y*cosf);
  cc := cplx(ca.x*cosf-ca.y*sinf,  ca.x*sinf+ca.y*cosf);
  { abs.value of stator current and phase}
  ACs := Cabs(ca);
  alpha := Cangle(ca);
  gam := Cangle(volt);  {phase of stator voltage}
  {calculate rotor current}
  wLr := X2+X3;         {rotor reactance}
  wLsr := X3;
  apu := Rr*Rr+slip*slip*wLr*wLr;
  if (apu > 0) then begin
    ACr := ACs*slip*wLsr/sqrt(apu);     {abs.value of rotor current}
  end else begin
    ACr := ACs;                         {just some value}
  end;
  { phase of rotor current w.r.t the general direction}
  apu := Rr*tan(alpha-gam)-slip*wLr;
  apv := -Rr-slip*wLr*sin(alpha-gam);
  cpu := cplx(apu,apv);
  beta := Cangle(cpu){+pi}+gam;
  { components of rotor current }
  x := ACr*cos(beta);
  y := ACr*sin(beta);
  Cr[1] := cplx(x, y);
  Cr[2] := cplx(x*cosf+y*sinf, -x*sinf+y*cosf);
  Cr[3] := cplx(x*cosf-y*sinf,  x*sinf+y*cosf);
  {park transformed rotor current}
  Csqumul(pP,@Cr,@Cpr,n3,1);
  freemem(pP,n3*n3*SizeOf(complex));   {free memory for Park's matrix}
  {currents to output}
  statcur.x := -ca.x;
  statcur.y := -cb.x;
  statcur.z := -cc.x;
  rotcur.x := Cpr[1].x;
  rotcur.y := Cpr[2].x;
  rotcur.z := 0;
end;


procedure AsynchMotATP(ipPar : longint);
{add the lines for an asynchronous motor in the ATP *.atp format
The as.motor model for ATP is rather complicated:

                   (V) voltage source
                    |
                    . (j,k,l)
                 switch
                    |
           (w)      |         (r,s,t)             (a,b,c)
(DC-source)-.-(M)---'-<switch>---.---<tr.line>------o  input node
            |     (g,h,i)        |
         capacitance             Zapp
            |                    |
            |                    . (d,e,f)
          Ground              switch
                                 |
                              Ground

(the lower case letters in parenthese refer to node names
  for example: (r,s,t) = namer, names, namet)
The DC-source and capacitance are for the rotational speed
The switches, voltage source and impedance Zapp are for initialization
The transmission line is needed, if there are more than one as.motor in the net
}
var
  apu,apv  : single;     {temporary}
  apw,apx  : single;     {temporary}
  curr     : complex;    {motor current}
  err      : integer;    {error flag}
  errMess  : string;     {error message}
  fii      : single;     {phase angle of voltage}
  freq     : single;     {frequency (50 or 60 Hz)}
  ii       : integer;    {index}
  ipSet    : longint;    {pointer to the "common" as integer}
  ix,iy    : integer;    {temporary}
  len      : integer;    {length of string}
  lbl      : string;     {label (name) of the component}
  linea    : string;     {one line of data}
  lineb    : string;     {one line of data}
  linec    : string;     {one line of data}
  namea,nameb,namec : string;  {names of input nodes}
  named,namee,namef : string;  {names for nodes of apparent impedance}
  nameg,nameh,namei : string;  {names for motor input nodes}
  namej,namek,namel : string;  {names for voltage source}
  namer,names,namet : string;  {names for end of tr.line}
  namey    : string;     {name of neutral node}
  namew    : string;     {name of node for rotating speed}
  nu,nn    : integer;    {upstream node, neutral nodes}
  omega    : single;     {angular frequency}
  omegam   : single;     {angular speed}
  Pow      : single;     {power}
  pComp    : TpComp;     {pointer to the component}
  pPar     : TpEdat;     {pointer to the data}
  pSet     : TpCompSet;  {pointer to the "common"}
  RL       : single;     {resistance corresponds to load}
  rotcur   : Ttrois;     {initial rotor currents}
  Rs,Rr    : single;     {stastor, rotor resistance}
  R0       : single;     {impedance (=resistance) of transmission line}
  slip     : single;     {slip of the motor}
  sn       : single;     {nominal slip}
  statcur  : Ttrois;     {initial stator currents}
  swtime   : string;     {switching time for switches}
  texta    : string;     {temporary}
  textb    : string;
  textc    : string;
  trlen    : single;     {length of transmission line}
  Vn       : single;     {nominal phase to phase voltage}
  volt     : complex;    {motor voltage}
  X1,X2,X3 : single;     {motor impedances}
  Xapp     : single;     {imag.part of apparent impedance}
  Xd       : single;     {transient reactance}
  XL,YC    : single;     {impedance, susceptance of tr.line/length}
  Zapp     : complex;    {apparent impedance}
  zneg     : complex;    {neg.seq.impedance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);        {pointer to data}
    ipSet := pPar^.ipSet;         {pointer to the common, as integer}
    pSet := TpCompSet(ipSet);     {pointer to the common}
    pComp := pPar^.pComp;         {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);   {name}
    pComp^.getNode(1,nu,err);     {upstream node}
    pComp^.getNode(3,nn,err);     {downstream node}
    freq := pSet^.appl_Hz;        {net frequency}
    omega := 2*pi*freq;          {angular freq}
    AsMotPar(pComp, Rs,Rr,X1,X2,X3,Vn,Pow,apu,apv,apw,apx,errMess); {get and calculate parameters}
    {do not care about an eventual error message, it has been handled in Elplex.pas}
    {get the current and voltage of the motor, for tr.line, for initial values}
    pComp^.GetVar(gAsmot_Ir,curr.x);    {real part of current}
    pComp^.GetVar(gAsmot_Ii,curr.y);    {imag. part of current}
    pComp^.GetVar(gAsmot_Vr,volt.x);    {real part of voltage}
    pComp^.GetVar(gAsmot_Vi,volt.y);    {imag. part of voltage}
    {apparent impedance of the motor}
    if (Cabs(curr) > 0.1) and (Cabs(volt) > 0.1) then begin
      Zapp := Cdiv(volt,curr);
    end else begin
      Zapp := cplx(1,1);        {just some value}
    end;
    {calculate the  slip}
    pComp^.GetPar(gASmot_sn,sn);           {nominal slip}
    AsMotLoad(Vn,Pow,sn,cplx(Rs,X1),cplx(Rr,X2),cplx(0,X3), RL,slip, err);
    zneg := Znegseq(Rs,Rr,X1,X2,X3,slip);   {neg.seq.impedance, for info only}
    omegam := omega*(1-slip);     {angular speed}
    Xd := X1+X2*X3/(X2+X3);       {transient reactance, for info only}
    pPar^.pAlist^.add('C  asynchronous machine  '+lbl+'  - - - - - - - - - - - ');
    {node names as text}
    NodeNames(nu,namea,nameb,namec);  {nominal motor input node}
    NodeNames(nn,namey,textb,textc);  {name of star point (textb,textc not needed)}
    {if there are more than one induction motor, must connect the motor using
     a transmission line. => another set of node names needed.}
    len := length(namea);
    if (pPar^.nAsMot >= 2 ) then begin
      {construct names for motor input nodes}
      namer := 'R';
      names := 'S';
      namet := 'T';
      for ii := 2 to len do begin
        namer := namer+namey[ii];
        names := names+namey[ii];
        namet := namet+namey[ii];
      end;
    end else begin    {only one motor, one set of node names needed}
      namer := namea;
      names := nameb;
      namet := namec;
    end;
    {construct names for additional nodes}
    named := 'D';
    namee := 'E';
    namef := 'F';
    nameg := 'G';
    nameh := 'H';
    namei := 'I';
    namej := 'J';
    namek := 'K';
    namel := 'L';
    for ii := 2 to len do begin
      named := named+namey[ii];
      namee := namee+namey[ii];
      namef := namef+namey[ii];
      nameg := nameg+namey[ii];
      nameh := nameh+namey[ii];
      namei := namei+namey[ii];
      namej := namej+namey[ii];
      namek := namek+namey[ii];
      namel := namel+namey[ii];
    end;
    {construct name for speed node}
    len := length(namey);
    namew := 'W';
    for ii := 2 to len do namew := namew+namey[ii];
    {first line with general parameters}
    linea :=  ' 3 1 1001'+namew+'       1                                        .1';
    linea := linea+'    '+TenChar(freq);
    pPar^.pAlist^.add(linea);
    {lines for speed, X3}
    pPar^.pAlist^.add('C Magnetization inductances');
    linea := '    '+TenChar(omegam)+'   '+TenChar(X3/omega)+' 0';
    lineb := '                 '+TenChar(X3/omega)+' 0';
    pPar^.pAlist^.add(linea);
    pPar^.pAlist^.add(lineb);
    {stator windings}
    pPar^.pAlist^.add('C Stator windings');
    linea := '                            '+nameg+namey+'      1';
    lineb := '    '+TenChar(Rs)+'    '+TenChar(X1/omega)+nameh+namey+'      1';
    linec := '    '+TenChar(Rs)+'    '+TenChar(X1/omega)+namei+namey+'      1';
    {add still the initial currents}
    InitialCurrent(Rr,Rs,X2,X3,slip,volt,curr, statcur, rotcur);
    linea := linea+'    '+TenChar(statcur.x);
    lineb := lineb+'    '+TenChar(statcur.y);
    linec := linec+'    '+TenChar(statcur.z);
    pPar^.pAlist^.add(linea);
    pPar^.pAlist^.add(lineb);
    pPar^.pAlist^.add(linec);
    {rotor windings}
    pPar^.pAlist^.add('C Rotor windings');
    linea := '    '+TenChar(Rr)+'    '+TenChar(X2/omega)+'                  1';
    lineb := linea;
    {add initial currents }
    linea := linea+'    '+TenChar(rotcur.x);
    lineb := lineb+'    '+TenChar(rotcur.y);
    pPar^.pAlist^.add(linea);
    pPar^.pAlist^.add(lineb);
    {apparent and neg.seq. impedance, for information only}
    pPar^.pAList^.add('C apparent  impedance '+TenChar(Zapp.x)+'  i'+TenChar(Zapp.y));
    pPar^.pAList^.add('C neg.seq.  impedance '+TenChar(zneg.x)+'  i'+TenChar(zneg.y));
    pPar^.pAlist^.add('C transient impedance '+TenChar(Rs)+'  i'+TenChar(Xd));
    {large resistance from neutral to ground}
    pPar^.pIlist^.add('C resistance to ground from as.mot '+lbl);
    texta := '  '+namey+'      ';
    textb := '              1.E5                                               0';
    pPar^.pIlist^.add(texta+textb);
    {voltage source to describe the constant rotating speed}
    pPar^.pVlist^.add('C speed of as.motor '+lbl);
    linea := '11'+namew+' 0'+TenChar(omegam)+
             '                                               -1.     1000.';
    pPar^.pVlist^.add(linea);
    {a capacitance parallel with the source (does not work otherwise)}
    pPar^.pIlist^.add('C speed of as.motor '+lbl);
    linea := '  '+namew+'      '+'      '+'      '+'      '+'      '+' 3.0e5';
    pPar^.pIlist^.add(linea);
    {if there are more than one motor, must add a transmission line
    at motor input, to separate from other motors}
    if (pPar^.nAsMot >= 2) then begin
      pPar^.pIlist^.add('C transmission line separates motor from other motors');
      {minimize disturbance, select jwL = jwC}
      Xapp := Zapp.y;                {apparent reactance}
      apu := Xapp*Xapp;              {reactance squared}
      apv := apu*apu;                {reactance to fourth power}
      apw := Cabs(Zapp);
      apw := apw*apw;
      apw := apw*apw;                {impedance to fourth power}
      R0 := sqrt(sqrt(apv+apw)-apu); {impedance of transmission line}
      XL := R0;
      YC := 1/R0;                    {susceptance in uS/length}
      {length of the line, just long enough}
      trlen := 1.01*omega*pPar^.dt/sqrt(XL*YC);
      linea := '-1'+namer+namea+
{              '                     1.0  1.e6 .0032 0 1                         0';}
               '                  '+SixChar(XL)+SixChar(YC*1e6)+SixChar(trlen)+
               ' 0 1                         0';
      lineb := '-2'+names+nameb+
               '                  '+SixChar(XL)+SixChar(YC*1e6)+SixChar(trlen)+
               ' 0 1                         0';
{              '                     1.0  1.e6 .0032 0 1                         0';}
      linec := '-3'+namet+namec+
               '                                                                 0';
      pPar^.pIlist^.add('C if time step (dT) is changed, change      <leng>'+
                        ' (length of line), 2 times !');
      pPar^.pIlist^.add('C                               < X  ><Y/e6><leng>');
      pPar^.pIlist^.add(linea);
      pPar^.pIlist^.add(lineb);
      pPar^.pIlist^.add(linec);
    end;
    {all swithces, for the initialization}
    {closing time = -1 for those that are closet at start
     switch time = swtime for all
     opening time = 1000 for those that are closed at swtime}
    pPar^.pSlist^.add('C  switches for the initialization of as.motor '+lbl);
    {the components for the initialization are used only for a motor, not a generator}
    if (Pow >= 0) then swtime := ' 0.5      '       {ten char.}
    else               swtime := '-0.5      ';      {addtional comp. not in use = switches do not work}
    {from Zapp to ground}
    linea := '  '+named+'      '+' -1.0     '+swtime+' 1.e6     ';
    lineb := '  '+namee+'      '+' -1.0     '+swtime+' 1.e6     ';
    linec := '  '+namef+'      '+' -1.0     '+swtime+' 1.e6     ';
    pPar^.pSlist^.add(linea);
    pPar^.pSlist^.add(lineb);
    pPar^.pSlist^.add(linec);
    {from tr.line (or input) to motor}
    linea := '  '+namer+nameg+swtime+' 1000.0   '+' 1.e6     ';
    lineb := '  '+names+nameh+swtime+' 1000.0   '+' 1.e6     ';
    linec := '  '+namet+namei+swtime+' 1000.0   '+' 1.e6     ';
    pPar^.pSlist^.add(linea);
    pPar^.pSlist^.add(lineb);
    pPar^.pSlist^.add(linec);
    {from voltage source to motor}
    linea := '  '+namej+nameg+' -1.0     '+swtime+' 1.e6     ';
    lineb := '  '+namek+nameh+' -1.0     '+swtime+' 1.e6     ';
    linec := '  '+namel+namei+' -1.0     '+swtime+' 1.e6     ';
    pPar^.pSlist^.add(linea);
    pPar^.pSlist^.add(lineb);
    pPar^.pSlist^.add(linec);
    {apparent impedance  to switch to ground}
    pPar^.pIlist^.add('C  apparent impedance for the initialization of as.motor '+lbl);
    ImpedanceLines(namer, names, namet, named, namee, namef, false,Zapp.x,Zapp.y,0.0,
                   linea, lineb, linec);
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
    {ATP does not like loose ends, add a large resistance to ground}
    pPar^.pIlist^.add('C  avoid loose ends, apparent impedance to ground');
    texta := '      ';          {ground}
    ImpedanceLines(named, namee, namef, texta, texta, texta, false,1.0e5,0.0,0.0,
                   linea, lineb, linec);
    {add the lines to the output list for impedances}
    pPar^.pIlist^.add(linea);
    pPar^.pIlist^.add(lineb);
    pPar^.pIlist^.add(linec);
    {generate lines for the initialization source}
    apu := Cabs(volt);              {voltage}
    fii := Cangle(volt)*180/pi;     {and its phase}
    linea := '14'+namej+'  '+TenChar(apu)+TenChar(freq)+TenChar(fii)+
             '          '+'          '+' -1.0     '+' 32.0     ';   {long enough ?}
    lineb := '14'+namek+'  '+TenChar(apu)+TenChar(freq)+TenChar(fii-120.0)+
             '          '+'          '+' -1.0     '+' 32.0     ';
    linec := '14'+namel+'  '+TenChar(apu)+TenChar(freq)+TenChar(fii+120.0)+
             '          '+'          '+' -1.0     '+' 32.0     ';
    pPar^.pVlist^.add('C  voltage source for the initialization of as.motor  '+lbl);
    pPar^.pVList^.add(linea);
    pPar^.pVList^.add(lineb);
    pPar^.pVlist^.add(linec);
  end;
end;

procedure NeutralName(jn : integer; Rg,Xg : single; var namen : string);
{generates the name for a neutran node. The name is six char. long and
 begins with 'XX'. If the grounding impedance Rg,Xg is zero, name = '      '
 jn     the node ID (number)
 Rg,Xg  grounding impedance
 namen  the name of the node}
var
  ii          : integer; {index}
  len         : integer; {length of text}
  txa,txb,txc : string;  {temporary}
begin
  if (Rg <> 0) or (Xg <> 0) then begin
    NodeNames(jn,txa,txb,txc);
    len := length(txa);
    if (len > 1) then begin      {shoult be exactly 6}
      namen := 'XX';
      for ii := 2 to len-1 do namen := namen+txa[ii];
    end else begin
      namen := txa;              {just some name}
    end;
  end else begin                 {connected directly to earth}
    namen := '      ';
  end;
end;

procedure TrafoATP(ipPar : longint; connect : string);
{generate the ATP-lines for a 2-winding trafo
 ipPar    pointer to the lines etc, as integer
 connect  the connection of the trafo YY, YD, DY, DD}
var
  apu,apv     : single;    {temporary}
  err         : integer;   {error flag}
  flux        : single;    {flux in Weber turns (?)}
  freq        : single;    {net frequency}
  Ion,Io      : single;    {nominal and final open circuit prim.current}
  ipSet       : longint;   {pointer to the "common" as integer}
  ix,iy       : integer;   {temporary}
  jpc,jsc     : integer;   {prim., sec. current node}
  jpn,jsn     : integer;   {prim., sec. neutral node}
  jpv,jsv     : integer;   {prim., sec. voltage node}
  lbl         : string;    {label of the trafo}
  line        : string;    {a line for the file}
  namepa      : string;    {name of primary (voltage) node}
  namepb      : string;    {name of primary (voltage) node}
  namepc      : string;    {name of primary (voltage) node}
  namepda     : string;    {name of primary downstream node in phase a}
  namepdb     : string;
  namepdc     : string;
  npla,nplb   : string;    {name of prim. "nonlinearity" node}
  nplc        : string;
  namepn      : string;    {name of primary neutral node}
  namesa      : string;    {name of sec. (voltage) node}
  namesb      : string;    {name of sec. (voltage) node}
  namesc      : string;    {name of sec. (voltage) node}
  namesda     : string;    {name of secondary downstream node in phase a}
  namesdb     : string;
  namesdc     : string;
  nsla,nslb   : string;    {name of sec. "nonlinearity" node}
  nslc        : string;
  namesn      : string;    {name of sec. neutral node}
  omega       : single;    {angular frequency}
  pComp       : TpComp;    {pointer to the trafo}
  pPar        : TpEdat;
  pSet        : TpCompSet; {pointer to the "common"}
//r01         : single;    {zero seq. impedance / pos.seq. impedance}
  R1n         : single;    {nominal primary resistance}
  R1,R2       : single;    {prim., sec. resistance}
  R1g,R2g     : single;    {prim., sec. grounding resistance}
  R0          : single;    {zero seq. resistance}
  Rz          : single;    {zero sequence (apparent) resistance}
  tap         : single;    {tap changer position, as fraction}
  V1n         : single;    {nominal primary voltage}
  V1,V2       : single;    {prim., sec. phase to phase voltage}
  XLn,XL      : single;    {nominal and final leakage reactance}
  X1,X2       : single;    {prim., sec. (leakage) reactance}
  X1g,X2g     : single;    {prim., sec. groudnig reactance}
  X0          : single;    {zero seq.reactance}
  Xz          : single;    {zero sequence (apparent) reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);          {pointer to data}
    ipSet := pPar^.ipSet;           {pointer to the common, as integer}
    pSet := TpCompSet(ipSet);       {pointer to the common}
    freq := pSet^.appl_Hz;          {net frequency}
    omega := 2*pi*freq;             {angular freq}
    pComp := pPar^.pComp;           {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);     {name}
    {nodes}
    pComp^.getNode(1,jpv,err);      {prim.voltage node}
    pComp^.getNode(2,jsv,err);      {sec. voltage node}
    pComp^.getNode(3,jpc,err);      {prim.current node}
    pComp^.getNode(4,jsc,err);      {sec. current node}
    pComp^.getNode(5,jpn,err);      {prim.neutral node}
    pComp^.getNode(6,jsn,err);      {sec. neutral node}
    {parameters}
    pComp^.GetPar(gTrafo_V1,apu);   {get value of prim. voltage}
    if (connect = 'YY') or (connect = 'YD') then V1n := apu/sqrt(3)
    else V1n := apu;
    pComp^.GetPar(gTrafo_V2,apu);   {get value of sec.  voltage}
    if (connect = 'YY') or (connect = 'DY') then V2 := apu/sqrt(3)
    else V2 := apu;
    pComp^.GetPar(gTrafo_X,XLn);    {get value of leak. reactance}
    if (connect = 'DY') or (connect = 'DD') then XLn := 3*XLn; {sqrt3 squared 11.12.2003}
    pComp^.GetPar(gTrafo_Io,Ion);   {open circuit prim. current}
    pComp^.GetPar(gTrafo_R1,R1n );  {get value of prim. resistance}
    if (connect = 'DY') or (connect = 'DD') then R1n := 3*R1n;
    R2 := 0;
    {take tap changer position into account}
    {tap := TrafoTap(pComp,connect);      {position of tap changer}
    pComp^.GetPref(tap,apu);
    V1 := V1n/(1+tap);
    R1 := R1n/(1+tap);
    XL := XLn/(1+tap);
    Io := Ion/(1+tap);
    {reactances cannot be zero, divide leakage reactance to primary and secondary}
    if (XL <= 0) then begin
      X1 := 0.00001;                {just some small value}
      X2 := 0.00001;
    end else begin
      if (V1 > 0) and (V2 > 0) then begin  {should be}
        apu := V1/V2;
        apu := apu*apu;             {impedance transformation ratio}
        apv := 0.875;               {fraction of reactance in primary}
        X1 := XL*apv;               {primary ractance}
        X2 := XL*(1-apv)/apu;       {secondary reactance}
      end else begin
        X1 := XL;
        X2 := 0.00001;
      end;
    end;
//    {ratio of zero sequence impedance to pos.seq. impedance: modeled as earthing impedance}
//    pComp^.GetPar(gTrafo_01,r01);    {zero seq. imped / pos.seq.imped}
//    if (r01 <= 0) then r01 := 1;
//    Rz := (r01-1)*R1/3;           {include zero seq. imped. into grounding imped}
//    Xz := (r01-1)*XL/3;
    {zero seq. impedance: modeled as part of earthing impedance}
    pComp^.GetPar(gTrafo_X0,X0);
    pComp^.GetPar(gTrafo_R0,R0);
    Rz := (R0-R1)/3;           {include zero seq. imped. into grounding imped}
    Xz := (X0-XL)/3;
    {names of (voltage) nodes}
    NodeNames(jpv,namepa,namepb,namepc); {primary}
    NodeNames(jsv,namesa,namesb,namesc); {secondary}
    {names of neutral nodes for Y-connected windings}
    if (connect = 'YY') or (connect = 'YD') then begin
      {get grounding impedances for Y-connected primary}
      pComp^.GetPar(gTrafo_R1g,R1g);     {prim. resistance to ground}
      pComp^.GetPar(gTrafo_X1g,X1g);     {prim. reactance to ground}
      NeutralName(jpn, R1g+Rz, X1g+Xz, namepn); {include zero seq. impedance}
    end else begin
      R1g := 0;
      X1g := 0;
      namepn := '      ';             {actually not needed}
    end;
   {Y-connected secondary}
    if (connect = 'YY') or (connect = 'DY') then begin
      pComp^.GetPar(gTrafo_R2g,R2g);     {sec.  resistance to ground}
      pComp^.GetPar(gTrafo_X2g,X2g);     {sec.  reactance to ground}
      if (connect = 'DY') then begin
        NeutralName(jsn, R2g+Rz, X2g+Xz, namesn); {include zero seq.imped.}
      end else begin
        NeutralName(jsn, R2g, X2g, namesn); {zero seq. imped in on the primary side}
      end;
    end else begin
      R2g := 0;
      X2g := 0;
      namesn := '      ';
    end;
    {names of "nonlinearity" nodes, only formally}
    NodeNames(jpc,npla,nplb,nplc);       {primary nonlin. node}
    NodeNames(jsc,nsla,nslb,nslc);       {second. nonlin. node}
    {names of downstream nodes}
    if (connect = 'YY') then begin  {Y-conn. primary}
      namepda := namepn;
      namepdb := namepn;
      namepdc := namepn;
      namesda := namesn;
      namesdb := namesn;
      namesdc := namesn;
    end else if (connect = 'YD') then begin
      namepda := namepn;
      namepdb := namepn;
      namepdc := namepn;
      namesda := namesb;                  {D-connected primary}
      namesdb := namesc;
      namesdc := namesa;
    end else if (connect = 'DY') then begin
      namepda := namepc;                  {D-connected primary}
      namepdb := namepa;
      namepdc := namepb;
      namesda := namesn;
      namesdb := namesn;
      namesdc := namesn;
    end else begin                        {DD}
      namepda := namepb;                  {D-connected primary}
      namepdb := namepc;
      namepdc := namepa;
      namesda := namesb;                  {D-connected secondary}
      namesdb := namesc;
      namesdc := namesa;
    end;
    {produce the ATP lines}
    flux := V1/omega;
    pPar^.pIlist^.add('C Transformer '+lbl+'- - - - - - - - - - - - - - ');
    line := '  TRANSFORMER             '+SixChar(Io)+SixChar(flux)+npla+
            '                                   0';
    pPar^.pIlist^.add(line);
    pPar^.pIlist^.add('          '+SixChar(Io)+'          '+SixChar(flux));
    {linear "characteristics"}
    pPar^.pIlist^.add('            9999');
    line := ' 1'+namepa+namepda+'            '+SixChar(R1)+SixChar(X1)+SixChar(V1);
    pPar^.pIlist^.add(line);             {prim. impedance+voltage}
    line := ' 2'+namesa+namesda+'            '+SixChar(R2)+SixChar(X2)+SixChar(V2);
    pPar^.pIlist^.add(line);             {sec.impedance+voltage}
    pPar^.pIlist^.add('  TRANSFORMER '+npla+'                  '+nplb);
    pPar^.pIlist^.add(' 1'+namepb+namepdb);
    pPar^.pIlist^.add(' 2'+namesb+namesdb);
    pPar^.pIlist^.add('  TRANSFORMER '+npla+'                  '+nplc);
    pPar^.pIlist^.add(' 1'+namepc+namepdc);
    pPar^.pIlist^.add(' 2'+namesc+namesdc);
    {earthing impedances for Y-connected windings, if impdance > 0}
    if (connect = 'YY') or (connect = 'YD') then begin       {y-connected primary}
      if (R1g+Rz <> 0) or (X1g+Xz <> 0) then begin
        pPar^.pIlist^.add('C  prim. earthing impedance of trafo '+lbl);
        pPar^.pIlist^.add('  '+namepn+'      '+'      '+'      '+
                          SixSignedChar(R1g+Rz)+SixSignedChar(X1g+Xz));
      end;
    end;
    if (connect = 'YY')  then begin            {y-connected secondary}
      if (R2g <> 0) or (X2g <> 0) then begin
        pPar^.pIlist^.add('C  sec. earthing impedance of trafo '+lbl);
        pPar^.pIlist^.add('  '+namesn+'      '+'      '+'      '+SixChar(R2g)+SixChar(X2g));
      end;
    end else if (connect = 'DY') then begin
      if (R2g+Rz <> 0) or (X2g+Xz <> 0) then begin     {include zero seq. impedance}
        if (V1 > 0) and (V2 > 0) then apu := V2/V1 else apu := 1;
        apu := apu*apu;
        pPar^.pIlist^.add('C  sec. earthing impedance of trafo '+lbl);
        pPar^.pIlist^.add('  '+namesn+'      '+'      '+'      '+
                          SixSignedChar(R2g+Rz*apu)+SixSignedChar(X2g+Xz*apu));
      end;
    end;
    {for delta windings, add a large impedance to ground, to avoid floating networks}
    if (connect = 'DY') then begin
      if (namepa <> '      ') then begin        {not ground}
        pPar^.pIlist^.add('C  large impedances to ground from D-winding ');
        pPar^.pIlist^.add('  '+namepa+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namepb+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namepc+'      '+'      '+'      '+'1.0e+6');
      end;
    end else if (connect = 'YD') then begin
      if (namesa <> '      ') then begin        {not ground}
        pPar^.pIlist^.add('C  large impedances to ground from D-winding ');
        pPar^.pIlist^.add('  '+namesa+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namesb+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namesc+'      '+'      '+'      '+'1.0e+6');
      end;
    end else if (connect = 'DD') then begin
      if (namepa <> '      ') then begin        {not ground}
        pPar^.pIlist^.add('C  large impedances to ground from D-windings ');
        pPar^.pIlist^.add('  '+namepa+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namepb+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namepc+'      '+'      '+'      '+'1.0e+6');
      end;  
      if (namesa <> '      ') then begin        {not ground}
        pPar^.pIlist^.add('C  large impedances to ground from D-windings ');
        pPar^.pIlist^.add('  '+namesa+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namesb+'      '+'      '+'      '+'1.0e+6');
        pPar^.pIlist^.add('  '+namesc+'      '+'      '+'      '+'1.0e+6');
      end;  
    end;
  end;
end;

procedure TrafoYYATP(ipPar : longint);
{generate the ATP-lines for a YY-trafo}
begin
  TrafoATP(ipPar, 'YY');
end;

procedure TrafoYDATP(ipPar : longint);
{generate the ATP-lines for a YY-trafo}
begin
  TrafoATP(ipPar, 'YD');
end;

procedure TrafoDYATP(ipPar : longint);
{generate the ATP-lines for a YY-trafo}
begin
  TrafoATP(ipPar, 'DY');
end;

procedure TrafoDDATP(ipPar : longint);
{generate the ATP-lines for a YY-trafo}
begin
  TrafoATP(ipPar, 'DD');
end;

procedure Trafo3WATP(ipPar : longint);
{generate the ATP lines for a 3-winding transformer}
var
  apu,apv,apw : single;    {temporary}
  err         : integer;   {error flag}
  flux        : single;    {flux in Weber turns (?)}
  freq        : single;    {net frequency}
  Ion,Io      : single;    {nominal and final open circuit prim.current}
  ipSet       : longint;   {pointer to the "common" as integer}
  isYYD       : boolean;   {flag for trafo type}
  ix,iy       : integer;   {temporary}
  jpc,jsc,jtc : integer;   {prim., sec., tert. current node}
  jpn,jsn,jtn : integer;   {prim., sec., tert. neutral node}
  jpv,jsv,jtv : integer;   {prim., sec., tert. voltage node}
  lbl         : string;    {label of the trafo}
  line        : string;    {a line for the file}
  namepa      : string;    {name of primary (voltage) node of phase a}
  namepb      : string;    {name of primary (voltage) node}
  namepc      : string;    {name of primary (voltage) node}
  namepda     : string;    {name of primary downstream node in phase a}
  namepdb     : string;
  namepdc     : string;
  npla,nplb   : string;    {name of prim. "nonlinearity" node}
  nplc        : string;
  namepn      : string;    {name of primary neutral node}
  namesa      : string;    {name of sec. (voltage) node}
  namesb      : string;    {name of sec. (voltage) node}
  namesc      : string;    {name of sec. (voltage) node}
  namesda     : string;    {name of secondary downstream node in phase a}
  namesdb     : string;
  namesdc     : string;
  nsla,nslb   : string;    {name of sec. "nonlinearity" node}
  nslc        : string;
  namesn      : string;    {name of sec. neutral node}
  nameta      : string;    {name of tert. (voltage) node}
  nametb      : string;    {name of tert. (voltage) node}
  nametc      : string;    {name of tert. (voltage) node}
  nametda     : string;    {name of tert. downstream node in phase a}
  nametdb     : string;
  nametdc     : string;
  ntla,ntlb   : string;    {name of tert. "nonlinearity" node}
  ntlc        : string;
  nametn      : string;    {name of tert. neutral node}
  omega       : single;    {angular frequency}
  pComp       : TpComp;    {pointer to the trafo}
  pPar        : TpEdat;
  pSet        : TpCompSet; {pointer to the "common"}
  r01,r02,r03 : single;    {zero seq. impedance / pos.seq. impedance}
  R1,R2,R3    : single;    {resistance of prim, sec, tert. windings}
  R1g,R2g,R3g : single;    {prim., sec., tert. grounding resistance}
  Rz1,Rz2,Rz3 : single;    {zero sequence (apparent) resistance}
  tap         : single;    {tap changer position, as fraction}
  V1,V2,V3    : single;    {prim., sec., tert.  voltage over the winding}
  XLn,XL      : single;    {nominal and final leakage reactance}
  X1,X2,X3    : single;    {prim., sec., tert. (leakage) reactance}
  X12,X13,X23 : single;    {(total) leakage reactances of the winding pairs}
  X1g,X2g,X3g : single;    {prim., sec. groudnig reactance}
  Xz1,Xz2,Xz3 : single;    {zero sequence (apparent) reactance}
begin
  if (ipPar <> 0) then begin
    pPar := TpEdat(ipPar);          {pointer to data}
    ipSet := pPar^.ipSet;           {pointer to the common, as integer}
    pSet := TpCompSet(ipSet);       {pointer to the common}
    freq := pSet^.appl_Hz;          {net frequency}
    omega := 2*pi*freq;             {angular freq}
    pComp := pPar^.pComp;           {pointer to the impedance}
    pComp^.GetLabel(lbl,ix,iy);     {name}
    {nodes}
    pComp^.getNode(1,jpv,err);      {prim.voltage node}
    pComp^.getNode(2,jsv,err);      {sec. voltage node}
    pComp^.getNode(3,jtv,err);      {tert. voltage node}
    pComp^.getNode(4,jpc,err);      {prim.current node}
    pComp^.getNode(5,jsc,err);      {sec. current node}
    pComp^.getNode(6,jtc,err);      {tert. current node}
    pComp^.getNode(7,jpn,err);      {prim.neutral node}
    pComp^.getNode(8,jsn,err);      {sec. neutral node}
    pComp^.getNode(9,jtn,err);      {tert. neutral node or circulating current node}
    {voltages}
    pComp^.GetPar(g3W_type,apu);
    isYYD := (apu = 2);             {flag for trafo type is YYD}
    pComp^.GetPar(g3W_V1,apu);      {get value of prim. voltage}
    V1 := apu/sqrt3;                {phase to neutral voltage = voltage over the winding}
    pComp^.GetPar(g3W_V2,apu);      {get value of sec.  voltage}
    V2 := apu/sqrt3;
    pComp^.GetPar(g3W_V3,apu);      {get value of tert.  voltage}
    if (isYYD) then V3 := apu else V3 := apu/sqrt3;  {voltage over winding}
    {impedances}
    pComp^.GetPar(g3W_X1,X12);      {get value of leak. reactance X1->2}
    pComp^.GetPar(g3W_X3,X13);      {get value of leak. reactance X1->3}
    pComp^.GetPar(g3W_X2,X23);      {get value of leak. reactance X2->3}
    {must calculate leakage reactances for each winding separately...}
    apu := X23*V1*V1/(V2*V2);       {normalize to V1}
    X1 := (X12+X13-apu)/2;
    apv := (X12-X13+apu)/2;
    X2 := apv*V2*V2/(V1*V1);
    apw := (apu-X12+X13)/2;
    X3 := apw*V3*V3/(V1*V1);        {at voltage over winding}
    {reactances cannot be zero}
    if (X1 <= 0) then X1 := 0.00001;
    if (X2 <= 0) then X2 := 0.00001;
    if (X3 <= 0) then X3 := 0.00001;
    {resistances of the windings}
    pComp^.GetPar(g3W_R1,R1 );      {get value of prim. resistance}
    pComp^.GetPar(g3W_R2,R2 );      {get value of sec. resistance}
    pComp^.GetPar(g3W_R3,R3 );      {get value of tert. resistance}
(**)
   {ratio of zero sequence impedance to pos.seq. impedance: modeled as earthing impedance}
    pComp^.GetPar(g3W_zpp12,r01);   {zero seq. imped / pos.seq.imped}
    if (r01 <= 0) then r01 := 1;
    Rz1 := (r01-1)*R1/3;            {include zero seq. imped. into grounding imped}
    Xz1 := (r01-1)*X1/3;      {or X12, or ...?}
    pComp^.GetPar(g3W_zpp13,r02);   {zero seq. imped / pos.seq.imped}
    if (r02 <= 0) then r02 := 1;
    Rz2 := (r02-1)*R2/3;            {include zero seq. imped. into grounding imped}
    Xz2 := (r02-1)*X2/3;
 Xz2 := (r01-1)*X2/3;
    {ratio for the third winding only if Y-winding}
    if (not isYYD) then begin
      pComp^.GetPar(g3W_zpp23,r03);    {zero seq. imped / pos.seq.imped}
      if (r03 <= 0) then r03 := 1;
    end else begin
      r03 := 1;
    end;
    {impedances in any case (perhaps not needed)}
    Rz3 := (r03-1)*R3/3;             {include zero seq. imped. into grounding imped}
    Xz3 := (r03-1)*X3/3;
//Xz3 := (r01-1)*X3/3;
(**)
(*
    {temporarily, use just some values...?}
    Rz1 := 0;
    Xz1 := 0;
    Rz2 := 0;
    Xz2 := 0;
    Rz3 := 0;
    Xz3 := 0;
*)
    {names of (voltage) nodes}
    NodeNames(jpv,namepa,namepb,namepc); {primary}
    NodeNames(jsv,namesa,namesb,namesc); {secondary}
    NodeNames(jtv,nameta,nametb,nametc); {tertiary}
    {names of neutral nodes for Y-connected windings}
    pComp^.GetPar(g3W_R1g,R1g);                 {prim. resistance to ground}
    pComp^.GetPar(g3W_X1g,X1g);                 {prim. reactance to ground}
    NeutralName(jpn, R1g+Rz1, X1g+Xz1, namepn); {include zero seq. impedance}
    pComp^.GetPar(g3W_R2g,R2g);                 {sec. resistance to ground}
    pComp^.GetPar(g3W_X2g,X2g);                 {sec. reactance to ground}
    NeutralName(jsn, R2g+Rz2, X2g+Xz2, namesn); {include zero seq. impedance}
    if (not isYYD) then begin
      pComp^.GetPar(g3W_R3g,R3g);               {prim. resistance to ground}
      pComp^.GetPar(g3W_X3g,X3g);               {prim. reactance to ground}
      NeutralName(jtn,R3g+Rz3,X3g+Xz3, nametn); {include zero seq. impedance}
    end else begin
      R3g := 0;
      X3g := 0;
      nametn := '      ';                       {actually not needed}
    end;
    {names of "nonlinearity" nodes, only formally}
    NodeNames(jpc,npla,nplb,nplc);       {primary nonlin. node}
    NodeNames(jsc,nsla,nslb,nslc);       {second. nonlin. node}
    NodeNames(jtc,ntla,ntlb,ntlc);       {tert.   nonlin. node}
    {names of downstream nodes}
    namepda := namepn;
    namepdb := namepn;
    namepdc := namepn;
    namesda := namesn;
    namesdb := namesn;
    namesdc := namesn;
    if (not isYYD) then begin           {YYY-trafo}
      nametda := nametn;
      nametdb := nametn;
      nametdc := nametn;
    end else begin                      {YYD-trafo}
      nametda := nametb;                {D-connected primary}
      nametdb := nametc;
      nametdc := nameta;
    end;
    {produce the ATP lines}
    flux := V1/omega;
    Io := pComp^.var2_s;                {saved in trafo3wimp of elplex.pas}
    pPar^.pIlist^.add('C Transformer '+lbl+'- - - - - - - - - - - - - - ');
    line := '  TRANSFORMER             '+SixChar(Io)+SixChar(flux)+npla+
            '                                   0';
    pPar^.pIlist^.add(line);
    pPar^.pIlist^.add('          '+SixChar(Io)+'          '+SixChar(flux));
    {linear "characteristics"}
    pPar^.pIlist^.add('            9999');
    line := ' 1'+namepa+namepda+'            '+SixChar(R1)+SixChar(X1)+SixChar(V1);
    pPar^.pIlist^.add(line);             {prim. impedance+voltage}
    line := ' 2'+namesa+namesda+'            '+SixChar(R2)+SixChar(X2)+SixChar(V2);
    pPar^.pIlist^.add(line);             {sec.impedance+voltage}
    line := ' 3'+nameta+nametda+'            '+SixChar(R3)+SixChar(X3)+SixChar(V3);
    pPar^.pIlist^.add(line);             {tert.impedance+voltage}
    pPar^.pIlist^.add('  TRANSFORMER '+npla+'                  '+nplb);
    pPar^.pIlist^.add(' 1'+namepb+namepdb);
    pPar^.pIlist^.add(' 2'+namesb+namesdb);
    pPar^.pIlist^.add(' 3'+nametb+nametdb);
    pPar^.pIlist^.add('  TRANSFORMER '+npla+'                  '+nplc);
    pPar^.pIlist^.add(' 1'+namepc+namepdc);
    pPar^.pIlist^.add(' 2'+namesc+namesdc);
    pPar^.pIlist^.add(' 3'+nametc+nametdc);
    {earthing impedances for Y-connected windings, if impedance > 0}
    {primary winding}
    if (R1g+Rz1 <> 0) or (X1g+Xz1 <> 0) then begin
      pPar^.pIlist^.add('C  prim. earthing impedance of trafo '+lbl);
      pPar^.pIlist^.add('  '+namepn+'      '+'      '+'      '+
                        SixSignedChar(R1g+Rz1)+SixSignedChar(X1g+Xz1));
    end;
    {secondary winding}
{?} if (R2g+Rz2 <> 0) or (X2g+Xz2 <> 0) then begin
      pPar^.pIlist^.add('C  sec. earthing impedance of trafo '+lbl);
      pPar^.pIlist^.add('  '+namesn+'      '+'      '+'      '+
{?}                     SixChar(R2g+Rz2)+SixChar(X2g+Xz2));     {Rz2, Xz2 included? (NOT in YY-trafo?}
    end;
    {tertiary winding}
    if (not isYYD) then begin    {YYY-trafo}
{?}   if (R3g+Rz3 <> 0) or (X3g+Xz3 <> 0) then begin
        pPar^.pIlist^.add('C  tert. earthing impedance of trafo '+lbl);
        pPar^.pIlist^.add('  '+nametn+'      '+'      '+'      '+
{?}                     SixChar(R3g+Rz3)+SixChar(X3g+Xz3));     {Rz3, Xz3 included? (NOT in YY-trafo?}
      end;
    end;  
  end;
end;




procedure AtpFile(ipSet : longint; FileName : string);
{produce the ATP-file of the diagram
 ipSet     pointer to the "common" as integer
 FileName  name of file where data is written}
const
  timestep = ' 50.0e-6';     {time step as string, must be 8 char.}
var
  apu,apv,apw : single;         {temporary}
  AsmotList   : TStringList;   {lines for asynch.motors}
  ATPdat      : TEdat;         {for transferring the parameters}
  AtpPro      : TCompProc;     {procedure for the *.atp file of each type of comp}
{  dt          : single;}
  errc        : integer;       {error flag}
  EarthList   : TList;         {list of earthing components}
  FaultList   : TStringList;   {lines for the fault(s) = swithc(es) to the *.atp file}
  GenerList   : TStringList;   {lines for the *.atp file from generators}
  ii          : integer;       {index}
  ImpedList   : TStringList;   {lines for the *.atp file from impedances}
  ipar        : integer;       {no of parameters}
  jcomp       : integer;       {counts components}
  len         : integer;       {length of text}
  line        : string;
  name        : string;        {name (type) of a component}
  nAsMot      : integer;       {no of asynchronous motors}
  nCurrT      : integer;       {no of current trafos}
  nGenMot     : integer;       {no of synchronous machines}
  nPQRel      : integer;       {PGQ-relays}
  nRelay      : integer;       {no of relays}
  nShort      : integer;       {no of faults}
  nSource     : integer;       {no of sources (feeders)}
  Outfile     : TextFile;      {file where written}
  pComp       : TpComp;        {pointer to a component}
  pSet        : TpCompSet;     {pointer to the "common"}
  SourcList   : TStringList;   {lines for sources}
  StarList    : TList;         {list of wye-termination comp}
  st1,st2     : string;        {temporary}
  st3,st4     : string;        {temporary}
  TacsList    : TStringList;   {list for "Tacs" for the *.atp file}
  text        : string;
begin
  pSet := TpCompSet(ipSet); {pointer from integer}
  {count certain types of components (especially asynch.motors)}
  CountComp(ipSet, nGenMot, nRelay, nShort,nCurrT,nAsMot, nSource,nPQRel);
//if (pSet^.p1SchComp <> nil) then begin   {there are components}
  if (pSet^.pCompList.count > 0) then begin
    GenerList := TStringList.create;
    ImpedList := TStringList.create;
    FaultList := TStringList.create;
    SourcList := TStringList.create;
    AsmotList := TStringList.create;
    TacsList  := TStringList.create;
    StarList  := TList.Create;       {list of wye-termination components}
    EarthList := TList.Create;       {list oe earthing impedances}
    {determine nodes (to be sure) = "Click" button "NodeBut"on form "Imagoo"}
(*  Imagoo.NodeButClick(Imagoo.NodeBut); {TURHA, koska laskettava ensin kuitenkin !} *)
    {collect data for the *.dat procedure}
    AtpDat.pIlist := @ImpedList;
    AtpDat.pGlist := @GenerList;
    AtpDat.pSlist := @FaultList;
    AtpDat.pVlist := @SourcList;
    AtpDat.pAlist := @AsmotList;
    AtpDat.pTlist := @TacsList;
    AtpDat.StarList := StarList;
    AtpDat.EarthList := EarthList;
    AtpDat.ipSet := ipSet;              {pointer to the "common"}
    AtpDat.branch := 0;                 {initialize for transmission lines}
    AtpDat.dt := 1;                     {initialize, just some value}
    AtpDat.nAsMot := nAsMot;            {no of asynch.motors}
    AtpDat.nGenMot := nGenMot;          {no of synch.machines}
    ValuesFromLine(timestep , 1, false,ipar, AtpDat.dt, apu, apv, apw,st1,st2,st3,st4, errc); {time step}
    {scan all components and search for stars and earthing impedances}
    pComp := nil;
    errc := 0;
    jcomp := 0;
    while(errc = 0) do begin                 {scan all components}
      pSet^.NextComp(pComp,errc,jcomp,true); {get pointer to next component}
      if (errc = 0) then begin
        {fill lists of Stars and earthing imp.}
        pComp^.GetName(name);
        if (name = gStar) then begin
          StarList.add(pComp);
        end else if (name = gGrImped) then begin
          EarthList.add(pComp);
          pComp^.int1_s := 0;         {reset, used as a flag}
        end else begin
          pComp^.int1_s := -1;        {initialize "pointes" to the star comp.}
          pComp^.int2_s := -1;
          pComp^.int3_s := -1;
        end;
      end;
    end;
    {scan all components}
    pComp := nil;
    errc := 0;
    jcomp := 0;
    while(errc = 0) do begin                 {scan all components}
      pSet^.NextComp(pComp,errc,jcomp,true); {get pointer to next component}
      if (errc = 0) then begin
        FindStar(pComp, StarList);  {star connected to this component ?}
        AtpDat.pComp := pComp;      {copy also the pointer to the component}
        pComp^.GetAtpPro(AtpPro);   {get procedure for constructing *dat file}
        AtpPro(longint(@AtpDat));   {*.dat lines of this component}
      end;
    end;
    {write the collected data to file}
   {open the given file,BUT ERRORS ?}
    AssignFile(Outfile, FileName);
    rewrite(outfile);
    writeln(outfile,'BEGIN NEW DATA CASE');     {file must start with this}
    writeln(outfile,'C --------------------------------------------------------');
    {the title, but at most 66 char. long (could be longer ?, with ATP, I think)}
    text := 'C Title: '+Imagoo.NetTitle.text;
    len := length(text);
    if (len <  66) then begin         {too short, must add spaces}
      line := text;
      for ii := len+1 to 66 do begin
        line := line+' ';
      end;
    end else if (len = 66) then begin {exactly right length}
      line := text;
    end else begin                    {too long, must cut}
      line := '';
      for ii := 1 to 66 do begin
        line := line+text[ii];
      end;
    end;
    writeln(outfile,line);            {the title}
    writeln(outfile,'C --------------------------------------------------------');
    {must reserve space, if more than three asynchronous machines}
    if (nAsMot > 3) then begin
      line := 'ABSOLUTE U.M. DIMENSIONS, '+inttostr(5*nAsMot)+', '+inttostr(nAsMot)+
              ', '+inttostr(6*nAsMot)+', '+inttostr(20+nAsMot)+',';
      writeln(outfile,line);
    end;
    writeln(outfile,'ALLOW EVEN PLOT FREQUENCY');        {unexplained}
    writeln(outfile,'C NOTE: if dT is changed, change length of line of each induction motor !');
    writeln(outfile,'C  dT  >< Tmax >< Xopt >< Copt >'); {explanations}
    line := '  '+SixChar(pSet^.appl_hz);
    writeln(outfile,timestep+'    3.00'+  line +  line);
    writeln(outfile,'C out:h>< plot:h');
    writeln(outfile,'      10       5       1       1       1       0       0       1       0');
    writeln(outfile,'C        1         2         3         4         5         6         7         8');
    writeln(outfile,'C 345678901234567890123456789012345678901234567890123456789012345678901234567890');
    {impedances - - - - - - - - - - - - - - - - - -}
    if (TacsList.count > 0) then begin
      writeln(outfile,'C  TACS for the voltage of synchronous generators');
      writeln(outfile,'TACS HYBRID');
      writeln(outfile,'/TACS');
      for ii := 0 to TacsList.count-1 do begin
        writeln(outfile,TacsList.Strings[ii]);
      end;
    writeln(outfile,'BLANK TACS');
    end;

    {impedances - - - - - - - - - - - - - - - - - -}
    writeln(outfile,'C impedances - - - - - - - - - - - - - - - - - - - - -');
    writeln(outfile,'/BRANCH');
    if (ImpedList.count > 0) then begin
      for ii := 0 to ImpedList.count-1 do begin
        writeln(outfile,ImpedList.Strings[ii]);
      end;
    end;
    { a 'blank' line terminates branch data}
    writeln(outfile,'BLANK BRANCH');
    {swithc(es) - - - - - - - - - - - - - - - - - }
    writeln(outfile,'/SWITCH');
    {start with an explanatory text}
    if (FaultList.count > 0) then begin
      line := 'C <node><node>< tclose >< topen  >';
      writeln(outfile,line);
      for ii := 0 to FaultList.count-1 do begin
        writeln(outfile,FaultList.Strings[ii]);
      end;
    end;
    {"empty" line, in any case, to indicate end of switch data}
    writeln(outfile,'BLANK SWITCH');
    {sources}
    writeln(outfile,'C   Sources ');
    writeln(outfile,'/SOURCE');
    writeln(outfile,'C < n 1><>< Ampl.  >< Freq.  ><Phase/T0><   A1   ><   T1   >< TSTART >< TSTOP  >');
    if (SourcList.count > 0) then begin
      for ii := 0 to SourcList.count-1 do begin
        writeln(outfile,SourcList.Strings[ii]);
      end;
    end;
    {generators}
    if (GenerList.count > 0) then begin
      for ii := 0 to GenerList.count-1 do begin
        writeln(outfile,GenerList.Strings[ii]);
      end;
    end;
    {asynch.motors}
    if (AsmotList.count > 0) then begin
      writeln(outfile,'C universal machines (induction motors) - - - - - -');
      writeln(outfile,'19 UM');
      writeln(outfile,'00            0');
      writeln(outfile,'BLANK general UM specification');
      for ii := 0 to AsmotList.count-1 do begin
        writeln(outfile,AsmotList.Strings[ii]);
      end;
      writeln(outfile,'BLANK UM');
    end;
    writeln(outfile,'BLANK SOURCE');
{   writeln(outfile,'$ end of machine data - - - - - - - - - - - - ');}
    {voltage output}
    writeln(outfile,'/OUTPUT');
    writeln(outfile,'C voltage saved in following nodes');
    writeln(outfile,'C node1 node2 node3 etc.  six charaters/node ');

    writeln(outfile,'BLANK OUTPUT');
    {end of data}
    writeln(outfile,'BEGIN NEW DATA CASE');
    writeln(outfile,'BLANK');
    CloseFile(Outfile);
    {finally, free the lists}
    GenerList.free;
    ImpedList.free;
    FaultList.free;
    SourcList.free;
    AsmotList.free;
    TacsList.free;
    StarList.free;
    EarthList.free;
  end;
end;


end.
