unit Cmatmul;
interface
  uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, StdCtrls, Buttons,
  Compdef,Utildef;

(*type
  pSingle = ^Single;
  pInteger = ^Integer;*)

procedure Csqumul(paa,pbb,pcc:pComplex; nn,nj: integer);
procedure putCmat(paa:pComplex;i,j,m,n:Integer;xx:Complex);
procedure getCmat(paa:pComplex;i,j,m,n:Integer;var xx:Complex);
procedure putCvec(paa:pComplex;i,m:Integer;xx:Complex);
procedure getCvec(paa:pComplex;i,m:Integer;var xx:Complex);
procedure copyCvec(paa,pbb:pComplex;nn:Integer);
procedure copyCvecCol(pbb,paa : pComplex; jj,nn : integer);
procedure clrCrow(paa:pComplex; i,nn :Integer);
procedure putImat(paa:pInteger;i,j,m,n:Integer;xx:Integer);
procedure getImat(paa:pInteger;i,j,m,n:Integer;var xx:Integer);
procedure CLUdecomp(paa ,pvv:pComplex;
                   Nn:Integer; pindx:pInteger;
                   var Dodd:Single; var Sngl: Integer;
                   var Tout:Complex);
procedure CLUbacksub(paa  : pComplex; Nn :Integer;
                     pindx: pInteger;
                     pbb  : pComplex);
function comptostr(cpu : complex): string;
procedure copyCmat(paa,pzz : pComplex; nn: integer);
procedure imgMat(paa : pComplex; pzz : pSingle; nn: integer);
procedure copyCrow(paa,pzz : pComplex; ir,nn : integer);
procedure clearCmat(paa : pComplex; nn : integer);
procedure clearCvec(pbb : pComplex; nn : integer);
function maxCvec(paa : pComplex; nn : integer) : complex;
procedure showCmatElements(paa : pComplex; mm,nn : integer);

implementation
{uses
  matest;}

procedure Csqumul(paa,pbb,pcc:pComplex; nn,nj: integer);
{this procedure multiplies a complex matrix or vector
                by a square complex matrix
    cc = aa*bb, where aa(nn,nn) is the square matrix
                      bb(nn,nj) is a more general matrix or vector
                      cc(nn,nj) is the result matrix (or vector
 NOTE: no checks of the dimensions are done, be sure that
       the dimensions (especially ) of cc are correct !!!
 paa,pbb,pcc  pointers to square matrices aa,bb,cc (in)
              but actually aa, bb are inputs, result to cc
 nn           dimension of the matrices (in)
 NOTE: pComplex is a pointer to a record of type "complex"
       *.x is the real part, and *.y is the imaginary part
       it is defined in unit matdef.pas}
var
  pai,pbj,paij,pbij,pcij: pComplex; {local pointers to rows, columns, elements}
  i,j,k: Integer;  {indices}
  oo   : Complex;  {used as zomplex zero}
  apu  : complex;  {temporary}
begin
  {define complex zero}
  oo.x := 0; oo.y := 0;
  {initialize local pointers to first element in the matrices}
  pai := paa;
  pbj := pbb;
  pcij := pcc;
  {loop over rows}
  for i:=1 to nn do begin
    {loop over columns}
    for j:=1 to nj do begin
      apu := oo;
      pbij := pbj;  {pointer to first element in column j of bb}
      paij := pai;  {pointer first element in row i of aa}
      for k := 1 to nn do begin
        apu := Cadd(apu,Cmult(paij^,pbij^)); {apu := apu+aa(i,k)*bb(k,j)}
        inc(paij);              {next element in aa = next in row i}
        inc(pbij,nj);           {next element in bb = next in column j}
      end;
      pcij^:= apu; {cc[i,j] := apu}
      inc(pcij);   {next element in cc}
      inc(pbj);    {next column in bb}
    end;
    pbj := pbb;    {again first column in bb}
    inc(pai,nn);   {next row in aa}
  end;
end;

procedure CLUdecomp(paa :pComplex;   {pointer to the matrix to be be decomposed
                                     original in, decomposed out}
                   pvv  :pComplex;   {pointer to working space,
                                     for implicit scaling of each row}
                   Nn   :Integer;   {dimension of the matrix, in}
                   pindx:pInteger;  {pointer to an array recording the
                                     row permutations}
                   var Dodd:Single; { -1, if odd number of permutaions, else 1}
                   var Sngl:Integer; {flag for singular matrix}
                   var Tout:Complex); {test-output}
{LUdecomp  decomposes (in place) a matrix into LU-form}
const
  tiny=1.0e-10;
var
  apu      : Real;     {temporary}
  big      : Complex;  {max. element, temporary}
  Cone     : Complex;  {complex one}
  Ctiny    : Complex;  {small complex number}
  i,j,k    : Integer;  {indices}
  imax     : Integer;  {max index}
  oo       : Complex;  {complex zero}
  pi,pj,pij: pComplex; {pointers to row i, column j, element i,j of matrix a}
  pjj      : pComplex; {pointer to diagonal}
  pik,pkj  : pComplex; {pointers to element ik, kj  in matrix a}
  pjk      : pComplex; {pointer to element jk in matrix a}
  pindxj   : pInteger; {pointer to element j in indx}
  pimaxk   : pComplex; {pointer to element  imax,k in matrix a}
  prj      : pComplex; {pointer to row j}
  pvi      : pComplex; {pointer to element i in vv}
  pvimax   : pComplex; {pointer to element imax in vv}
  rbig     : Single;   {max abs. value}
  rdum     : Single;   {temporary, real}
  sum, dum : Complex;  {temporary}

begin
  try
  oo.x := 0; oo.y := 0;           {define complex zero}
  Cone.x := 1; Cone.y := 0;       {define complex one}
  Ctiny.x := tiny;                {define a small....}
  Ctiny.y := 0;                   {... complex number}
  Dodd := 1;
  Sngl := 0;
  {initialize vector vv, just for clarity}
  pvi := pvv;
  for i:=1 to Nn do begin
    {vv[i-1] := 0.0;} pvi^:= oo;
    inc(pvi);
  end;
  {check for zero rows}
  pij := paa;  {initialize pointer to first element in matrix a}
  pvi := pvv;
  for i:=1 to Nn do begin
    big := oo;
    for j:=1 to Nn do begin
      if (Cabs(pij^) > Cabs(big)) then big := pij^;
      inc(pij);  {pointer to next element in matrix a}
    end; {end of loop over j,  = over elements in a row}
    {preliminary check of obvious singularity}
    if (Cabs(big) <= tiny{0.0}) then begin
      Sngl := Sngl+1;
    end else begin
      {vv[i-1] := 1.0/big;} pvi^:= Cdiv(Cone,big);
    end;
    inc(pvi);
  end;   {end of loop over i,  = over rows}
  {proceed only if not obviously singlular matrix}
  if (Sngl = 0) then begin
    {start the decomposition}
    pj := paa;       {initialize column pointer to first column}
    pjj := paa;      {initialize pointer to diagonal element}
    prj := paa;      {initialize pointer to row j}
    pindxj := pindx; {initialize pointer to vector indx}
    for j:=1 to Nn do begin
      if (j > 1) then begin
        pi := paa;
        pij := pj;
        for i:=1 to j-1 do begin
          sum := pij^;
          pik := pi;
          pkj := pj;
          if (i > 1) then begin
            for k:=1 to i-1 do begin
              if ((pik^.x <> 0) or (pik^.y <> 0)) and
              ((pkj^.x <> 0) or (pkj^.y <> 0)) then            {test, may speedup}
                sum := Csub(sum,Cmult(pik^,pkj^));  { a(i,k)*a(k,j) }
              inc(pik);
              inc(pkj,Nn);
            end;
            pij^:= sum;
          end;
          inc(pij,Nn);
          inc(pi,Nn); {next row}
        end;
      end;
      rbig := 0.0;
      pij := pjj; {alkaa (j,j)st}
      pi  := prj; {paa; inc(pi,(j-1)*Nn);} {alkaa (j,1):st}
      pvi := pvv; inc(pvi,(j-1));
      imax := j;  {initialize 11.1.2002. Needed, if singular !}
      for i:=j to Nn do begin
        sum := pij^;
        if (j > 1) then begin
          pik :=  pi;  {alkaa (j,1).st}
          pkj :=  pj;  {alkaa (1,j):st}
          for k:= 1 to j-1 do begin
            if ((pik^.x <> 0) or (pik^.y <> 0)) and
            ((pkj^.x <> 0) or (pkj^.y <> 0)) then            {test, may speedup}
              sum := Csub(sum,Cmult(pik^,pkj^));
            inc(pik);
            inc(pkj,Nn);
          end;
          pij^:= sum;
        end;
        rdum := Cabs(pvi^)*Cabs(sum);
        if (rdum > rbig) then begin
          rbig := rdum;
          imax := i;
        end;
        inc(pvi);
        inc(pi,Nn);
        inc(pij,Nn);
      end;
      if (j <> imax) then begin
        pjk := paa;    inc(pjk,(j-1)*Nn);      {alkaa (j,1):st}
        pimaxk := paa; inc(pimaxk,(imax-1)*Nn); {alkaa (imax,1)st}
        for k:=1 to Nn do begin
          dum := pimaxk^;     {dum := a[imax,k] }
          pimaxk^ := pjk^;    {a[imax,k] := a[j,k] }
          pjk^ := dum;        {a[j,k] := dum  }
          inc(pimaxk);
          inc(pjk);
        end;
        Dodd := -Dodd;
        pvi := pvv; inc(pvi,j-1);
        pvimax := pvv; inc(pvimax,imax-1);
        {vv[imax-1] := vv[j-1];}pvimax^:= pvi^;
      end;
      pindxj^:= imax;  {indx[j-1] := imax;}
      if (j <> Nn) then begin
        if (Cabs(pjj^)< Cabs(Ctiny)) then begin
          pjj^:= Ctiny;                 {if a[j,j] = 0.0 then a[j,j] := tiny;}
          inc(sngl);
        end;
        dum := Cdiv(Cone,pjj^);         {dum := 1.0/a[j,j];}
        pij := pjj; inc(pij,Nn);         {alkaa (j+1,j):st }
        for i:=j+1 to Nn do begin
          if (pij^.x <> 0) or (pij^.y <> 0) then
            pij^ := Cmult(pij^,dum); {a[i,j] := a[i,j]*dum;}
          inc(pij,Nn);
        end;
      end;
      inc(pindxj);   {pointer to next element in indx}
      inc(pjj,Nn+1); {pointer to next diagonal element (jj)}
      inc(pj);       {pointer to next column}
      inc(prj,Nn);    {pointer to next row j}
    end;
    pjj := paa;
    inc(pjj,Nn*Nn-1);
    if (Cabs(pjj^) < Cabs(Ctiny)) then begin
      pjj^ := Ctiny;  {if a[n,n] = 0 then a[n,n] := tiny}
      inc(sngl);
    end;
  end;
  finally
  end;
end;
                
procedure CLUbacksub(paa   : pComplex;   {pointer to the matrix aa in LU-form}
                     Nn    : Integer;    {dimension of the matrixes}
                     pindx : pInteger;   {pointer to the array storing
                                          the row exchanges}
                     pbb   : pComplex);  {pointer to the right hand side
                                             AND solution}
{CLUbacksub solves the equation  AX = b, when A is given in the LU-form
 paa    is the pointer to the first element in A  (in)
 pbb    is the pointer to the right hand side b (in),
        and returns the solution X (out)
 pindx  is the pointer to the array storing the row exhanges,
        calculated in LUdecomp (in)}
 const
   tiny = 1.0e-10;
 var
  apu      : Single;    {temporary}
  j,ip,ii,i: Integer;   {indices}
  sum      : Complex;   {temporary sum}
  pbi      : pComplex;  {pointer to bb[i]}
  pbid     : pComplex;  {pointer to bb[...]}
  pbj      : pComplex;  {pointer to bb[j]}
  pdi      : pInteger;  {pointer to indx[i]}
  pij      : pComplex;  {pointer to element (ij) in array aa}
  pri      : pComplex;  {pointer to row i in array aa}
  pii      : pComplex;  {pointer to element ii in aa}
 begin
   try
   ii := 0;
   pri := paa;
   pdi := pindx; {pointer to indx, starts from 1}
   pbi := pbb;   {pointer to bb[i], starts from 1}
   for i:=1 to Nn do begin
     ip := pdi^;                    {ip := indx[i];}
     pbid := pbb;
     if (ip > 0) then inc(pbid,ip-1);   {pointer to bb[indx[i]]}
     sum := pbid^;                  {sum := bb[ip1];}
     pbid^:= pbi^;                  {bb[ip] := bb[i];}
     if (ii <> 0) then begin
       pij := pri; inc(pij,ii-1);   {starts from i,ii}
       pbj := pbb; inc(pbj,ii-1);   {starts from ii}
       for j := ii to i-1 do begin
         if ((pij^.x <> 0) or (pij^.y <> 0)) and ((pbj^.x <> 0) or (pbj^.y <> 0)) then
           sum := Csub(sum, Cmult(pij^,pbj^));     {sum-bb[j]}
         inc(pij);
         inc(pbj);
       end;
     end else if (sum.x <> 0) or (sum.y <> 0) then begin
       ii := i;
     end;
     pbi^ := sum;                   {bb[i] := sum}
     inc(pri,Nn);
     inc(pbi);
     inc(pdi);
   end;
   pii := paa; inc(pii,Nn*Nn-1); {pointer to last element in aa}
   pri := pii; dec(pri,Nn-1);    {pointer to last row in aa}
   pbi := pbb; inc(pbi,Nn-1);    {pointer to b[i], starts from Nn}
   for i:=Nn downto 1 do begin
     sum := pbi^;                {sum := bb[i];}
     if (i < Nn) then begin
       pij := pri; inc(pij,i);   {starts from i,i+1}
       pbj := pbb; inc(pbj,i);   {starts from i+1}
       for j:=i+1 to Nn do begin
         if ((pij^.x <> 0) or (pij^.y <> 0)) and ((pbj^.x <> 0) or (pbj^.y <> 0)) then
         sum := Csub(sum,Cmult(pij^,pbj^));   {sum := sum-pij^*bb[j];}
         inc(pij);
         inc(pbj);
       end;
     end;
     apu :=Cabs(pii^);
     if (apu > tiny {0.9e-20}) then begin {avoid division by "zero"}
       pbi^:= Cdiv(sum,pii^);            {bb[i] := sum/pii^;}
     end else begin
       try
       pbi^.x := sum.x*tiny{1.e20};
       pbi^.y := sum.y*tiny{1.e20};
       finally
       end;
     end;
     dec(pbi);
     dec(pri,Nn);
     dec(pii,Nn+1);
   end;
   finally
   end;
 end;

 procedure putCmat(paa: pComplex;  {pointer to the first element of the matrix}
                  i,j: Integer;    {indices of the element}
                  m,n: Integer;    {the dimensions of the matrix}
                  xx : Complex);   {the value to be put into the matrix}
 {putmat copies the value of xx to the location i,j of the complex matrix
         with dimensions (m rows * n columns)}
 var
   pij: pComplex;  {pointer to the element i,j}
 begin
   {some checks, just to be sure}
   if (i>0) and (i<=m) and (j>0) and (j<=n) then begin
     pij := paa;
     inc(pij,n*(i-1)+j-1);
     pij^:= xx;
   end;
 end;

 procedure clrCrow(paa: pComplex; i,nn :Integer);
 {clear row i in matrix  paa with dimension nn*nn}
 var
   j   : integer;   {index}
   pij : pComplex;  {pointer to the element i,j}
 begin
   {some checks, just to be sure}
   if (i>0) and (i<=nn) then begin
     pij := paa;
     inc(pij,nn*(i-1));
     for j:=1 to nn do begin
       pij^.x := 0;
       pij^.y := 0;
       inc(pij);
     end;
   end;
 end;

 procedure putImat(paa: pInteger; {pointer to the first element of the matrix}
                   i,j: Integer;  {indices of the element}
                   m,n: Integer;  {the dimensions of the matrix}
                   xx : Integer); {the value to be put into the matrix}
 {putImat copies the value of the INTEGER xx to the location i,j of tha matrix
         with dimensions (m rows * n columns)}
 var
   pij: pInteger;  {pointer to the element i,j}
 begin
   {some checks, just to be sure}
   if (i>0) and (i<=m) and (j>0) and (j<=n) then begin
     pij := paa;
     inc(pij,n*(i-1)+j-1);
     pij^:= xx;
   end;
 end;

 procedure putCvec(paa: pComplex; i,m:Integer;xx:Complex);
 var
   pi: pComplex;
 begin
   if ((i>0) and (i<=m)) then begin
     pi := paa;
     inc(pi,i-1);
     pi^:= xx;
   end
 end;

 procedure getCvec(paa: pComplex; i,m:Integer;var xx:Complex);
 var
   pi: pComplex;
 begin
   if ((i>0) and (i<=m)) then begin
     pi := paa;
     inc(pi,i-1);
     xx := pi^;
   end
 end;

 procedure copyCvec(paa,pbb: pComplex; nn:Integer);
 {copy vector paa^ to pbb^}
 var
   i    : integer;
   pa,pb: pComplex;
 begin
   if ((nn>0)) then begin
     pa := paa;
     pb := pbb;
     for i:=1 to nn do begin
       pb^ := pa^;
       inc(pa);
       inc(pb);
     end;
   end
 end;

 procedure copyCvecCol(pbb,paa : pComplex; jj,nn : integer);
 {copy vector pbb^ to column jj of square matrix paa^
 pbb   pointer to vector to be copied
 paa   pointer to matrix, where copied
 ii    index of row
 nn    dim of paa }
 var
   ii    : integer;
   pa,pb : pComplex;
 begin
   if (nn > 0) and (jj > 0) and (jj <= nn) then begin
     pa := paa;
     pb := pbb;
     inc (pa,jj-1);
     for ii := 1 to nn do begin
       pa^ := pb^;
       inc(pa,nn);
       inc(pb);
     end;
   end;
 end;



 procedure getCmat(paa: pComplex;  {pointer to the first element of the matrix}
                   i,j: Integer;   {indices of the element}
                   m,n: Integer;   {the dimensions of the matrix}
               var xx : Complex);  {the value to get the value of aa(i,j)}
 {getmat copies the value of xx to the location i,j of tha matrix
         with dimensions (m rows * n columns)}
 var
   pij: pComplex;  {pointer to the element i,j}
 begin
   {some checks, just to be sure}
   if (i>0) and (i<=m) and (j>0) and (j<=n) then begin
     pij := paa;
     inc(pij,n*(i-1)+j-1);
     xx := pij^;
   end;
 end;

 procedure getImat(paa: pInteger;{pointer to the first element of the matrix}
                  i,j: Integer;  {indices of the element}
                  m,n: Integer;  {the dimensions of the matrix}
              var xx : Integer); {the value to get the value of aa(i,j)}
 {getmat copies the value of the INTEGER xx to the location i,j of the matrix
         with dimensions (m rows * n columns)}
 var
   pij: pInteger;  {pointer to the element i,j}
 begin
   {some checks, just to be sure}
   if (i>0) and (i<=m) and (j>0) and (j<=n) then begin
     pij := paa;
     inc(pij,n*(i-1)+j-1);
     xx := pij^;
   end;
 end;

function comptostr(cpu : complex): string;
{change a complex number to string (for display)}
begin
  comptostr := floattostr(cpu.x)+'  '+floattostr(cpu.y);
end;

procedure copyCmat(paa,pzz : pComplex; nn : integer);
{copy the complex matrix paa to matrix pzz,
 both are square matrices with dimension nn}
var
  ii  : longint;
  n2  : longint;
  pa  : pComplex;
  pz  : pComplex;
begin
  if (paa <> nil) and (pzz <> nil) and (nn > 0) then begin
    pa := paa;
    pz := pzz;
    n2 := nn*nn;
    for ii:=1 to n2 do begin
      pz^:= pa^;
      inc(pz);
      inc(pa);
    end;
  end;
end;

procedure imgMat(paa : pComplex; pzz : pSingle; nn : integer);
{copy the imaginary part of the complex matrix paa to real matrix pzz,
 both are square matrices with dimension nn}
var
  ii  : longint;
  n2  : longint;
  pa  : pComplex;
  pz  : pSingle;
begin
  if (paa <> nil) and (pzz <> nil) and (nn > 0) then begin
    pa := paa;
    pz := pzz;
    n2 := nn*nn;
    for ii:=1 to n2 do begin
      pz^:= pa^.y;
      inc(pz);
      inc(pa);
    end;
  end;
end;

procedure copyCrow(paa,pzz : pComplex; ir,nn : integer);
{copy the row ir of complex matrix paa to matrix pzz,
 both are square matrices with dimension nn}
var
  ii  : longint;
  pa  : pComplex;
  pz  : pComplex;
begin
  if (paa<>nil) and (pzz<>nil) and (nn >0) and (ir >0) and (ir <=nn) then begin
    pa := paa;
    inc(pa,(ir-1)*nn);
    pz := pzz;
    inc(pz,(ir-1)*nn);
    for ii:=1 to nn do begin
      pz^:= pa^;
      inc(pz);
      inc(pa);
    end;
  end;
end;

procedure clearCmat(paa :pComplex; nn : integer);
{put zeroes in all entries of the square matrix at paa, with dimension nn}
var
  cpu : complex;
  ii  : longint;
  n2  : longint;
  pa  : pComplex;
begin
  if (paa <> nil) and (nn > 0) then begin
    cpu.x := 0; cpu.y := 0;
    pa := paa;
    n2 := nn*nn;
    for ii:=1 to n2 do begin
      pa^:= cpu;
      inc(pa);
    end;
  end;
end;

procedure clearCvec(pbb :pComplex; nn : integer);
{put zeroes in all entries of the vector at pbb, with dimension nn}
var
  cpu : complex;
  ii  : longint;
  pb  : pComplex;
begin
  if (pbb <> nil) and (nn > 0) then begin
    cpu := cplx(0,0);
    pb := pbb;
    for ii:=1 to nn do begin
      pb^:= cpu;
      inc(pb);
    end;
  end;
end;


function maxCvec(paa : pComplex; nn : integer) : complex;
{find the largest element by absolute value}
var
  amx : single;
  apu : single;
  cmx : complex;
  ii  : integer;
  pa  : pComplex;
begin
  if (paa <> nil) and (nn > 0) then begin
    amx := 0;
    cmx := cplx(0.0,0.0);
    pa := paa;
    for ii := 1 to nn do begin
      apu := Cabs(pa^);
      if (apu > amx) then begin
        cmx := pa^;
        amx := apu;
      end;
      inc(pa);
    end;
    maxCvec := cmx;
  end else begin
    maxCvec := cplx(0.0,0.0);
  end;
end;

procedure showCmatElements(paa : pComplex; mm,nn : integer);
{display the elements of paa}
var
  pa  : pComplex;
  i,j : integer;
begin
  if (mm > 0) and (nn > 0) then begin
    pa := paa;
    for i:=1 to mm  do begin
      for j := 1 to nn do begin
        showmessage(inttostr(i)+'  '+inttostr(j)+'  '+comptostr(pa^));
        inc(pa);
      end;
    end;
  end else begin
    showmessage('wrong dimension(s)');
  end;
end;


end.
