unit CurrPhas;

interface

uses
  SysUtils, 
  utildef, compdef, skemadef, imagodef, skemproc, phasor, CTParam,
  dialogs;

procedure ShowCurrPhasors(ipComp : longint); far;  {show currents using phasors}
procedure RemoveCTphasor (ipComp : longint); far;  {remove the phasor of a curr.trafo}
procedure RemoveAllPhasors;                        {used in skemadef}
procedure UpdateCurrPhasors;
procedure GetCurrents(ipComp : longint; var CurrA,AngleA,CurrB,AngleB,CurrC,AngleC : single);
procedure GetVoltages(ipComp : longint; phphV: boolean;
                      var VoltA,AngleA,VoltB,AngleB,VoltC,AngleC : single);
procedure GetCompCurrents(ipComp : longint; wind: integer;
                      var CurrA,AngleA,CurrB,AngleB,CurrC,AngleC : single);                      

implementation

uses
  imagopas, specialCT, RelCalc;

procedure GetCurrents(ipComp : longint; var CurrA,AngleA,CurrB,AngleB,CurrC,AngleC : single);
{get the currents of a current trafo in polar form
 ipComp  pointer to the trafo as integer
 CurrA, AngleA   magnitude and angle of phase a (phase in degrees)
 CurrB, AngleB   magnitude and angle of phase b (phase in degrees)
 CurrC, AngleC   magnitude and angle of phase c (phase in degrees)}
var
  curr     : single;      {max. current}
  c3,f3    : Ttrois;      {current, phase}
  Ia,Ib,Ic : complex;     {the currents of phases a,b,c}
  ll       : longint;     {temporary}
  nvar     : integer;     {no of output variables}
  pComp    : TpComp;      {pointer to the trafo}
  pSet     : TpCompSet;   {pointer to the common}
begin
  if (ipComp <> 0) then begin
    pComp := TpComp(ipComp);
    pComp^.GetipSet(ll);
    pSet := TpCompSet(ll);                {pointer to the common}
    pComp^.GetNvar(nvar);                 {no of variables}
    {initial currents, or currents at some given time}
    if (pSet^.displ_time <= 0) then begin {at initial time}
      pComp^.GetVar(gCurrT_Ir,Ia.x);        {real part of current, phase a}
      pComp^.GetVar(gCurrT_Ii,Ia.y);        {imag.part of current}
      pComp^.GetVar(gCurrT_Ir+nvar,Ib.x);   {real part of current, phase b}
      pComp^.GetVar(gCurrT_Ii+nvar,Ib.y);   {imag.part of current, phase b}
      pComp^.GetVar(gCurrT_Ir+2*nvar,Ic.x); {real part of current, phase c}
      pComp^.GetVar(gCurrT_Ii+2*nvar,Ic.y); {imag.part of current, phase c}
      {get magnitude and phase}
      CurrA  := Cabs(Ia);
      AngleA := Cangle(Ia)*180/pi;
      CurrB  := Cabs(Ib);
      AngleB := Cangle(Ib)*180/pi;
      CurrC  := Cabs(Ic);
      AngleC := Cangle(Ic)*180/pi;
    end else begin                        {at given time}
      {find current at given time, in Skemproc.pas}
      GetCTcurr(pComp,pSet^.displ_time,curr,c3,f3);  {in skemproc.pas}
      CurrA := c3.x;                      {magnitude of current in phase A}
      CurrB := c3.y;
      CurrC := c3.z;
      AngleA := f3.x;                     {phase of current in phase A}
      AngleB := f3.y;
      AngleC := f3.z;
    end;
  end else begin                          {error case}
    CurrA  := 0;
    AngleA := 0;
    CurrB  := 0;
    AngleB := 0;
    CurrC  := 0;
    AngleC := 0;
  end;
end;

procedure DeltaCurrentsIf(pComp : TpComp; CurrA,AngleA,CurrB,AngleB,CurrC,AngleC : single;
                                     var  CurDA,AnglDA,CurDB,AnglDB,CurDC,AnglDC: single);
{Calculates the secondary currents of a current trafo, if the secondary is used.
 (the secondary is used, if the parameter conn <> 0).The secondary may be connected
 as Y or Delta. If delta, then either a-b or a-c - connection. The transformation
 ratio is taken to be one. The parameters (connection and mode) are read from the CT
 pComp   pointer to the current trafo
 inputs (primary current):
 CurrA, AngleA   magnitude and angle of phase a (phase in degrees)
 CurrB, AngleB   magnitude and angle of phase b (phase in degrees)
 CurrC, AngleC   magnitude and angle of phase c (phase in degrees)
 outputs (secondary current):
 CurDA, AnglDA   magnitude and angle of phase a (phase in degrees)
 CurDB, AnglDB   magnitude and angle of phase b (phase in degrees)
 CurDC, AnglDC   magnitude and angle of phase c (phase in degrees)}
const
  DtoR = pi/180;       {degrees to radians}
var
 apu      : single;    {temporary}
 ca,cb,cc : complex;   {input currents as complex}
 da,db,dc : complex;   {output currents as complex}
 comp     : boolean;   {compensate the delta connection, so that transformation ratio holds}
 conn     : single;    {connection, Y or delta a-b, or delta a-c}
 Ipr,Isc  : single;    {nominal prim, sec. currents of the current trafo}
 mode     : single;    {flag for the use of prim. or sec. current}
 ratio    : single;    {transformation ratio}
 txt      : string;
begin
  if (pComp <> nil) then begin
    GetCTmainPar(pComp, ratio, Ipr, Isc, conn, mode);  {parameters of the CT}
//  CurrGrid.GetDelt(longint(pComp),conn);
//  if {(mode <> 0) and} (conn <> 0) then begin          {secondary delta currents used}
    if (CurrPhasorForm.SecondaryItem.checked) then begin
      {currents from polar to complex}
      apu := AngleA*DtoR;
      ca.x := CurrA*cos(apu);
      ca.y := CurrA*sin(apu);
      apu := AngleB*DtoR;
      cb.x := CurrB*cos(apu);
      cb.y := CurrB*sin(apu);
      apu := AngleC*DtoR;
      cc.x := CurrC*cos(apu);
      cc.y := CurrC*sin(apu);
      {output, secondary currents, but with ratio = 1}
      {if connected in Y, no need to compensate transformation ratio for delta-connection}
      if ((conn <> 1) and (conn <> 2)) then comp := false else comp := true;
      CurrentsFromDelta(ca,cb,cc, conn, 1{mode}, 1{ratio}, 1{Isec},comp{true},da,db,dc); {in skemproc.pas}
      {..false = do not compensate the effect of delta connection in the transformation ratio}
      {change to polar form}
      CurDA := Cabs(da);
      AnglDA := Cangle(da)/DtoR;
      CurDB := Cabs(db);
      AnglDB := Cangle(db)/DtoR;
      CurDC := Cabs(dc);
      AnglDC := Cangle(dc)/DtoR;
    end else begin                   {not delta or not secondary, copy to output}
      CurDA := CurrA;
      AnglDA :=AngleA;
      CurDB := CurrB;
      AnglDB :=AngleB;
      CurDC := CurrC;
      AnglDC :=AngleC;
    end;
  end else begin                     {prim.curr, or no current trafo? (error?)}
    CurDA := CurrA;
    AnglDA :=AngleA;
    CurDB := CurrB;
    AnglDB :=AngleB;
    CurDC := CurrC;
    AnglDC :=AngleC;
  end;
end;


procedure GetVoltages(ipComp : longint; phphV: boolean; var VoltA,AngleA,VoltB,AngleB,VoltC,AngleC : single);
{get the voltages of a bus in polar form
 ipComp  pointer to the trafo as integer
 phphV   phase-phase voltages used
 VoltA, AngleA   magnitude and angle of phase a (phase in degrees)
 VoltB, AngleB   magnitude and angle of phase b (phase in degrees)
 VoltC, AngleC   magnitude and angle of phase c (phase in degrees)}
var
  Va,Vb,Vc    : complex;  {the voltages of phases a,b,c}
  Uab,Ubc,Uca : complex;  {phase-phase voltages}
  nvar        : integer;  {no of output variables}
  pComp       : TpComp;   {pointer to the bus}
begin
  if (ipComp <> 0) then begin
    pComp := TpComp(ipComp);
    pComp^.GetNvar(nvar);               {no of variables}
    pComp^.GetVar(gBus_Vr,Va.x);        {real part of voltage, phase a}
    pComp^.GetVar(gBus_Vi,Va.y);        {imag.part }
    pComp^.GetVar(gBus_Vr+nvar,Vb.x);   {real part, phase b}
    pComp^.GetVar(gBus_Vi+nvar,Vb.y);   {imag.part, phase b}
    pComp^.GetVar(gBus_Vr+2*nvar,Vc.x); {real part, phase c}
    pComp^.GetVar(gBus_Vi+2*nvar,Vc.y); {imag.part, phase c}
    {phase-phase or phase-neutral voltages}
    if (phphV{VoltMode_p = PhasePhase}) then begin
      Uab := Csub(Va,Vb);        {phase-phase voltages}
      Ubc := Csub(Vb,Vc);
      Uca := Csub(Vc,Va);
      {get magnitude and phase}
      VoltA  := Cabs(Uab);
      AngleA := Cangle(Uab)*180/pi;
      VoltB  := Cabs(Ubc);
      AngleB := Cangle(Ubc)*180/pi;
      VoltC  := Cabs(Uca);
      AngleC := Cangle(Uca)*180/pi;
    end else begin               {phase-neutral voltages}
      {get magnitude and phase}
      VoltA  := Cabs(Va);
      AngleA := Cangle(Va)*180/pi;
      VoltB  := Cabs(Vb);
      AngleB := Cangle(Vb)*180/pi;
      VoltC  := Cabs(Vc);
      AngleC := Cangle(Vc)*180/pi;
    end;
  end else begin                          {error case}
    VoltA  := 0;
    AngleA := 0;
    VoltB  := 0;
    AngleB := 0;
    VoltC  := 0;
    AngleC := 0;
  end;
end;

procedure GetCompCurrents(ipComp : longint; wind: integer; var CurrA,AngleA,CurrB,AngleB,CurrC,AngleC : single);
{get the currents of a transformer in polar form
 ipComp  pointer to the trafo as integer
 wind    winding (prim, sec, maybe tert(?))
 CurrA, AngleA   magnitude and angle of phase a (phase in degrees)
 CurrB, AngleB   magnitude and angle of phase b (phase in degrees)
 CurrC, AngleC   magnitude and angle of phase c (phase in degrees)}
var
  curr     : single;      {max. current}
  c3,f3    : Ttrois;      {current, phase}
  Ia,Ib,Ic : complex;     {the currents of phases a,b,c}
  ll       : longint;     {temporary}
  name     : string;      {name = type of the component}
  nvar     : integer;     {no of output variables}
  pComp    : TpComp;      {pointer to the trafo}
  pSet     : TpCompSet;   {pointer to the common}
begin
  if (ipComp <> 0) then begin
    pComp := TpComp(ipComp);
    pComp^.GetName(name);
    pComp^.GetipSet(ll);
    pSet := TpCompSet(ll);                {pointer to the common}
    pComp^.GetNvar(nvar);                 {no of variables}
    {only initial currents available}
    if (name = gTrafo3w) then begin           {3-winding trafo}
      if (wind = 1) then begin                {primary winding}
        pComp^.GetVar(g3w_I1r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(g3w_I1i,Ia.y);          {imag.part of current}
        pComp^.GetVar(g3w_I1r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(g3w_I1i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(g3w_I1r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(g3w_I1i+2*nvar,Ic.y);   {imag.part of current, phase c}
      end else if (wind = 2) then  begin      {secondary winding}
        pComp^.GetVar(g3w_I2r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(g3w_I2i,Ia.y);          {imag.part of current}
        pComp^.GetVar(g3w_I2r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(g3w_I2i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(g3w_I2r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(g3w_I2i+2*nvar,Ic.y);   {imag.part of current, phase c}
      end else if (wind = 3) then  begin      {tertiary winding}
        pComp^.GetVar(g3w_I3r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(g3w_I3i,Ia.y);          {imag.part of current}
        pComp^.GetVar(g3w_I3r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(g3w_I3i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(g3w_I3r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(g3w_I3i+2*nvar,Ic.y);   {imag.part of current}
      end else begin
        Ia := cplx(0,0);
        Ib := cplx(0,0);
        Ic := cplx(0,0);
      end;
    end else if (name = gAtrafo) then begin       {autotrafo}
      if (wind = 1) then begin                {primary winding}
        pComp^.GetVar(gAtrafo_I1r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(gAtrafo_I1i,Ia.y);          {imag.part of current}
        pComp^.GetVar(gAtrafo_I1r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(gAtrafo_I1i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(gAtrafo_I1r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(gAtrafo_I1i+2*nvar,Ic.y);   {imag.part of current, phase c}
      end else if (wind = 2) then  begin      {secondary winding}
        pComp^.GetVar(gAtrafo_I2r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(gAtrafo_I2i,Ia.y);          {imag.part of current}
        pComp^.GetVar(gAtrafo_I2r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(gAtrafo_I2i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(gAtrafo_I2r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(gAtrafo_I2i+2*nvar,Ic.y);   {imag.part of current, phase c}
      end else if (wind = 3) then  begin      {tertiary winding}
        pComp^.GetVar(gAtrafo_I3r,Ia.x);          {real part of current, phase a}
        pComp^.GetVar(gAtrafo_I3i,Ia.y);          {imag.part of current}
        pComp^.GetVar(gAtrafo_I3r+nvar,Ib.x);     {real part of current, phase b}
        pComp^.GetVar(gAtrafo_I3i+nvar,Ib.y);     {imag.part of current, phase b}
        pComp^.GetVar(gAtrafo_I3r+2*nvar,Ic.x);   {real part of current, phase c}
        pComp^.GetVar(gAtrafo_I3i+2*nvar,Ic.y);   {imag.part of current}
      end else begin
        Ia := cplx(0,0);
        Ib := cplx(0,0);
        Ic := cplx(0,0);
      end;
    end else begin                             {2-winding trafo}
      if (wind = 1) then begin                 {primary winding}
        pComp^.GetVar(gTrafo_I1r,Ia.x);        {real part of current, phase a}
        pComp^.GetVar(gTrafo_I1i,Ia.y);        {imag.part of current}
        pComp^.GetVar(gTrafo_I1r+nvar,Ib.x);   {real part of current, phase b}
        pComp^.GetVar(gTrafo_I1i+nvar,Ib.y);   {imag.part of current, phase b}
        pComp^.GetVar(gTrafo_I1r+2*nvar,Ic.x); {real part of current, phase c}
        pComp^.GetVar(gTrafo_I1i+2*nvar,Ic.y); {imag.part of current, phase c}
      end else if (wind = 2) then  begin       {secondary winding}
        pComp^.GetVar(gTrafo_I2r,Ia.x);        {real part of current, phase a}
        pComp^.GetVar(gTrafo_I2i,Ia.y);        {imag.part of current}
        pComp^.GetVar(gTrafo_I2r+nvar,Ib.x);   {real part of current, phase b}
        pComp^.GetVar(gTrafo_I2i+nvar,Ib.y);   {imag.part of current, phase b}
        pComp^.GetVar(gTrafo_I2r+2*nvar,Ic.x); {real part of current, phase c}
        pComp^.GetVar(gTrafo_I2i+2*nvar,Ic.y); {imag.part of current, phase c}
      end else begin                           {not yet...}
        Ia := cplx(0,0);
        Ib := cplx(0,0);
        Ic := cplx(0,0);
      end;
    end;
    {get magnitude and phase}
    CurrA  := Cabs(Ia);
    AngleA := Cangle(Ia)*180/pi;
    CurrB  := Cabs(Ib);
    AngleB := Cangle(Ib)*180/pi;
    CurrC  := Cabs(Ic);
    AngleC := Cangle(Ic)*180/pi;
  end else begin                          {error case}
    CurrA  := 0;
    AngleA := 0;
    CurrB  := 0;
    AngleB := 0;
    CurrC  := 0;
    AngleC := 0;
  end;
end;

procedure ShowCurrPhasors(ipComp : longint);
{show the current(s) using phasors, if allowed (button down)
 show also the parameter window (if allowed)
 ipComp  pointer to the current trafo as integer}
var
  aa,ab,ac : single;  {phase angles of currents, in degrees}
  added    : boolean; {true, when trafo added to phasor list}
  ca,cb,cc : single;  {magnitudes of currents}
  curr     : Tmultix; {magnitude and phase of the 3 phase currents}
  ix,iy    : integer; {temporary (coordinates)}
  lbl      : string;  {label of the component}
  name     : string;
  pComp    : TpComp;
  pa,pb,pc : single;  {magnitudes of primary currents}
  ratio    : single;  {transformation ratio of the current trafo}
begin
  pComp := TpComp(ipComp);
  pComp^.GetName(name);
  pComp^.GetLabel(lbl,ix,iy);
  if (imagoo.CurrPhasorBut.down) then begin
    {read the currents from the component}
    GetCurrents(ipComp,pa{ca},aa,pb{cb},ab,pc{cc},ac);
    {show secondary currents, if selected, but Y or delta determined separately}
    if (CurrPhasorForm.SecondaryItem.checked) then begin
      pComp^.GetPar(gCT_ratio,ratio);        {transformation ratio}
      ca := pa/ratio;
      cb := pb/ratio;
      cc := pc/ratio;
    end else begin                           {show primary currents}
      ca := pa;
      cb := pb;
      cc := pc;
    end;
    {show secondary, possibly delta currents (but with ratio = 1), if selected}
    DeltaCurrentsIf(pComp, ca,aa,cb,ab,cc,ac,curr.u,curr.v,curr.w,curr.x,curr.y,curr.z);
    added := CurrGrid.ShowPhasors(ipComp, cRead,@curr,lbl,''); {read from current trafos}
    if (added) then begin    {continue, if this trafo added}
      SeqGrid.ShowPhasors(ipComp, cCopy, nil,lbl,'');        {copy from CurrGrid}
      pComp^.SetPWR(true);                                {set flag}
    end;
    CurrPhasorForm.PointEdit.text := '';
  end;
  {add to list of special relays, if not special relay}
  if (SummingForm.visible) and not (name = gSpecCT) then
    AddCurrTrafoToList(ipComp); {in specialCT}
  {send currents to matrix calculator (Thn joku ehto ??) - - - - - - - - - - - - - - - - }
  imagoo.SendCurrents(ipComp);
  {show parameter window}
  if (name <> gSpecCT) then begin
    CTForm.caption := 'Current Transf. '+lbl;
  end else begin
    CTForm.caption := 'Special Curr. Transf. '+lbl;
  end;
  CTForm.CTParamShow(ipComp);      {in CTParam.pas}
  {put to relay calculator, if asked for}
  if (RelayCalcForm.MeasureBut.down) then begin
    RelayCalcForm.SetMeasureCT(ipComp);
  end else if (RelayCalcForm.ReferenceBut.down) then begin
    RelayCalcForm.SetReferenceCT(ipComp);         
  end;
end;

procedure RemoveCTphasor(ipComp : longint);
{remove (delete) a current trafo from the grid of CT:s for phasor display
 ipComp  pointer to the current trafo, as integer}
var
  row : integer;  {the row of the identifier in the grid}
begin
  if (CurrGrid.isThere(ipComp,row,'RemoveCTphasor')) then begin
    CurrGrid.DeleteCTphasor(row);
    SeqGrid.DeleteCTphasor(row);
    CurrGrid.DrawPhasors;
  end;
end;

procedure RemoveAllPhasors;
{clear the grid, tables and phasor picture}
begin
  CurrGrid.RemoveAllPhasors;
  SeqGrid.RemoveAllPhasors;
  CurrPhasorForm.visible := false;
end;

procedure UpdateCurrPhasors;
{update the phasors (after some new calculation)}
var
  aa,ab,ac : single;  {phase angles of currents (degrees)}
  ca,cb,cc : single;  {magnitudes of currents}
  curr     : Tmultix; {currents = magnitude and phase (in degrees)}
  ii       : integer; {index}
  ipComp   : longint; {pointer to the current trafos, as integer}
  nCT      : integer; {number of current trafos}
  pa,pb,pc : single;  {magnitudes of primary currents}
  pComp    : TpComp;  {pointer to the current trafo}
  ratio    : single;  {transformation ratio, prim/sec}
begin
  {read the currents from the current trafos in the grid of the graph}
  nCT := CurrGrid.NoOfItems;                   {number of current trafos}
  if (nCT > 0) then begin
    for ii := 1 to nCT do begin                {all current trafos}
      ipComp := CurrGrid.GetIp(ii);
      pComp := TpComp(ipComp);
      GetCurrents(ipComp,pa{ca},aa,pb{cb},ab,pc{cc},ac); {read currents}
      {show secondary currents, if selected, but Y or delta determined separately}
      if (CurrPhasorForm.SecondaryItem.checked) then begin
        pComp^.GetPar(gCT_ratio,ratio);        {transformation ratio}
        ca := pa/ratio;
        cb := pb/ratio;
        cc := pc/ratio;
      end else begin                           {show primary currents}
        ca := pa;
        cb := pb;
        cc := pc;
      end;
//      if (CurrPhasorForm.DeltaItem.checked) then begin
//      pComp := TpComp(ipComp);
        DeltaCurrentsIf(pComp, ca,aa,cb,ab,cc,ac,curr.u,curr.v,curr.w,curr.x,curr.y,curr.z);
//      end else begin
//        curr.u := ca;
//        curr.v := aa;
//        curr.w := cb;
//        curr.x := ab;
//        curr.y := cc;
//        curr.z := ac;
//      end;
      CurrGrid.UpdatePhasors(cRead,@curr,ii);  {update the phasor graphs, in phasor.pas}
    end;
    SeqGrid.UpdatePhasors(cCopy,nil,0);        {update also the sequence currents}
  end;
end;

end.
