unit linetest;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls;

type
  TForm1 = class(TForm)
    MainImg: TImage;
    PointUp: TImage;
    PointDw: TImage;
    DrawBut: TButton;
    ClearBut: TButton;
    procedure FormCreate(Sender: TObject);
    procedure PointUpMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PointDwMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure DrawButClick(Sender: TObject);
    procedure ClearButClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

const
  topmarg = 10;
  botmarg = 15;
  leftmarg = 30;
  rightmarg = 20;

procedure TForm1.FormCreate(Sender: TObject);
begin
  with MainImg.canvas do begin
    {a rectangle around the image}
    pen.color := clBlack;
    brush.style := bsSolid;
    brush.color := clWhite;
    pen.width := 1;
    rectangle(0,0,MainImg.width,MainImg.height);
    {a rectangle along the margins}
    pen.color := clRed;
    brush.style := bsClear;
    rectangle(leftmarg,topmarg,MainImg.width-rightmarg,MainImg.height-botmarg);
  end;
  {a cross to the "points"}
  with PointUp.canvas do begin
    pen.color := clWhite;
    brush.style := bsSolid;
    brush.color := clWhite;
    pen.width := 1;
    rectangle(0,0,PointUp.width,PointUp.height);
    pen.color := clRed;
    moveTo(PointUp.width div 2,0);
    lineTo(pointUp.width div 2,PointUp.height);
    moveTo(0,PointUp.height div 2);
    lineTo(pointUp.width, PointUp.height div 2);
  end;
    with PointDw.canvas do begin
    pen.color := clWhite;
    brush.style := bsSolid;
    brush.color := clWhite;
    pen.width := 1;
    rectangle(0,0,PointDw.width,PointDw.height);
    pen.color := clBlue;
    moveTo(PointDw.width div 2,0);
    lineTo(pointDw.width div 2,PointDw.height);
    moveTo(0,PointDw.height div 2);
    lineTo(pointDw.width, PointDw.height div 2);
  end;
end;

procedure PartLine(xl,xr,yt,yb,x1,y1,x2,y2: integer; var xs,ys,xe,ye: integer;
                   var draw: boolean);
{determines which part of given line is inside a given rectangle
 The given line is defined by its endpoints (x1,y1) and (x2,y2)
 The rectangle is defined by its left, right, top and bottom edges xl,xr,yt,yb
 (Note: the top coordinate yt must be smaller than the bottom coordinate yb!)
 The part line is defined by calculating its start- and endpoints (xs,ys) and (xe,ye)
 draw  true, when a line can be drawn}
var
  aa,bb   : single;  {coefficients for the line y = aa*x +bb}
  apu,apv : single;  {temporary}
  ffound  : boolean; {true, when first intersection found}
  in1,in2 : boolean; {true, when point 1 or 2 inside the rectangle}
  sfound  : boolean; {second intersection found}
  xin,yin : integer; {coordinates of the point inside the rectangle}
  xut,yut : integer; {coordinates of the point outside the rectangle}
  xb,xt   : single;  {intersections with the boundaries}
  yl,yr   : single;  {y at the left, right intersection}

begin
  {points inside the rectangle?}
  in1 := (x1 >= xl) and (x1<= xr) and (y1 >= yt) and (y1 <= yb);
  in2 := (x2 >= xl) and (x2<= xr) and (y2 >= yt) and (y2 <= yb);
  {both points inside the rectangle}
  if (in1 and in2) then begin
    draw := true;
    xs := x1;
    ys := y1;
    xe := x2;
    ye := y2;
  {one point inside the rectangle - - - - - - - -  - - - - - - - - - }
  end else if (in1 or in2) then begin
    draw := true;
    {define the inside point and outside point}
    if (in1) then begin   {point 1 inside}
      xin := x1;
      yin := y1;
      xut := x2;
      yut := y2;
    end else begin        {point 2 inside}
      xin := x2;
      yin := y2;
      xut := x1;
      yut := y1;
    end;
    {start = inside point}
    xs := xin;
    ys := yin;
    {intersection at the edges}
    if (xin = xut) then begin           {vertical line}
      xe := xin;                        {endpoint has same x-coord}
      {inside point higher or lower than the outside point}
      if (yin <= yut) then ye := yb else ye := yt;
    end else if (yin = yut) then begin  {horizontal line}
      ye := yin;
      if (xin <= xut) then xe := xr else xe := xl;
    end else begin                      {oblique: not vertical, not horizontal}
      {define the line as y = ax+b}
      aa := (yut-yin)/(xut-xin);
      bb := yin-aa*xin;
      {eight different cases}
      if (yut <= yt) then begin            {above the rectangle}
        if (xut <= xl) then begin          {upper left corner}
          yl := aa*xl+bb;                  {y at the left intersection}
          if (yl >= yt) then begin         {below top}
            ye := round(yl);
            xe := xl;
          end else begin                   {above top}
            xe := round((yt-bb)/aa);
            ye := yt;
          end;
        end else if (xut >= xr) then begin {upper right corner}
          yr := aa*xr+bb;                  {y at the right intersection}
          if (yr >= yt) then begin         {below top}
            ye := round(yr);
            xe := xr;
          end else begin                   {above top, intersects x-line}
            xe := round((yt-bb)/aa);
            ye := yt;
          end;
        end else begin                     {upper middle box}
          xe := round((yt-bb)/aa);
          ye := yt;
        end;
      end else if (yut >= yb) then begin   {below the rectangle}
        if (xut <= xl) then begin          {lower left corner}
          yl := aa*xl+bb;                  {y at the left intersection}
          if (yl <= yb) then begin         {above bottom line}
            ye := round(yl);
            xe := xl;
          end else begin                   {below bottom line}
            xe := round((yb-bb)/aa);
            ye := yb;
          end;
        end else if (xut >= xr) then begin {lower right corner}
          yr := aa*xr+bb;                  {y at the right intersection}
          if (yr <= yb) then begin         {above bottom line}
            ye := round(yr);
            xe := xr;
          end else begin                   {below bottom line, intersects x-line}
            xe := round((yb-bb)/aa);
            ye := yb;
          end;
        end else begin                     {lower middle box}
          xe := round((yb-bb)/aa);
          ye := yb;
        end;
      end else begin                       {about same level as the rectangle}
        if (xut <= xl) then begin          {left box}
          ye := round(aa*xl+bb);
          xe := xl;
        end else if (xut >= xr) then begin {right box}
          ye := round(aa*xr+bb);
          xe := xr;
        end else begin                     {error case, in the rectangle}
          xe := xut;
          ye := yut;
        end;
      end;
    end;
  {both points outside the rectangle - - - - - - - - - - - - - - - - - -}
  end else begin
    {if both points on the same side, cannot draw}
    if ((y1 < yt) and (y2 < yt)) or
       ((y1 > yb) and (y2 > yb)) or
       ((x1 < xl) and (x2 < xl)) or
       ((x1 > xr) and (x2 > xr)) then begin
      draw := false;              {cannot draw}
      xs := 0; ys := 0;           {just some values}
      xe := 0; ye := 0;
    end else begin                {can perhaps be drawn, must study more}
      {horizontal or vertical?}
      if (x1 = x2) then begin                  {vertical}
        if (x1 > xl) and (x1 < xr) then begin  {line goes over the rectangle}
          draw := true;
          xs := x1;
          ys := yb;
          xe := x1;
          ye := yt;
        end else begin                         {outside the rectangle}
          draw := false;
          xs := 0; ys := 0;                    {just some values}
          xe := 0; ye := 0;
        end;
      end else if (y1 = y2) then begin         {horizontal}
        if (y1 > yt) and (y1 < yb) then begin  {line goes over the rectangle}
          draw := true;
          xs := xl;
          ys := y1;
          xe := xr;
          ye := y1;
        end else begin                         {outside the rectangle}
          draw := false;
          xs := 0; ys := 0;                    {just some values}
          xe := 0; ye := 0;
        end;
      end else begin                           {oblique, not horiz., not vertic.}
        {equation for the line y = a*x+b}
        aa := (y2-y1)/(x2-x1);
        bb := y1-aa*x1;
        {calculate intersections with the boundary lines}
        xb := (yb-bb)/aa;
        yl := aa*xl+bb;
        xt := (yt-bb)/aa;
        yr := aa*xr+bb;
        {find which of them are on the boundaries of the rectangle, if any}
        ffound := false;
        sfound := false;
        if (xb >= xl) and (xb <= xr) then begin
          ffound := true;
          xs := round(xb);
          ys := yb;
        end;
        if (yl >= yt) and (yl <= yb) then begin
          if (ffound) then begin     {first point already found}
            xe := xl;
            ye := round(yl);
            sfound := true;
          end else begin
            ffound := true;
            xs := xl;
            ys := round(yl);
          end;
        end;
        if (xt >= xl) and (xt <= xr) then begin
          if (ffound) then begin     {first point already found}
            xe := round(xt);
            ye := yt;
            sfound := true;
          end else begin
            ffound := true;
            xs := round(xt);
            ys := yt;
          end;
        end;
        if (yr >= yt) and (yr <= yb) then begin
          if (ffound) then begin     {first point already found}
            xe := xr;
            ye := round(yr);
            sfound := true;
          end else begin
            ffound := true;
            xs := xr;
            ys := round(yr);
          end;
        end;
        {line can be drawn, if two intersections found}
        draw := (ffound and sfound);
      end;
    end;
  end;
end;
procedure TForm1.PointUpMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
{move the point}
const
  x0  : integer = 0;    {old coordinates}
  y0  : integer = 0;
begin
  if (ssLeft in Shift) then begin        {left button pressed}
    PointUp.left := PointUp.left+X-x0;
    PointUp.top  := PointUp.top +Y-y0;
  end else begin
    x0 := X;
    y0 := Y;
  end;
end;

procedure TForm1.PointDwMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
{move the point}
const
  x0  : integer = 0;    {old coordinates}
  y0  : integer = 0;
begin
  if (ssLeft in Shift) then begin        {left button pressed}
    PointDw.left := PointDw.left+X-x0;
    PointDw.top  := PointDw.top +Y-y0;
  end else begin
    x0 := X;
    y0 := Y;
  end;

end;

procedure TForm1.DrawButClick(Sender: TObject);
{draw a line from point to point, but only inside the margins}
var
  draw  : boolean;
  x1,y1 : integer;
  x2,y2 : integer;
  xs,ys : integer;
  xe,ye : integer;
begin
  {the points in image coordinates}
  x1 := PointDw.left+(PointDw.width  div 2)-MainImg.left;
  y1 := PointDw.top +(PointDw.height div 2)-MainImg.top;
  x2 := PointUp.left+(PointUp.width  div 2)-MainImg.left;
  y2 := PointUp.top +(PointUp.height div 2)-MainImg.top;
  {find the start and end points of the line within the margins}
  PartLine(leftmarg,MainImg.width-rightmarg,topmarg,MainImg.height-botmarg,
           x1,y1,x2,y2,xs,ys,xe,ye,draw);
  {draw the line, if possible}
  if (draw) then begin
    with MainImg.canvas do begin
      pen.color := clRed;
      pen.width := 1;
      MoveTo(xs,ys);
      LineTo(xe,ye);
    end;
  end;
end;

procedure TForm1.ClearButClick(Sender: TObject);
begin
  with MainImg.canvas do begin
    {a rectangle around the image}
    pen.color := clBlack;
    brush.style := bsSolid;
    brush.color := clWhite;
    pen.width := 1;
    rectangle(0,0,MainImg.width,MainImg.height);
    {a rectangle along the margins}
    pen.color := clRed;
    brush.style := bsClear;
    rectangle(leftmarg,topmarg,MainImg.width-rightmarg,MainImg.height-botmarg);
  end;
end;  

end.
