unit marker;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids, ExtCtrls, Routines, Skemadef, Buttons, Utildef;

procedure DrawMark(Toile :TCanvas; cx, cy, yof, itype : integer; scale : single);
procedure DeleteAllMarkers;
procedure display_message(Form : TForm; But : TBitBtn; text : string);

type
  TMarkerForm = class(TForm)
    CloseBut: TButton;
    MarkerGrid: TStringGrid;
    TypePanel: TPanel;
    TypeImage: TImage;
    MarkerList: TListBox;
    AddBut: TButton;
    MarkerEdit: TEdit;
    MsgBut: TBitBtn;
    DelBut: TButton;
    ModifyBut: TButton;
    procedure CloseButClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure MarkerGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure TypeImageMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure AddButClick(Sender: TObject);
    procedure MsgButClick(Sender: TObject);
    procedure DelButClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure MarkerListClick(Sender: TObject);
    procedure ConstructLine(var line: string; var ertot: integer);
    procedure ModifyButClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

procedure GetMarkerParam(var VerticT,HorizT,SizeT,MargT : integer);
procedure MarkerMenu(pImage : TpImage; pPanel : TpPanel; pGrid : TpGrid; row : integer);
procedure MarkerIndexToCell(x,y: integer; pGrid: TpGrid; pPanel: TpPanel; row : integer);

var
  MarkerForm: TMarkerForm;



implementation

uses Relcurr, Skemproc, Impmark, Imagopas;

{$R *.DFM}

const
  ActRow : integer = 0;  {selected row}
  NameRow = 0;
  CurrRow = 1;
  TimeRow = 2;
  VoltRow = 3;
  TypeRow = 4;
(*VerticT = 3;    {no of rows of marker types}
  HorizT  = 5;    {no of columns of marker types}
  SizeT   = 24;   {size of the square for one marker type}
  MargT   = 4; *)

var
  d_p : char;  {decimal point: '.' or ','}

procedure GetMarkerParam(var VerticT,HorizT,SizeT,MargT : integer);
{get parameters for the marker}
begin
  VerticT := 3;    {no of rows of marker types}
  HorizT  := 5;    {no of columns of marker types}
  SizeT   := 24;   {size of the square for one marker type}
  MargT   := 4;
end;

procedure TMarkerForm.CloseButClick(Sender: TObject);
begin
  MarkerForm.visible := false;
end;

procedure display_message(Form : TForm; But : TBitBtn; text : string);
{show a message using the message button
 But    the button, NOTE a Bit Button!
 Form   the form on which the button is
 text = the message to be shown (must not be too long !}
var
  mestext : string;  {the complete text to be shown}
  mesw    : integer; {width of the message}
  mesh    : integer; {height of the message}
begin
  {add 'OK' to the error message}
  mestext := text+'  -OK-  ';
  But.Glyph.canvas.font.size := But.font.size;
  But.Glyph.canvas.font.name := But.font.name;
  But.Glyph.canvas.font.color := clRed;
  mesw := But.Glyph.canvas.textwidth(mestext);
  mesh := But.Glyph.canvas.textheight(mestext);
  But.Glyph.height := mesh+4;
  But.Glyph.width := mesw;
  But.width := mesw+16;
  But.height := mesh+8;
  But.Glyph.canvas.textout(1,1,mestext);
  But.BringToFront;
  But.top := Form.height div 2;     {put it in the middle of the screen}
  But.left :=Form.width div 2 - But.width div 2;
  But.visible := true;
  But.SetFocus;
end;

procedure DrawMark(Toile :TCanvas; cx, cy, yof, itype : integer; scale : single);
{draw a marker
 Toile  the canvas on which drawn
 cx,cy  coordinates of the centre of the marker
 yof    offset in y-direction
 itype  type +color of the marker (how it does not look like, one of a set of given markers
 scale  scale of the marker}
const
  d3 = 3;
  d4 = 4;  {radius (or other basic dimension) of the unscaled marker}
var
  apu     : single;
  HorizT  : integer;  {no of columns of marker types}
  ipu,ipv : integer;  {temporary}
  ipw     : integer;  {temporary}
  jcolor  : integer;  {index of color}
  jtype   : integer;  {type of the marker}
  dia     : integer;  {diameter of the symbol}
  ix,iy   : integer;  {coordinates}
begin
  GetMarkerParam(ipu,HorizT,ipv,ipw);
  {type and color of the marker are coded in itype, extract them}
  jcolor := (itype-1) div HorizT;    {color index}
  jtype := itype-jcolor*HorizT;      {type index}
  if      (jcolor = 0) then toile.pen.color := clRed
  else if (jcolor = 1) then toile.pen.color := clBlue
  else                       toile.pen.color := clBlack;
  toile.brush.style := bsClear;
  toile.pen.width := 1;
  if (jtype = 1) then begin               {a plain circle}
    ix := cx-round(scale*d4);
    iy := cy-round(scale*d4);
    dia := round(scale*2*d4);
    toile.ellipse(ix,iy+yof,ix+dia,iy+dia+yof);
  end else if (jtype = 2) then begin      {a square}
    ix := cx-round(scale*d4);
    iy := cy-round(scale*d4);
    dia := round(scale*2*d4);
    toile.rectangle(ix,iy+yof,ix+dia,iy+dia+yof);
  end else if (jtype = 3) then begin      {circle with a cross}
    ix := cx-round(scale*d3);
    iy := cy-round(scale*d3);
    dia := round(scale*2*d3);
    toile.ellipse(ix,iy+yof,ix+dia,iy+dia+yof);
    ipu := round(scale*1.5*d3);
    toile.moveto(cx-ipu,cy+yof);
    toile.lineto(cx+ipu,cy+yof);
    toile.moveto(cx,cy-ipu+yof);
    toile.lineto(cx,cy+ipu+yof);
  end else if (jtype = 4) then begin      {square with a cross}
    ix := cx-round(scale*d3);
    iy := cy-round(scale*d3);
    dia := round(scale*2*d3);
    toile.rectangle(ix,iy+yof,ix+dia,iy+dia+yof);
    ipu := round(scale*1.5*d3);
    toile.moveto(cx-ipu,cy+yof);
    toile.lineto(cx+ipu,cy+yof);
    toile.moveto(cx,cy-ipu+yof);
    toile.lineto(cx,cy+ipu+yof);
  end else if (jtype = 5) then begin      {circle with a X}
    ix := cx-round(scale*d3);
    iy := cy-round(scale*d3);
    dia := round(scale*2*d3);
    toile.ellipse(ix,iy+yof,ix+dia,iy+dia+yof);
    ipu := round(scale*1.5*d3);
    toile.moveto(cx-ipu,cy-ipu+yof);
    toile.lineto(cx+ipu,cy+ipu+yof);
    toile.moveto(cx+ipu,cy-ipu+yof);
    toile.lineto(cx-ipu,cy+ipu+yof);
  end;
end;

procedure MarkerMenu(pImage : TpImage; pPanel : TpPanel; pGrid : TpGrid;
                     row : integer);
{prepare the marker menu: an image on a panel. The image contains small
 images of the markers. The panel is put on the grid near to the row for markers
 pImage  pointer to the image on which the pictures of the markers are drawn
 pPanel  pointer to the panel, where the image is
 pGrid   pointer to the grid for marker parameters
 row     the row for the marker type-parameter}
var
  cx,cy   : integer;   {coordinates of the center of the marker}
  ii,jj   : integer;   {indices}
  itype   : integer;   {index for the type of the marker}
  HorizT  : integer;    {no of columns of marker types}
  MargT   : integer;    {margin}
  SizeT   : integer;    {size of the square for one marker type}
  VerticT : integer;    {no of rows of marker types}
begin
  GetMarkerParam(VerticT,HorizT,SizeT,MargT);
  pImage^.width := HorizT*SizeT;
  pImage^.height := VerticT*SizeT;
  pImage^.top := MargT;
  pImage^.left := MargT;
  pPanel^.width := pImage^.width+2*MargT;
  pPanel^.height := pImage^.height+2*MargT;
  pPanel^.top := pGrid^.top+Row*(pGrid^.DefaultRowHeight+
                   pGrid^.GridLineWidth)-pPanel^.height;
  pPanel^.left := pGrid^.left+pGrid^.DefaultColWidth;
  {draw the markers on the "menu"}
  for ii := 1 to verticT do begin
    for jj := 1 to horizT do begin
      itype := (ii-1)*horizT+jj;
      cx := (jj-1)*SizeT+(sizeT div 2);
      cy := (ii-1)*SizeT+(sizeT div 2);
      DrawMark(pImage^.canvas, cx, cy,0, itype , 2);
    end;
  end;
end;

procedure TMarkerForm.FormCreate(Sender: TObject);
{initialize the form}
begin
  {determine the char.for the decimal point}
  DecimalPoint(d_p);  {in routines.pas}
  {initialize the grid for the parameters of the markers}
  with MarkerGrid do begin
    {first, the captions}
    cells[0,NameRow] := 'name';
    cells[0,CurrRow] := 'current (A)';
    cells[0,TimeRow] := 'time (s)';
    cells[0,VoltRow] := 'voltage (V)';
    cells[0,TypeRow] := 'type';
    {then the values}
    cells[1,NameRow] := 'marker 1';
    cells[1,CurrRow] := '1000';
    cells[1,TimeRow] := '1';
    cells[1,VoltRow] := '1000';
    cells[1,TypeRow] := '1';
  end;
  {prepare image and its background panel for selecting the type of the marker}
  MarkerMenu(@TypeImage, @TypePanel, @MarkerGrid, TypeRow);
  {add flag that form created}
  Imagoo.ReadyEdit.text := Imagoo.ReadyEdit.text+'m';
end;

procedure TMarkerForm.FormShow(Sender: TObject);
{copy the voltage level of the relay on display to the parameters of the marker}
var
  apu  : single;    {temporary}
  ii   : integer;   {index}
  len  : integer;   {length of text}
  ll   : longint;   {temporary}
  line : string;    {temporary}
  text : string;    {temporary}
  Volt : single;    {the voltage level}
begin
  RelayCharact.GetVoltLev(Volt);          {get the voltage of the relay}
  text := format('%8.0f',[Volt]);
  {remove leading spaces}
  len := length(text);
  line := '';
  if (len > 0) then begin
    for ii:=1 to len do begin             {length depends on format}
      if (text[ii] <> ' ') then begin     {skip all spaces}
        line := line+text[ii];            {append "non-space" characters}
      end;
    end;
  end;
  MarkerGrid.cells[1,VoltRow] := line;
  {first guess for the position of the marker }
  if (MarkerList.Items.count = 0) then begin
    apu := RelayCharact.EndCurr;
    ll := round(apu/2);
    MarkerGrid.Cells[1,CurrRow] := inttostr(ll);
  end;
  MarkerGrid.Cells[1,TimeRow] := '1';
  MarkerGrid.Cells[1,TypeRow] := '1';
  ModifyBut.enabled := false;
end;

procedure TMarkerForm.MarkerGridSelectCell(Sender: TObject; Col,
  Row: Longint; var CanSelect: Boolean);
begin
  ActRow := row;
  {if row for marker types selected, show the "menu"}
  if (row = TypeRow) then begin
    {cannot edit the row for the marker type}
    MarkerGrid.options := MarkerGrid.options-[goEditing,goAlwaysShowEditor];
    {show the menu}
    TypePanel.bringtofront;
    TypePanel.visible := true;
  end else begin
    MarkerGrid.options := MarkerGrid.options+[goEditing,goAlwaysShowEditor];
    TypePanel.visible := false;
  end;
end;

procedure MarkerIndexToCell(x,y : integer; pGrid : TpGrid; pPanel : TpPanel;
                             row : integer);
{put the index of the marker to the cell of the marker grid, hide panel
 x,y     mouse coordinates
 pGrid   pointer to the string grid
 pPanel  pointer to the panel, where the marker types are displayed}
var
  ii,jj,kk     : integer;
  HorizT,SizeT : integer; {parameters for the markers}
begin
  GetMarkerParam(ii,HorizT,SizeT,jj);
  ii := (X div SizeT)+1;
  jj := (Y div SizeT)+1;
  kk := (jj-1)*HorizT+ii;                 {type index}
  pGrid^.cells[1,Row] := inttostr(kk);
  pPanel^.visible := false;
end;

procedure TMarkerForm.TypeImageMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
{get index for marker type from X and Y}
begin
  MarkerIndexToCell(x,y, @MarkerGrid, @TypePanel,TypeRow);
end;

procedure TMarkerForm.FormKeyPress(Sender: TObject; var Key: Char);
begin
  {accept only numbers + '.' or ',', when entering data}
  if (ActRow > 0) then begin             {first row is for text}
    if ((key = ',') or (key = '.')) then key := d_p
    {only numbers and . or , accepted}
    else if not(((key >= '0') and (key <= '9')) or
{                (key  = ',') or  (key  = '.')  or }
{                (key  = 'e') or  (key  = 'E')  or }
                 (key  = #8 ) or
                 (key =  #3 ) or (key = #22)) then   {"copy" and "paste"}
      key := #0;
  end else if (key = ';') then begin   {row for the name of the marker}
    key := #0;                         {';' used as separator}
  end;
end;

procedure TMarkerForm.ConstructLine(var line: string; var ertot: integer);
{construct the line with the parameters of the marker
 line    the line with the parameters
 ertot   error flag}
var
  dp    : Char;      {decimal point: (,) or (.)}
  err   : integer;  {error flag}
{ ertot : integer;  {total error}
  ii,jj : integer;
{ line  : string;}
  pComp : TpComp;   {pointer to the active relay}
  txt   : string;   {temporary}
begin
  {check first that the numbers are correct}
  ertot := 0;
  for ii := 1 to MarkerGrid.RowCount-1 do begin
    line := MarkerGrid.cells[1,ii];
    chkfloat(line, err,txt);
    ertot := ertot+err;
    if (err <> 0) then begin
      display_message(MarkerForm,MsgBut,'error in row '+inttostr(ii+1));
    end;
  end;
  {add only if no errors}
  if (ertot = 0) then begin
    line :=         MarkerGrid.cells[1,0]+';'+
            ' I = '+MarkerGrid.cells[1,1]+'A;'+
            ' t = '+MarkerGrid.cells[1,2]+'s;'+
            ' V = '+MarkerGrid.cells[1,3]+'V;'+
            '   ( '+MarkerGrid.cells[1,4]+' )';
  end else begin
    line := '';
  end;
end;

procedure TMarkerForm.AddButClick(Sender: TObject);
{add a new marker to the list of markers}
var
  dp    : Char;      {decimal point: (,) or (.)}
  err   : integer;  {error flag}
  ertot : integer;  {total error}
  ii,jj : integer;
  line  : string;
  pComp : TpComp;   {pointer to the active relay}
begin
  ConstructLine(line,ertot);    {construct line with parameters}
(*
  {check first that the numbers are correct}
  ertot := 0;
  for ii := 1 to MarkerGrid.RowCount-1 do begin
    line := MarkerGrid.cells[1,ii];
    chkfloat(line, err);
    ertot := ertot+err;
    if (err <> 0) then begin
      display_message(MarkerForm,MsgBut,'error in row '+inttostr(ii+1));
    end;
  end;
  {add only if no errors}
  if (ertot = 0) then begin
    line :=         MarkerGrid.cells[1,0]+';'+
            ' I = '+MarkerGrid.cells[1,1]+'A;'+
            ' t = '+MarkerGrid.cells[1,2]+'s;'+
            ' V = '+MarkerGrid.cells[1,3]+'V;'+
            '   ( '+MarkerGrid.cells[1,4]+' )';
*)
  if (ertot = 0) then begin
    MarkerList.items.add(line);
    pComp := TpComp(RelayCharact.tag);        {recall pointer to the relay}
    if (pComp <> nil) then Relay_show(pComp,'real'); {redraw the curves}
  end;
end;

procedure TMarkerForm.MsgButClick(Sender: TObject);
begin
  MsgBut.visible := false;
end;

procedure TMarkerForm.DelButClick(Sender: TObject);
{delete the selected marker}
var
  iMark : integer;
  pComp : TpComp;   {pointer to the (last clicked) relay of the graph}
begin
  iMark := MarkerList.ItemIndex;    {index of selected marker}
  if (iMark >= 0) then begin        {some item selected}
    MarkerList.Items.delete(iMark); {delete it}
    RelayCharact.DeleteMarkerIDCoord(iMark);  {update list of ID-coordinates}
    pComp := TpComp(RelayCharact.tag);        {recall pointer to the relay}
    if (pComp <> nil) then Relay_show(pComp,'real'); {redraw the curves}
  end;
end;

procedure DeleteAllMarkers;
{delete the selected marker}
begin
  MarkerForm.MarkerList.Items.clear;
  MarkerForm.visible := false;
end;

procedure TMarkerForm.FormActivate(Sender: TObject);
begin
  {first guess for the position of the marker }
{ if (MarkerList.Items.count = 0) then begin
    apu := RelayCharact.EndCurr;
    ll := round(apu/2);
    MarkerGrid.Cells[1,CurrRow] := inttostr(ll);
  end;}   {moved to formshow}
end;



procedure TMarkerForm.MarkerListClick(Sender: TObject);
{copy the data from the selected line to the grid}
var
  line    : string;  {line with the data of the marker}
  txa,txb : string;  {temporary}
begin
  line := MarkerList.items[MarkerList.ItemIndex];
  {copy parameters one by one}
  ExtractParam(1,';',line,txa);         {identifier}
  MarkerGrid.Cells[1,NameRow] := txa;   {identifier taken as such}
  {current}
  ExtractParam(2,';',line,txa);         {current}
  GetNumericalString(txa, txb);         {extract numerical value as text}
  MarkerGrid.Cells[1,CurrRow] := txb;   {current to grid}
  {time}
  ExtractParam(3,';',line,txa);         {time}
  GetNumericalString(txa, txb);         {extract numerical value as text}
  MarkerGrid.Cells[1,TimeRow] := txb;   {time to grid}
  {voltage}
  ExtractParam(4,';',line,txa);         {voltage}
  GetNumericalString(txa, txb);         {extract numerical value as text}
  MarkerGrid.Cells[1,VoltRow] := txb;   {voltage to grid}
  {marker type}
  ExtractParam(5,';',line,txa);         {index of marker type}
  GetNumericalString(txa, txb);         {extract numerical value as text}
  MarkerGrid.Cells[1,TypeRow] := txb;   {type index to grid}
  ModifyBut.enabled := true;
end;



procedure TMarkerForm.ModifyButClick(Sender: TObject);
{modify the selected marker = replace parameters in the list and in the description
 with the parameters in the grid}
var
  ertot    : integer;  {error flag}
  DataLine : string;   {data for the marker in one line}
  iMark    : integer;  {index of the marker}
  pComp    : TpComp;   {pointer to the relay}
begin
  ertot := 0;                                 {initialize}
  ConstructLine(DataLine, ertot);             {collect the parameters to one line}
  iMark := MarkerList.ItemIndex;              {index of selected marker}
  if (ertot  = 0) and                         {no errors}
     (iMark >= 0) and (iMark < MarkerList.items.count) then begin  {reasonable index}
    {replace the data in the list visible on the lower part of the marker form}
    MarkerList.Items.Delete(iMark);           {delete the old line}
    MarkerList.Items.Insert(iMark,DataLine);  {add the new line}
    pComp := TpComp(RelayCharact.tag);        {recall pointer to the relay}
    if (pComp <> nil) then Relay_show(pComp,'real'); {redraw the curves}
  end;
  ModifyBut.enabled := false;
end;

end.
