unit matrix;

interface

uses

  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, StdCtrls, UAlnEdit, ExtCtrls, Buttons, Menus,Compdef, ShellApi,
  phasor,utildef, JpgImg;

const
  nVect = 6;      {number of vectors}
  nUndo = 90{72}; {size of the "undo" array}
  marg  = 3;      {the dimensions of the rectangle = 2*marg larger than the button}

type          {for receiving currents from elplek}
  TSixSingle = packed record
  x: single;
  y: single;
  z: single;
  u: single;
  v: single;
  w: single;
end;

type
  TPhasorData = record
  phi : integer;  {index}
  ia  : integer;  {color index}
  ib  : integer;  {color index}
  ic  : integer;  {color index}
  x1  : integer;  {relative coordinate of identifier}
  y1  : integer;
  x2  : integer;
  y2  : integer;
  x3  : integer;
  y3  : integer;
  ind : integer;  {index for scaling the graph}
  txt : string;   {the caption, or title}
end;

{a type defining the sheet type}
type
  Tsheet = (matr, vect);

type
  TpAlEdit = ^TAlignEdit; {pointer to align edit}
  TpEdit = ^TEdit;        {pointer to an edit}
  TpForm = ^TForm;
  TpGrid = ^TStringGrid;  {pointer to an string grid}
  TpCombo = ^TComboBox;   {pointer to combo box}
  M3x3 = array[0..8] of double;   {for permanent matrices}
  TpM3x3 = ^M3x3;
  TpPhasorData = ^TPhasorData;
  TpSaveDialog = ^TSaveDialog;
  TpTextFile = ^TextFile;
  TpAlEditArr = array[1..nVect] of TpAlEdit; {pointers to the constants}
  TpGridArr = array[1..nVect] of TpGrid;     {pointers to the vectors}
  TstrArr = array[1..nUndo] of string;       {for undo}

const
  {the compensation matrices, without the multiplier}
  m00 : M3x3 = ( 1, 0, 0, {}  0, 1, 0, {}  0, 0, 1);
  m01 : M3x3 = ( 1,-1, 0, {}  0, 1,-1, {} -1, 0, 1);
  m02 : M3x3 = ( 1,-2, 1, {}  1, 1,-2, {} -2, 1, 1);
  m03 : M3x3 = ( 0,-1, 1, {}  1, 0,-1, {} -1, 1, 0);
  m04 : M3x3 = (-1,-1, 2, {}  2,-1,-1, {} -1, 2,-1);
  m05 : M3x3 = (-1, 0, 1, {}  1,-1, 0, {}  0, 1,-1);
  m06 : M3x3 = (-2, 1, 1, {}  1,-2, 1, {}  1, 1,-2);
  m07 : M3x3 = (-1, 1, 0, {}  0,-1, 1, {}  1, 0,-1);
  m08 : M3x3 = (-1, 2,-1, {} -1,-1, 2, {}  2,-1,-1);
  m09 : M3x3 = ( 0, 1,-1, {} -1, 0, 1, {}  1,-1, 0);
  m10 : M3x3 = ( 1, 1,-2, {} -2, 1, 1, {}  1,-2, 1);
  m11 : M3x3 = ( 1, 0,-1, {} -1, 1, 0, {}  0,-1, 1);
  m12 : M3x3 = ( 2,-1,-1, {} -1, 2,-1, {} -1,-1, 2);
  onesq3 = 0.577350269;    {1/sqrt(3)}
  one3rd = 0.333333333;    {1/3}
  {the multipliers of the compensation matrices}
  mcoeff : array[0..12] of double =
           (1.0,onesq3,one3rd,onesq3,one3rd,onesq3,one3rd,
               onesq3,one3rd,onesq3,one3rd,onesq3,one3rd);


function  TwoDecimalsAtMost(rr : double): string;
procedure DecimalPoint(var decp: Char);
function  ManyChars(rr : double) : string;
function  spaceless(text : string): string;
function  TenChar(txt: string): string;
{define an object for calculations and operations}
type
  TMatrCalc = class(TObject)
  procedure initialize(pLOImg: TpImage);
  procedure PutR11(pra,prb,prc,prd,pre,prf: TpAlEdit);
  procedure PutR12(pra,prb,prc,prd,pre,prf: TpAlEdit);
  procedure PutR21(pra,prb,prc,prd,pre,prf: TpAlEdit);
  procedure PutR22(pra,prb,prc,prd,pre,prf: TpAlEdit);
  procedure PutRes(pra,prb,prc,prd,pre,prf: TpAlEdit);
  procedure PutpGrid(pa,pb,pc,pd,pe,pf,pch: TpGrid);
  procedure GetpGrid(var pGrid : TpGridArr);
  procedure PutpMgrid(pa,pb,pc,pd,pe,pf: TpGrid);
  procedure PutpMatr(pMatr: TpGrid);
  procedure PutpErrEd(pErEd: TpEdit);
  procedure MarkLastOperation(ButLeft,ButTop,ButWidth,ButHeight: integer);
  procedure BeforeLastOper;
  procedure PutVectIndex(ind : integer);
  procedure GetVectIndex(var ind : integer);
  procedure ClearVect;
  procedure ClearScalar;
  procedure ClearAll;
  procedure SaveState;
  procedure UndoLastOper;
  procedure SaveGrid(ig: integer);
  procedure ValuesFromLine(line: String; nstrt :Integer; var ipar: Integer;  {in routines.pas}
                           var pa1,pa2,pa3,pa4 : double;
                           var errc, ierr :Integer);
  procedure GetVector(ig: integer; var Vmagn, Vangl: array of double;
                      var era,erm: array of integer );
  procedure CalcButClick(pr11,pr12,pr21,pr22,pRes: TpAlEdit; ig: integer;
                      var someerr: boolean);
  procedure PutAngle(angl : double; ii,dec : integer);
  procedure PutMagn(magn : double; ii : integer);
  procedure GetReImVector(var Vreim: array of Dcomplex; var ertot : integer);
  procedure PutReImVector(Vreim: array of Dcomplex);
  procedure PhaseOrSeq(ps: integer);
  procedure CurrToMultix(var curr :Tmultix; ii : integer; var ertot: integer);
  procedure CopyRow(fromrow,torow: integer; CopyBut: TSpeedButton);
  procedure SwapRows(row1,row2: integer; SwapBut: TSpeedButton);
  procedure RotateRows(up: boolean; RotBut: TSpeedButton);
  procedure PutAddEdit(pAddEdit: TpAlEdit);
  procedure AddAngle(row: integer; AngleBut: TSpeedButton; var err: integer);
  procedure ConjButClick(ConjBut: TSpeedButton);
  procedure NegButClick(NegBut: TSpeedButton);
  procedure HandleCopyDataRecord(copyDataStruct: PCopyDataStruct);
  procedure SetSheet(sheet: Tsheet);
  procedure SaveButClick(abc: char; SaveBut: TSpeedButton; var err: integer );
  procedure HideVector;
  procedure MoveAndShowVector(iv,ix,iy: integer);
  procedure PutVect(ii: integer; m1,m2,m3,a1,a2,a3: string);
  procedure CoeffFromEdit(pr11,pr12,pr21,pr22: TpAlEdit;
                       var r11, r12, r21, r22: double;
                       var er1, er2, er3, er4: integer);
  procedure Set3I03I2(used: boolean);
  procedure SavePhasors;
  procedure RestorePhasors;
  procedure WritePhasorPar(pFile: TpTextFile; nCT : integer);
  procedure ReadPhasorPar(pFile: TpTextFile);
  procedure PutForm(pForm: TpForm);
  procedure PhasorsToFile(pFile: TpTextFile);
  procedure PhasorsFromFile(pFile: TpTextFile);
  procedure SaveSheet(pDlg: TpSaveDialog);
  procedure SaveVectors(pFile :TpTextFile);
  procedure MarkVectorError(pVGrid: TpGrid; kk,ii : integer);

private
  backup_c   : TstrArr;     {array 1..nUndo for undo}
  coefcol_c  : Tcolor;      {color of the coefficient edit boxes}
  LOleft_c   : integer;     {left-coord of the latest operation (before the present one)}
  LOtop_c    : integer;
  LOWidth_c  : integer;
  LOheight_c : integer;
  nVect_c    : integer;     {number of vectors}
  pAddEdit_c : TpAlEdit;    {pointer to edit-comp. for adding an angle}
  pchGrid_c  : TpGrid;      {pointer to the check grid}
  pErrEd_c   : TpEdit;      {pointer to error edit}
  pForm_c    : TpForm;      {pointer to the form}
  pGrid_c    : TpGridArr;   {pointers to the vectors}
  pMatr_c    : TpGrid;      {pointer to the matrix on the matrix edit sheet}
  pMgrid_c   : TpGridArr;   {pointers to the grids on MxEditForm}
  pLOImg_c   : TpImage;     {pointer to the marker of the last operation}
  pr11_c     : TpAlEditArr; {pointers to the coefficients 11, array 1..nVect}
  pr12_c     : TpAlEditArr; {pointers to the coefficients 12}
  pr21_c     : TpAlEditArr; {pointers to the coefficients 11}
  pr22_c     : TpAlEditArr; {pointers to the coefficients 22}
  pres_c     : TpAlEditArr; {pointers to the resulting coefficients}
  sheet_c    : Tsheet;      {defines sheet type}
  rdVect_c   : integer;     {index of vector selected with the radio buttons}
  seq3_c     : boolean;     {3I0, 3I2 in use}
public
  nPhasor_c  : integer;     {number of phasors}
  PhaDat_c   : array[1..maxCT] of TPhasorData;  {for saving the phasor, maxCT in Phasor.pas}
  SeqDat_c   : array[1..maxCT] of TPhasorData;  {for saving the seq.phasor, maxCT in Phasor.pas}
  Vect_lbl   : array[1..nVect] of string;  
end;

type
  TMatrForm = class(TForm)
    RaEdit11: TAlignEdit;
    RaEdit12: TAlignEdit;
    RaEdit21: TAlignEdit;
    RaEdit22: TAlignEdit;
    RaMult1: TLabel;
    RaMult2: TLabel;
    RaLab1: TLabel;
    RaLab2: TLabel;
    RaLine: TShape;
    RaLab3: TLabel;
    RaLab4: TLabel;
    RaLab5: TLabel;
    RaLab6: TLabel;
    RaResEdit: TAlignEdit;
    RaCalcBut: TSpeedButton;
    VaGrid: TStringGrid;
    CheckGrid: TStringGrid;
    CheckButA: TButton;
    RbMult1: TLabel;
    RbMult2: TLabel;
    RbLab1: TLabel;
    RbLab2: TLabel;
    RbLine: TShape;
    RbLab3: TLabel;
    RbLab4: TLabel;
    RbLab5: TLabel;
    RbLab6: TLabel;
    RbCalcBut: TSpeedButton;
    RbEdit11: TAlignEdit;
    RbEdit12: TAlignEdit;
    RbEdit21: TAlignEdit;
    RbEdit22: TAlignEdit;
    RbResEdit: TAlignEdit;
    VbGrid: TStringGrid;
    CheckButB: TButton;
    VaLabel: TLabel;
    VbLabel: TLabel;
    MxMenu: TMainMenu;
    FileMenu: TMenuItem;
    ExitItem: TMenuItem;
    ABshape: TShape;
    RcMult1: TLabel;
    RcMult2: TLabel;
    RcLab1: TLabel;
    RcLab2: TLabel;
    RcLine: TShape;
    RcLab3: TLabel;
    RcLab4: TLabel;
    RcLab5: TLabel;
    RcLab6: TLabel;
    RcCalcBut: TSpeedButton;
    RcEdit11: TAlignEdit;
    RcEdit12: TAlignEdit;
    RcEdit21: TAlignEdit;
    RcEdit22: TAlignEdit;
    RcResEdit: TAlignEdit;
    VcGrid: TStringGrid;
    CheckButC: TButton;
    VcLabel: TLabel;
    RdMult1: TLabel;
    RdMult2: TLabel;
    RdLab1: TLabel;
    RdLab2: TLabel;
    RdLine: TShape;
    RdLab3: TLabel;
    RdLab4: TLabel;
    RdLab5: TLabel;
    RdLab6: TLabel;
    RdCalcBut: TSpeedButton;
    RdEdit11: TAlignEdit;
    RdEdit12: TAlignEdit;
    RdEdit21: TAlignEdit;
    RdEdit22: TAlignEdit;
    RdResEdit: TAlignEdit;
    VdGrid: TStringGrid;
    CheckButD: TButton;
    VdLabel: TLabel;
    CDshape: TShape;
    ReMult1: TLabel;
    ReMult2: TLabel;
    ReLab1: TLabel;
    ReLine: TShape;
    ReLab3: TLabel;
    ReLab4: TLabel;
    ReLab5: TLabel;
    ReLab6: TLabel;
    ReCalcBut: TSpeedButton;
    ReEdit11: TAlignEdit;
    ReEdit12: TAlignEdit;
    ReEdit21: TAlignEdit;
    ReEdit22: TAlignEdit;
    ReResEdit: TAlignEdit;
    VeGrid: TStringGrid;
    CheckButE: TButton;
    VeLabel: TLabel;
    RfMult1: TLabel;
    RfMult2: TLabel;
    RfLab1: TLabel;
    RfLab2: TLabel;
    RfLine: TShape;
    RfLab3: TLabel;
    RfLab4: TLabel;
    RfLab5: TLabel;
    RfLab6: TLabel;
    RfCalcBut: TSpeedButton;
    RfEdit11: TAlignEdit;
    RfEdit12: TAlignEdit;
    RfEdit21: TAlignEdit;
    RfEdit22: TAlignEdit;
    RfResEdit: TAlignEdit;
    VfGrid: TStringGrid;
    CheckButF: TButton;
    VfLabel: TLabel;
    BotShape: TShape;
    EFshape: TShape;
    MiddleShape: TShape;
    AddBoxA: TComboBox;
    PlusALab: TLabel;
    AddBoxB: TComboBox;
    PlusBlab: TLabel;
    AddBoxC: TComboBox;
    PlusClab: TLabel;
    AddBoxD: TComboBox;
    PlusElab: TLabel;
    AddBoxE: TComboBox;
    SumBut: TSpeedButton;
    RadioButA: TRadioButton;
    RadioButB: TRadioButton;
    RadioButC: TRadioButton;
    RadioButD: TRadioButton;
    RadioButE: TRadioButton;
    RadioButF: TRadioButton;
    DelVectBut: TSpeedButton;
    DelScalBut: TSpeedButton;
    DelAllBut: TSpeedButton;
    FillBut: TButton;
    UndoBut: TSpeedButton;
    NegBut: TSpeedButton;
    ConjBut: TSpeedButton;
    SeqBut: TSpeedButton;
    PhaseBut: TSpeedButton;
    SelectMatrBox: TComboBox;
    MatMultBut: TSpeedButton;
    ElplekBut: TSpeedButton;
    LastOperImg: TImage;
    PhasorBut: TSpeedButton;
    ViewMenu: TMenuItem;
    PhasorItem: TMenuItem;
    BotShapeL: TShape;
    CopyBut21: TSpeedButton;
    CopyBut31: TSpeedButton;
    CopyBut12: TSpeedButton;
    CopyBut32: TSpeedButton;
    CopyBut13: TSpeedButton;
    CopyBut23: TSpeedButton;
    SwapBut12: TSpeedButton;
    SwapBut13: TSpeedButton;
    SwapBut23: TSpeedButton;
    BotLeftShape: TShape;
    RotUpBut: TSpeedButton;
    RotDownBut: TSpeedButton;
    AddAngleEdit: TAlignEdit;
    AddAngleBut1: TSpeedButton;
    AddAngleBut2: TSpeedButton;
    AddAngleBut3: TSpeedButton;
    SaveBox: TComboBox;
    SaveBut: TSpeedButton;
    MatrEditMenu: TMenuItem;
    ReLab2: TLabel;
    OptionsMenu: TMenuItem;
    Seq3Item: TMenuItem;
    Seq3Label: TLabel;
    SaveItem: TMenuItem;
    OpenItem: TMenuItem;
    LineItem: TMenuItem;
    SaveFormDlg: TSaveDialog;
    ExportSheetItem: TMenuItem;
    NumberItem: TMenuItem;
    ErrEdit: TEdit;
    NotesItem: TMenuItem;
    ButtonItem: TMenuItem;
    HelpItem: TMenuItem;
    HelpTopics: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure RaCalcButClick(Sender: TObject);
    procedure SaveOrRestore(pFile: TpTextFile; save: boolean);
    { - - - - - - - - }
    procedure DrawCell(pGrid: TpGrid; Col, Row: Longint;
              Rect: TRect; State: TGridDrawState);
    procedure SelectCell(pGrid: TpGrid; Col, Row: Longint;
              var CanSelect: Boolean);
    procedure m12mult;
    procedure UpdatePhasors;
{    procedure VxPointers;}
    { - - - - - }
    procedure VaGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure VaGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure CheckButAMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure CheckButAMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure CheckGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure CheckGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure VbGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure RbCalcButClick(Sender: TObject);
    procedure VbGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure ExitItemClick(Sender: TObject);
    procedure RcCalcButClick(Sender: TObject);
    procedure VcGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure VcGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure RdCalcButClick(Sender: TObject);
    procedure VdGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure VdGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure ReCalcButClick(Sender: TObject);
    procedure VeGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure VeGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure RfCalcButClick(Sender: TObject);
    procedure VfGridDrawCell(Sender: TObject; Col, Row: Longint;
      Rect: TRect; State: TGridDrawState);
    procedure VfGridSelectCell(Sender: TObject; Col, Row: Longint;
      var CanSelect: Boolean);
    procedure SumButClick(Sender: TObject);
    procedure RadioButAClick(Sender: TObject);
    procedure RadioButBClick(Sender: TObject);
    procedure RadioButCClick(Sender: TObject);
    procedure RadioButDClick(Sender: TObject);
    procedure RadioButEClick(Sender: TObject);
    procedure RadioButFClick(Sender: TObject);
    procedure DelVectButClick(Sender: TObject);
    procedure DelScalButClick(Sender: TObject);
    procedure DelAllButClick(Sender: TObject);
    procedure FillButClick(Sender: TObject);
    procedure UndoButClick(Sender: TObject);
    procedure NegButClick(Sender: TObject);
    procedure ConjButClick(Sender: TObject);
    procedure SeqButClick(Sender: TObject);
    procedure PhaseButClick(Sender: TObject);
    procedure MatMultButClick(Sender: TObject);
    procedure ElplekButClick(Sender: TObject);
    procedure PhasorButClick(Sender: TObject);
    procedure PhasorItemClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure CopyBut21Click(Sender: TObject);
    procedure CopyBut31Click(Sender: TObject);
    procedure CopyBut12Click(Sender: TObject);
    procedure CopyBut32Click(Sender: TObject);
    procedure CopyBut13Click(Sender: TObject);
    procedure CopyBut23Click(Sender: TObject);
    procedure SwapBut12Click(Sender: TObject);
    procedure SwapBut13Click(Sender: TObject);
    procedure SwapBut23Click(Sender: TObject);
    procedure RotUpButClick(Sender: TObject);
    procedure RotDownButClick(Sender: TObject);
    procedure AddAngleBut1Click(Sender: TObject);
    procedure AddAngleBut2Click(Sender: TObject);
    procedure AddAngleBut3Click(Sender: TObject);
    procedure SaveButClick(Sender: TObject);
    procedure MatrEditMenuClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Seq3ItemClick(Sender: TObject);
    procedure SaveItemClick(Sender: TObject);
    procedure OpenItemClick(Sender: TObject);
    procedure ExportSheetItemClick(Sender: TObject);
    procedure NumberItemClick(Sender: TObject);
    procedure NotesItemClick(Sender: TObject);
    procedure ButtonItemClick(Sender: TObject);
    procedure HelpTopicsClick(Sender: TObject);

  private
    { Private declarations }
    Caption_main : string; {the original, proper caption (needed for transferring data from elplek}
    Orig_Width     : integer;
    Orig_height    : integer;
    procedure WMCopyData(var Msg: TWMCopyData); message WM_COPYDATA;
    procedure HandleCopyDataRecord(copyDataStruct: PCopyDataStruct);
//  procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public declarations }
    NumbersInitialDir: string;               {for temporary storage}
    PhasorsInitialDir: string;               {for temporary storage}
    NotesInitialDir: string;
    Comp_Matr : array[0..12] of M3x3;        {all compensation matrices}
  end;

var
  MatrForm  : TMatrForm;
  Matr_calc : TMatrCalc;  {object for calculations and operations}
{cpf : TCurrPhasorForm;}

const
  LightBlue = $FFCC99;
  pi = 3.1415926535897;
var
  dec_p  : char;  {decimal point character '.' or ','}
implementation

uses Mxedit, numbers, Notes;

{$R *.DFM}

{var
  Vect_lbl : array[1..nVect] of string;}
const
  iSavedGrid      : integer = 0;   {index of the saved grid}
  iSavedGrid_prev : integer = 0;   {previous index}
  left_prev       : integer = 0;   {left-coord. of the previous pos. of the check grid}
  top_prev        : integer = 0;


procedure DecimalPoint(var decp: Char);
{determines the character for decimal point, is it a full stop (.),
 or a comma (,), and copies it to dec_p}
var
  apu   : Single;    {temporary}
  i     : Integer;   {counter, or index}
  iapu  : Integer;   {length of line, temporary}
  line  : String;    {apu as string, temporary}
  tx    : Char;      {one character, temporary}
begin
  {determine, is  comma (,) or period (.) used as decimal point}
  apu := 0.5;                   {need a value with decimal point}
  line := floattostr(apu);      {change it to string}
  iapu := length(line);         {length of the string (should be 3 !)}
  tx := ',';                    {default decimal point}
  for i:=1 to iapu do begin
    if (line[i] = '.') or (line[i] = ',') then begin
      tx := line[i];            {pick . or ,}
      break;
    end;
  end;
  decp := tx;                   {copy the decimal point to output}
end;

function TwoDecimalsAtMost(rr : {single}double): string;
{express a small single as a string with at most two decimals}
var
  found : boolean;  {true, when the decimal point found}
  id    : integer;   {counts numbers after dec.point}
  ii    : integer;   {index}
  len   : integer;   {length of string}
  txa   : string;    {temporary}
  txt   : string;    {temporary string}
begin
  txt := floattostr(rr);
  len := length(txt);
  id := 0;
  found := false;
  {scan and copy, copy at most two decimals}
  txa := '';
  for ii := 1 to len do begin
    txa := txa+txt[ii];
    if (found) then begin        {decimal point has been found, count characters}
      inc(id);
      if (id >= 2) then break;   {at most two char. after the decimal point}
    end;
    if (txt[ii] = '.') or (txt[ii] = ',') then found := true; {decimal point}
  end;
  TwoDecimalsAtMost := txa;
end;
(*
procedure TMatrForm.CreateParams(var Params: TCreateParams);
{this is needed in order to keep forms always on top
 (fsStayOnTop does not work, when two or more forms)}
begin
  inherited CreateParams(Params);
  with Params do begin
    ExStyle := ExStyle or WS_EX_TOPMOST;
    WndParent := GetDesktopwindow;
  end;
end;
*)
procedure TMatrForm.FormCreate(Sender: TObject);
var
  ExeFile : string;    {name of this file, without the extension}
  ExeName : string;    {name of this file, with the extension}
  ii      : integer;   {index}
  ipu     : integer;   {temporary}
  ix,iy   : integer;   {coordinates}
  path    : String;    {path to the help file}
begin
{cpf := TCurrPhasorForm.create(MatrForm);}
{cpf.parent := MatrForm;                 }
  Orig_Width  := width;     {save the original size}
  Orig_height := height;
  DecimalPoint(dec_p);      {determine the char. for the decimal point}
  Caption_main := caption;  {save the original caption}
  Matr_calc := TMatrCalc.create;   {creatr object for operations}
  Matr_Calc.initialize(@LastOperImg);
  Matr_Calc.SetSheet(vect);        {define sheet type}
  {pointers of the coefficients to the arrays}
  Matr_Calc.PutR11(@RaEdit11,@RbEdit11,@RcEdit11,@RdEdit11,@ReEdit11,@RfEdit11);
  Matr_Calc.PutR12(@RaEdit12,@RbEdit12,@RcEdit12,@RdEdit12,@ReEdit12,@RfEdit12);
  Matr_Calc.PutR21(@RaEdit21,@RbEdit21,@RcEdit21,@RdEdit21,@ReEdit21,@RfEdit21);
  Matr_Calc.PutR22(@RaEdit22,@RbEdit22,@RcEdit22,@RdEdit22,@ReEdit22,@RfEdit22);
  Matr_Calc.PutRes(@RaResEdit,@RbResEdit,@RcResEdit,@RdResEdit,@ReResEdit,@RfResEdit);
  Matr_Calc.PutpGrid(@VaGrid,@VbGrid,@VcGrid,@VdGrid,@VeGrid,@VfGrid,@CheckGrid);
  Matr_Calc.PutpMatr(nil);
  Matr_Calc.PutForm(@MatrForm);
  {pointers to vectors in matrix edit form done in Mxedit.pas}
  Matr_Calc.PutAddEdit(@AddAngleEdit);
  Matr_Calc.Set3I03I2(Seq3Item.checked);
  Matr_Calc.PutpErrEd(@ErrEdit);
  {initialize radio-index of the vectors}
  if      (RadioButA.checked) then {rd_Vect} ipu := 1
  else if (RadioButB.checked) then {rd_Vect} ipu := 2
  else if (RadioButC.checked) then {rd_Vect} ipu := 3
  else if (RadioButD.checked) then {rd_Vect} ipu := 4
  else if (RadioButE.checked) then {rd_Vect} ipu := 5
  else if (RadioButF.checked) then {rd_Vect} ipu := 6
  else begin  {no button checked}
    RadioButA.checked := true;
    {rd_Vect} ipu := 1;
  end;
  Matr_calc.rdVect_c := ipu{rd_vect};        {initialize (change!!?)}
  {background color of the coefficients}
  RaEdit11.color := LightBlue;
  RaEdit12.color := LightBlue;
  RaEdit21.color := LightBlue;
  RaEdit22.color := LightBlue;
  RbEdit11.color := LightBlue;
  RbEdit12.color := LightBlue;
  RbEdit21.color := LightBlue;
  RbEdit22.color := LightBlue;
  RcEdit11.color := LightBlue;
  RcEdit12.color := LightBlue;
  RcEdit21.color := LightBlue;
  RcEdit22.color := LightBlue;
  RdEdit11.color := LightBlue;
  RdEdit12.color := LightBlue;
  RdEdit21.color := LightBlue;
  RdEdit22.color := LightBlue;
  ReEdit11.color := LightBlue;
  ReEdit12.color := LightBlue;
  ReEdit21.color := LightBlue;
  ReEdit22.color := LightBlue;
  RfEdit11.color := LightBlue;
  RfEdit12.color := LightBlue;
  RfEdit21.color := LightBlue;
  RfEdit22.color := LightBlue;
  CheckGrid.color := {clWhite}LightBlue;
  {construct the compensation matrices as one two-dimensional matrix}
  m12mult;
  {initialize the vectors and coefficients = clear}
  DelAllButClick(nil);
(*{labels of vectors, for phasors}
  Vect_lbl[1] := 'Va';
  Vect_lbl[2] := 'Vb';
  Vect_lbl[3] := 'Vc';
  Vect_lbl[4] := 'Vd';
  Vect_lbl[5] := 'Ve';
  Vect_lbl[6] := 'Vf'; *)
  NumbersInitialDir := 'C:\temp';   {temporary, change when reading the *.ini file..}
  PhasorsInitialDir := 'C:\temp';   {..in Mxedit.pas}
  NotesInitialDir := 'C:\temp';
  {find the help-file}
  ExeName := ExtractFileName(paramStr(0));
  ExeFile := ChangeFileExt(ExeName, '');   {remove the extension}
  path := ExtractFilePath(paramStr(0)) + ExeFile+'.hlp';
  if FileExists(path) then Application.HelpFile := path
  else begin
    showmessage('help file not found');
    HelpItem.enabled := false;
  end;
end;

procedure TMatrForm.FormKeyPress(Sender: TObject; var Key: Char);
{check that only numerals and comma or full stop are accepted at the Edit}
begin
{showmessage(activeControl.name);}
  {only numbers and . or , accepted}
  if not(((key >= '0') and (key <= '9')) or
          (key  = ',') or  (key  = '.')  or
          (key  = 'e') or  (key  = 'E')  or
          (key  = '-') or  (key  = '+')  or
          (key  = #8 ) or
          (key =  #3 ) or (key = #22)) then       {"copy" and "paste"}
           key := #0;                             {neglect other keys }
  {check correct char for decimal point}
  if      (dec_p = ',') and (key = '.') then key := ','
  else if (dec_p = '.') and (key = ',') then key := '.';
  {hide the error indicator (if visible)}
  ErrEdit.visible := false;
  ErrEdit.left := 0;
  ErrEdit.top := 0;
end;

function spaceless(text : string): string;
{remove spaces from text}
var
  ii    : integer;
  len   : integer;
  txout : string;
begin
  len := length(text);
  txout := '';
  if (len > 0)  then begin
    for ii := 1 to len do begin
      if (text[ii] <> ' ') then txout := txout+text[ii];
    end;
  end;
  spaceless := txout;
end;

function ManyChars(rr : double) : string;
{write a positive or negative real using several places, fill with leading spaces (?)
 rr    the real
 text  the outputted text}
const
  small = 1e-8;
var
  aa       : double;     {abs value of rr}
  first    : boolean;
  ii,jj,kk : integer;
  len      : integer;    {length of line}
  line     : string;     {text under construction}
  ss       : string;     {space or minus sign}
  started  : boolean;    {nonzero number found}
  text     : string;
begin
  aa := abs(rr);
  if (aa < small) then begin
    text := ' 0.0      ';
  end else if (aa < 0.0001) then
    text := format('%9.7g',[aa])
  else if (aa < 1.0) then begin
    line := format('%10.8f',[aa]);
    {remove trailing zeroes}
    len := length(line);
    if (len > 1) then begin
      text := '';
      for ii := 1 to len do begin
        text := text+' ';          {lengt same as line, but only spaces}
      end;
      started := false;            {not yet found a nonzero number}
      kk := len;                   {last char. in text}
      for ii := 1 to len do begin
        jj := len-ii+1;
        if (line[jj] = '0') and (not started) then begin
          {zeroes not copied}
        end else begin
          text[kk] := line[jj];
          dec(kk);
          started := true;         {a nonzero number found}
        end;
      end;
    end else begin
      text := line;
    end;
  end else if (aa < 1.e5) then begin
    if (aa = round(aa)) then text := format('%7.6g',[aa])+'.0'  {integer}
    else text := format('%9.8g',[aa]);
  end else if (aa < 1.e20) then
    text := format('%9.8g',[aa])
  else
    text := '1.0e20    ';
  len := length(text);
  line := '';
  if (len > 0) then begin
    {add the minus sign for a negative number}
    if (aa >= small) then begin    {no minus sign to zero}
      if (rr < 0) then ss := '-'
      else             ss := ' ';
      first := true;
      for ii := 1 to len do begin
        if (text[ii] <> ' ') and first then begin
          line := line+ss;
          first := false;
        end;
        line := line+text[ii];
      end;
    end else begin
      line := text;      {zero...}
    end;
  end;
  ManyChars := line;
end;


procedure chkfloat(line: String; var ercod: Integer);    {in routines.pas !!}
{checks, whether the line corresponds to a valid floating point
 (or integer) number
 line    the string that should be a number (in)
 ercod   error code,  = 0, if correct number}
var
  dp   : Char;      {decimal point: (,) or (.)}
  epart: string;    {exponential part}
  i    : integer;   {index, or counter}
  nch  : integer;   {number of characters in the string to be studied}
  ph   : integer;   {phases, depending what has been found
                     0 = start
                     1 = only numbers found
                     2 = + or -  found
                     3 = , (or .) found
                     4 = , (r .) and numbers found
                     5 = e or E found
                     6 = + or - after e or E found
                     7 = one number found after e or E or e+,e-,E+E-
                     8 = two numbers found after e ...}
  tx   : char;      { a single character}
  begin
    DecimalPoint(dp);             {get the decimal point character}
    ph := 0;                      {initialize phase}
    ercod := 0;                   {initialize error code}
    epart := '';                  {initialize exponential part}
    nch := length(line);          {number of characters}
    for i:=1 to nch do begin      {scan the whole line}
      tx := line[i];              {copy to local}
      if (ph = 0) then begin      {just starting}
      {first character must be only + - , e E  a number  space}
        if      (tx = ' ')                 then ph := 0
        else if (tx >='0') and (tx <= '9') then ph := 1
        else if (tx ='+') or (tx = '-')    then ph := 2
        else if (tx = dp)                  then ph := 3
        else if (tx = 'e') or (tx = 'E')   then ph := 5
        else begin
          ercod := 1;             {set error code}
          break;                  {errror found, stop}
        end;
      end else if (ph = 1) then begin
      {after a number only  another number  ,  e E allowed}
        if      (tx>='0') and (tx <= '9')  then ph := 1
        else if (tx = dp)                  then ph := 3
        else if (tx = 'e') or (tx = 'E')   then ph := 5
        else  begin
          ercod := 2;
          break;
        end;
      end else if (ph = 2) then begin
      {after + or - only  a number   ,  e E allowed}
        if (tx >='0') and (tx <= '9')      then ph := 1
        else if (tx = dp)                  then ph := 3
        else if (tx = 'e') or (tx = 'E')   then ph := 5
        else begin
          ercod := 3;
          break;
        end;
      end else if (ph = 3) then begin
      {after a comma only a number allowed or e or E}
        if (tx >= '0') and (tx <='9')      then ph := 4
        else if (tx = 'e') or (tx = 'E')   then ph := 5
        else begin
          ercod := 5;
          break;
        end;
       end else if (ph = 4) then begin
       {after a comma and numbers only  nunbers  e   E allowed}
         if (tx >= '0') and (tx <='9')     then ph := 4
         else if (tx = 'e') or (tx = 'E')  then ph := 5
         else begin
           ercod := 5;
           break;
         end;
       end else if (ph = 5) then begin
       {after e or E only + -  numbers allowed}
         if (tx ='+') or (tx = '-')         then ph := 6
         else if (tx >='0') and (tx <= '9') then begin
           epart := tx;
           ph := 7;
         end else begin
           ercod := 6;
           break;
         end;
       end else if (ph = 6) then begin
       {after e+ etc only numbers allowed}
         if (tx >='0') and (tx <= '9')      then begin
           epart := tx;                     {save exp.part}
           ph := 7
         end else begin
           ercod := 7;
           break;
         end;
       end else if (ph = 7) then begin
       {after  e+ only two numbers allowed}
         if (tx >='0') and (tx <= '9')      then begin
           epart := epart+tx;               {complete exponential part}
           if (epart < '37') then           ph := 8
           else begin
             ercod := 11;
             break;
           end;
         end else begin
           ercod := 8;
           break;
         end;
       end else if (ph = 8) then begin
       {after e+$$ only spaces allowed}
         if (tx = ' ')                      then ph := 8
         else begin
           ercod := 9;
           break;
         end;
       end else begin
       {there should not be any other phases}
         ercod := 10;
         break;
      end;
    end;
  end;


procedure TMatrCalc.ValuesFromLine(line: String; nstrt :Integer; var ipar: Integer;  {in routines.pas}
                                   var pa1,pa2,pa3,pa4 : double;
                                   var errc, ierr :Integer);
{extracts (numeric) values of parameters from a text line of numbers
          values (words) separated by space(s)
          at most 4 parameters
 line     the text line with the numbers (in)
 nstrt    start copying of parameters from parameter nstrt (at most 4 par.)
 ipar     number of parameters found in the line (out)
 pa1,pa2,pa3,pa4    the numeric values (out)
 errc     error code (defined in chckfloat), = 0, if no errors
 ierr     index of the (last) parameter with the error, = 0, if no errors}
var
  apu   : double;   {temporary}
  dec_p : Char;     {decimal point}
  err   : integer;  {error flag}
  j     : Integer;  {index, or counter}
  ll    : Integer;  {length of line}
  numb  : boolean;
  tx    : char;     {one character}
  word  : String;   {one word in the line}
  wfnd  : boolean;  {word found}
begin
  {first determine the character used as decimal point}
  DecimalPoint(dec_p);
  {extract the parameter values from the line, separated by spaces etc}
  ll := length(line);             {length of the line}
  {check that there are numbers}
  numb := false;
  if (ll > 0) then begin
    for j := 1 to ll do begin
      if (line[j] >= '0') and (line[j] <= '9') then begin
        numb := true;
        break;
      end;
    end;
  end;
  ipar := 0;                        {initialize parameter count}
  wfnd := false;                    {initialize flag for finding a word}
  word := '';                       {clear the word}
  errc := 0;                        {initialize error flag}
  ierr := 0;
  if (ll > 0) and (numb) then begin {there are numbers on the line}
    for j:=1 to ll do begin         {purpose is to read to end of line}
      tx := line[j];                {copy one character, study it}
      if (tx=' ') or (j=ll) then begin  {a space or end of line => end of word}
        if (j=ll) then begin        {last char. treated separately}
          if (tx <> ' ') then begin {last char. not a space}
            word := word+tx;        {append last character}
            wfnd := true;           {a non-space char. found (number?)}
          end;
        end;
        if (wfnd) then begin        {was constructing a word}
          ipar := ipar+1;           {increment parameter counter}
          wfnd := false;            {for next time}
          chkfloat(word,err);       {check is the word a number}
          if (err <> 0) then ierr := ipar;  {index of the (latest) parameter with the error}
          errc := errc+err;         {"total" error}
          if (err <> 0) then begin  {not a number}
{            show_message('parameter '+inttostr(ipar)+' is not a number,'
                        +  ' error code '+inttostr(errc)+'   '+word);}
            word := '';
          end else begin            {is a number}
            if (word[1] = 'e') or (word[1] = 'E') then
                 apu := strtofloat('1'+word) {'e' actually not allowed as first}
            else apu := strtofloat(    word);{change to number}
            if      (ipar = nstrt)   then pa1 := apu   {copy to component}
            else if (ipar = nstrt+1) then pa2 := apu
            else if (ipar = nstrt+2) then pa3 := apu
            else if (ipar = nstrt+3) then pa4 := apu;  {at most 4 params.}
            word := '';             {reinitialize the word}
          end;
        end else begin              {not a word, spaces in a row}
          wfnd := false;
          {apu := 0;}
        end;
      end else begin                {not a space, assumed is part of a number}
        wfnd := true;               {found or constructing a word}
        {accept both . and , as decimal point, eventually must change in order
         to correspond to system default}
         if      (dec_p = ',') and (tx = '.') then tx := ','
         else if (dec_p = '.') and (tx = ',') then tx := '.';
         word := word+tx             {append this char. to word}
      end;
    end;
  end else begin                     {no text on line}
    pa1 := 0.0;
    pa2 := 0.0;
    pa3 := 0.0;
    pa4 := 0.0;
    ipar := 0;
    errc := 10;
  end;
end;

procedure TMatrCalc.MarkVectorError(pVGrid: TpGrid; kk,ii : integer);
{mark the cell of a grid in case of an error in the contents of the cell
 pVGrid  pointer to the grid
 kk, ii  column, row}
var
  i1  : integer;
  ipu : integer;  {temporary}
begin
  {calculate position of the erroneous cell}
  ipu := pVGrid^.left;
  for i1 := 0 to kk-1 do ipu := ipu+pVGrid^.ColWidths[i1]+pVGrid^.GridLineWidth;
  {indicate the error with a colored line}
  {(Use TEdit, because bringtofront does not work for images or shapes over a grid !?)}
  pErrEd_c.left := ipu;
  pErrEd_c.top := pVGrid^.top+(ii+1)*(pVGrid^.defaultrowheight+pVGrid^.GridLineWidth)-2;
  pErrEd_c.width := pVGrid^.ColWidths[kk];
  pErrEd_c.visible := true;
  pErrEd_c.BringToFront;
end;

procedure TMatrCalc.GetVector(ig: integer; var Vmagn, Vangl: array of double;
                              var era,erm: array of integer );
{read the magnitude and angle of the vector given as strings in a grid
 pVgrid   pointer to the grid with the vector 3 rows and 3 columns, middle column not used
 Vmagn    magnitude of the vector components, array [0..2]
 Vangl    phase angle of the vector components, array [0..2], angle in degrees
 era,erm  error flags for angle, magnitude, arrays [0..2]}
var
  apu,apv : double;  {temporary}
  apw,apx : double;  {temporary}
  ierr    : integer; {index of error}
  ii      : integer; {index}
  ipu     : integer; {temporary}
  pVGrid  : TpGrid;  {pointer to the grid = vector}
begin
  if (ig > 0) and (ig <= nVect_c) then begin
    pVGrid := pGrid_c[ig];
    for ii := 0 to 2 do begin
      {magnitude first}
      if (pVgrid^.Cells[0,ii] = '') then begin    {if cell is empty, use 1}
        Vmagn[ii] := 0;
        erm[ii] := 0;
      end else begin
        ValuesFromLine(pVgrid^.Cells[0,ii], 1, ipu, apu, apv, apw, apx, erm[ii],ierr);
        if (erm[ii] = 0) then Vmagn[ii] := apu
        else MarkVectorError(pVGrid, 0,ii);
      end;
      {then phase angle, needed for checks only}
      if (pVgrid^.Cells[2,ii] = '') then begin    {if cell is empty, use 0}
        Vangl[ii] := 0;
        era[ii] := 0;
      end else begin
        ValuesFromLine(pVgrid^.Cells[2,ii], 1, ipu, apu, apv, apw, apx, era[ii],ierr);
        if (era[ii] = 0) then Vangl[ii] := apu;
      end;
    end;
  end;
end;

procedure TMatrCalc.SaveGrid(ig: integer);
{save the contents of grid[ig] to the check grid.
 Put the check grid near to the saved grid
 ig   index of the pointer to the grid to be saved}
var
  ii,jj  : integer;  {indices}
  pVgrid : TpGrid;   {pointer to the grid = vector}
  pr11   : TpAlEdit; {pointer to edit near to the grid, for position of the check grid}
begin
  iSavedGrid_prev := iSavedGrid;             {previous index}
  iSavedGrid := ig;                          {save index, needed for checkBut}
  if (ig > 0) and (ig <= nVect_c) then begin
    pVgrid := pGrid_c[ig];
    pr11 := pr11_c[ig];
    {copy vector to CheckGrid, for checks}
    for ii := 0 to 2 do begin
      for jj := 0 to 2 do begin
        pchGrid_c^.Cells[ii,jj] := pVgrid^.cells[ii,jj];
      end;
    end;
    {put the check grid on the coefficients (no room elsewhere}
    left_prev := pchGrid_c^.left;      {save for undo}
    top_prev := pchGrid_c^.top;
    pchGrid_c^.left := pr11^.left;
    pchGrid_c^.top := pr11^.top-8;
    pchGrid_c^.col := 1;
  end;
end;

procedure TmatrCalc.CoeffFromEdit(pr11,pr12,pr21,pr22: TpAlEdit;
                               var r11, r12, r21, r22: double;
                               var er1, er2, er3, er4: integer);
{get the coefficients (multipliers, divisors) from the edit-boxes
 pr11...  pointers to the edit-boxes
 r11 ...  the coefficients
 er1 ...  error flags}
var
  apu,apv   : double;   {temporary}
  apw,apx   : double;   {temporary}
  ierr      : integer;  {index of error}
  ipu       : integer;  {temporary}
begin
  {if edit text is empty, use 1}
  if (pr11^.text = '') then begin
    r11 := 1;
    er1 := 0;
  end else begin
    ValuesFromLine(pr11^.text, 1, ipu, r11, apv, apw, apx, er1,ierr);
  end;
  {second multiplier}
  if (pr12^.text = '') then begin
    r12 := 1;
    er2 := 0;
  end else begin
    ValuesFromLine(pr12^.text, 1, ipu, r12, apv, apw, apx, er2,ierr);
  end;
  {first divisor}
  if (pr21^.text = '') then begin
    r21 := 1;
    er3 := 0;
  end else begin
    ValuesFromLine(pr21^.text, 1, ipu, r21, apv, apw, apx, er3,ierr);
  end;
  {second divisor}
  if (pr22^.text = '') then begin
    r22 := 1;
    er4 := 0;
  end else begin
    ValuesFromLine(pr22^.text, 1, ipu, r22, apv, apw, apx, er4,ierr);
  end;
end;

procedure TMatrCalc.Set3I03I2(used: boolean);
{set or reset the flag "3I0, 3I2 in use}
begin
  seq3_c := used;
end;



procedure TMatrCalc.CalcButClick(pr11,pr12,pr21,pr22,pRes: TpAlEdit; ig: integer;
                             var someerr: boolean);
{calculate the coefficient using values from four edits, display it,
 and multiply a vector with it
 pr11,pr12   pointers to the multiplier edits
 pr21,pr22   pointers to the divisor edits
 pRes        pointer to the edit for the result
 ig          index of the grid = vector
 someerr     error flag}
var
  apu,apv   : double;   {temporary}
  apw,apx   : double;   {temporary}
  clr       : Tcolor;   {color of coefficient boxes}
  dpu,dpv   : double;   {temporary}
  er1,er2   : integer;  {error flag}
  er3,er4   : integer;  {error flag}
  ii,jj     : integer;  {index}
  ipu       : integer;  {temporary}
  ra        : double;   {the coefficient}
  ra11,ra12 : double;   {multipliers of the coefficient}
  ra21,ra22 : double;   {divisors of the coefficient}
  raerr     : boolean;  {error flag}
  Restxt    : string;   {result text}
  txt       : string;   {temporary text}
  Verr      : boolean;  {error flag}
  era       : array[0..2] of integer; {error flag for phase angle}
  erm       : array[0..2] of integer; {error flag for magnitude}
  Vangl     : array[0..2] of double;  {phase angle}
  Vmagn     : array[0..2] of double;  {magnitudes of the vector}
  Vres      : array[0..2] of double;  {magnitudes of the coeff.*vector}
begin
  someerr := false;
  {get the coefficients from the edit-components}
  CoeffFromEdit(pr11,pr12,pr21,pr22,
                ra11, ra12, ra21, ra22, er1, er2, er3, er4);
  {continue, if proper coefficients}
  if (er1 = 0) and (er2 = 0) and (er3 = 0) and (er4 = 0) and
     (ra21 <> 0) and (ra22 <> 0) then begin
    raerr := false;                {no errors}
    dpu := ra11*ra12;
    dpv := ra21*ra22;
    ra := dpu/dpv;
    txt := ManyChars(ra);          {result to text}
    restxt := spaceless(txt);      {remove spaces}
{   pRes^.text := restxt;          {result to display}  {below!}
    {no errors = normal colors}
    if (sheet_c = vect) then clr := LightBlue else clr := LightViolet3;
    pr11^.color := clr{LightBlue};
    pr12^.color := clr{LightBlue};
    pr21^.color := clr{LightBlue};
    pr22^.color := clr{LightBlue};
  end else begin
    raerr := true;
    {indicate errors with colors}
    if (er1 <> 0) then pr11^.color := clYellow;
    if (er2 <> 0) then pr12^.color := clYellow;
    if (er3 <> 0) or (ra21 = 0) then pr21^.color := clYellow;
    if (er4 <> 0) or (ra22 = 0) then pr22^.color := clYellow;
  end;
  {find the magnitudes and angles of the components of the vector}
  GetVector(ig, Vmagn, Vangl, era,erm);
  {continue, if no errors}
  Verr := false;                 {initialize to no error}
  for ii := 0 to 2 do begin
    if (erm[ii] <> 0) or (era[ii] <> 0) then begin
      Verr := true;
      {must indicate the error somehow}
      break;
    end;
  end;
  {carry out the calculations, if no errors}
  if (not raerr) and (not Verr) then begin   {no errors in the coefficient, or matrix}
    SaveState;       {save for undo}
    SaveGrid(ig);    {save grid for checks}
    pRes^.text := restxt;          {resulting coefficient to display}
    {carry out the calculations}
    for ii := 0 to 2 do begin
      dpu := ra*Vmagn[ii];
      {magnitude should be positive}
      if (dpu >= 0) then begin
        Vres[ii] := dpu;
      end else begin             {for negative magnitude, change phase angle}
        Vres[ii] := -dpu;
        dpv := Vangl[ii]+180;
        if (dpv >= 360) then begin
          Vangl[ii] := dpv-360;
        end else begin
          Vangl[ii] := dpv;
        end;
      end;
      txt := ManyChars(Vres[ii]);    {result to text}
      restxt := spaceless(txt);      {remove spaces}
      pgrid_c[ig]^.Cells[0,ii] := restxt; {result to display}
    end;
    {check the angle. If small,  write a zero}
    for ii := 0 to 2 do begin
      if (abs(Vangl[ii]) < 0.01) then begin
        pgrid_c[ig]^.Cells[2,ii] := '0';
      end else begin
        txt := ManyChars(Vangl[ii]);    {result to text}
        restxt := spaceless(txt);       {remove spaces}
        pgrid_c[ig]^.Cells[2,ii] := restxt;  {result to display}
      end;
    end;
  end else begin
    someerr := true;
  end;
{ UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;}
end;


procedure TMatrForm.RaCalcButClick(Sender: TObject);
{calculate the coefficient ra, and multiply vector Va with it}
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@raEdit11,@raEdit12,@raEdit21,@raEdit22,@raResEdit,1,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(RaCalcBut.Left,RaCalcBut.Top,RaCalcBut.Width,RaCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.SelectCell(pGrid: TpGrid; Col, Row: Longint;
  var CanSelect: Boolean);
{the middle column (with the angle symbol) cannot be edited}
begin
  if (col = 1) then begin
    pGrid^.options := pGrid^.options-[goEditing,goAlwaysShowEditor];
  end else begin
    pGrid^.options := pGrid^.options+[goEditing,goAlwaysShowEditor];
  end;
end;

procedure TMatrForm.VaGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
{the middle column (with the angle symbol) cannot be edited}
begin
  SelectCell(@VaGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.VaGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
{draw the angle symbols in the middle column}
begin
  DrawCell(@VaGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.DrawCell(pGrid: TpGrid; Col, Row: Longint;
          Rect: TRect; State: TGridDrawState);
{draw the angle symbols in the middle column
 pGrid  pointer to the grid}
begin
  if (col = 1) then begin     {middle column}
    pGrid^.canvas.font.size := 10;
    pGrid^.canvas.font.name := 'symbol';
    pGrid^.canvas.TextOut(rect.left,rect.top,' ');
  end;
end;

procedure TMatrForm.CheckButAMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (((sender as Tcomponent).name = 'CheckButA') and (iSavedGrid = 1)) or
     (((sender as Tcomponent).name = 'CheckButB') and (iSavedGrid = 2)) or
     (((sender as Tcomponent).name = 'CheckButC') and (iSavedGrid = 3)) or
     (((sender as Tcomponent).name = 'CheckButD') and (iSavedGrid = 4)) or
     (((sender as Tcomponent).name = 'CheckButE') and (iSavedGrid = 5)) or
     (((sender as Tcomponent).name = 'CheckButF') and (iSavedGrid = 6)) then begin
    {show the initial vector}
    CheckGrid.visible := true;
    CheckGrid.BringToFront;
    ActiveControl := nil; {ei auta}
  end;
end;

procedure TMatrForm.CheckButAMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CheckGrid.visible := false;
end;

procedure TMatrForm.CheckGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
{draw the symbol of an angle}
begin
  if (col = 1) then begin     {middle column}
    CheckGrid.canvas.font.size := 10;
    CheckGrid.canvas.font.name := 'symbol';
    CheckGrid.canvas.TextOut(rect.left,rect.top,' ');
  end;
end;

procedure TMatrForm.CheckGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
{ei taida vaikuttaa mitn}
begin
  if (col = 1) then begin
    CheckGrid.options := CheckGrid.options-[goEditing,goAlwaysShowEditor];
  end else begin
    CheckGrid.options := CheckGrid.options+[goEditing,goAlwaysShowEditor];
  end;
end;

procedure TMatrForm.VbGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  DrawCell(@VbGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.RbCalcButClick(Sender: TObject);
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@rbEdit11,@rbEdit12,@rbEdit21,@rbEdit22,@rbResEdit,2,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(RbCalcBut.Left,RbCalcBut.Top,RbCalcBut.Width,RbCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.VbGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
  SelectCell(@VbGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.ExitItemClick(Sender: TObject);
begin
  MatrForm.close{Application.Terminate}; {carries out FormClose, terminates application}
end;

procedure TMatrForm.RcCalcButClick(Sender: TObject);
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@rcEdit11,@rcEdit12,@rcEdit21,@rcEdit22,@rcResEdit,3,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(RcCalcBut.Left,RcCalcBut.Top,RcCalcBut.Width,RcCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;  
end;

procedure TMatrForm.VcGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  DrawCell(@VcGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.VcGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
  SelectCell(@VcGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.RdCalcButClick(Sender: TObject);
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@rdEdit11,@rdEdit12,@rdEdit21,@rdEdit22,@rdResEdit,4,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(RdCalcBut.Left,RdCalcBut.Top,RdCalcBut.Width,RdCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;  
end;

procedure TMatrForm.VdGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  DrawCell(@VdGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.VdGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
  SelectCell(@VdGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.ReCalcButClick(Sender: TObject);
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@reEdit11,@reEdit12,@reEdit21,@reEdit22,@reResEdit,5,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(ReCalcBut.Left,ReCalcBut.Top,ReCalcBut.Width,ReCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;  
end;

procedure TMatrForm.VeGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  DrawCell(@VeGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.VeGridSelectCell(Sender: TObject; Col, Row: Longint;
  var CanSelect: Boolean);
begin
  SelectCell(@VeGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.RfCalcButClick(Sender: TObject);
var
  err: boolean;
begin
  Matr_Calc.CalcButClick(@rfEdit11,@rfEdit12,@rfEdit21,@rfEdit22,@rfResEdit,6,err);
  if (not err) then begin
    Matr_Calc.MarkLastOperation(RfCalcBut.Left,RfCalcBut.Top,RfCalcBut.Width,RfCalcBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;  
end;

procedure TMatrForm.VfGridDrawCell(Sender: TObject; Col, Row: Longint;
  Rect: TRect; State: TGridDrawState);
begin
  DrawCell(@VfGrid, Col, Row, Rect, State);
end;

procedure TMatrForm.VfGridSelectCell(Sender: TObject; Col,
  Row: Longint; var CanSelect: Boolean);
begin
  SelectCell(@VfGrid, Col, Row, CanSelect);
end;

procedure TMatrForm.SumButClick(Sender: TObject);
{calculate the sum of the selected vectors}
const
  nV = 5; {number of vectors}
var
  abc     : char;        {identifying character}
  dpu,dpv : double;      {temporary}
  ertot   : integer;     {total error}
  ii,jj   : integer;     {index}
  ll      : longint;     {temporary}
  restxt  : string;      {result as text}
  txt     : string;      {temporary}
  era     : array[0..2]  of integer; {error flag for phase angle}
  erm     : array[0..2]  of integer; {error flag for magnitude}
  pVcombo : array[1..nV] of TpCombo; {pointers to comboboxes for selecting the vectors}
  Vangl   : array[0..2]  of double;  {phase angle}
  Vmagn   : array[0..2]  of double;  {magnitudes of the vector}
  Vsum    : array[0..2]  of Dcomplex; {sum of the vectors}
begin
  {pointers to the and combo boxes}
  pVcombo[1] := @AddBoxA;
  pVcombo[2] := @AddBoxB;
  pVcombo[3] := @AddBoxC;
  pVcombo[4] := @AddBoxD;
  pVcombo[5] := @AddBoxE;
  ertot := 0;
  {calculate the sum of selected vectors  (ord('a') = 97,...,ord('e') = 101)}
  for jj := 0 to 2 do Vsum[jj] := Dcplx(0,0);         {initialize}
  for ii := 1 to nV do begin
    if (length(pVcombo[ii]^.text) >=3) then begin
      abc := pVcombo[ii]^.text[3];                    {third char. identifies the vector}
      ll := ord(abc)-96;                              {character as number, to index}
      if (ll >= 1) and (ll <= nV) then begin          {reasonable index of a vector}
        Matr_Calc.GetVector(ll, Vmagn, Vangl, era,erm); {get magnitude and angle}
        {continue, if no errors}
        ertot := era[0]+era[1]+era[2]+erm[0]+erm[1]+erm[2];
        if (ertot = 0) then begin
{         Matr_calc.SaveState;        {save for undo}
{         Matr_calc.SaveGrid(6);      {save grid for checks}
          for jj := 0 to 2 do Vangl[jj] := pi*Vangl[jj]/180;  {angle to radians}
          {add to sum}
          for jj := 0 to 2 do Vsum[jj] := Dadd(Vsum[jj],DPolarToCplx(Vmagn[jj],Vangl[jj]));
        end else begin
          break;
        end;
      end;
    end;
  end;
  if (ertot = 0) then begin           {continue only if no errors}
    Matr_calc.SaveState;              {save for undo}
    Matr_calc.SaveGrid(6);            {save grid for checks}
    {result to vector Vf in polar form}
    for jj := 0 to 2 do begin
      dpu := Dabs(Vsum[jj]);          {magnitude}
      dpv := Dangle(Vsum[jj])*180/pi; {phase angle in degrees}
      txt := ManyChars(dpu);          {magnitude to text}
      restxt := spaceless(txt);       {remove spaces}
      Vfgrid.Cells[0,jj] := restxt;   {magnitude to display}
      {angle not defined for very small magnitude}
      if (dpu > 0.01) then begin
        if (abs(dpv) > 0.01) then begin   {angle not very small}
          txt := ManyChars(dpv);          {angle to text}
          restxt := spaceless(txt);       {remove spaces}
          Vfgrid.Cells[2,jj] := restxt;   {angle to display}
        end else begin
          Vfgrid.Cells[2,jj] := '';
        end;
      end else begin
        Vfgrid.Cells[2,jj] := '';
      end;
    end;
    Matr_Calc.MarkLastOperation(SumBut.Left,SumBut.Top,SumBut.Width,SumBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.RadioButAClick(Sender: TObject);
begin
{ rd_Vect := 1;  {update index of selected vector}
  Matr_Calc.PutVectIndex(1);
end;

procedure TMatrForm.RadioButBClick(Sender: TObject);
begin
{ rd_Vect := 2;  {update index of selected vector}
  Matr_Calc.PutVectIndex(2);
end;

procedure TMatrForm.RadioButCClick(Sender: TObject);
begin
{ rd_Vect := 3;  {update index of selected vector}
  Matr_Calc.PutVectIndex(3);
end;

procedure TMatrForm.RadioButDClick(Sender: TObject);
begin
{ rd_Vect := 4;  {update index of selected vector}
  Matr_Calc.PutVectIndex(4);
end;

procedure TMatrForm.RadioButEClick(Sender: TObject);
begin
{ rd_Vect := 5;  {update index of selected vector}
  Matr_Calc.PutVectIndex(5);
end;

procedure TMatrForm.RadioButFClick(Sender: TObject);
begin
{ rd_Vect := 6;  {update index of selected vector}
  Matr_Calc.PutVectIndex(6);
end;

procedure TMatrForm.DelVectButClick(Sender: TObject);
{clear the selected vector}
var
  ii : integer;  {index}
begin
  Matr_calc.SaveState;                {save for undo}
  Matr_Calc.ClearVect;      {clear the selected vector}
  Matr_Calc.MarkLastOperation(DelVectBut.Left,DelVectBut.Top,DelVectBut.Width,DelVectBut.Height);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.DelScalButClick(Sender: TObject);
{clear the selected coefficients}
begin
  Matr_calc.SaveState;                {save for undo}
  Matr_Calc.ClearScalar;    {clear the coeff. of the selected vector}
  Matr_Calc.MarkLastOperation(DelScalBut.Left,DelScalBut.Top,DelScalBut.Width,DelScalBut.Height);
end;

procedure TMatrCalc.ClearAll;
{clear all vectors and coefficients}
var
  ii,jj : integer;
begin
  for ii := 1 to nVect_c do begin
    {scalars first}
    pr11_c[ii]^.text := '';
    pr12_c[ii]^.text := '';
    pr21_c[ii]^.text := '';
    pr22_c[ii]^.text := '';
    pres_c[ii]^.text := '';
    {then the vectors}
    for jj := 0 to 2 do begin
      pGrid_c[ii]^.cells[0,jj] := '0';
      pGrid_c[ii]^.cells[2,jj] := '0';
    end;
  end;
end;

procedure TMatrForm.DelAllButClick(Sender: TObject);
{clear all vectors and coefficients
 At initialization, Sender = nil}
var
  ii,jj : integer;
begin
  if (sender <> nil) then Matr_calc.SaveState;     {save for eventual "undo"}
  Matr_Calc.ClearAll;
  if (sender <> nil) then begin   {not at initialization}
    Matr_Calc.MarkLastOperation(DelAllBut.Left,DelAllBut.Top,DelAllBut.Width,DelAllBut.Height);
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.FillButClick(Sender: TObject);
{fill all fields with some values}
var
  ii,jj : integer;
begin
  Matr_calc.SaveState;                {save for undo}
  for ii := 1 to nVect do begin
    {scalars first}
    Matr_calc.pr11_c[ii]^.text := inttostr(ii+1);
    Matr_calc.pr12_c[ii]^.text := inttostr(ii+2);
    Matr_calc.pr21_c[ii]^.text := inttostr(ii+3);
    Matr_calc.pr22_c[ii]^.text := inttostr(ii+4);
    Matr_calc.pres_c[ii]^.text := inttostr(ii+5);
    {then the vectors}
    for jj := 0 to 2 do begin
      Matr_calc.pgrid_c[ii]^.cells[0,jj] := inttostr(ii+3*jj);
      Matr_calc.pgrid_c[ii]^.cells[2,jj] := inttostr(10*ii+120*jj);
    end;
  end;
end;

procedure TMatrCalc.SaveState;
{save all coefficients and vectors as strings}
var
  ia    : integer;
  ii,jj : integer;
  kk    : integer;
begin
  {the coefficients first}
  for ii := 1 to nVect_c do begin
    if (pr11_c[ii] <> nil) then backup_c[ii          ] := pr11_c[ii]^.text;
    if (pr12_c[ii] <> nil) then backup_c[ii+  nVect_c] := pr12_c[ii]^.text;
    if (pr21_c[ii] <> nil) then backup_c[ii+2*nVect_c] := pr21_c[ii]^.text;
    if (pr22_c[ii] <> nil) then backup_c[ii+3*nVect_c] := pr22_c[ii]^.text;
    if (pres_c[ii] <> nil) then backup_c[ii+4*nVect_c] := pres_c[ii]^.text;
  end;
  {next, the vectors}
  for ii := 1 to nVect_c do begin
    for jj := 0 to 2 do begin
      backup_c[nVect_c*(ii+4)+2*jj+1] := pGrid_c[ii]^.cells[0,jj];
      backup_c[nVect_c*(ii+4)+2*jj+2] := pGrid_c[ii]^.cells[2,jj];
    end;
  end;
  {finally, the "check" vector}
  for jj := 0 to 2 do begin
    backup_c[nVect_c*11+2*jj+1] := pchGrid_c^.cells[0,jj];
    backup_c[nVect_c*11+2*jj+2] := pchGrid_c^.cells[2,jj];
  end;
  {after that, still the matrix, if it exists}
  ia := nVect_c*11+2*2+2+1;
  if (pMatr_c <> nil) then begin
    for ii := 0 to 2 do begin
      for jj := 1 to 6 do begin
        kk := ColIndex[jj];                    {excludes columns with the angle symbol}
        backup_c[ia] := pMatr_c^.cells[kk,ii]; {save the entries of the matrix}
        inc(ia);
      end;
    end;
  end;
end;

procedure TMatrCalc.UndoLastOper;
{return all coefficients and vectors as strings to the state before the last action}
var
  ia    : integer;
  ii,jj : integer;
  kk    : integer;
begin
  {the coefficients first}
  for ii := 1 to nVect_c do begin
    pr11_c[ii]^.text := backup_c[ii          ];
    pr12_c[ii]^.text := backup_c[ii+  nVect_c];
    pr21_c[ii]^.text := backup_c[ii+2*nVect_c];
    pr22_c[ii]^.text := backup_c[ii+3*nVect_c];
    pres_c[ii]^.text := backup_c[ii+4*nVect_c];
  end;
  {next, the vectors}
  for ii := 1 to nVect_c do begin
    for jj := 0 to 2 do begin
      pGrid_c[ii]^.cells[0,jj] := backup_c[nVect_c*(ii+4)+2*jj+1];
      pGrid_c[ii]^.cells[2,jj] := backup_c[nVect_c*(ii+4)+2*jj+2];
    end;
  end;
  {finally, the "check" vector}
  for jj := 0 to 2 do begin
    pchGrid_c^.cells[0,jj] := backup_c[nVect_c*11+2*jj+1];
    pchGrid_c^.cells[2,jj] := backup_c[nVect_c*11+2*jj+2];
  end;
  iSavedGrid := iSavedGrid_prev;
  pchGrid_c^.left := left_prev;      {previous pos. of the check grid}
  pchGrid_c^.top := top_prev;
  {after that, still the matrix, if it exists}
  ia := nVect_c*11+2*2+2+1;
  if (pMatr_c <> nil) then begin
    for ii := 0 to 2 do begin
      for jj := 1 to 6 do begin
        kk := ColIndex[jj];                    {excludes columns with the angle symbol}
        pMatr_c^.cells[kk,ii] := backup_c[ia]; {save the entries of the matrix}
        inc(ia);
      end;
    end;
  end;
end;

procedure TMatrForm.UndoButClick(Sender: TObject);
begin
  Matr_Calc.UndoLastOper;
  Matr_Calc.BeforeLastOper;   {mark the operation before the last operation}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.PutAngle(angl : double; ii,dec : integer);
{write the phase angle of component ii to the selected vector grid
 angl  the angle in degrees
 ii   index of the component, 0..2
 dec  index for decimals}
var
  dpv    : double;  {temporary}
  restxt : string;  {result as text}
  txt    : string;  {temporary}
begin
  if (ii >= 0) and (ii <= 2) then begin
    if (angl >= 359.99) then dpv := angl-360
    else if (angl <= -359.99) then dpv := angl+360
    else dpv := angl;
    if (abs(dpv) > 0.01) then begin             {angle not very small}
      if (dec = 2) then
        txt := TwoDecimalsAtMost(dpv)           {angle to text, with two decimals}
      else
        txt := ManyChars(dpv);                  {angle to text}
      restxt := spaceless(txt);                 {remove spaces}
      pGrid_c[rdVect_c]^.Cells[2,ii] := restxt;  {angle to display}
    end else begin
      pgrid_c[rdVect_c].Cells[2,ii] := '0';
    end;
  end;
end;

procedure TMatrCalc.PutMagn(magn : double; ii : integer);
{write the magnitude of component ii to the selected vector grid
 angl  the angle in degrees
 ii   index of the component, 0..2}
var
  dpv    : double;  {temporary}
  restxt : string;  {result as text}
  txt    : string;  {temporary}
begin
  if (ii >= 0) and (ii <= 2) then begin
    dpv := magn;
    if (abs(dpv) > 0.001) then begin            {magnitude not very small}
      txt := ManyChars(dpv);                    {angle to text}
      restxt := spaceless(txt);                 {remove spaces}
      pgrid_c[rdVect_c]^.Cells[0,ii] := restxt;  {angle to display}
    end else begin
      pgrid_c[rdVect_c].Cells[0,ii] := '0';
    end;
  end;
end;

procedure TMatrForm.NegButClick(Sender: TObject);
{change the sign of the selected vector}
begin
  Matr_Calc.NegButClick(NegBut);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.NegButClick(NegBut: TSpeedButton);
{change the sign of the selected vector}
var
  dpv     : double;      {temporary}
  ertot   : integer;     {total error}
  ii      : integer;     {index}
  era     : array[0..2]  of integer; {error flag for phase angle}
  erm     : array[0..2]  of integer; {error flag for magnitude}
  Vangl   : array[0..2]  of double;  {phase angle}
  Vmagn   : array[0..2]  of double;  {magnitudes of the vector}
begin
  GetVector(rdVect_c, Vmagn, Vangl, era,erm); {get magnitude and angle}
  {continue, if no errors}
  ertot := era[0]+era[1]+era[2];  {only angle needed}
  if (ertot = 0) then begin
    SaveState;                    {save for undo}
    SaveGrid(rdVect_c);           {save grid for checks}
    for ii := 0 to 2 do begin     {all three components of the vector}
      dpv := Vangl[ii]+180;       {change sign = angle + 180 degrees}
      if (abs(Vmagn[ii]) < 0.01) then dpv := 0;
      PutAngle(dpv,ii,0);         {put the angle to the selected vector}
    end;
    MarkLastOperation(NegBut.Left,NegBut.Top,NegBut.Width,NegBut.Height);
  end;
end;

procedure TMatrForm.ConjButClick(Sender: TObject);
{calculate the complex conjugate of the selected vector}
begin
  Matr_Calc.ConjButClick(ConjBut);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.ConjButClick(ConjBut: TSpeedButton);
{calculate the complex conjugate of the selected vector}
var
  dpv     : double;      {temporary}
  ertot   : integer;     {total error}
  ii      : integer;     {index}
  era     : array[0..2]  of integer; {error flag for phase angle}
  erm     : array[0..2]  of integer; {error flag for magnitude}
  Vangl   : array[0..2]  of double;  {phase angle}
  Vmagn   : array[0..2]  of double;  {magnitudes of the vector}
begin
  GetVector(rdVect_c, Vmagn, Vangl, era,erm); {get magnitude and angle}
  {continue, if no errors}
  ertot := era[0]+era[1]+era[2];  {only angle needed}
  if (ertot = 0) then begin
    SaveState;                    {save for undo}
    SaveGrid(rdVect_c);           {save grid for checks}
    for ii := 0 to 2 do begin     {all three components of the vector}
      dpv := -Vangl[ii];          {compl.conj = change sign of the phase angle}
      if (abs(Vmagn[ii]) < 0.01) then dpv := 0;
      PutAngle(dpv,ii,0);         {put the angle to the selected vector}
    end;
    MarkLastOperation(ConjBut.Left,ConjBut.Top,ConjBut.Width,ConjBut.Height);
  end;
end;

procedure PhaseToSeqD(Pa,Pb,Pc : Dcomplex; var Ps,Ns,Zs : Dcomplex);
{converts the phase variables to symmetrical components
 Ps   pos. seq. component
 Ns   neg. seq. component
 Zs   zero seq. component
 Pa,Pb,Pc  the phase variables
}
const
  eifii  : Dcomplex = (x : -0.5; y: 0.8660254); {exp(i fii), phase difference}
  emifii : Dcomplex = (x : -0.5; y:-0.8660254); {exp(-i fii) phase difference}
  one3rd = 0.333333333;
var
  cpu  : Dcomplex;  {temporary}
begin
  cpu := Dadd(Pa,Dadd(Pb,Pc));
  Zs  := Drmult(one3rd,cpu);
  cpu := Dadd(Dmult(eifii,Pb),Dmult(emifii,Pc));
  Ps  := Drmult(one3rd,Dadd(cpu,Pa));
  cpu := Dadd(Dmult(emifii,Pb),Dmult(eifii,Pc));
  Ns  := Drmult(one3rd,Dadd(cpu,Pa));
end;

procedure SeqToPhaseD(Ps,Ns,Zs : Dcomplex; var Pa,Pb,Pc : Dcomplex);
{converts the symmetrical components to phase variables
 Ps   pos. seq. component
 Ns   neg. seq. component
 Zs   zero seq. component
 Pa,Pb,Pc  the phase variables }
const
  eifii : Dcomplex = (x : -0.5; y: 0.8660254); {exp(i fii), phase difference}
  emifii: Dcomplex = (x : -0.5; y:-0.8660254); {exp(-i fii) phase difference}
var
  cpu  : Dcomplex;  {temporary}
begin
  Pa  := Dadd(Dadd(Ps,Ns),Zs);
  cpu := Dadd(Dmult(emifii,Ps),Dmult(eifii,Ns));
  Pb  := Dadd(cpu,Zs);
  cpu := Dadd(Dmult(eifii,Ps),Dmult(emifii,Ns));
  Pc  := Dadd(cpu,Zs);
end;

procedure TMatrCalc.GetReImVector(var Vreim: array of Dcomplex; var ertot : integer);
{get the selected vector in component form (real, imag. components
 Save state and vector for undo and checks, if no errors
 Vreim   the vector in real, imag form
 ertot   error flag, from reading the vector components}
var
  dpv     : double;     {temporary}
  jj      : integer;    {index}
  era     : array[0..2] of integer;  {error flag for phase angle}
  erm     : array[0..2] of integer;  {error flag for magnitude}
  Vangl   : array[0..2] of double;   {phase angle}
  Vmagn   : array[0..2] of double;   {magnitudes of the vector}
begin
  GetVector(rdVect_c, Vmagn, Vangl, era,erm); {get magnitude and angle}
  ertot := era[0]+era[1]+era[2]+erm[0]+erm[1]+erm[2];
  {continue, if total error = 0}
  if (ertot = 0) then begin
    SaveState;                                     {save for undo}
    SaveGrid(rdVect_c);                            {save grid for checks}
    {real and imag.components of the vector}
    for jj := 0 to 2 do begin
      dpv := pi*Vangl[jj]/180;                     {angle in radians}
      Vreim[jj] := DPolarToCplx(Vmagn[jj],dpv);
    end;
  end;
end;

procedure TMatrCalc.PutReImVector(Vreim: array of Dcomplex);
{write the complex vector Vreim in polar form to the selected vector (angle in degrees)}
var
  dpv     : double;     {temporary}
  ii      : integer;    {index}
  Vangl   : array[0..2] of double;   {phase angle}
  Vmagn   : array[0..2] of double;   {magnitudes of the vector}
begin
  {change to phase and angle}
  for ii := 0 to 2 do begin
    Vmagn[ii] := Dabs(Vreim[ii]);
    Vangl[ii] := 180*Dangle(Vreim[ii])/pi;
  end;
  {display = write back to selected vector grid}
  for ii := 0 to 2 do begin
    putMagn(Vmagn[ii],ii);
    {a (near) zero vector has no phase}
    if (abs(Vmagn[ii]) > 0.01) then dpv := Vangl[ii] else dpv := 0;
    putAngle(dpv,ii,0);
  end;
end;

procedure TMatrCalc.PhaseOrSeq(ps: integer);
{calculate the sequence or phase components of the selected vector
 ps   if = 0, calculate sequence components, else phase components}
var
  apu     : double;      {temporary}
  ertot   : integer;     {total error}
  Vreim   : array[0..2] of Dcomplex; {real and imag. components of the vector}
  Vsequ   : array[0..2] of Dcomplex; {sequence components, or phase components}
begin
  GetReImVector(Vreim, ertot);
  if (ertot = 0) then begin
    {3I0, 3I2 in use ?}
    if (seq3_c) then apu := 3 else apu := 1;
    {change phase to sequence, or sequence to phase}
    if (ps = 0) then begin
      PhaseToSeqD(Vreim[0],Vreim[1],Vreim[2],Vsequ[0],Vsequ[1],Vsequ[2]);
      Vsequ[1] := Drmult(apu,Vsequ[1]);
      Vsequ[2] := Drmult(apu,Vsequ[2]);
    end else begin
      SeqToPhaseD(Vreim[0],Drmult(1/apu,Vreim[1]),Drmult(1/apu,Vreim[2]),
                  Vsequ[0],Vsequ[1],Vsequ[2]);
    end;
    {display in polar form}
    PutReImVector(Vsequ);
  end;
end;

procedure TMatrForm.SeqButClick(Sender: TObject);
{calculate the sequence components of the selected vector}
begin
  Matr_Calc.PhaseOrSeq(0);     {calculate sequence components}
  Matr_Calc.MarkLastOperation(SeqBut.Left,SeqBut.Top,SeqBut.Width,SeqBut.Height);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.PhaseButClick(Sender: TObject);
begin
  Matr_Calc.PhaseOrSeq(1);     {calculate phase components}
  Matr_Calc.MarkLastOperation(PhaseBut.Left,PhaseBut.Top,PhaseBut.Width,PhaseBut.Height);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TmatrForm.m12mult;
{collect the compensation matrices in one two-dimensional array}
(*
const
  {the compensation matrices, without the multiplier}
  m00 : M3x3 = ( 1, 0, 0, {}  0, 1, 0, {}  0, 0, 1);
  m01 : M3x3 = ( 1,-1, 0, {}  0, 1,-1, {} -1, 0, 1);
  m02 : M3x3 = ( 1,-2, 1, {}  1, 1,-2, {} -2, 1, 1);
  m03 : M3x3 = ( 0,-1, 1, {}  1, 0,-1, {} -1, 1, 0);
  m04 : M3x3 = (-1,-1, 2, {}  2,-1,-1, {} -1, 2,-1);
  m05 : M3x3 = (-1, 0, 1, {}  1,-1, 0, {}  0, 1,-1);
  m06 : M3x3 = (-2, 1, 1, {}  1,-2, 1, {}  1, 1,-2);
  m07 : M3x3 = (-1, 1, 0, {}  0,-1, 1, {}  1, 0,-1);
  m08 : M3x3 = (-1, 2,-1, {} -1,-1, 2, {}  2,-1,-1);
  m09 : M3x3 = ( 0, 1,-1, {} -1, 0, 1, {}  1,-1, 0);
  m10 : M3x3 = ( 1, 1,-2, {} -2, 1, 1, {}  1,-2, 1);
  m11 : M3x3 = ( 1, 0,-1, {} -1, 1, 0, {}  0,-1, 1);
  m12 : M3x3 = ( 2,-1,-1, {} -1, 2,-1, {} -1,-1, 2);
  onesq3 = 0.577350269;    {1/sqrt(3)}
  one3rd = 0.333333333;    {1/3}
  {the multipliers of the compensation matrices}
  mcoeff : array[0..12] of double =
           (1.0,onesq3,one3rd,onesq3,one3rd,onesq3,one3rd,
               onesq3,one3rd,onesq3,one3rd,onesq3,one3rd);
*)
var
  ii  : integer;
{ Comp_Matr : array[0..12] of M3x3;  {all compensation matrices}
begin
  {collect the compensation matrices in one array
   the first index is the index of the matrix, 0..12,
   the second index is the index of an element
      the index of element row i, column j,
      index(i,j) = 3*i+j  when i = 0..2 and j = 0..2}
  for ii := 0 to 8 do begin
    Comp_Matr[ 0,ii] := mcoeff[ 0]*m00[ii];
    Comp_Matr[ 1,ii] := mcoeff[ 1]*m01[ii];
    Comp_Matr[ 2,ii] := mcoeff[ 2]*m02[ii];
    Comp_Matr[ 3,ii] := mcoeff[ 3]*m03[ii];
    Comp_Matr[ 4,ii] := mcoeff[ 4]*m04[ii];
    Comp_Matr[ 5,ii] := mcoeff[ 5]*m05[ii];
    Comp_Matr[ 6,ii] := mcoeff[ 6]*m06[ii];
    Comp_Matr[ 7,ii] := mcoeff[ 7]*m07[ii];
    Comp_Matr[ 8,ii] := mcoeff[ 8]*m08[ii];
    Comp_Matr[ 9,ii] := mcoeff[ 9]*m09[ii];
    Comp_Matr[10,ii] := mcoeff[10]*m10[ii];
    Comp_Matr[11,ii] := mcoeff[11]*m11[ii];
    Comp_Matr[12,ii] := mcoeff[12]*m12[ii];
  end;
end;

procedure TMatrForm.MatMultButClick(Sender: TObject);
{multiply the selected vector with the selected compensation matrix}
var
  apu,apv  : double;      {temporary}
  apw,apx  : double;      {temporary}
  dpu      : Dcomplex;    {temporary}
  err      : integer;     {error flag}
  ertot    : integer;     {error flag}
  ervec    : integer;     {error flag}
  ia       : integer;     {index in the saved matrices}
  ierr     : integer;     {index of an error}
  ii,jj,kk : integer;     {index}
  ip       : integer;     {index of a saved matrix}
  ipu,ipv  : integer;     {temporary}
  mm       : integer;     {index of the matrix, 0..12}
  parr     : TpMatrStrArr;  {pointer to array of a saved matrix}
  rr       : double;      {calculated coefficient}
  txt,txx  : string;      {temporary}
  txy,txz  : string;      {temporary}
  matr     : array[0..9] of Dcomplex; {the multiplier matrix}
  Vreim    : array[0..2] of Dcomplex; {real and imag. components of the selected vector}
  Vprod    : array[0..2] of Dcomplex; {real and imag. components of the product vector}
begin
  {determine index of matrix from the text of the combo box}
  mm := SelectMatrBox.Items.IndexOf(SelectMatrBox.text);  {from 0 to no of matrices-1}
  if (mm < 0) then begin                    {error, use identity matrix, just to be sure}
    for ii := 0 to 8 do matr[ii] := Dcplx(0,0);
    matr[0] := Dcplx(1,0);
    matr[4] := Dcplx(1,0);
    matr[8] := Dcplx(1,0);
    ertot := 1;                             {set error flag}
    rr := 1;
  end else if (mm <= 12) then begin         {fixed matrix}
    for ii := 0 to 8 do begin               {copy matrix to local}
      matr[ii] := Dcplx(Comp_Matr[mm,ii],0); {fixed matrices are actually real matrices}
    end;
    rr := 1;                                {coefficient}
    ertot := 0;                             {no error}
  end else begin                            {edited, saved matrix}
    ip := mm-13;                            {index of the matrix}
    {read the matrix}
    parr := MxEditForm.Matr_list.items[ip]; {pointer to the array for storing the matrix}
    ia := 1;                                {first entry (0 = caption)}
    ertot := 0;
    for ii := 0 to 8 do begin
      txt := parr[ia];                      {magnitude of an element as a string}
      inc(ia);
      txx := parr[ia];                      {phase angle of the element}
      inc(ia);
      {change magnitude and angle from strings to numbers}
      Matr_Calc.ValuesFromLine(txt+' '+txx, 1, ipu, apu, apv, apw, apx, err,ierr);
      if (err = 0) then begin
        apw := pi*apv/180;                  {angle in radians}
        matr[ii] := DPolarToCplx(apu,apw);
      end else begin
        ertot := ertot+err;                 {error flag}
      end;
     end;
     {read the coefficients}
     txt := parr[ia]; if (txt = '') or (txt = ' ') then txt := '1';
     inc(ia);
     txx := parr[ia]; if (txx = '') or (txx = ' ') then txx := '1';
     inc(ia);
     txy := parr[ia]; if (txy = '') or (txy = ' ') then txy := '1';
     inc(ia);
     txz := parr[ia]; if (txz = '') or (txz = ' ') then txz := '1';
     inc(ia);
     txt := txt+' '+txx+' '+txy+' '+txz;   {all coefficients in one string}
     Matr_Calc.ValuesFromLine(txt, 1, ipu, apu, apv, apw, apx, err,ierr);
     if (err = 0) then begin
       if (apw <> 0) and (apx <> 0) then begin   {divisors must not be zeroes}
         rr := apu*apv/(apw*apx);
       end else begin
         rr := 1;
         ertot := ertot+1;
       end;
     end else begin
       ertot := ertot+err;
     end;
  end;
  {carry out the multiplication, if no errors}
  if (ertot = 0) then begin
    Matr_Calc.GetReImVector(Vreim, ervec);  {get selected vector, save for undo}
    if (ervec = 0) then begin               {continue, if no errors}
      for ii := 0 to 2 do begin             {all components}
        dpu := Dcplx(0,0);
        for jj := 0 to 2 do begin
          ipu := 3*ii+jj;                   {index of element ii,jj in the matrix}
          dpu := Dadd(dpu,Dmult(Matr[ipu],Vreim[jj])); {matrix multiplication}
        end;
        dpu := Drmult(rr,dpu);              {multiply with the coefficients}
        Vprod[ii] := dpu;
      end;
      {display the result}
      Matr_Calc.PutReImVector(Vprod);
      Matr_Calc.MarkLastOperation(MatMultBut.Left,MatMultBut.Top,MatMultBut.Width,MatMultBut.Height);
      UpdatePhasors;
      if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
    end;
  end;
end;

procedure TMatrForm.HandleCopyDataRecord(copyDataStruct: PCopyDataStruct);
{receive data from Elplek to the selected vector}
begin
  Matr_Calc.HandleCopyDataRecord(copyDataStruct);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.HandleCopyDataRecord(copyDataStruct: PCopyDataStruct);
var
  dpv    : double;     {temporary}
  ii     : integer;    {index}
  Restxt : string;     {result text}
  txt    : string;     {temporary text}
  Vmagn  : array[0..2] of double;  {magnitude}
  Vangl  : array[0..2] of double;  {phase angle, in degrees}
begin
  Vmagn[0] := TSixSingle(copyDataStruct.lpData^).x;
  Vangl[0] := TSixSingle(copyDataStruct.lpData^).y;
  Vmagn[1] := TSixSingle(copyDataStruct.lpData^).z;
  Vangl[1] := TSixSingle(copyDataStruct.lpData^).u;
  Vmagn[2] := TSixSingle(copyDataStruct.lpData^).v;
  Vangl[2] := TSixSingle(copyDataStruct.lpData^).w;
  {save for undo and checks}
  SaveState;                              {save for undo}
  SaveGrid(rdVect_c);                     {save grid for checks}
  {write to the selected vector}
  for ii := 0 to 2 do begin
    PutMagn(Vmagn[ii],ii);
    if (abs(Vmagn[ii]) > 0.01) then dpv := Vangl[ii] else dpv := 0;
    putAngle(dpv,ii,2);
  end;
end;

procedure TMatrForm.WMCopyData(var Msg: TWMCopyData);
{receive data from elplek}
begin
  {Handle of the Sender}
  HandleCopyDataRecord(Msg.CopyDataStruct);
  {Send something back}
  msg.Result := 1;
  BringToFront;
end;



procedure TMatrForm.ElplekButClick(Sender: TObject);
{change the caption of the form, in order to allow/deny transfer of data from elplek
 (ekplek looks for the correct caption. If none found, data is not transferred}
begin
  if (ElplekBut.down) then MatrForm.caption := Caption_main+' '
  else MatrForm.caption := Caption_main;
end;

procedure TMatrCalc.CurrToMultix(var curr :Tmultix; ii : integer; var ertot: integer);
{read the currents from the vector ii and copy to curr as magnitude and phase}
var
  era     : array[0..2] of integer;  {error flag for phase angle}
  erm     : array[0..2] of integer;  {error flag for magnitude}
  Vangl   : array[0..2] of double;   {phase angle}
  Vmagn   : array[0..2] of double;   {magnitudes of the vector}
begin
  if (ii > 0) and (ii <= nVect) then begin  {reasonable index}
    GetVector(ii, Vmagn, Vangl, era,erm); {get magnitude and angle}
    ertot := era[0]+era[1]+era[2]+erm[0]+erm[1]+erm[2];
    {continue, if total error = 0}
    if (ertot = 0) then begin
      Curr.u := Vmagn[0];
      Curr.v := Vangl[0];
      Curr.w := Vmagn[1];
      Curr.x := Vangl[1];
      Curr.y := Vmagn[2];
      Curr.z := Vangl[2];
    end;
  end else begin
    ertot := -1;
  end;
end;

procedure TMatrForm.PhasorButClick(Sender: TObject);
{add the selected vector to the phasor display}
var
  Curr    : TMultix;    {currents in one variable}
  ertot   : integer;
  era     : array[0..2] of integer;  {error flag for phase angle}
  erm     : array[0..2] of integer;  {error flag for magnitude}
  Vangl   : array[0..2] of double;   {phase angle}
  Vmagn   : array[0..2] of double;   {magnitudes of the vector}
begin
  Matr_Calc.CurrToMultix(curr, Matr_Calc.rdVect_c, ertot);  {copy currents from vector }
  if (ertot = 0) and (CurrPhasorForm.visible) then begin
    CurrGrid.ShowPhasors(Matr_Calc.rdVect_c, cRead, @Curr,Matr_Calc.Vect_lbl[Matr_Calc.rdVect_c]);
    SeqGrid.ShowPhasors(Matr_Calc.rdVect_c,cCopy,nil,Matr_Calc.Vect_lbl[Matr_Calc.rdVect_c]);
    CurrPhasorForm.BringToFront;
  end;
  {just to be sure, update all phasors}
  UpdatePhasors;
end;

procedure TMatrForm.UpdatePhasors;
{update the phasors (after some new calculation)}
var
  curr   : Tmultix;   {currents = magnitude and phase (in degrees)}
  ertot  : integer;   {error flag}
  ii,jj  : integer;   {index}
  nCT    : integer;   {number of currents in the grid}
begin
  {read the currents from the vectors that are listed in the grid of the graph}
  nCT := CurrGrid.NoOfItems;                   {number of current trafos}
  if (nCT > 0) then begin
    for ii := 1 to nCT do begin                {all current trafos}
      jj := CurrGrid.GetIp(ii);
      if (MatrForm.visible) then Matr_Calc.CurrToMultix(curr, jj, ertot)
      else Mx_Calc.CurrToMultix(curr, jj, ertot);                {copy currents from vectors }
      if (ertot = 0) then CurrGrid.UpdatePhasors(cRead,@curr,ii);  {update the phasor graphs}
    end;
    SeqGrid.UpdatePhasors(cCopy,nil,0);        {update also the sequence currents}
  end;
end;

procedure TMatrForm.PhasorItemClick(Sender: TObject);
const
  top0      : integer = -1;  {position}
var
  dir0,dir1 : string;        {directories}
begin
  CurrPhasorForm.visible := not CurrPhasorForm.visible;
  {set initial position}
  if (CurrPhasorForm.visible) and (top0 < 0) then begin
    CurrPhasorForm.top := top;   {same top as the main form}
    top0 := top;
  end;
  {set the initial directory, if not yet set}
  CurrGrid.GetInitDir(dir0,dir1);   {original and present initial directories}
  if (dir0 = dir1) and (PhasorsInitialDir <> '') then
    CurrGrid.PutInitDir(PhasorsInitialDir);
end;

procedure TMatrForm.FormActivate(Sender: TObject);
begin
{ if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;}
end;

procedure TMatrCalc.CopyRow(fromrow,torow: integer; CopyBut: TSpeedButton);
{copy text from fromrow to torow in the selected vector
 CopyBut  button starting the copying}
begin
  if (fromrow >= 0) and (fromrow <= 2) and (torow >= 0) and (torow <= 2) then begin
    SaveState;                                     {save for undo}
    SaveGrid(rdVect_c);                            {save grid for checks}
    MarkLastOperation(CopyBut.Left,CopyBut.top,CopyBut.Width,CopyBut.Height);
    pgrid_c[rdVect_c]^.cells[0,torow] := pgrid_c[rdVect_c]^.cells[0,fromrow];
    pgrid_c[rdVect_c]^.cells[2,torow] := pgrid_c[rdVect_c]^.cells[2,fromrow];
{   MatrForm.UpdatePhasors; }
{   if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront; }
  end;
end;

procedure TMatrForm.CopyBut21Click(Sender: TObject);
{copy row 2 to row 1 in the selected vector}
begin
  Matr_Calc.CopyRow(1,0,CopyBut21);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.CopyBut31Click(Sender: TObject);
begin
  Matr_Calc.CopyRow(2,0,CopyBut31);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.CopyBut12Click(Sender: TObject);
begin
  Matr_Calc.CopyRow(0,1,CopyBut12);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.CopyBut32Click(Sender: TObject);
begin
  Matr_Calc.CopyRow(2,1,CopyBut32);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.CopyBut13Click(Sender: TObject);
begin
  Matr_Calc.CopyRow(0,2,CopyBut13);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.CopyBut23Click(Sender: TObject);
begin
  Matr_Calc.CopyRow(1,2,CopyBut23);  {indices go from 0 to 2}
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.SwapRows(row1,row2: integer; SwapBut: TSpeedButton);
{swap texts of row1 and row2 in the selected vector
 SwapBut  button starting the swapping}
var
  txt1,txt2 : string; {temporary}
begin
  if (row1 >= 0) and (row1 <= 2) and (row2 >= 0) and (row2 <= 2) then begin
    SaveState;                                     {save for undo}
    SaveGrid(rdVect_c);                            {save grid for checks}
    MarkLastOperation(SwapBut.Left,SwapBut.top,SwapBut.Width,SwapBut.Height);
    txt1 := pgrid_c[rdVect_c]^.cells[0,row1];
    txt2 := pgrid_c[rdVect_c]^.cells[2,row1];
    pgrid_c[rdVect_c]^.cells[0,row1] := pgrid_c[rdVect_c]^.cells[0,row2];
    pgrid_c[rdVect_c]^.cells[2,row1] := pgrid_c[rdVect_c]^.cells[2,row2];
    pgrid_c[rdVect_c]^.cells[0,row2] := txt1;
    pgrid_c[rdVect_c]^.cells[2,row2] := txt2;
{   UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;}
  end;
end;

procedure TMatrForm.SwapBut12Click(Sender: TObject);
{exchange rows 1 and 2 in the selected vector}
begin
  Matr_Calc.SwapRows(0,1,SwapBut12);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.SwapBut13Click(Sender: TObject);
begin
  Matr_Calc.SwapRows(0,2,SwapBut13);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.SwapBut23Click(Sender: TObject);
begin
  Matr_Calc.SwapRows(1,2,SwapBut23);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.RotateRows(up: boolean; RotBut: TSpeedButton);
{Rotate rows of the selected vector up or down
 up      if true, rotate up
 RotBut  button starting the rotation}
var
  txt1,txt2 : string; {temporary}
begin
  SaveState;                                     {save for undo}
  SaveGrid(rdVect_c);                            {save grid for checks}
  MarkLastOperation(RotBut.Left,RotBut.top,RotBut.Width,RotBut.Height);
  txt1 := pgrid_c[rdVect_c]^.cells[0,0];
  txt2 := pgrid_c[rdVect_c]^.cells[2,0];
  if (up) then begin                             {rotate up}
    pgrid_c[rdVect_c]^.cells[0,0] := pgrid_c[rdVect_c]^.cells[0,1];
    pgrid_c[rdVect_c]^.cells[2,0] := pgrid_c[rdVect_c]^.cells[2,1];
    pgrid_c[rdVect_c]^.cells[0,1] := pgrid_c[rdVect_c]^.cells[0,2];
    pgrid_c[rdVect_c]^.cells[2,1] := pgrid_c[rdVect_c]^.cells[2,2];
    pgrid_c[rdVect_c]^.cells[0,2] := txt1;
    pgrid_c[rdVect_c]^.cells[2,2] := txt2;
  end else begin                                 {rotate down}
    pgrid_c[rdVect_c]^.cells[0,0] := pgrid_c[rdVect_c]^.cells[0,2];
    pgrid_c[rdVect_c]^.cells[2,0] := pgrid_c[rdVect_c]^.cells[2,2];
    pgrid_c[rdVect_c]^.cells[0,2] := pgrid_c[rdVect_c]^.cells[0,1];
    pgrid_c[rdVect_c]^.cells[2,2] := pgrid_c[rdVect_c]^.cells[2,1];
    pgrid_c[rdVect_c]^.cells[0,1] := txt1;
    pgrid_c[rdVect_c]^.cells[2,1] := txt2;
  end;
{ UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;}
end;

procedure TMatrForm.RotUpButClick(Sender: TObject);
begin
  Matr_Calc.RotateRows(true,RotUpBut);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrForm.RotDownButClick(Sender: TObject);
begin
  Matr_Calc.RotateRows(false,RotDownBut);
  UpdatePhasors;
  if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
end;

procedure TMatrCalc.AddAngle(row: integer; AngleBut: TSpeedButton; var err: integer);
{add the value in AddAngleEdit to the given row of the selected vector
 row     index of the row (0...2)
 AddBut  button starting the addition
 err     error flag}
var
  aa,av     : double;  {angles}
  apu,apv   : double;  {temporary}
  apw       : double;  {temporary}
  era,erv   : integer; {error flag}
  ierr      : integer; {index of the error}
  ipu       : integer; {temporary}
  txt1,txt2 : string;  {temporary}
begin
  if (row >= 0) and (row <= 2) then begin
    {get the selected angle of the selected vector, to av}
    ValuesFromLine(pgrid_c[rdVect_c]^.Cells[2,row], 1, ipu, av,apu, apv, apw, erv,ierr);
    {get the angle to be added}
    ValuesFromLine(pAddEdit_c^.text, 1, ipu, aa, apu, apv, apw, era,ierr);
    err := era+erv;
    {continue, if both values correct}
    if (err = 0) then begin
      SaveState;                                     {save for undo}
      SaveGrid(rdVect_c);                            {save grid for checks}
      MarkLastOperation(AngleBut.Left,AngleBut.top,AngleBut.Width,AngleBut.Height);
      apu := av+aa;
      PutAngle(apu, row, 2);       {write the angle to the selected vector}
{     UpdatePhasors;
      if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront; }
    end;
  end;
end;

procedure TMatrForm.AddAngleBut1Click(Sender: TObject);
var
  err : integer;
begin
  Matr_Calc.AddAngle(0,AddAngleBut1,err);
  if (err = 0) then begin
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.AddAngleBut2Click(Sender: TObject);
var
  err : integer;
begin
  Matr_Calc.AddAngle(1,AddAngleBut2,err);
  if (err = 0) then begin
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.AddAngleBut3Click(Sender: TObject);
var
  err : integer;
begin
  Matr_Calc.AddAngle(2,AddAngleBut3,err);
  if (err = 0) then begin
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrCalc.SaveButClick(abc: char; SaveBut: TSpeedButton; var err: integer );
{copy the selected vector to the vector indicated by the character (from the save box)
 abc      the identifying character a..f  for vectors on the vector sheed,
          or u..z  for vectors on the matrix edit sheet
 SaveBut  button clicked to save the vector
 err      error flag}
var
  ii,jj : integer;      {index}
  ll    : longint;      {temporary}
  mx    : boolean;      {matrix edit form}
begin
  ll := ord(abc);         {ascii code of the third char}
  if (ll >= 97) and (ll <= 102) then begin           {a...f}
    ii := ll-96;          {index of the vector, 1..6}
    if (sheet_c = vect) then mx := false else mx := true;   {vector form}
  end else if (ll >= 117) and (ll <= 122) then begin {u...z}
    ii := ll-116;         {index of the vector in the matrix sheet, 1..6}
    if (sheet_c = vect) then mx := true else mx := false;   {matrix edit form}
  end else begin
    ii := -1;             {error case}
    mx := false;
  end;
  {if reasonable index, copy}
  if (ii > 0) then begin
    err := 0;
    SaveState;                          {save for undo}
    if (not mx) then SaveGrid(ii);      {save grid for checks}
    MarkLastOperation(SaveBut.Left,SaveBut.top,SaveBut.Width,SaveBut.Height);
    if (not mx) then begin                        {vector sheet}
      for jj := 0 to 2 do begin
        pgrid_c[ii]^.cells[0,jj] := pgrid_c[rdVect_c]^.cells[0,jj];
        pgrid_c[ii]^.cells[2,jj] := pgrid_c[rdVect_c]^.cells[2,jj];
      end;
    end else begin          {matrix edit sheet}
      for jj := 0 to 2 do begin
        pMgrid_c[ii]^.cells[0,jj] := pgrid_c[rdVect_c]^.cells[0,jj];
        pMgrid_c[ii]^.cells[2,jj] := pgrid_c[rdVect_c]^.cells[2,jj];
      end;
    end;
{   UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront; }
  end else begin
    err := 1;
  end;
end;


procedure TMatrForm.SaveButClick(Sender: TObject);
{copy the selected vector to the vector indicated by the SaveBox}
var
  abc   : char;         {identifying character}
  err   : integer;      {error flag}
begin
  abc := SaveBox.text[3]; {the third char. identifies the vector}
  Matr_Calc.SaveButClick(abc, SaveBut, err);
  if (err = 0) then begin
    UpdatePhasors;
    if (CurrPhasorForm.visible) then CurrPhasorForm.BringToFront;
  end;
end;

procedure TMatrForm.MatrEditMenuClick(Sender: TObject);
begin
  {save and clear phasor display, to give space for phasors in the matrix sheet}
  Matr_Calc.SavePhasors;
  CurrGrid.RemoveAllPhasors;         {initially, clear all phasors}
  SeqGrid.RemoveAllPhasors;
  CurrPhasorForm.visible := false;
  {hide the vector sheet, show the matrix sheet}
  MatrForm.visible := false;
  MxEditForm.top := top;
  MxEditForm.left := left;
  MxEditForm.visible := true;
  {restore the phasors of the matrix sheet}
  Mx_Calc.RestorePhasors;
end;

procedure TMatrCalc.initialize(pLOImg: TpImage);
begin
  pLOImg_c := pLOImg;
  {initialize the previous position and size}
  LOleft_c   := pLOImg_c^.left;
  LOtop_c    := pLOImg_c^.top;
  LOWidth_c  := pLOImg_c^.width;
  LOheight_c := pLOImg_c^.height;
  nVect_c := high(pr11_c);         {no of vectors = dimension, same for all}
  nPhasor_c := 0;                  {no phasors initially}
end;

procedure TMatrCalc.MarkLastOperation(ButLeft,ButTop,ButWidth,ButHeight: integer);
{draw a colored rectangle around the button last pressed
 ButLeft,  ButTop     position of the button
 ButWidth, ButHeight  size of the button}
var
  ix,iy : integer;   {coordinates of the rectangle}
  hh,ww : integer;   {height and width of the rectangle}
begin
  {save the previous position and size}
  LOleft_c   := pLOImg_c^.left;
  LOtop_c    := pLOImg_c^.top;
  LOWidth_c  := pLOImg_c^.width;
  LOheight_c := pLOImg_c^.height;
  {determine size and position of the rectangle}
  pLOImg_c^.picture := nil;
  ww := ButWidth+2*marg;
  hh := ButHeight+2*marg;
  ix := ButLeft-marg;
  iy := ButTop-marg;
  pLOImg_c^.width := ww;
  pLOImg_c^.height := hh;
  pLOImg_c^.left := ix;
  pLOImg_c^.top := iy;
  {draw the rectangle}
  pLOImg_c^.canvas.pen.color := clYellow;
  pLOImg_c^.canvas.brush.color := clYellow;
  pLOImg_c^.canvas.brush.style := bsSolid;
  pLOImg_c^.canvas.rectangle(0,0,ww,hh);
  pLOImg_c^.SendToBack;
  pLOImg_c^.visible := true;
end;

procedure TMatrCalc.BeforeLastOper;
{indicate the previous last operation}
begin
  pLOImg_c^.picture := nil;
  pLOImg_c^.left := LOleft_c;
  pLOImg_c^.top := LOtop_c;
  pLOImg_c^.width := LOWidth_c ;
  pLOImg_c^.height := LOheight_c;
  pLOImg_c^.canvas.pen.color := clYellow;
  pLOImg_c^.canvas.brush.color := clYellow;
  pLOImg_c^.canvas.brush.style := bsSolid;
  pLOImg_c^.canvas.rectangle(0,0,pLOImg_c^.width,pLOImg_c^.height);
end;

procedure TMatrCalc.PutR11(pra,prb,prc,prd,pre,prf: TpAlEdit);
{fill the arrays to pointers to the coefficients 11}
begin
  pr11_c[1] := pra;
  pr11_c[2] := prb;
  pr11_c[3] := prc;
  pr11_c[4] := prd;
  pr11_c[5] := pre;
  pr11_c[6] := prf;
end;

procedure TMatrCalc.PutR12(pra,prb,prc,prd,pre,prf: TpAlEdit);
{fill the arrays to pointers to the coefficients 12}
begin
  pr12_c[1] := pra;
  pr12_c[2] := prb;
  pr12_c[3] := prc;
  pr12_c[4] := prd;
  pr12_c[5] := pre;
  pr12_c[6] := prf;
end;

procedure TMatrCalc.PutR21(pra,prb,prc,prd,pre,prf: TpAlEdit);
{fill the arrays to pointers to the coefficients 21}
begin
  pr21_c[1] := pra;
  pr21_c[2] := prb;
  pr21_c[3] := prc;
  pr21_c[4] := prd;
  pr21_c[5] := pre;
  pr21_c[6] := prf;
end;

procedure TMatrCalc.PutR22(pra,prb,prc,prd,pre,prf: TpAlEdit);
{fill the arrays to pointers to the coefficients 22}
begin
  pr22_c[1] := pra;
  pr22_c[2] := prb;
  pr22_c[3] := prc;
  pr22_c[4] := prd;
  pr22_c[5] := pre;
  pr22_c[6] := prf;
end;

procedure TMatrCalc.PutRes(pra,prb,prc,prd,pre,prf: TpAlEdit);
{fill the arrays to pointers to the coefficients 22}
begin
  pres_c[1] := pra;
  pres_c[2] := prb;
  pres_c[3] := prc;
  pres_c[4] := prd;
  pres_c[5] := pre;
  pres_c[6] := prf;
end;

procedure TMatrCalc.PutpGrid(pa,pb,pc,pd,pe,pf,pch: TpGrid);
{fill the array of pointers to the vectors}
begin
  pGrid_c[1] := pa;
  pGrid_c[2] := pb;
  pGrid_c[3] := pc;
  pGrid_c[4] := pd;
  pGrid_c[5] := pe;
  pGrid_c[6] := pf;
  pchGrid_c := pch;  {check grid}
end;

procedure TMatrCalc.GetpGrid(var pGrid : TpGridArr);
{get the array of pointers to the vector grids}
begin
  pGrid := pGrid_c;
end;

procedure TMatrCalc.PutpMgrid(pa,pb,pc,pd,pe,pf: TpGrid);
{fill the array of pointers to the vectors in the matrix edit sheet}
begin
  pMgrid_c[1] := pa;
  pMgrid_c[2] := pb;
  pMgrid_c[3] := pc;
  pMgrid_c[4] := pd;
  pMgrid_c[5] := pe;
  pMgrid_c[6] := pf;
end;

procedure TMatrCalc.PutpMatr(pMatr: TpGrid);
begin
  pMatr_c := pMatr;
end;  

procedure TMatrCalc.PutpErrEd(pErEd: TpEdit);
begin
  pErrEd_c := pErEd;
end;

procedure TMatrCalc.PutForm(pForm: TpForm);
begin
  pForm_c := pForm;
end;  

procedure TMatrCalc.PutAddEdit(pAddEdit: TpAlEdit);
begin
  pAddEdit_c := pAddEdit;
end;

procedure TMatrCalc.PutVectIndex(ind : integer);
begin
  rdVect_c := ind;
end;

procedure TMatrCalc.GetVectIndex(var ind : integer);
begin
  ind := rdVect_c;
end;

procedure TMatrCalc.ClearVect;
{clear the selected vector}
var
  ii : integer;
begin
  for ii := 0 to 2 do begin
    pGrid_c[rdVect_c]^.cells[0,ii] := '0';
    pGrid_c[rdVect_c]^.cells[2,ii] := '0';
  end;
end;

procedure TMatrCalc.ClearScalar;
{clear the coefficients of the selected vector}
begin
  pr11_c[rdVect_c]^.text := '';
  pr12_c[rdVect_c]^.text := '';
  pr21_c[rdVect_c]^.text := '';
  pr22_c[rdVect_c]^.text := '';
  pres_c[rdVect_c]^.text := '';
end;

procedure TMatrCalc.SetSheet(sheet: Tsheet);
begin
  sheet_c := sheet;
  {labels of vectors, for phasors}
  if (sheet = vect) then begin
    Vect_lbl[1] := 'Va';
    Vect_lbl[2] := 'Vb';
    Vect_lbl[3] := 'Vc';
    Vect_lbl[4] := 'Vd';
    Vect_lbl[5] := 'Ve';
    Vect_lbl[6] := 'Vf';
  end else begin
    {labels of vectors, for phasors}
    Vect_lbl[1] := 'Vu';
    Vect_lbl[2] := 'Vv';
    Vect_lbl[3] := 'Vw';
    Vect_lbl[4] := 'Vx';
    Vect_lbl[5] := 'Vy';
    Vect_lbl[6] := 'Vz';
  end;
end;

procedure TMatrCalc.HideVector;
{hide the presently selected vector}
begin
  pGrid_c[rdVect_c]^.visible := false;
end;

procedure TMatrCalc.MoveAndShowVector(iv,ix,iy: integer);
{move the given vector (index iv) to the given position (ix, iy)
 show it, and make it the selected vector
 iv     index of the vector
 ix,iy  x- and y-coordinates}
begin
  if (iv > 0) and (iv <= nVect_c) then begin
    rdVect_c := iv;
    pGrid_c[rdVect_c]^.left := ix;
    pGrid_c[rdVect_c]^.top := iy;
    pGrid_c[rdVect_c]^.visible := true;
  end;
end;

procedure TMatrCalc.PutVect(ii: integer;m1,m2,m3,a1,a2,a3: string);
{put the given magnitudes and angles in the vector
 ii  index of the vector
 m1,m2,m3   magnitudes, as strings
 a1,a2,a3   angles, as strings (in degrees)}
begin
  if (ii > 0) and (ii <= nVect_c) then begin
    pgrid_c[ii]^.cells[0,0] := m1;
    pgrid_c[ii]^.cells[0,1] := m2;
    pgrid_c[ii]^.cells[0,2] := m3;
    pgrid_c[ii]^.cells[2,0] := a1;
    pgrid_c[ii]^.cells[2,1] := a2;
    pgrid_c[ii]^.cells[2,2] := a3;
  end;
end;

procedure TMatrCalc.SavePhasors;
{save (to a local array) the data for phasors, so that they can be restored}
var
  ii : integer;
begin
  nPhasor_c := CurrGrid.NoOfItems;         {number of phasors in the graph}
  if (nPhasor_c > 0) then begin
    for ii := 1 to nPhasor_c do begin
      PhaDat_c[ii].phi := CurrGrid.Getip(ii);                   {index of the phasor}
      {phase components}
      CurrGrid.GetColorIndex(ii,PhaDat_c[ii].ia,PhaDat_c[ii].ib,
                                PhaDat_c[ii].ic);               {colors of phasors}
      CurrGrid.GetIDRelPos(ii,PhaDat_c[ii].x1,PhaDat_c[ii].y1,
                              PhaDat_c[ii].x2,PhaDat_c[ii].y2,
                              PhaDat_c[ii].x3,PhaDat_c[ii].y3); {relative position of identifiers}
      CurrGrid.GetScaleIndex(PhaDat_c[ii].ind);                 {index for scaling the graph}
      CurrGrid.GetCaption(ii,PhaDat_c[ii].txt);                 {caption, or title}
      {sequence components, similarly}
      SeqDat_c[ii].phi := SeqGrid.Getip(ii);                    {index of the phasor}
      SeqGrid.GetColorIndex(ii,SeqDat_c[ii].ia,SeqDat_c[ii].ib,
                               SeqDat_c[ii].ic);                {colors of phasors}
      SeqGrid.GetIDRelPos(ii,SeqDat_c[ii].x1,SeqDat_c[ii].y1,
                             SeqDat_c[ii].x2,SeqDat_c[ii].y2,
                             SeqDat_c[ii].x3,SeqDat_c[ii].y3);  {relative position of identifiers}
      SeqGrid.GetScaleIndex(SeqDat_c[ii].ind);                  {index for scaling the graph}
      SeqGrid.GetCaption(ii,SeqDat_c[ii].txt);                  {caption, or title}
    end;
  end;
end;

procedure TMatrCalc.RestorePhasors;
{restore the phasors from the saved data}
var
  curr  : TMultix;   {currents in one variable}
  ertot : integer;   {error flag}
  ii    : integer;
begin
  for ii := 1 to nPhasor_c do begin                            {add all to the phasor list}
    {reasonable index}
    if (PhaDat_c[ii].phi > 0) and (PhaDat_c[ii].phi <= nVect) then begin
      CurrToMultix(curr, PhaDat_c[ii].phi, ertot);   {copy currents from vector }
      CurrGrid.ShowPhasors(PhaDat_c[ii].phi, cRead, @Curr,PhaDat_c[ii].txt);   {add phasors}
      CurrGrid.PutColorIndex(ii, PhaDat_c[ii].ia,PhaDat_c[ii].ib,PhaDat_c[ii].ic);                  {set colors}
      CurrGrid.PutIDRelPos(ii,PhaDat_c[ii].x1,PhaDat_c[ii].y1,
                              PhaDat_c[ii].x2,PhaDat_c[ii].y2,
                              PhaDat_c[ii].x3,PhaDat_c[ii].y3); {coordinates of ID}
      CurrGrid.PutScaleIndex(PhaDat_c[ii].ind);                 {index for scaling the graph}
      CurrGrid.DrawPhasors;                                     {update graph}
    end;
    {the same for sequence components}
    {reasonable index}
    if (PhaDat_c[ii].phi > 0) and (PhaDat_c[ii].phi <= nVect) then begin
      SeqGrid.ShowPhasors(PhaDat_c[ii].phi,cCopy,nil,PhaDat_c[ii].txt);  {add seq. phasors}
      SeqGrid.PutColorIndex(ii, SeqDat_c[ii].ia,SeqDat_c[ii].ib,SeqDat_c[ii].ic);                   {set colors}
      SeqGrid.PutIDRelPos(ii,SeqDat_c[ii].x1,SeqDat_c[ii].y1,
                             SeqDat_c[ii].x2,SeqDat_c[ii].y2,
                             SeqDat_c[ii].x3,SeqDat_c[ii].y3);  {coordinates of ID}
      SeqGrid.PutScaleIndex(SeqDat_c[ii].ind);                  {index for scaling the graph}
      SeqGrid.DrawPhasors;                                      {update graph}
    end;
  end;
end;

procedure TMatrForm.FormDestroy(Sender: TObject);
begin
{ ShellExecute(Handle, 'open', 'c:\imago\imagopro.exe', nil, nil, SW_SHOWNORMAL) ;}
{  Matr_calc.free;}  {moved to FormClose}
{  Mx_Calc.free;  }
{cpf.free;}
end;


procedure TMatrForm.FormClose(Sender: TObject; var Action: TCloseAction);
{write the .ini-file = saved matrices, etc.}
var
  ExeFile  : string;    {name of this *.exe file, without the extension}
  ExeName  : string;    {name of this *.exe-file, with the extension}
  FileName : string;    {name of .ini file}
  ii,jj    : integer;   {index}
  IniFile  : TextFile;  {initialization file}
  pArr     : TpMatrStrArr; {temporary pointer for saving the matrix}
begin
  {open (or create) the *.ini-file}
  ExeName := ExtractFileName(paramStr(0));
  ExeFile := ChangeFileExt(ExeName, '');   {remove the extension}
  FileName := ExtractFilePath(paramStr(0)) + ExeFile+'.ini'; {complete name}
  AssignFile(IniFile, FileName);
  {$I-}
  rewrite(IniFile);                        {open the file}
  {$I+}
  if IOResult = 0 then begin               {successful open file}
    try
      {save the matrices}
      if (MxEditForm.Matr_List.count > 0) then begin
        for ii := 0 to MxEditForm.Matr_List.count-1 do begin
          writeln(IniFile,'!');                      {mark matrices}
          pArr := MxEditForm.Matr_List.items[ii];    {pointer to a saved matrix}
          for jj := 0 to high(pArr^) do begin
            writeln(IniFile,pArr^[jj]);
          end;
        end;
      end;
      {write the initial directories}
      writeln(IniFile,'#');                 {flag for the initial dir.}
      writeln(IniFile,MatrForm.SaveFormDlg.InitialDir);
      writeln(IniFile,MxEditForm.SaveFormDlg.InitialDir);
      writeln(IniFile,NumberForm.SaveNumbersDlg.InitialDir);
      writeln(IniFile,CurrPhasorForm.PhasorDialog.InitialDir);
      writeln(IniFile,NotesForm.SaveDlg.InitialDir);
    finally
      CloseFile(IniFile);
    end;
  end else begin                         {open file failed}
    MessageDlg('Cannot update *.ini file', mtWarning, [mbOk], 0);
  end;
  Matr_calc.free;
  Mx_Calc.free;
end;

procedure TMatrForm.Seq3ItemClick(Sender: TObject);
begin
  Seq3Item.checked := not Seq3Item.checked;
  Seq3Label.visible := Seq3Item.checked;
  Matr_Calc.Set3I03I2(Seq3Item.checked);
end;

procedure TMatrCalc.ReadPhasorPar(pFile: TpTextFile);
{read the data of the phasor graphs from file
 pFile  pointer to the file where written}
var
  ii,kk : integer;
  pDat  : TpPhasorData;   {pointer to the data}
begin
  for ii := 1 to nPhasor_c do begin            {all phasors = all current trafos}
    for kk := 1 to 2 do begin
      if (kk = 1) then pDat := @PhaDat_c[ii]
      else             pDat := @SeqDat_c[ii];
      readln(pFile^,pDat^.phi); {read index from the file}
      readln(pFile^,pDat^.ia);  {read color index from the file}
      readln(pFile^,pDat^.ib);  {read color index from the file}
      readln(pFile^,pDat^.ic);  {read color index from the file}
      readln(pFile^,pDat^.x1);  {read rel.pos of the identifiers}
      readln(pFile^,pDat^.y1);
      readln(pFile^,pDat^.x2);
      readln(pFile^,pDat^.y2);
      readln(pFile^,pDat^.x3);
      readln(pFile^,pDat^.y3);
      readln(pFile^,pDat^.ind);
      readln(pFile^,pDat^.txt);
    end;
  end;
end;

procedure TMatrCalc.WritePhasorPar(pFile: TpTextFile; nCT: integer);
{write the data of the phasor graphs to file
 pFile  pointer to the file where written}
var
  ii,kk : integer;
  pDat  : TpPhasorData;   {pointer to the data}
begin
  for ii := 1 to nCT do begin            {all phasors = all current trafos}
    for kk := 1 to 2 do begin
      if (kk = 1) then pDat := @PhaDat_c[ii]
      else             pDat := @SeqDat_c[ii];
      Writeln(pFile^,inttostr(pDat^.phi)); {read index from the file}
      Writeln(pFile^,inttostr(pDat^.ia));  {read color index from the file}
      Writeln(pFile^,inttostr(pDat^.ib));  {read color index from the file}
      Writeln(pFile^,inttostr(pDat^.ic));  {read color index from the file}
      Writeln(pFile^,inttostr(pDat^.x1));  {read rel.pos of the identifiers}
      Writeln(pFile^,inttostr(pDat^.y1));
      Writeln(pFile^,inttostr(pDat^.x2));
      Writeln(pFile^,inttostr(pDat^.y2));
      Writeln(pFile^,inttostr(pDat^.x3));
      Writeln(pFile^,inttostr(pDat^.y3));
      Writeln(pFile^,inttostr(pDat^.ind));
      Writeln(pFile^,pDat^.txt);
    end;
  end;
end;

procedure SaveOrReadLastOperation(pFile : TpTextFile; save: boolean; pLOimg: TpImage);
{save or read the marker of the last operation
 pFile   pointer to file where written/read
 save    flag for saving or restoring
 pLOimg  pointer to the image of the last operation}
var
  line: string;
begin
  if (save) then begin
    Writeln(pFile^,'last operation in the matrix sheet');
    Writeln(pFile^,inttostr(pLOimg.left));
    Writeln(pFile^,inttostr(pLOimg.top));
    Writeln(pFile^,inttostr(pLOimg.width));
    Writeln(pFile^,inttostr(pLOimg.height));
  end else begin
    readln(pFile^,line);    {the caption}
    pLOimg.picture := nil;
    readln(pFile^,line);
    pLOimg.left := strtoint(line);
    readln(pFile^,line);
    pLOimg.top := strtoint(line);
    readln(pFile^,line);
    pLOimg.width := strtoint(line);
    readln(pFile^,line);
    pLOimg.height := strtoint(line);
    {draw the rectangle}
    pLOimg.canvas.pen.color := clYellow;
    pLOimg.canvas.brush.color := clYellow;
    pLOimg.canvas.brush.style := bsSolid;
    pLOimg.canvas.rectangle(0,0,pLOimg.width,pLOimg.height);
    pLOimg.SendToBack;
    pLOimg.visible := true;
  end;
end;

procedure TMatrCalc.PhasorsToFile(pFile: TpTextFile);
{save the phasor data to file, for restoring}
begin
  if (pForm_c^.visible) then begin     {phasors of MatrForm in graph}
    SavePhasors;                     {save to local array..}
  end;                               {..if not visible, then already saved}
  {write to file, if there are any phasors}
  Writeln(pFile^,'phasors');
  if (nPhasor_c > 0) then begin
    Writeln(pFile^,inttostr(nPhasor_c)); {write number of phasors}
    WritePhasorPar(pFile,nPhasor_c);     {write data for phasors to file}
  end else begin
    Writeln(pFile^,'0');                 {write 0 = number of phasors}
  end;
  if (CurrPhasorForm.visible) and (pForm_c^.visible) then Writeln(pFile^,'1')
  else Writeln(pFile^,'0');
end;


procedure TMatrCalc.PhasorsFromFile(pFile: TpTextFile);
{read the phasor data from file, and save it to the local array}
var
  line: string;
begin
  readln(pFile^,line);               {read the caption}
  readln(pFile^,line);               {number of phasors}
  if (line = '0') then begin
    {no phasors, do nothing }
    nPhasor_c := 0;
  end else begin
    nPhasor_c := strtoint(line);     {number of current transformers = phasors}
    ReadPhasorPar(pFile);            {read phasor data from the file}
  end;
  {remove old phasors, restore the saved phasors, if this form visible}
  if (pForm_c^.visible) then begin
    CurrGrid.RemoveAllPhasors;       {initially, clear all phasors}
    SeqGrid.RemoveAllPhasors;
    if (nPhasor_c > 0) then begin
      RestorePhasors;
    end;
  end;
  readln(pFile^,line);               {phasor form visible/not}
  if (line = '1') and (pForm_c^.visible) then CurrPhasorForm.visible := true
  else CurrPhasorForm.visible := false;
end;

procedure TMatrForm.SaveOrRestore(pFile: TpTextFile; save: boolean);
{save or restore the state
 pFile     pointer to the file where saved, or from where restored
 save      if true, save, else restore}
var
  Curr     : TMultix;  {currents in one variable}
  ertot    : integer;  {error flag}
  ia,ib,ic : integer;  {color indices}
  ii,jj,kk : integer;
  ind      : integer;  {scaling index}
  ipu      : integer;
  nCT      : integer;  {number of current trafos}
  line     : string;   {one line read from the file}
  phi      : integer;  {index of the phasor}
  txt      : string;   {the caption}
  x1,x2,x3 : integer;  {x-coordinates of identifiers}
  y1,y2,y3 : integer;  {y-coordinates of identifiers}
begin
  {the coefficients in the vector sheet - - - - - - - - - - - - - - - }
  if (save)then begin
    for ii := 1 to nVect do begin
      Writeln(pFile^,Matr_Calc.pr11_c[ii]^.text);
      Writeln(pFile^,Matr_Calc.pr12_c[ii]^.text);
      Writeln(pFile^,Matr_Calc.pr21_c[ii]^.text);
      Writeln(pFile^,Matr_Calc.pr22_c[ii]^.text);
      Writeln(pFile^,Matr_Calc.pres_c[ii]^.text);
    end;
  end else begin
    for ii := 1 to nVect do begin
      readln(pFile^,line);
      Matr_Calc.pr11_c[ii]^.text := line;
      readln(pFile^,line);
      Matr_Calc.pr12_c[ii]^.text := line;
      readln(pFile^,line);
      Matr_Calc.pr21_c[ii]^.text := line;
      readln(pFile^,line);
      Matr_Calc.pr22_c[ii]^.text := line;
      readln(pFile^,line);
      Matr_Calc.pres_c[ii]^.text := line;
    end;
  end;
  {the vectors in the vector sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    for ii := 1 to nVect do begin
      for jj := 0 to 2 do begin
        Writeln(pFile^,Matr_Calc.pGrid_c[ii]^.cells[0,jj]);
        Writeln(pFile^,Matr_Calc.pGrid_c[ii]^.cells[2,jj]);
      end;
    end;
  end else begin
    for ii := 1 to nVect do begin
      for jj := 0 to 2 do begin
        readln(pFile^,line);
        Matr_Calc.pGrid_c[ii]^.cells[0,jj] := line;
        readln(pFile^,line);
        Matr_Calc.pGrid_c[ii]^.cells[2,jj] := line;
      end;
    end;
  end;
 {the check grid in the vector sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    for jj := 0 to 2 do begin
      Writeln(pFile^,CheckGrid.cells[0,jj]);
      Writeln(pFile^,CheckGrid.cells[2,jj]);
    end;
    {the position of the check grid}
    Writeln(pFile^,inttostr(CheckGrid.left));
    Writeln(pFile^,inttostr(CheckGrid.top));
  end else begin
    for jj := 0 to 2 do begin
      readln(pFile^,line);
      CheckGrid.cells[0,jj] := line;
      readln(pFile^,line);
      CheckGrid.cells[2,jj] := line;
    end;
    {the position of the check grid}
    readln(pFile^,line);
    CheckGrid.left := strtoint(line);
    readln(pFile^,line);
    CheckGrid.top := strtoint(line);
  end;
  {texts on edits and other boxes on the vector sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    Writeln(pFile^,AddAngleEdit.text);
    Writeln(pFile^,SelectMatrBox.text);
    Writeln(pFile^,SaveBox.text);
  end else begin
    readln(pFile^,line);
    AddAngleEdit.text := line;
    readln(pFile^,line);
    SelectMatrBox.text := line;
    readln(pFile^,line);
    SaveBox.text := line;
  end;
  {image marking the last operation - - - - - - - - - - - - - - - }
  SaveOrReadLastOperation(pFile, save, @LastOperImg);
  {index of the selected vector - - - - - - - - - - - - - - - }
  if (save) then begin
    Writeln(pFile^,inttostr(Matr_Calc.rdVect_c));
  end else begin
    readln(pFile^,line);
    ipu := strtoint(line);
    if (ipu > 0)  and (ipu <= nVect) then begin
      Matr_Calc.rdVect_c := ipu;
      if      (ipu = 1) then RadioButA.checked := true
      else if (ipu = 2) then RadioButB.checked := true
      else if (ipu = 3) then RadioButC.checked := true
      else if (ipu = 4) then RadioButD.checked := true
      else if (ipu = 5) then RadioButE.checked := true
      else if (ipu = 6) then RadioButF.checked := true
    end;
  end;
  {the 3I0, 3I2 option - - - - - - - - - - - - - - - }
  if (save)then begin
    if (Seq3Item.checked) then Writeln(pFile^,'1')
    else Writeln(pFile^,'0');
  end else begin
    readln(pFile^,line);
    if (line = '1') then begin     {option selected}
      Seq3Item.checked := true;
      Seq3Label.visible := true;
    end else begin
      Seq3Item.checked := false;
      Seq3Label.visible := false;
    end;
  end;
  {the current transformer (elplek) button - - - - - - - - - - - - - - - }
  if (save) then begin
    if (ElplekBut.down) then Writeln(pFile^,'1')
    else Writeln(pFile^,'0');
  end else begin
    readln(pFile^,line);
    if (line = '1') then ElplekBut.down := true
    else ElplekBut.down := false;
  end;
  {the position of the calculator (this form) - - - - - - - - - - - - - - - }
  if (save) then begin
    Writeln(pFile^,inttostr(left));
    Writeln(pFile^,inttostr(top));
  end else begin
    readln(pFile^,line);
    {left := strtoint(line);}
    readln(pFile^,line);
    {top := strtoint(line);}
  end;
  {phasors in the graph of the vector sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    Matr_Calc.PhasorsToFile(pFile);
  end else begin                         {restoring the state of the calculator}
    Matr_Calc.PhasorsFromFile(pFile);
  end;
  {data in the matrix edit form - - - - - - - - - - - - - - - - - - - - - }
  {coefficients for the vector}
  if (save) then begin
    writeln(pFile^,'vector coefficients on the matrix sheet');
    Writeln(pFile^,Mx_Calc.pr11_c[1]^.text);
    Writeln(pFile^,Mx_Calc.pr12_c[1]^.text);
    Writeln(pFile^,Mx_Calc.pr21_c[1]^.text);
    Writeln(pFile^,Mx_Calc.pr22_c[1]^.text);
    Writeln(pFile^,Mx_Calc.pres_c[1]^.text);
  end else begin                           {reading data}
    readln(pFile^,line);                   {caption}
    readln(pFile^,line);
    Mx_Calc.pr11_c[1]^.text := line;
    readln(pFile^,line);
    Mx_Calc.pr12_c[1]^.text := line;
    readln(pFile^,line);
    Mx_Calc.pr21_c[1]^.text := line;
    readln(pFile^,line);
    Mx_Calc.pr22_c[1]^.text := line;
    readln(pFile^,line);
    Mx_Calc.pres_c[1]^.text := line;
  end;
  {the vectors in the matrix sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    Writeln(pFile^,'vectors in the matrix sheet');
    for ii := 1 to nVect do begin
      for jj := 0 to 2 do begin
        Writeln(pFile^,Mx_Calc.pGrid_c[ii]^.cells[0,jj]);
        Writeln(pFile^,Mx_Calc.pGrid_c[ii]^.cells[2,jj]);
      end;
    end;
  end else begin
    readln(pFile^,line);      {the caption}
    for ii := 1 to nVect do begin
      for jj := 0 to 2 do begin
        readln(pFile^,line);
        Mx_Calc.pGrid_c[ii]^.cells[0,jj] := line;
        readln(pFile^,line);
        Mx_Calc.pGrid_c[ii]^.cells[2,jj] := line;
      end;
    end;
  end;
  {the check grid in the matrix sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    writeln(pFile^,'check grid in the matrix sheet');
    for jj := 0 to 2 do begin
      Writeln(pFile^,MxEditForm.CheckGrid.cells[0,jj]);
      Writeln(pFile^,MxEditForm.CheckGrid.cells[2,jj]);
    end;
    {the position of the check grid}
    Writeln(pFile^,inttostr(MxEditForm.CheckGrid.left));
    Writeln(pFile^,inttostr(MxEditForm.CheckGrid.top));
  end else begin
    readln(pFile^,line);       {the caption}
    for jj := 0 to 2 do begin
      readln(pFile^,line);
      MxEditForm.CheckGrid.cells[0,jj] := line;
      readln(pFile^,line);
      MxEditForm.CheckGrid.cells[2,jj] := line;
    end;
    {the position of the check grid}
    readln(pFile^,line);
    MxEditForm.CheckGrid.left := strtoint(line);
    readln(pFile^,line);
    MxEditForm.CheckGrid.top := strtoint(line);
  end;
  {coefficients for the matrix - - - - - - - - - - - - - - - - }
  if (save) then begin
    writeln(pFile^,'matrix coefficients on the matrix sheet');
    Writeln(pFile^,MxEditForm.RgEdit11.text);
    Writeln(pFile^,MxEditForm.RgEdit12.text);
    Writeln(pFile^,MxEditForm.RgEdit21.text);
    Writeln(pFile^,MxEditForm.RgEdit22.text);
  end else begin                           {reading data}
    readln(pFile^,line);                   {caption}
    readln(pFile^,line);
    MxEditForm.RgEdit11.text := line;
    readln(pFile^,line);
    MxEditForm.RgEdit12.text := line;
    readln(pFile^,line);
    MxEditForm.RgEdit21.text := line;
    readln(pFile^,line);
    MxEditForm.RgEdit22.text := line;
  end;
  {the matrix itself - - - - - - - - - - - - - - - - - }
  if (save) then begin
    writeln(pFile^,'the matrix');
    for ii := 0 to 2 do begin
      for jj := 1 to 6 do begin
        kk := ColIndex[jj];
        writeln(pFile^,MxEditForm.MatrGrid.cells[kk,ii]);
      end;
    end;
  end else begin
    readln(pFile^,line);             {the caption}
    for ii := 0 to 2 do begin
      for jj := 1 to 6 do begin
        kk := ColIndex[jj];
        readln(pFile^,line);
        MxEditForm.MatrGrid.cells[kk,ii] := line;
      end;
    end;
  end;
  {index of the selected vector - - - - - - - - - - - - - - - }
  if (save) then begin
    writeln(pFile^,'index of the selected vector');
    Writeln(pFile^,inttostr(Mx_Calc.rdVect_c));
  end else begin
    readln(pFile^,line);   {the caption}
    readln(pFile^,line);
    ipu := strtoint(line);
    if (ipu > 0)  and (ipu <= nVect) then begin
      Mx_Calc.rdVect_c := ipu;
      if      (ipu = 1) then MxEditForm.RadioButU.checked := true
      else if (ipu = 2) then MxEditForm.RadioButV.checked := true
      else if (ipu = 3) then MxEditForm.RadioButW.checked := true
      else if (ipu = 4) then MxEditForm.RadioButX.checked := true
      else if (ipu = 5) then MxEditForm.RadioButY.checked := true
      else if (ipu = 6) then MxEditForm.RadioButZ.checked := true
    end;
  end;
  {texts on edits and other boxes on the matrix sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    Writeln(pFile^,'texts on the matrix sheet');
    Writeln(pFile^,MxEditForm.MatrixBox.text);
    Writeln(pFile^,MxEditForm.StoreEdit.text);
    Writeln(pFile^,MxEditForm.SaveBox.text);
    Writeln(pFile^,MxEditForm.VeditLabel.caption);
    Writeln(pFile^,MxEditForm.RaLab2.caption);
    Writeln(pFile^,MxEditForm.RaLab5.caption);
  end else begin
    readln(pFile^,line);       {the caption}
    readln(pFile^,line);
    MxEditForm.MatrixBox.text := line;
    readln(pFile^,line);
    MxEditForm.StoreEdit.text := line;
    readln(pFile^,line);
    MxEditForm.SaveBox.text := line;
    readln(pFile^,line);
    MxEditForm.VeditLabel.caption := line;
    readln(pFile^,line);
    MxEditForm.RaLab2.caption := line;
    readln(pFile^,line);
    MxEditForm.RaLab5.caption := line;
  end;
  {image marking the last operation - - - - - - - - - - - - - - - }
  SaveOrReadLastOperation(pFile, save, @MxEditForm.LastOperImg);
  {phasors in the graph of the matrix edit sheet - - - - - - - - - - - - - - - }
  if (save) then begin
    Mx_Calc.PhasorsToFile(pFile);
  end else begin                         {restoring the state of the calculator}
    Mx_Calc.PhasorsFromFile(pFile);
  end;
  {the notes in the note-memo - - - - - - - - - - - - - - - - - - - }
  if (save) then begin
    writeln(pFile^,'Notes');
    kk := NotesForm.NoteMemo.lines.count;
    writeln(pFile^,inttostr(kk));
    if (kk > 0) then begin
      for ii := 0 to kk-1 do writeln(pFile^,NotesForm.NoteMemo.lines[ii]);
    end;
  end else begin
    NotesForm.NoteMemo.lines.clear;         {clear old lines first}
    readln(pFile^,line);                   {the caption}
    readln(pFile^,line);
    kk := strtoint(line);                  {no of lines in the note-memo}
    if (kk > 0) then begin
      for ii := 0 to kk-1 do begin
        readln(pFile^,line);
        NotesForm.NoteMemo.lines.add(line);
      end;
    end;
  end;
  {the selections in the sum in the vector sheet}
  if (save) then begin
    writeln(pFile^,'the Add-boxes');
    writeln(pFile^,AddBoxA.text);
    writeln(pFile^,AddBoxB.text);
    writeln(pFile^,AddBoxC.text);
    writeln(pFile^,AddBoxD.text);
    writeln(pFile^,AddBoxE.text);
  end else begin   {restoring}
    {must check end of file, because this item added later}
    if (not (EOF(pFile^))) then begin
      readln(pFile^,line);  {the caption}
      readln(pFile^,line);
      AddBoxA.text := line; {text in the first box of the sum}
      readln(pFile^,line);
      AddBoxB.text := line;
      readln(pFile^,line);
      AddBoxC.text := line;
      readln(pFile^,line);
      AddBoxD.text := line;
      readln(pFile^,line);
      AddBoxE.text := line;
    end;
  end;
end;

procedure TMatrForm.SaveItemClick(Sender: TObject);
{save the state so that one can restart in exactly (?) the same state}
var
  ExeFile  : string;    {name of this *.exe file, without the extension}
  ExeName  : string;    {name of this *.exe-file, with the extension}
  FileName : string;    {name of .ini file}
  ii,jj    : integer;   {index}
  SaveFile : TextFile;  {file for saving}
begin
  {open (or create) the save-file}
  ExeName := ExtractFileName(paramStr(0));
  ExeFile := ChangeFileExt(ExeName, '');   {remove the extension}
  FileName := ExtractFilePath(paramStr(0)) + ExeFile+'.txt'; {complete name}
  AssignFile(SaveFile, FileName);
  {$I-}
  rewrite(SaveFile);                       {open the file}
  {$I+}
  if IOResult = 0 then begin               {successful open file}
    try
      {save all data}
      SaveOrRestore(@SaveFile,true);
    finally
      CloseFile(SaveFile);
    end;
  end else begin                           {open file failed}
    MessageDlg('Cannot write the save-file', mtWarning, [mbOk], 0);
  end;
end;

procedure TMatrForm.OpenItemClick(Sender: TObject);
{read and restore the saved state}
var
  ExeFile  : string;       {name of this *.exe file, without the extension}
  ExeName  : string;       {name of this *.exe-file, with the extension}
  FileName : string;       {name of .ini file}
  ii       : integer;
  SaveFile : TextFile;     {initialization file}
begin
  ExeName := ExtractFileName(paramStr(0));
  ExeFile := ChangeFileExt(ExeName, '');   {remove the extension}
  FileName := ExtractFilePath(paramStr(0)) + ExeFile+'.txt';
  {open it, if it exists}
  if FileExists(FileName) then begin
    AssignFile(SaveFile,FileName);
    Reset(SaveFile);                      {open the file}
    try
      {read the state}
      SaveOrRestore(@SaveFile,false);     {false = restore (=read)}
    finally
      CloseFile(SaveFile);
    end;
  end else begin                          {file does not exist, give warning}
    MessageDlg('Cannot read the save-file', mtWarning, [mbOk], 0);
  end;
end;

procedure SaveJpegImage(pForm: TpForm; pictbit: Tbitmap; FileName: string);
{save a bitmap in the Jpeg form to a file
 pForm     pointer to the form where the bitmap is defined (needed as owner and parent)
 pictbit   the bitmap to be saved
 FileName  name of the file where saved}
var
  FormJpeg    : TJpegImage; {for saving the bitmap as a jpeg-image}
begin
  FormJpeg := TJpegImage.create(pForm^);
  FormJpeg.parent := pForm^;
  FormJpeg.quality := 100;                   {default = 75, but now the file is larger}
  FormJpeg.picture.bitmap.assign(pictbit);   {copy in jpeg-format}
  FormJpeg.savetofile(FileName);
  FormJpeg.free;
end;

procedure TMatrCalc.SaveSheet(pDlg: TpSaveDialog);
{save the form (sheet) to file as bmp, jpg, or wmf (or emf?)}
var
  FormImage : TBitmap;         {image of the form}
  FormMetaf : TMetafile;       {image of the form as a metafile}
  MFCanvas  : TMetaFileCanvas;
begin
  if (pDlg^.execute) then begin
    FormImage := pForm_c^.GetFormImage;
    if (pDlg^.FilterIndex = 1) then begin     {save as a bitmap file}
      FormImage.savetofile(pDlg^.FileName);
    end else if (pDlg^.FilterIndex = 2) then begin {save as a jpeg-file}
      SaveJpegImage(pForm_c, FormImage, pDlg^.FileName);
    end else begin                   {windows metafile}
      FormMetaf := Tmetafile.create;
      FormMetaf.width := pForm_c^.clientwidth;
      FormMetaf.height := pForm_c^.clientheight;
      MFCanvas := TMetafileCanvas.Create(FormMetaf, 0);
      MFCanvas.Draw(0, 0, FormImage);
      MFCanvas.Free;
      FormMetaf.savetofile(pDlg^.FileName);
      FormMetaf.free;
    end;
    FormImage.free;
    pDlg^.InitialDir := ExtractFileDir(pDlg^.FileName); {save directory}
  end;
end;

function TenChar(txt: string): string;
{fill with spaces so that the total length is 10 char}
const
  ten = 10;
var
  ii  : integer;
  len : integer;  {length of the string}
  txx : string;
  txy : string;
begin
  {a space for the sign}
  if (txt[1] = '-') or (txt[1] = '+') then txx := txt else txx := ' '+txt;
  len := length(txx);
  if (len >= ten) then begin
    TenChar := txx;
  end else begin
    txy := txx;
    for ii := 1 to (ten-len) do txy := txy+' ';
    TenChar := txy;
  end;
end;

procedure TMatrCalc.SaveVectors(pFile :TpTextFile);
{save the vectors to the given file}
var
  ii,jj : integer;
  line  : string;
begin
  for ii := 1 to nVect_c do begin           {all vectors}
    writeln(pFile^,' '+Vect_lbl[ii]);
    for jj := 0 to 2 do begin             {all components}
      line := TenChar(pGrid_c[ii]^.cells[0,jj])+' < '+
              TenChar(pGrid_c[ii]^.cells[2,jj]);
      writeln(pFile^,line);
    end;
    writeln(pFile^,' ');
  end;
end;

procedure TMatrForm.ExportSheetItemClick(Sender: TObject);
{export the image of the sheet as a bitmap, jpeg, or wmf (or emf?)}
begin
  Matr_Calc.SaveSheet(@SaveFormDlg);
end;

procedure TMatrForm.NumberItemClick(Sender: TObject);
begin
  NumberForm.visible := true;
end;

procedure TMatrForm.NotesItemClick(Sender: TObject);
begin
  NotesForm.visible := true;
end;

procedure TMatrForm.ButtonItemClick(Sender: TObject);
{show/hide the buttons for tests}
begin
  ButtonItem.checked := not ButtonItem.checked;
  FillBut.visible := ButtonItem.checked;
  CheckButA.visible := ButtonItem.checked;
  CheckButB.visible := ButtonItem.checked;
  CheckButC.visible := ButtonItem.checked;
  CheckButD.visible := ButtonItem.checked;
  CheckButE.visible := ButtonItem.checked;
  CheckButF.visible := ButtonItem.checked;
end;

procedure TMatrForm.HelpTopicsClick(Sender: TObject);
begin
  Application.HelpCommand(HELP_FINDER, 0);
end;

end.
