unit numbers;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, matrix, Mxedit;

type
  TNumberForm = class(TForm)
    VabcBox: TCheckBox;
    MatrixBox: TCheckBox;
    VxyzBox: TCheckBox;
    Label1: TLabel;
    SaveNumbersDlg: TSaveDialog;
    CancelBut: TButton;
    ExportBut: TButton;
    UserMatrixBox: TCheckBox;
    procedure CancelButClick(Sender: TObject);
    procedure ExportButClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  NumberForm: TNumberForm;

implementation

{$R *.DFM}

procedure TNumberForm.CancelButClick(Sender: TObject);
begin
  NumberForm.visible := false;
end;

procedure TNumberForm.ExportButClick(Sender: TObject);
var
  dpu      : double;
  ii,jj,kk : integer;
  ip       : integer;
  ipu      : integer;
  ll       : integer;
  line     : string;
  Outfile  : TextFile;     {file where written}
  pArr     : TpMatrStrArr; {temporary pointer for saving the matrix}
  pFix     : TpM3x3;       {pointer to one fixed matrix}
  pGrid    : TpGridArr;    {array of pointers to the vector grids}
begin
  if (SaveNumbersDlg.execute) then begin
    AssignFile(Outfile, SaveNumbersDlg.filename);    {open the given file,BUT ERRORS ?}
    rewrite(outfile);
    try
      writeln(outfile,'Differential protection calculator');      {some caption}
      writeln(outfile,' ');
      if (VabcBox.checked) then begin           {vectors Va...Vf written - - - - - - - }
        writeln(outfile,'Vectors Va...Vf - - - - - - - - - - - -');
        Matr_Calc.SaveVectors(@Outfile);
      end;
      if (VxyzBox.checked) then begin           {vectors Vu...Vz written - - - - - - - }
        writeln(outfile,'Vectors Vu...Vz - - - - - - - - - - - -');
        Mx_Calc.SaveVectors(@Outfile);
      end;
      if (UserMatrixBox.checked) then begin         {user defined matrices written}
        {save the matrices}
        {the matrices edited (generated) by the user - - - - - - - - - - - - - - }
        if (MxEditForm.Matr_List.count > 0) then begin
          writeln(outfile,' ');
          writeln(outfile,'The matrices edited by the user - - - - - - - - - - - - ');
          writeln(outfile,' ');
          for ii := 0 to MxEditForm.Matr_List.count-1 do begin
            pArr := MxEditForm.Matr_List.items[ii];    {pointer to a saved matrix}
            {first, the caption, or name of the matrix}
            writeln(outfile,' ');                      {one empty line}
            writeln(outfile,pArr^[0]);
            ll := 1;                                   {index in the array}
            for jj := 1 to 3 do begin                  {one row of the matrix}
              line := '';
              for kk := 1 to 6 do begin
                line := line+TenChar(pArr^[ll]);
                {add the symbol for the angle}
                if (kk = 1) or (kk = 3) or (kk = 5) then line := line+' > '
                else line := line+'   ';
                inc(ll);
              end;
              writeln(outfile,line);
            end;
            {the coefficients, first the multipliers}
            line := 'the multipliers';
            for kk := 1 to 2 do begin
              if (pArr^[ll] <> '') then
                line := line+TenChar(pArr^[ll])+' ';
              inc(ll);
            end;
            writeln(outfile,line);
            {then the divisors}
            line := 'the divisors   ';
            for kk := 1 to 2 do begin
              if (pArr^[ll] <> '') then
                line := line+TenChar(pArr^[ll])+' ';
              inc(ll);
            end;
            writeln(outfile,line);
          end;
        end;
      end;
      {the fixed matrices - - - - - - - - - - - - - - - - - - - }
      if (MatrixBox.checked) then begin
        writeln(outfile,' ');
        writeln(outfile,'The fixed matrices - - - - - - - - - - - - ');
        writeln(outfile,' ');
        for ip := 0 to high(MxEditForm.pFix_Matr) do begin
          pFix := MxEditForm.pFix_Matr[ip];           {pointer to the selected fixed matrix}
          writeln(outfile,'[M'+inttostr(ip)+']');     {name of the matrix}
          for ii := 0 to 2 do begin                   {the rows}
            line := '';
            for jj := 0 to 2 do begin
              kk := 3*jj;
              dpu := pFix^[ii*3+jj];
              ipu := round(dpu);
              line  := line+TenChar(inttostr(ipu))+' '; {magnitude (angle = 0, always)}
            end;
            writeln(outfile,line);
          end;
          writeln(outfile,'The multiplier '+ManyChars(mcoeff[ip]));
          writeln(outfile,' ');
        end;
      end;
    finally
      CloseFile(Outfile);
      SaveNumbersDlg.InitialDir := ExtractFileDir(SaveNumbersDlg.FileName); {save directory}
    end;
  end;
  NumberForm.visible := false;
end;

procedure TNumberForm.FormCreate(Sender: TObject);
begin
  SaveNumbersDlg.InitialDir := MatrForm.NumbersInitialDir;
end;

end.
