unit perunit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, UAlnEdit, Routines, Utildef ;

type
  TPuForm = class(TForm)
    VoltLab: TLabel;
    VoltEdit: TAlignEdit;
    PowEdit: TAlignEdit;
    PowerLab: TLabel;
    Z1Lab: TLabel;
    ZpuEdit: TAlignEdit;
    ZohmEdit: TAlignEdit;
    EquLab: TLabel;
    OhmLab: TLabel;
    VprevBut: TButton;
    PprevBut: TButton;
    Z2Lab: TLabel;
    Z2puEdit: TAlignEdit;
    Equ2Lab: TLabel;
    Z2ohmEdit: TAlignEdit;
    Label1: TLabel;
    PlusBut: TButton;
    MinusBut: TButton;
    SqrtPbut: TButton;
    SqrtMbut: TButton;
    ResEdit: TAlignEdit;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure VoltEditKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure PowEditKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ZpuEditKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure VoltEditExit(Sender: TObject);
    procedure PowEditExit(Sender: TObject);
    procedure ZpuEditExit(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure VprevButClick(Sender: TObject);
    procedure PprevButClick(Sender: TObject);
    procedure Z2puEditKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Z2puEditExit(Sender: TObject);
    procedure Values(var Ohm1, Ohm2 : single);
    procedure PlusButClick(Sender: TObject);
    procedure MinusButClick(Sender: TObject);
    procedure SqrtPbutClick(Sender: TObject);
    procedure SqrtMbutClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure UpdateOhms(pBase: TpAlEdit);
  end;

procedure PuToOhms;

const
  Vref_p : string = '1';  {old reference voltage}
  Pref_p : string = '1';

var
  PuForm   : TPuForm;
  dec_p    : Char;    {decimal point - character}


implementation
uses
  EMTP, Imagopas, Skemadef;

{$R *.DFM}

procedure TPuForm.FormCreate(Sender: TObject);
begin
  KeyPreview := true;                {for checking keys}
  DecimalPoint(dec_p);
  {add flag that form created}
  Imagoo.ReadyEdit.text := Imagoo.ReadyEdit.text+'u';
end;

procedure TPuForm.FormKeyPress(Sender: TObject; var Key: Char);
{check that only numerals and comma or full stop are accepted at the Edit}
begin
  {only numbers and . or , accepted}
  if not(((key >= '0') and (key <= '9')) or
          (key  = ',') or  (key  = '.')  or
          (key  = #8 ) or
          (key =  #3 ) or (key = #22)) then       {"copy" and "paste"}
           key := #0;                             {neglect other keys }
  {check correct char for decimal point}
  if      (dec_p = ',') and (key = '.') then key := ','
  else if (dec_p = '.') and (key = ',') then key := '.';
  {allow only one decimal point}
  if  (VoltEdit.tagB <> 0) or (PowEdit.tagB  <> 0) or
      (ZpuEdit.tagB  <> 0) or (Z2puEdit.tagB <> 0) then begin
    if (key = dec_p) then key := #0;
  end;
end;

procedure PuToOhms;
{convert the impedance in pu-units to ohms}
var
  err1,err2 : integer; {error flag}
  err3,err4 : integer;
  apu,apv   : single;  {temporary}
  apw       : single;  {temporary}
  ipu       : integer; {temporary}
  Pref      : single;  {reference power}
  st1,st2   : string;  {temporary}
  st3,st4   : string;  {temporary}
{ text      : string;  {temprary}
  Vref      : single;  {reference voltage}
  Zabs      : single;  {impedance in ohms}
  Z1pu,Z2pu : single;  {impedance in pu}
  Zref      : single;  {reference impedance}
begin
  ValuesFromLine(PuForm.PowEdit.text , 1,false, ipu, Pref, apu, apv, apw,st1,st2,st3,st4, err1);
  ValuesFromLine(PuForm.VoltEdit.text, 1,false, ipu, Vref, apu, apv, apw,st1,st2,st3,st4, err2);
  ValuesFromLine(PuForm.ZpuEdit.text , 1,false, ipu, Z1pu, apu, apv, apw,st1,st2,st3,st4, err3);
  ValuesFromLine(PuForm.Z2puEdit.text, 1,false, ipu, Z2pu, apu, apv, apw,st1,st2,st3,st4, err4);
  if (err1 = 0) and (err2 = 0) and
     (Pref > 0) and (Vref > 0) then begin
    Zref := Vref*Vref/Pref;
    if (err3  = 0) and (Z1pu >= 0) then begin
      Zabs := Z1pu*Zref*0.01;
      PuForm.ZohmEdit.text := SixPlaces(Zabs);
    end;
    if (err4  = 0) and (Z2pu >= 0) then begin
      Zabs := Z2pu*Zref*0.01;
      PuForm.Z2ohmEdit.text := SixPlaces(Zabs);
    end;
  end;
end;

procedure TPuForm.UpdateOhms(pBase: TpAlEdit);
{update the ohm-values after a change in the base value (MVA or kV)
 pBase  pointer to the edit for base value}
var
  ipu  : integer;
  text : string;
begin
  ipu := 0;
  {first, set flag, if there is a decimal point}
  text := pBase^.text;
  FlagDecimalPoint(text,ipu);
  pBase^.tagB := ipu;
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
end;

procedure TPuForm.VoltEditKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  ipu  : integer;
  text : string;
begin
  UpdateOhms(@VoltEdit);
(*  ipu := 0;
  {first, set flag, if there is a decimal point}
  text := VoltEdit.text;
  FlagDecimalPoint(text,ipu);
  VoltEdit.tagB := ipu;
  {change pu-value to ohms}
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
*)
end;

procedure TPuForm.PowEditKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  ipu  : integer;
  text : string;
begin
  UpdateOhms(@PowEdit);
(*ipu := 0;
  {first, set flag, if there is a decimal point}
  text := PowEdit.text;
  FlagDecimalPoint(text,ipu);
  PowEdit.tagB := ipu;
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
*)
end;

procedure TPuForm.ZpuEditKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  ipu  : integer;
  text : string;
begin
  ipu := 0;
  {first, set flag, if there is a decimal point}
  text := ZpuEdit.text;
  FlagDecimalPoint(text,ipu);
  ZpuEdit.tagB := ipu;
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
end;

procedure TPuForm.VoltEditExit(Sender: TObject);
begin
  VoltEdit.tagB := 0;
end;

procedure TPuForm.PowEditExit(Sender: TObject);
begin
  PowEdit.tagB := 0;
end;

procedure TPuForm.ZpuEditExit(Sender: TObject);
begin
  ZpuEdit.tagB := 0;
end;

procedure TPuForm.FormHide(Sender: TObject);
{save reference values as strings}
{var
  pSet : TpCompSet;}
begin
{ pSet := TpCompSet(ipSetg);      {pointer to the "common"}
  if (tag = 0) then begin         { <> 0  =  cancel-button was pressed}
    Vref_p := VoltEdit.text;
    Pref_p := PowEdit.text;
  end;
  tag := 0;
end;

procedure TPuForm.VprevButClick(Sender: TObject);
var
  ipu : integer;
begin
  FlagDecimalPoint(Vref_p, ipu);
  VoltEdit.tagB := ipu;        {set flag for decimal point}
  VoltEdit.text := Vref_p;     {get the previous reference voltage}
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
end;

procedure TPuForm.PprevButClick(Sender: TObject);
var
  ipu : integer;
begin
  FlagDecimalPoint(Pref_p, ipu);
  PowEdit.tagB := ipu;        {set flag for decimal point}
  PowEdit.text := Pref_p;     {get the previous reference voltage}
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}
end;

procedure TPuForm.Z2puEditKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var
  ipu  : integer;
  text : string;
begin
  ipu := 0;
  {first, set flag, if there is a decimal point}
  text := Z2puEdit.text;
  FlagDecimalPoint(text,ipu);
  Z2puEdit.tagB := ipu;
  PuToOhms;
  ResEdit.text := '';        {result of calculation no more valid}  
end;

procedure TPuForm.Z2puEditExit(Sender: TObject);
begin
  Z2puEdit.tagB := 0;
end;

procedure TPuForm.Values(var Ohm1, Ohm2 : single);
{get the values of impedances in ohms}
var
  apu,apv : single;  {temporary}
  apw     : single;
  er1,er2 : integer; {error flag}
  ipu     : integer; {no of parameters, temporaray}
  st1,st2 : string;  {temporary}
  st3,st4 : string;  {temporary}
  Z1,Z2   : single;  {impedances in ohms}
begin
  {read the impedances in ohm}
  ValuesFromLine(ZohmEdit.text , 1,false, ipu, Z1, apu, apv, apw,st1,st2,st3,st4, er1);
  ValuesFromLine(Z2ohmEdit.text, 1,false, ipu, Z2, apu, apv, apw,st1,st2,st3,st4, er2);
  if (er1 = 0) then Ohm1 := Z1 else Ohm1 := -1;    {-1 as a flag}
  if (er2 = 0) then Ohm2 := Z2 else Ohm2 := -1;    {-1 as a flag}
end;

procedure TPuForm.PlusButClick(Sender: TObject);
{add the impedances in ohms}
var
  sum   : single;
  Z1,Z2 : single;
begin
  Values(Z1,Z2);                        {get impedances in ohms}
  if (Z1 >= 0) and (Z2 >= 0) then begin {negative is flag for erroneous input}
    sum := Z1+Z2;
    ResEdit.text := SixPlaces(sum);
  end else begin
    ResEdit.text := '***';
  end;
end;

procedure TPuForm.MinusButClick(Sender: TObject);
{subtract the impedances}
var
  diff  : single;
  Z1,Z2 : single;
begin
  Values(Z1,Z2);                        {get impedances in ohms}
  if (Z1 >= 0) and (Z2 >= 0) then begin {negative is flag for erroneous input}
    diff := Z1-Z2;
    if (diff >= 0) then ResEdit.text := SixPlaces(diff)
    else                ResEdit.text := '- '+SixPlaces(abs(diff));
  end else begin
    ResEdit.text := '***';
  end;
end;

procedure TPuForm.SqrtPbutClick(Sender: TObject);
{calculate sqrt(Z1*Z1 + Z2*Z2) }
var
  sum   : single;
  Z1,Z2 : single;
begin
  Values(Z1,Z2);                        {get impedances in ohms}
  if (Z1 >= 0) and (Z2 >= 0) then begin {negative is flag for erroneous input}
    sum := sqrt(Z1*Z1+Z2*Z2);
    ResEdit.text := SixPlaces(sum);
  end else begin
    ResEdit.text := '***';
  end;

end;

procedure TPuForm.SqrtMbutClick(Sender: TObject);
{calculate sqrt(Z1*Z1 + Z2*Z2) }
var
  diff  : single;
  Z1,Z2 : single;
begin
  Values(Z1,Z2);                         {get impedances in ohms}
  if (Z1 >= Z2) and (Z2 >= 0) then begin {negative is flag for erroneous input}
    diff := sqrt(Z1*Z1-Z2*Z2);
    ResEdit.text := SixPlaces(diff);
  end else begin
    ResEdit.text := '****';
  end;


end;

procedure TPuForm.FormActivate(Sender: TObject);
begin
  {clear all decimal point-flags}
  VoltEdit.tagB := 0;
  PowEdit.tagB := 0;
  ZpuEdit.tagB := 0;
  Z2puEdit.tagB := 0;
end;

end.
